/*

XKnotJob V1.0

Copyright (C) 2017 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.awt.Toolkit;
import java.io.*;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Objects;
import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 *
 * @author dirk
 */
public class XKnotJob {

    /**
     * @param args the command line arguments
     */
    
    public static void main(String[] args) {
        // TODO code application logic here
        JFrame frame = new JFrame("XKnotJob");
        frame.setSize(800,600);
        frame.setResizable(false);
        frame.setLocationRelativeTo(null);
        frame.add(new LinkJob(frame));
        frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);
    }
    
}

@SuppressWarnings("serial")
class Card extends JPanel {
    CardLayout cards = new CardLayout();
    
    public Card() {
        setLayout(cards);
    }
}

@SuppressWarnings("serial")
class LinkJob extends JComponent {
    
    JPanel panel = new JPanel();
    LinkData currentData;
    DefaultListModel<String> listModel = new DefaultListModel<String>();
    JList<String> list = new JList<String>(listModel);
    ArrayList<LinkData> theLinks = new ArrayList<LinkData>();
    Card cards = new Card();
    boolean click = false;
    File loadLinksFrom = null;
    File saveKhovanov = null;
    
    public LinkJob(JFrame frame) {
        int [] torus_crossings = {-2,3,3};
        int [][] torus_paths = {{1,2,4,5},{3,6,2,1},{5,4,6,3}};
        int [] sphere_crossings = {3,-2,3,-2,3,2};
        int [][] sphere_paths = {{1,11,2,10},{7,6,3,2},{10,3,9,4},{4,5,12,1},{5,9,6,8},{11,12,8,7}};
        int [] pre_crossings = {-2,2,2};
        int [][] pre_paths = {{1,2,5,6},{3,4,2,1},{6,5,4,3}};
        int [] torus55_crossings = {2,2,-2,1,1,2,-2,2,2,2,1,-1};
        int [][] torus55_paths = {{1,2,7,6},{2,3,11,10},{10,13,8,7},{14,4,15,3},{1,19,5,14},{6,9,18,19},{4,5,20,21},{15,16,12,11},{13,12,22,23},{23,24,9,8},{21,17,22,16},{20,18,24,17}};
        int [] torus74_crossings = {5,2,1,1,2,3,-2,2,2,1};

        int [][] torus74_paths = {{12,18,16,11},{3,2,1,16},{14,15,3,18},{7,14,12,13},{4,13,11,9},{5,7,4,6},{9,1,17,10},{2,19,20,17},{15,5,8,19},{8,6,10,20}};
        
        link torus43 = new link(torus_crossings,torus_paths);
        link torus54 = new link(sphere_crossings,sphere_paths);
        link torus74 = new link(torus74_crossings,torus74_paths);
        link torus55 = new link(torus55_crossings,torus55_paths);
        link pretzel = new link(pre_crossings,pre_paths);
        LinkData torus = new LinkData("Torus(4,3)",torus43,"Also the P(-2,3,3) pretzel.",frame);
        LinkData sphere = new LinkData("Torus(5,4)",torus54,"According to some Reidemeister moves.",frame);
        LinkData torus_74 = new LinkData("Torus(7,4)",torus74,"A glued diagram for the torus knot (7,4).",frame);
        LinkData torus_55 = new LinkData("Torus(5,5)",torus55,"A glued diagram for the torus link (5,5).",frame);
        LinkData pretz = new LinkData("Pretzel(-2,2,2)",pretzel,"Apart from orientations this is the (3,3)-torus link.",frame);
        
        theLinks.add(torus);
        theLinks.add(sphere);
        theLinks.add(torus_74);
        theLinks.add(torus_55);
        theLinks.add(pretz);
        
        for (LinkData Lin : theLinks) listModel.addElement(Lin.name.getText());
        
        list.setFixedCellWidth(140);
        list.setVisibleRowCount(25);
        list.setSelectedIndex(0);
        
        currentData = torus;
        
        list.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent le) {
                @SuppressWarnings("unchecked")
                JList<String> list = (JList<String>) le.getSource();
                int u = list.getSelectedIndex();
                if ( u >= 0 ) {
                    for (LinkData Lin : theLinks) Lin.panel.setVisible(false);
                    theLinks.get(u).panel.setVisible(true);
                }
            }
        });
        
        Container box = Box.createVerticalBox();
        
        box.add(new JScrollPane(list));
        JPanel buttons = new JPanel(new GridLayout(4,1));
        JButton addLink = new JButton("Add Link");
        buttons.add(addLink);
        JButton removeLink = new JButton("Remove Links");
        buttons.add(removeLink);
        JButton loadLink = new JButton("Load Links");
        buttons.add(loadLink);
        JButton saveLink = new JButton("Save Links");
        buttons.add(saveLink);
        box.add(buttons);
        panel.add(box);
        setLayout(new BorderLayout());   
        add(panel, BorderLayout.WEST);
        for (LinkData Lin : theLinks) cards.add(Lin.panel,Lin.name.getText());
        add(cards,BorderLayout.CENTER);
        addLink.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                JDialog fram = new JDialog(new JFrame(),"Create Link", true);
                fram.setBounds(frame.getX()+200, frame.getY()+150,400,300);
                fram.setResizable(false);
                fram.setLayout(new GridLayout(4,2));
                JButton enterDiag = new JButton("Glued Diagram");
                JButton enterDT = new JButton("DT Code");
                JButton enterBraid = new JButton("Braid Code");
                JButton enteralpDT = new JButton("alphabetical DT Code");
                JButton conCaten = new JButton("Concatenate Links");
                JButton disjUnion = new JButton("Disjoint Union");
                JButton torLink = new JButton("Torus Link");
                JButton cancel = new JButton("Cancel");
                ArrayList<JPanel> Panels = new ArrayList<JPanel>();
                for (int i = 0; i < 8; i++) {
                    JPanel test = new JPanel();
                    Panels.add(test);
                }
                Panels.get(0).add(enterDiag);
                Panels.get(1).add(enterDT);
                Panels.get(2).add(enteralpDT);
                Panels.get(3).add(enterBraid);
                Panels.get(4).add(conCaten);
                Panels.get(5).add(disjUnion);
                Panels.get(6).add(torLink);
                Panels.get(7).add(cancel);
                for (int i = 0; i < 8; i++) fram.add(Panels.get(i));
                cancel.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                    }
                });
                torLink.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        enterTorusLink(frame);
                    }
                });
                enterDT.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        enterDTCode(frame);
                    }
                });
                enteralpDT.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        enteralpDTCode(frame);
                    }
                });
                enterBraid.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        enterBraid(frame);
                    }
                });
                enterDiag.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        int [] lame = {0,0,0};
                        int [][] bore = {{0,0,0,0},{0,0,0,0},{0,0,0,0}};
                        LinkData Knott = enterGluedDiagram(frame,"","",lame,bore);
                        if (Knott != null) {
                            theLinks.add(Knott);
                            listModel.addElement(Knott.name.getText());
                            cards.add(Knott.panel, Knott.name.getText());
                            if (list.getModel().getSize()>1) 
                                list.setSelectedIndex(list.getModel().getSize()-1);
                        }
                        
                    }
                });
                disjUnion.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        UnionKnot L = new UnionKnot(frame,"Disjoint Union", false);
                        if (L.Knot != null) {
                            theLinks.add(L.Knot);
                            listModel.addElement(L.Knot.name.getText());
                            cards.add(L.Knot.panel,L.Knot.name.getText());
                            if (list.getModel().getSize()>1) 
                                list.setSelectedIndex(list.getModel().getSize()-1);
                        }
                    }
                });
                conCaten.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        fram.setVisible(false);
                        UnionKnot L = new UnionKnot(frame,"Concatenate Links", true);
                        if (L.Knot != null) {
                            theLinks.add(L.Knot);
                            listModel.addElement(L.Knot.name.getText());
                            cards.add(L.Knot.panel,L.Knot.name.getText());
                            if (list.getModel().getSize()>1) 
                                list.setSelectedIndex(list.getModel().getSize()-1);
                        }
                    }
                });
                fram.setVisible(true);
            }
        });
        removeLink.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                if (list.getSelectedIndex()>=0) removeLinks(frame);
                list.setSelectedIndex(0);
            }
        });
        loadLink.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                JFileChooser chooser = new JFileChooser();
                if (loadLinksFrom != null) chooser.setCurrentDirectory(loadLinksFrom);
                FileNameExtensionFilter filter = new FileNameExtensionFilter("Glued Codes (*.gld)", "gld");
                FileNameExtensionFilter filtes = new FileNameExtensionFilter("DT-Codes (*.dtc)", "dtc");
                FileNameExtensionFilter filtet = new FileNameExtensionFilter("Alphabetic DT-Codes (*.adc)", "adc");
                chooser.setFileFilter(filter);
                chooser.addChoosableFileFilter(filtes);
                chooser.addChoosableFileFilter(filtet);
                int val = chooser.showOpenDialog(frame);
                if (val == JFileChooser.APPROVE_OPTION) {
                    try {
                        File file = chooser.getSelectedFile();
                        loadLinksFrom = chooser.getCurrentDirectory();
                        String fname = file.getAbsolutePath();
                        JTextField fieldmin = new JTextField();
                        JTextField fieldmax = new JTextField();
                        JCheckBox combineCr = new JCheckBox();
                        JCheckBox improveReide = new JCheckBox();
                        click = chooseLinks(frame,fieldmin,fieldmax,combineCr,improveReide);
                        if (click) {
                            String mini = fieldmin.getText();
                            String maxi = fieldmax.getText();
                            int miny = 0,maxy = -1;
                            try {
                                miny = Integer.parseInt(mini);
                                maxy = Integer.parseInt(maxi);
                            }
                            catch (NumberFormatException e) {

                            }
                            if (maxy == -1 | maxy - miny > 500) maxy = miny + 500;
                            ArrayList<String> knotstrings = new ArrayList<String>();
                            FileReader fr = new FileReader(file);
                            BufferedReader in = new BufferedReader(fr);
                            if (fname.endsWith(".dtc")) {
                                String line;
                                String number = "";
                                String numbtwo = "";
                                String dtcode = "";
                                int pos = 0;
                                boolean keepreading = true;
                                while (keepreading) {
                                    line = in.readLine();
                                    if (line == null) keepreading = false;
                                    else {
                                        pos++;
                                        if (pos >= miny & (pos <= maxy | maxy < miny)) 
                                            knotstrings.add(line);
                                        if (pos > maxy & maxy >= 0) keepreading = false;
                                    }
                                }
                                fr.close();
                                for (String kline : knotstrings) {
                                    boolean linkok = true;
                                    number = "";
                                    int x = 0;
                                    while (x < kline.length() & kline.charAt(x) == ' ') x++;
                                    if (x >= kline.length()) linkok = false;
                                    else {
                                        boolean firstread = true;
                                        while (x < kline.length() & firstread) {
                                            number = number + kline.substring(x, x+1);
                                            x++;
                                            if (kline.charAt(x) == ' ') firstread = false;
                                        }
                                        while (x < kline.length() & kline.charAt(x) == ' ') x++;
                                        if (x >= kline.length()) linkok = false;
                                        else {
                                            boolean secondread = true;
                                            numbtwo = "";
                                            while (x < kline.length() & secondread) {
                                                numbtwo = numbtwo + kline.substring(x, x+1);
                                                x++;
                                                if (kline.charAt(x) == ' ') secondread = false;
                                            }
                                            while (x < kline.length() & kline.charAt(x) == ' ') x++;
                                            if (x >= kline.length()) linkok = false;
                                            else {
                                                dtcode = kline.substring(x);
                                            }
                                        }
                                    }
                                    if (linkok) {
                                        String name = number;
                                        try {
                                            int u = Integer.parseInt(numbtwo);
                                            if (dtcode.contains("-")) name = name+"n"+numbtwo;
                                            else name = name+"a"+numbtwo;
                                        }
                                        catch (NumberFormatException e) {
                                            name = name + numbtwo;
                                        }
                                        boolean combine = combineCr.isSelected();
                                        boolean improve = improveReide.isSelected();
                                        link Link = obtainLink(dtcode,improve,combine);
                                        if (Link != null) {
                                            LinkData LinkDa = new LinkData(name,Link,"",frame);
                                            theLinks.add(LinkDa);
                                            listModel.addElement(LinkDa.name.getText());
                                            cards.add(LinkDa.panel, LinkDa.name.getText());
                                        }
                                    }
                                }
                            }
                            if (fname.endsWith(".adc")) {
                                String line;
                                int pos = 0;
                                boolean keepreading = true;
                                int funny = list.getModel().getSize();
                                while (keepreading) {
                                    line = in.readLine();
                                    if (line == null) keepreading = false;
                                    else {
                                        pos++;
                                        if (pos >= miny & (pos <= maxy | maxy < miny)) {
                                            boolean combine = combineCr.isSelected();
                                            boolean improve = improveReide.isSelected();
                                            link Link = obtainLinkAlp(line,improve,combine);
                                            if (Link != null) {
                                                LinkData LinkDa = new LinkData(line,Link,"",frame);
                                                theLinks.add(LinkDa);
                                                listModel.addElement(LinkDa.name.getText());
                                                cards.add(LinkDa.panel, LinkDa.name.getText());
                                            }
                                        }
                                    }
                                }
                            }
                            if (fname.endsWith(".gld")) {
                                String line1,line2,line3,line4;
                                int pos = 0;
                                boolean keepreading = true;
                                int funny = list.getModel().getSize();
                                while (keepreading) {
                                    line1 = in.readLine();
                                    line2 = in.readLine();
                                    line3 = in.readLine();
                                    line4 = in.readLine();
                                    if (line1 == null | line2 == null | line3 == null | line4 == null) keepreading = false;
                                    else {
                                        pos++;
                                        if (pos >= miny & (pos <= maxy | maxy < miny)) {
                                            link Link = diagramCreator(line3,line4);
                                            if (Link != null) {
                                                boolean combine = combineCr.isSelected();
                                                if (combine) Link = combineCrossings(Link.crossings,Link.paths,true);
                                        
                                                LinkData LinkDa = new LinkData(line1,Link,line2,frame);
                                                theLinks.add(LinkDa);
                                                listModel.addElement(LinkDa.name.getText());
                                                cards.add(LinkDa.panel, LinkDa.name.getText());
                                            }
                                            //else System.out.println("Problem"+line1+" "+line2+" "+line3+" "+line4);
                                        }
                                    }
                                }
                                fr.close();
                                list.setSelectedIndex(funny);
                            }
                        }
                    }
                    catch (IOException io) {
                        
                    }
                }
            }
        });
        
        saveLink.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                if (list.isSelectionEmpty()) {
                    //System.out.println("Nothing chosen");
                }
                else {
                    JFileChooser chooser = new JFileChooser();
                    if (loadLinksFrom != null) chooser.setCurrentDirectory(loadLinksFrom);
                    FileNameExtensionFilter filter = new FileNameExtensionFilter("Glued Codes (*.gld)", "gld");
                    chooser.setFileFilter(filter);
                    int val = chooser.showSaveDialog(frame);
                    if (val == JFileChooser.APPROVE_OPTION) {
                        try {
                            File file = chooser.getSelectedFile();
                            loadLinksFrom = chooser.getCurrentDirectory();
                            String fname = file.getAbsolutePath();
                            if(!fname.endsWith(".gld") ) {
                                file = new File(fname + ".gld");
                            }
                            FileWriter fw = new FileWriter(file);
                            PrintWriter pw = new PrintWriter(fw);
                            for (int r : list.getSelectedIndices()) {
                                LinkData theLink = theLinks.get(r);
                                pw.println(theLink.name.getText());
                                pw.println(theLink.comments.getText());
                                String crossings = "";
                                for (int u = 0; u < theLink.Link.crossings.length-1; u++) 
                                    crossings = crossings + String.valueOf(theLink.Link.crossings[u])+",";
                                crossings = crossings + String.valueOf(theLink.Link.crossings[theLink.Link.crossings.length-1]);
                                pw.println(crossings);
                                String paths = "";
                                for (int u = 0; u < theLink.Link.paths.length; u++)
                                    for (int v = 0; v < 4; v++) {
                                        paths = paths + String.valueOf(theLink.Link.paths[u][v]);
                                        if (u != theLink.Link.paths.length-1 | v != 3)
                                            paths = paths +",";
                                    }
                                pw.println(paths);
                            }
                            fw.close();
                        }
                        catch (IOException e) {
                            
                        }
                    }
                }
            }
        });
    }
    
    class UnionKnot {
        LinkData Knot = null;
        int choiceOne = 0;
        int choiceTwo = 0;
        boolean choicOne = false;
        boolean choicTwo = false;
        
        public UnionKnot(JFrame frame, String title, boolean concat) {
            JDialog fram = new JDialog(new JFrame(), title, true);
            fram.setBounds(frame.getX()+175, frame.getY()+100,450,400);
            fram.setResizable(false);
            fram.setLayout(new BorderLayout());
            JButton ok = new JButton("OK");
            JButton cancel = new JButton("Cancel");
            ok.setEnabled(false);
            JLabel label = new JLabel("Please choose two links.", SwingConstants.CENTER);
            label.setPreferredSize(new Dimension(140,30));
            fram.add(label, BorderLayout.NORTH);
            JPanel buttons = new JPanel();
            buttons.add(ok);
            buttons.add(cancel);
            fram.add(buttons, BorderLayout.SOUTH);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            JList<String> listlinks = new JList<>(listModel);
            listlinks.setSelectedIndex(0);
            JScrollPane pane = new JScrollPane(listlinks);
            pane.setPreferredSize(new Dimension(140,300));
            pane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
            fram.add(pane, BorderLayout.WEST);
            JButton LinkOne = new JButton("--- Link 1 -->");
            JButton LinkTwo = new JButton("--- Link 2 -->");
            Container middleButtons = Box.createVerticalBox();//new JPanel(new BorderLayout());
            middleButtons.setPreferredSize(new Dimension(140,300));
            middleButtons.add(Box.createVerticalGlue());
            middleButtons.add(LinkOne);
            middleButtons.add(Box.createVerticalGlue());
            middleButtons.add(LinkTwo);
            middleButtons.add(Box.createVerticalGlue());
            fram.add(middleButtons, BorderLayout.CENTER);
            JLabel KnotOne = new JLabel(" ");
            JLabel KnotTwo = new JLabel(" ");
            JCheckBox mirrorOne = new JCheckBox("Mirror");
            JCheckBox mirrorTwo = new JCheckBox("Mirror");
            SpinnerNumberModel spmdOne = new SpinnerNumberModel(1,1,1,1);
            SpinnerNumberModel spmdTwo = new SpinnerNumberModel(1,1,1,1);
            JSpinner spinOne = new JSpinner(spmdOne);
            JSpinner spinTwo = new JSpinner(spmdTwo);
            spinOne.setPreferredSize(new Dimension(20,16));
            spinTwo.setPreferredSize(new Dimension(30,12));
            JLabel labOne = new JLabel("Component");
            JLabel labTwo = new JLabel("Component");
            Container compPanOne = Box.createHorizontalBox();//new JPanel(new GridLayout(1,2));
            Container compPanTwo = Box.createHorizontalBox();//new JPanel(new GridLayout(2,1));
            compPanOne.add(labOne);
            compPanOne.add(spinOne);
            compPanTwo.add(labTwo);
            compPanTwo.add(spinTwo);
            JPanel paneOne = new JPanel(new GridLayout(3,1));//Box.createVerticalBox();//new JPanel();
            JPanel paneTwo = new JPanel(new GridLayout(3,1));//Box.createVerticalBox();//new JPanel();
            paneOne.add(KnotOne);
            //paneOne.add(Box.createVerticalGlue());
            paneOne.add(mirrorOne);
            paneTwo.add(KnotTwo);
            //paneTwo.add(Box.createVerticalGlue());
            paneTwo.add(mirrorTwo);
            if (concat) {
                //paneOne.add(Box.createVerticalGlue());
                paneOne.add(compPanOne);
                //paneTwo.add(Box.createVerticalGlue());
                paneTwo.add(compPanTwo);
            }
            Container rightStuff = Box.createVerticalBox();
            rightStuff.add(Box.createVerticalGlue());
            rightStuff.add(paneOne);
            rightStuff.add(Box.createVerticalGlue());
            rightStuff.add(paneTwo);
            rightStuff.add(Box.createVerticalGlue());
            rightStuff.setPreferredSize(new Dimension(160,300));
            fram.add(rightStuff, BorderLayout.EAST);
            ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                    link Link1 = theLinks.get(choiceOne).Link;
                    link Link2 = theLinks.get(choiceTwo).Link;
                    int k = Link1.crossings.length + Link2.crossings.length;
                    int[] crossings = new int[k];
                    int[][] paths = new int[k][4];
                    int factor = 1;
                    if (mirrorOne.isSelected()) factor = -1;
                    for (int i = 0; i < Link1.crossings.length; i++) {
                        crossings[i] = factor * Link1.crossings[i];
                        for (int j = 0; j < 4; j++) paths[i][j] = Link1.paths[i][j];
                    }
                    if (mirrorTwo.isSelected()) factor = -1;
                    else factor = 1;
                    int m = Link1.crossings.length;
                    for (int i = 0; i < Link2.crossings.length; i++) {
                        crossings[m+i] = factor * Link2.crossings[i];
                        for (int j = 0; j < 4; j++) paths[m+i][j] = Link2.paths[i][j]+2*m;
                    }
                    String name = " U ";
                    boolean abort = false;
                    if (concat) {
                        name = " # ";
                        int comp1 = (int) spinOne.getValue() - 1;
                        int comp2 = (int) spinTwo.getValue() - 1;
                        if (Link1.components.isEmpty() | Link2.components.size()==0) abort = true;
                        else {
                            int number1 = Link1.components.get(comp1).get(0);
                            int number2 = Link2.components.get(comp2).get(0)+2*m;
                            boolean notonce = true;
                            boolean notwice = true;
                            int a1 = 0;
                            int a2 = 0;
                            while (notonce|notwice) {
                                if (paths[a1][a2] == number1) {
                                    if (notonce) notonce = false;
                                    else notwice = false;
                                }
                                if (notwice) {
                                    a2++;
                                    if (a2 == 4) {
                                        a2 = 0;
                                        a1++;
                                    }
                                }
                            }
                            paths[a1][a2] = number2;
                            a1 = m;
                            a2 = 0;
                            notonce = true;
                            notwice = true;
                            while (notonce|notwice) {
                                if (paths[a1][a2] == number2) {
                                    if (notonce) notonce = false;
                                    else notwice = false;
                                }
                                if (notwice) {
                                    a2++;
                                    if (a2 == 4) {
                                        a2 = 0;
                                        a1++;
                                    }
                                }
                            }
                            paths[a1][a2] = number1;
                        }
                    }
                    name = theLinks.get(choiceOne).name.getText()+name+theLinks.get(choiceTwo).name.getText();
                    if (!abort) {
                        link Link = new link(crossings,paths);
                        Knot = new LinkData(name,Link,"",frame);
                    }
                }
            });
            LinkOne.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    int u = listlinks.getSelectedIndex();
                    String name = theLinks.get(u).name.getText();
                    KnotOne.setText(name);
                    choicOne = true;
                    if (choicTwo) ok.setEnabled(true);
                    choiceOne = u;
                    if (concat) {
                        int max = theLinks.get(u).Link.components.size();
                        if (max < 1) max = 1;
                        spmdOne.setMaximum(max);
                        spmdOne.setValue(1);
                    }
                }
            });
            LinkTwo.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    int u = listlinks.getSelectedIndex();
                    String name = theLinks.get(u).name.getText();
                    KnotTwo.setText(name);
                    choicTwo = true;
                    if (choicOne) ok.setEnabled(true);
                    choiceTwo = u;
                    if (concat) {
                        int max = theLinks.get(u).Link.components.size();
                        if (max < 1) max = 1;
                        spmdTwo.setMaximum(max);
                        spmdTwo.setValue(1);
                    }
                }
            });
            fram.setVisible(true);
        }
    }
    
    boolean chooseLinks(JFrame frame, JTextField fieldmin, JTextField fieldmax, JCheckBox combine, JCheckBox improve) {
        JDialog fram = new JDialog(new JFrame(), "Choose Links", true);
        fram.setBounds(frame.getX()+200, frame.getY()+200,400,200);
        fram.setResizable(false);
        fram.setLayout(new BorderLayout());
        JButton ok = new JButton("OK");
        JButton cancel = new JButton("Cancel");
        JLabel label = new JLabel("Please choose the range of links.",SwingConstants.CENTER);
        fram.add(label, BorderLayout.NORTH);
        JPanel panec = new JPanel(new GridLayout(2,1));
        JPanel panet = new JPanel(new GridLayout(2,1));
        JLabel min = new JLabel("First Link : ", SwingConstants.RIGHT);
        JLabel max = new JLabel("Last Link : ",SwingConstants.RIGHT);
        label.setPreferredSize(new Dimension(80,32));
        fieldmin.setPreferredSize(new Dimension(50,20));
        panec.add(min);
        panet.add(fieldmin);
        panec.add(max);
        panet.add(fieldmax);
        JPanel centert = new JPanel();
        centert.add(panec);
        centert.add(panet);
        combine.setSelected(true);
        JPanel boxes = new JPanel(new GridLayout(2,1));
        boxes.add(combine);
        boxes.add(improve);
        JLabel combilab = new JLabel("Combine Crossings");
        JPanel text = new JPanel(new GridLayout(2,1));
        text.add(combilab);
        JLabel imprlab = new JLabel("Improve via R-III");
        text.add(imprlab);
        JPanel centerb = new JPanel();
        JPanel totalcenter = new JPanel(new GridLayout(2,1));
        centerb.add(boxes);
        centerb.add(text);
        totalcenter.add(centert);
        totalcenter.add(centerb);
        fram.add(totalcenter, BorderLayout.CENTER);
        JPanel panes = new JPanel();
        panes.add(ok);
        panes.add(cancel);
        fram.add(panes, BorderLayout.SOUTH);
        JLabel empty = new JLabel(" ");
        JLabel empti = new JLabel(" ");
        empti.setPreferredSize(new Dimension(120,40));
        empty.setPreferredSize(new Dimension(100,40));
        fram.add(empty, BorderLayout.WEST);
        fram.add(empti, BorderLayout.EAST);
        click = false;
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                click = true;
                fram.setVisible(false);
            }
        });
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        fram.setVisible(true);
        return click;
    }
    
    final link diagramCreator(String arg1, String arg2) {
    	link Link = null;
    	int fool = 0;
    	int [] crossings = new int[3];
    	int [][] paths = new int[3][4];
    	arg1 = arg1+",";
        int k = arg1.length();
        ArrayList<Integer> commas = new ArrayList<Integer>();
        int j = 1;
        while (j < k) {
            if (arg1.charAt(j) == ',') commas.add(j);
            j++;
        }
        crossings = new int [commas.size()];
        try {
            int pos = 0;
            for (int i = 0 ; i < commas.size(); i++) {
                crossings[i] = Integer.parseInt(arg1.substring(pos,commas.get(i)));
                pos = commas.get(i)+1;
            }
        }
        catch (NumberFormatException e) {
            fool = 4;
        }
        arg2 = arg2+",";
        k = arg2.length();
        commas.clear();
        j = 1;
        while (j < k) {
            if (arg2.charAt(j) == ',') commas.add(j);
            j++;
        }
        paths = new int [crossings.length][4];
        if (commas.size() == 4 * crossings.length) {
            try {
                int pos = 0;
                for (int i = 0; i < commas.size(); i++) {
                    paths[i/4][i%4] = Integer.parseInt(arg2.substring(pos,commas.get(i)));
                    pos = commas.get(i)+1;
                }
            }
            catch (Exception e) {
                fool = 5;
            }
        }
        else fool = 5;
        if (fool == 0) Link = new link(crossings,paths);
    	return Link;
    }
    
    void removeLinks(JFrame frame) {
        JDialog fram = new JDialog(new JFrame(),"Remove Links", true);
        fram.setBounds(frame.getX()+300,frame.getY()+250,200,100);
        fram.setResizable(false);
        JLabel labe = new JLabel("Are you sure?",SwingConstants.CENTER);
        fram.setLayout(new BorderLayout());
        fram.add(labe, BorderLayout.CENTER);
        JButton yesBut = new JButton("Yes");
        JButton noBut = new JButton("No");
        JPanel pane = new JPanel();
        pane.add(yesBut);
        pane.add(noBut);
        fram.add(pane, BorderLayout.SOUTH);
        noBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        yesBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                int[] u = list.getSelectedIndices();
                for (int v = u.length - 1; v >= 0; v--) {
                    cards.remove(u[v]);
                    theLinks.remove(u[v]);
                    listModel.remove(u[v]);
                }
                fram.setVisible(false);
            }
        });
        fram.setVisible(true);
    }
    
    LinkData Knot;
    //link theLink;
    //String dtCode;
    
    private class enterData{
        String theString = null;
        boolean combin = true;
        boolean improv = true;
        
        enterData(JFrame frame, String title, String code, boolean allow) {
            JDialog fram = new JDialog(new JFrame(), title, true);
            fram.setBounds(frame.getX()+200, frame.getY()+215, 400, 170);
            fram.setResizable(false);
            fram.setLayout(new BorderLayout());
            JTextField field = new JTextField();
            field.setPreferredSize(new Dimension(300,30));
            JLabel label = new JLabel(code);
            JPanel pane = new JPanel();
            JPanel buttons = new JPanel();
            JButton ok = new JButton("OK");
            JButton cancel = new JButton("Cancel");
            buttons.add(ok);
            buttons.add(cancel);
            pane.add(label);
            pane.add(field);
            JPanel boxes = new JPanel(new GridLayout(2,1));
            JCheckBox combine = new JCheckBox();
            JCheckBox improve = new JCheckBox();
            combine.setSelected(true);
            improve.setEnabled(allow);
            boxes.add(combine);
            boxes.add(improve);
            JLabel combilab = new JLabel("Combine Crossings");
            JPanel text = new JPanel(new GridLayout(2,1));
            text.add(combilab);
            JLabel imprlab = new JLabel("Improve via R-III");
            text.add(imprlab);
            JPanel centerb = new JPanel();
            centerb.add(boxes);
            centerb.add(text);
            fram.add(centerb, BorderLayout.CENTER);
            fram.add(pane,BorderLayout.NORTH);
            fram.add(buttons, BorderLayout.SOUTH);
            field.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    theString = field.getText();
                    improv = improve.isSelected();
                    combin = combine.isSelected();
                    fram.setVisible(false);
                }
            });
            ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    theString = field.getText();
                    improv = improve.isSelected();
                    combin = combine.isSelected();
                    fram.setVisible(false);
                }
            });
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            fram.setVisible(true);
        }
    }
    
    void enterBraid(JFrame frame) {
        enterData data = new enterData(frame,"Braid-Code", "Braid-Code : ", true);
        if (data.theString == null) return;
        link theLink = obtainBraid(data.theString,data.improv,data.combin);
        if (theLink == null) return;
        LinkData NData = enterGluedDiagram(frame,data.theString,"",theLink.crossings,theLink.paths);
        if (NData != null) {
            theLinks.add(NData);
            listModel.addElement(NData.name.getText());
            cards.add(NData.panel, NData.name.getText());
            if (list.getModel().getSize()>1) 
                list.setSelectedIndex(list.getModel().getSize()-1);
        }
    }
    
    void enterDTCode(JFrame frame) {
        enterData data = new enterData(frame,"DT-Code", "DT-Code : ", true);
        if (data.theString == null) return;
        link theLink = obtainLink(data.theString,data.improv,data.combin);
        if (theLink == null) return;
        LinkData NData = enterGluedDiagram(frame,data.theString,"",theLink.crossings,theLink.paths);
        if (NData != null) {
            theLinks.add(NData);
            listModel.addElement(NData.name.getText());
            cards.add(NData.panel, NData.name.getText());
            if (list.getModel().getSize()>1) 
                list.setSelectedIndex(list.getModel().getSize()-1);
        }
    }
    
    void enteralpDTCode(JFrame frame) {
        enterData data = new enterData(frame,"Alphabetical DT-Code", "DT-Code : ", true);
        if (data.theString == null) return;
        link theLink = obtainLinkAlp(data.theString,data.improv,data.combin);
        if (theLink == null) return;
        LinkData NData = enterGluedDiagram(frame,data.theString,"",theLink.crossings,theLink.paths);
        if (NData != null) {
            theLinks.add(NData);
            listModel.addElement(NData.name.getText());
            cards.add(NData.panel, NData.name.getText());
            if (list.getModel().getSize()>1) 
                list.setSelectedIndex(list.getModel().getSize()-1);
        }
    }
    
    link obtainLink(String dtcode, boolean rmover, boolean combine) {
        link Link = null;
        ArrayList<Integer> numbers = new ArrayList<Integer>();
        boolean ok = true;
        int u = 0;
        int v = 0;
        String number = "";
        dtcode = dtcode+" ";
        while (ok & u < dtcode.length()) {
            if (dtcode.charAt(u)== ' ') {
                number = dtcode.substring(v, u);
                v = u+1;
                if (number.length() > 0) {
                    try {
                        int numb = Integer.parseInt(number);
                        numbers.add(numb);
                    }
                    catch(NumberFormatException e) {
                        ok = false;
                    }
                }
            }
            u++;
        }
        if (!ok) return null;
        int k = numbers.size();
        ArrayList<Integer> check = new ArrayList<Integer>();
        for (int i = 0; i < 2*k; i++) check.add(i+1);
        int i = 0;
        while (i < k & ok) {
            if (!check.contains(Math.abs(numbers.get(i)))|numbers.get(i)%2 !=0) ok = false;
            else i++;
        }
        if (!ok) return null;
        int[] crossings = new int[numbers.size()];
        int[][] paths = new int[crossings.length][4];
        paths[0][0] = 1;
        paths[0][2] = 2;
        paths = modifyPaths(paths,2,numbers);
        if (paths == null) return null;
        for (int j = 0; j < numbers.size(); j++) {
            int odd = 2*j+1;
            int even = odd+1;
            boolean found = false;
            int w = 0;
            while (!found) {
                if (contains(paths[w],odd,even)) found = true;
                else w++;
            }
            if (paths[w][0] == odd | paths[w][2] == odd) {
                if (numbers.get(j) < 0) crossings[w] = 1;
                else crossings[w] = -1;
            }
            else {
                if (numbers.get(j) < 0) crossings[w] = -1;
                else crossings[w] = 1;
            }
        }
        Link = new link(crossings,paths);
        if (rmover) {
            Link = girthMinimize(getOptimized(Link));
        }
        else {
            if (combine) Link = combineCrossings(crossings,paths,true); 
        }
        return Link;
    }
    
    link linkFromRMove(link Link, ReidemeisterMove move) {
        link newLink = cloneLink(Link);
        int changeX = (move.u+1)%4;
        int changeY = (move.v+1)%4;
        int keepZX = move.t;
        int keepZY = move.s;
        boolean xBigger = ((4+move.t-move.s)%4 == 1);
        int moveZX, moveZY;
        if (xBigger) {
            moveZX = (keepZX+1)%4;
            moveZY = (keepZY+3)%4;
        }
        else {
            moveZX = (keepZX+3)%4;
            moveZY = (keepZY+1)%4;
        }
        if (contains(Link.paths[move.z],Link.paths[move.x][changeX])) {
            newLink.paths[move.x][changeX] = Link.paths[move.z][moveZX];
            newLink.paths[move.x][(changeX+2)%4] = Link.paths[move.x][changeX];
            newLink.paths[move.z][keepZX] = Link.paths[move.x][(changeX+2)%4];
            newLink.paths[move.z][moveZX] = Link.paths[move.x][changeX];
        }
        else {
            newLink.paths[move.x][(changeX+2)%4] = Link.paths[move.z][moveZX];
            newLink.paths[move.x][changeX] = Link.paths[move.x][(changeX+2)%4];
            newLink.paths[move.z][keepZX] = Link.paths[move.x][changeX];
            newLink.paths[move.z][moveZX] = Link.paths[move.x][(changeX+2)%4];
        }
        if (contains(Link.paths[move.z], Link.paths[move.y][changeY])) {
            newLink.paths[move.y][changeY] = Link.paths[move.z][moveZY];
            newLink.paths[move.y][(changeY+2)%4] = Link.paths[move.y][changeY];
            newLink.paths[move.z][keepZY] = Link.paths[move.y][(changeY+2)%4];
            newLink.paths[move.z][moveZY] = Link.paths[move.y][changeY];
        }
        else {
            newLink.paths[move.y][(changeY+2)%4] = Link.paths[move.z][moveZY];
            newLink.paths[move.y][changeY] = Link.paths[move.y][(changeY+2)%4];
            newLink.paths[move.z][keepZY] = Link.paths[move.y][changeY];
            newLink.paths[move.z][moveZY] = Link.paths[move.y][(changeY+2)%4];
        }
        return newLink;
    }
    
    link cloneLink(link Link) {
        int[] ncrossings = new int[Link.crossings.length];
        int[][] npaths = new int[Link.paths.length][4];
        for (int r = 0; r < ncrossings.length; r++) ncrossings[r] = Link.crossings[r];
        for (int r = 0; r < ncrossings.length; r++) {
            for (int s = 0; s < 4; s++) npaths[r][s] = Link.paths[r][s];
        }
        link clone = new link(ncrossings,npaths);
        return clone;
    }
    
    ArrayList<ReidemeisterMove> findReidemeister(link Link) {
        ArrayList<ReidemeisterMove> moves = new ArrayList<ReidemeisterMove>();
        int a = 0;
        while (a < Link.crossings.length-1) {
            if (Math.abs(Link.crossings[a]) == 1) {
                int b = a+1;
                while (b < Link.crossings.length) {
                    if (Math.abs(Link.crossings[b]) == 1) {
                        boolean found = false;
                        int u = 0;
                        int v = 0;
                        while (!found & u < 4) {
                            v = 0;
                            boolean abovea = (((u % 2) == 0) == (Link.crossings[a] < 0));
                            while (!found & v < 4) {
                                boolean aboveb = (((v % 2) == 0) == (Link.crossings[b] < 0));
                                if ((aboveb == abovea) & (Link.paths[a][u] == Link.paths[b][v])) found = true;
                                else v++;
                            }
                            if (!found) u++;
                        }
                        if (found) {// possible Reidemeister Move III (or II, but not checking for those)
                            for (int c = 0; c < Link.crossings.length; c++) {
                                if (c != a & c != b ) {
                                    boolean gfund = false;
                                    boolean order = true;
                                    int t = 0;
                                    int s = 1;
                                    while (!gfund & t < 3) {
                                        s = t+1;
                                        while (!gfund & s < 4) {
                                            int pa = Link.paths[c][t];
                                            int pb = Link.paths[c][s];
                                            if (contains(Link.paths[a],pa) & contains(Link.paths[b],pb)) gfund = true;
                                            else if (contains(Link.paths[b],pa) & contains(Link.paths[a],pb)) {
                                                gfund = true;
                                                order = false;
                                            }
                                            if (!gfund) s++;
                                        }
                                        if (!gfund) t++;
                                    }
                                    if (gfund) {
                                        ReidemeisterMove move;
                                        if (order) move = new ReidemeisterMove(a,b,c,u,v,t,s);
                                        else move = new ReidemeisterMove(a,b,c,u,v,s,t);
                                        moves.add(move);
                                    }
                                }
                            }
                        }
                    }
                    b++;
                }
            }
            a++;
        }
        return moves;
    }
    
    boolean contains(int[] aray, int u) {
        boolean inthere = false;
        int t = 0;
        while (t < aray.length & !inthere) {
            if (aray[t] == u) inthere = true;
            else t++;
        }
        return inthere;
    }
    
    link getOptimized(link Link) {
        link optimalLink = combineCrossings(Link.crossings,Link.paths,false);
        ReidemeisterMove emptyMove = new ReidemeisterMove();
        ArrayList<ReidemeisterMove> moves = findReidemeister(Link);
        for (ReidemeisterMove move : moves) {
            link newLink = linkFromRMove(Link,move);
            newLink = optimize(newLink,emptyMove);
            if (newLink.crossings.length < optimalLink.crossings.length) optimalLink = newLink;
        }
        return optimalLink;
    }
    
    link optimize(link Link, ReidemeisterMove move) {
        Link = combineCrossings(Link.crossings, Link.paths,false);
        link optLink = cloneLink(Link);
        ReidemeisterMove emptyMove = new ReidemeisterMove();
        ArrayList<ReidemeisterMove> moves = findReidemeister(Link);
        for (ReidemeisterMove mov : moves) {
            if (mov.compareTo(move) > 0) {
                link newLink = linkFromRMove(Link,mov);
                newLink = combineCrossings(newLink.crossings, newLink.paths,false);
                if (newLink.crossings.length < Link.crossings.length) newLink = optimize(newLink,emptyMove);
                else newLink = optimize(newLink,mov);
                if (newLink.crossings.length < optLink.crossings.length) optLink = newLink;
            }
        }
        return optLink;
    }
    
    link combineCrossings(int[] crossings, int[][] paths, boolean girthMin) {
        int i = 0;
        boolean sequence = false;
        int s1 = 0;
        int s2 = 0;
        int q1 = 0;
        int q2 = 0;//q's are position, s's are values
        while (i < crossings.length-1) {
            if (sequence) {
                int u = i+1;
                boolean found = false;
                while (u < crossings.length & !found) {
                    if (contains(paths[u],s1,s2,crossings[u])) {
                        found = true;
                        boolean fnd = false;
                        int y = 0;
                        while (!fnd) {
                            if (paths[u][y] == s2) fnd = true;
                            else y++;
                        }
                        int t1 = (y+1)%4;
                        int p1 = (y+2)%4;
                        int p2 = (y+3)%4;
                        int[] newcrossings = new int[crossings.length-1];
                        for (int j = 0; j < i; j++) newcrossings[j] = crossings[j];
                        if (t1%2 != 0) newcrossings[i] = crossings[i]-crossings[u];
                        else newcrossings[i] = crossings[i]+crossings[u];
                        for (int j = i+1; j < u; j++) newcrossings[j] = crossings[j];
                        for (int j = u; j < newcrossings.length; j++) newcrossings[j] = crossings[j+1];
                        int[][] newpaths = new int[paths.length-1][4];
                        for (int j = 0; j < i; j++) {
                            for (int k = 0; k < 4; k++) 
                                newpaths[j][k] = theNewNumber(paths[j][k],s1,s2,2*paths.length);
                        }
                        newpaths[i][q1] = theNewNumber(paths[i][q1],s1,s2,2*paths.length);
                        newpaths[i][q2] = theNewNumber(paths[i][q2],s1,s2,2*paths.length);
                        newpaths[i][(q1+1)%4] = theNewNumber(paths[u][p1],s1,s2,2*paths.length);
                        newpaths[i][(q1+2)%4] = theNewNumber(paths[u][p2],s1,s2,2*paths.length);
                        for (int j = i+1; j < u; j++) {
                            for (int k = 0; k < 4; k++)
                                newpaths[j][k] = theNewNumber(paths[j][k],s1,s2,2*paths.length);
                        }
                        for (int j = u; j < newpaths.length; j++) {
                            for (int k = 0; k < 4; k++)
                                newpaths[j][k] = theNewNumber(paths[j+1][k],s1,s2,2*paths.length);
                        }
                        paths = newpaths;
                        crossings = newcrossings;
                        s1 = paths[i][(q1+1)%4];
                        s2 = paths[i][(q1+2)%4];
                    }
                    u++;
                }
                if (!found) {
                    sequence = false;
                    i++;
                }
            }
            else {
                q2 = 0;
                q1 = 1;
                boolean found = false;
                while (!found & q2 < 4) {
                    s1 = paths[i][(q1+1)%4];
                    s2 = paths[i][(q1+2)%4];
                    int u = i+1;
                    while (!found & u < paths.length) {
                        if (Math.abs(crossings[i]) == 1 | (q1%2 == 0)) {
                            if (contains(paths[u],s1,s2,crossings[u])) found = true;
                        }
                        if (!found) u++;
                    }
                    if (!found) {
                        q2++;
                        q1 = (q2+1)%4;
                    }
                }
                if (found) {
                    sequence = true;
                    if (q2%2 ==0) {
                        int help = paths[i][0];
                        paths[i][0] = paths[i][1];
                        paths[i][1] = paths[i][2];
                        paths[i][2] = paths[i][3];
                        paths[i][3] = help;
                        crossings[i] = - crossings[i];
                        q1 = (q1+3)%4;
                        q2 = (q2+3)%4;
                    }
                }
                else i++;
            }
        }
        if (girthMin) {
            return girthMinimize(new link(crossings,paths));
        }
        else return new link(crossings,paths);
    }
    
    link girthMinimize(link Link) {
        int[] crossings = new int[Link.crossings.length];
        int[][] paths = new int[Link.paths.length][4];
        ArrayList<Integer> used = new ArrayList<Integer>();
        ArrayList<Integer> usedPaths = new ArrayList<Integer>();
        crossings[0] = Link.crossings[0];
        for (int i = 0; i < 4; i++) {
            paths[0][i] = Link.paths[0][i];
            usedPaths.add(Link.paths[0][i]);
        }
        used.add(0);
        int k = 1;
        while (k < crossings.length) {
            int candidate = 1;
            int overlap = -1;
            for (int j = 1; j < crossings.length; j++) {
                if (!used.contains(j)) {
                    int newov = overlap(usedPaths,Link.paths[j]); 
                    if (newov > overlap) {
                        overlap = newov;
                        candidate = j;
                    }
                }
            }
            used.add(candidate);
            for (int l = 0; l < 4; l++) {
                crossings[k] = Link.crossings[candidate];
                paths[k][l] = Link.paths[candidate][l];
                usedPaths.add(paths[k][l]);
            }
            k++;
        }
        link theLink = new link(crossings,paths);
        return theLink;
    }
    
    int overlap(ArrayList<Integer> paths, int[] path) {
        int lap = 0;
        for (int t : path) if (paths.contains(t)) lap++;
        return lap;
    }
    
    link obtainLinkAlp(String code, boolean rmover, boolean combine) {
        link Link = null;
        char[] chars = code.toCharArray();
        if (chars.length < 3) return null;
        int numberCross = Character.getNumericValue(chars[0])-9;
        if (numberCross < 1) return null;
        int numberComp = Character.getNumericValue(chars[1])-9;
        if (numberComp < 1) return null;
        if (chars.length < 2+numberComp) return null;
        int[] compCross = new int[numberComp];
        for (int e = 2; e < 2+numberComp ; e++) compCross[e-2] = Character.getNumericValue(chars[e])-9;
        int k = 0;
        int sum = 0;
        boolean bad = false;
        while (!bad & k < numberComp) {
            if (compCross[k] < 1) bad = true;
            sum = sum + compCross[k];
            k++;
        }
        if (sum != numberCross) bad = true;
        if (bad) return null;
        for (int e = 1; e < compCross.length; e++) compCross[e] = compCross[e]+compCross[e-1];
        if (chars.length != 2 + numberComp + numberCross) return null;
        int[] numbers = new int[numberCross];
        boolean[] used = new boolean[numberCross];
        for (int i = 2+numberComp; i < chars.length; i++) {
            int v = Character.getNumericValue(chars[i])-9;
            int factor = -1;
            if (Character.isLowerCase(chars[i])) factor = 1;
            numbers[i-2-numberComp] = 2 * v * factor;
            if (v >= 1 & v <= numberCross) used[v-1] = true;
        }
        boolean allgood = true;
        int y = 0;
        while (y < used.length & allgood) {
            allgood = used[y];
            y++;
        }
        if (!allgood) return null;
        int[][] paths = new int[numberCross][4];
        paths[0][0] = 1;
        paths[0][2] = 2;
        paths = modifyPaths(paths,compCross,2,numbers);
        if (paths == null) return null;
        int[] crossings = new int[numberCross];
        for (int j = 0; j < numbers.length; j++) {
            int odd = 2*j+1;
            int even = odd+1;
            boolean found = false;
            int w = 0;
            while (!found) {
                if (contains(paths[w],odd,even)) found = true;
                else w++;
            }
            if (paths[w][0] == odd | paths[w][2] == odd) {
                if (numbers[j] < 0) crossings[w] = 1;
                else crossings[w] = -1;
            }
            else {
                if (numbers[j] < 0) crossings[w] = -1;
                else crossings[w] = 1;
            }
        }
        Link = new link(crossings,paths);
        if (rmover) {
            Link = girthMinimize(getOptimized(Link));
        }
        else {
            if (combine) Link = combineCrossings(crossings,paths,true); 
        }
        return Link;
    }
    
    link obtainBraid(String braid, boolean rmover, boolean combine) {
        link Braid = null;
        char[] chars = braid.toCharArray();
        if (chars.length < 1) return null;
        ArrayList<Integer> crossins = new ArrayList<Integer>();
        ArrayList<ArrayList<Integer>> pathz = new ArrayList<ArrayList<Integer>>();
        int numberBraids = Character.getNumericValue(chars[0])-9;
        if (numberBraids < 2 | numberBraids > 10) return Braid;
        int[] ends = new int[numberBraids];
        int lastNumber = numberBraids;
        for (int i = 1; i <= numberBraids; i++) ends[i-1] = i;
        for (int j = 1; j < chars.length; j++) {
            int e = Character.getNumericValue(chars[j])-10;
            if (e < 0 | e >= numberBraids) return Braid;
            if (Character.isLowerCase(chars[j])) crossins.add(1);
            else crossins.add(-1);
            ArrayList<Integer> path = new ArrayList<Integer>();
            path.add(ends[e+1]);
            path.add(lastNumber+2);
            path.add(lastNumber+1);
            path.add(ends[e]);
            ends[e] = lastNumber+1;
            ends[e+1] = lastNumber+2;
            lastNumber = lastNumber+2;
            pathz.add(path);
        }
        ArrayList<Integer> loops = new ArrayList<Integer>();
        for (int r = 0; r < ends.length; r++) {
            if (ends[r] == r+1) loops.add(r+1);
        }
        int[] crossings = new int[crossins.size()+loops.size()];
        int[][] paths = new int[crossins.size()+loops.size()][4];
        for (int i = 0; i < crossins.size(); i++) {
            crossings[i] = crossins.get(i);
            for (int j = 0; j < 4; j++) paths[i][j] = pathz.get(i).get(j);
        }
        for (int[] path : paths) {
            for (int j = 0; j < 4; j++) {
                if (contains(ends,path[j])) {
                    for (int[] pat : paths) reduce(pat, path[j]);
                    int t = path[j];
                    path[j] = numberOf(ends,path[j]);
                    reduce(ends, t);
                    ends[path[j]-1] = -1;
                    lastNumber--;
                }
            }
        }
        for (int w = 0; w < loops.size(); w++) {
            crossings[w+crossins.size()] = 0;
            paths[w+crossins.size()][0] = loops.get(w);
            paths[w+crossins.size()][1] = lastNumber+1;
            paths[w+crossins.size()][2] = lastNumber+1;
            paths[w+crossins.size()][3] = loops.get(w);
            lastNumber++;
        }
        Braid = new link(crossings,paths);
        if (rmover) {
            Braid = girthMinimize(getOptimized(Braid));
        }
        else {
            if (combine) Braid = combineCrossings(crossings,paths,true); 
        }
        return Braid;
    }
    
    void reduce(int[] path, int value) {
        for (int j = 0; j < path.length; j++) if (path[j] > value) path[j] = path[j]-1;
    }
    
    int numberOf(int[] aray, int entry) {
        boolean found = false;
        int i = 0;
        int n = -1;
        while (!found & i < aray.length) {
            if (aray[i] == entry) found = true;
            else i++;
        }
        if (found) n = i+1;
        return n;
    }
    
    boolean contains(int[] path, int o, int e, int cros) {
        boolean found = false;
        for (int r = 0; r < 4; r++) {
            if (Math.abs(cros) == 1 | (r%2 != 0) ) {
                if (path[r] == o & path[(r+1)%4] == e) found = true;
                if (path[r] == e & path[(r+1)%4] == o) found = true;
            }
        }
        return found;
    }
    
    int theNewNumber(int old, int s1, int s2, int max) {
        int nu = old;
        int big = s2;
        int sma = s1;
        if (s1 > s2) {
            big = s1;
            sma = s2;
        }
        if (sma == 1 | big == max) {
            if (sma == 1) {
                if (old < big) nu = old - 1;
                else nu = old - 2;
            }
            else {
                if (old >= sma) nu = old - 1;
            }
        }
        else {
            if (old >= sma & old <= big) nu = old - 1;
            if (old > big) nu = old - 2;
        }
        return nu;
    }
    
    boolean contains(int[] path, int o, int e) {
        boolean conto = false;
        boolean conte = false;
        for (int r : path) {
            if (r == o) conto = true;
            if (r == e) conte = true;
        }
        return conto & conte;
    }
    
    int[][] modifyPaths(int[][] paths, int[] components, int runner, int[] numbers) {
        if (!isPlanar(paths,runner)) return null;
        if (runner > 2 * paths.length) return paths;
        int [][] copyPaths = copyOf(paths);
        int odd = 1;
        int compbegin = 1;
        int comp = 1;
        boolean cont = true;
        while (cont) {
            if (runner <= 2 * components[comp-1]) cont = false;
            else comp++;
            if (comp >= components.length) cont = false;
        }
        boolean lastInComp = false;
        if (runner == 2 * components[comp-1]) lastInComp = true;
        if (comp > 1) compbegin = 2 * components[comp-2] + 1;
        if (runner % 2 == 0) {
            boolean found = false;
            int i = 0;
            while (!found) {
                if (Math.abs(numbers[i]) == runner) found = true;
                else i++;
            }
            odd = 2*i + 1;
            if (odd > runner) {
                found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == 0 & copyPaths[u][1] == 0) found = true;
                    else u++;
                }
                copyPaths[u][0] = runner;
                if (!lastInComp) copyPaths[u][2] = runner+1;
                else copyPaths[u][2] = compbegin;
                return modifyPaths(copyPaths,components,runner+1,numbers);
            }
            else {
                found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == odd) found = true;
                    else u++;
                }
                copyPaths[u][1] = runner;
                runner++;
                int nextrun = runner;
                if (lastInComp) runner = compbegin;
                copyPaths[u][3] = runner;
                copyPaths = modifyPaths(copyPaths,components,nextrun,numbers);
                if (copyPaths != null) return copyPaths;
                else {
                    int [][] cop2Paths = copyOf(paths);
                    cop2Paths[u][3] = nextrun - 1;
                    cop2Paths[u][1] = runner;
                    return modifyPaths(cop2Paths,components,nextrun,numbers);
                }
            }
        }
        else {
            int i = (runner - 1)/2;
            int even = Math.abs(numbers[i]);
            if (even > runner) {
                boolean found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == 0 & copyPaths[u][1] == 0) found = true;
                    else u++;
                }
                copyPaths[u][0] = runner;
                if (!lastInComp) copyPaths[u][2] = runner+1;
                else copyPaths[u][2] = compbegin;
                return modifyPaths(copyPaths,components,runner+1,numbers);
            }
            else {
                boolean found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == even) found = true;
                    else u++;
                }
                copyPaths[u][1] = runner;
                runner++;
                int nextrun = runner;
                if (lastInComp) runner = compbegin;
                copyPaths[u][3] = runner;
                copyPaths = modifyPaths(copyPaths,components,nextrun,numbers);
                if (copyPaths != null) return copyPaths;
                else {
                    int [][] cop2Paths = copyOf(paths);
                    cop2Paths[u][3] = nextrun - 1;
                    cop2Paths[u][1] = runner;
                    return modifyPaths(cop2Paths,components,nextrun,numbers);
                }
            }
        }
    }
    
    int[][] modifyPaths(int[][] paths, int runner, ArrayList<Integer> numbers) {
        if (!isPlanar(paths,runner)) return null;
        if (runner > 2 * paths.length) return paths;
        int [][] copyPaths = copyOf(paths);
        int odd = 1;
        if (runner % 2 == 0) {
            boolean found = false;
            int i = 0;
            while (!found) {
                if (Math.abs(numbers.get(i)) == runner) found = true;
                else i++;
            }
            odd = 2*i + 1;
            if (odd > runner) {
                found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == 0 & copyPaths[u][1] == 0) found = true;
                    else u++;
                }
                copyPaths[u][0] = runner;
                copyPaths[u][2] = runner+1;
                return modifyPaths(copyPaths,runner+1,numbers);
            }
            else {
                found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == odd) found = true;
                    else u++;
                }
                copyPaths[u][1] = runner;
                runner++;
                int nextrun = runner;
                if (runner > 2*paths.length) runner = 1;
                copyPaths[u][3] = runner;
                copyPaths = modifyPaths(copyPaths,nextrun,numbers);
                if (copyPaths != null) return copyPaths;
                else {
                    int [][] cop2Paths = copyOf(paths);
                    cop2Paths[u][3] = nextrun - 1;
                    cop2Paths[u][1] = runner;
                    return modifyPaths(cop2Paths,nextrun,numbers);
                }
            }
        }
        else {
            int i = (runner - 1)/2;
            int even = Math.abs(numbers.get(i));
            if (even > runner) {
                boolean found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == 0 & copyPaths[u][1] == 0) found = true;
                    else u++;
                }
                copyPaths[u][0] = runner;
                copyPaths[u][2] = runner+1;
                return modifyPaths(copyPaths,runner+1,numbers);
            }
            else {
                boolean found = false;
                int u = 0;
                while (!found) {
                    if (copyPaths[u][0] == even) found = true;
                    else u++;
                }
                copyPaths[u][1] = runner;
                runner++;
                int nextrun = runner;
                if (runner > 2*paths.length) runner = 1;
                copyPaths[u][3] = runner;
                copyPaths = modifyPaths(copyPaths,nextrun,numbers);
                if (copyPaths != null) return copyPaths;
                else {
                    int [][] cop2Paths = copyOf(paths);
                    cop2Paths[u][3] = nextrun - 1;
                    cop2Paths[u][1] = runner;
                    return modifyPaths(cop2Paths,nextrun,numbers);
                }
            }
        }
    }
    
    boolean isPlanar(int[][] paths, int max) {
        boolean [][] checkedSurfaces = new boolean[paths.length][4];
        int discs = 0;
        int runner = 1;
        int pos1 = 0;
        int pos2 = 0;
        int twice = 0;
        
        //System.out.println("New guy");
        //for (int[] path : paths) {
        //    for (int p : path) System.out.print(p+" ");
        //    System.out.println();
        //}
        
        for (int i = 0; i < paths.length; i++) 
            if (paths[i][0] == 0 & paths[i][1] == 0) {
                for (int j = 0; j < 4; j++) checkedSurfaces[i][j] = true;
            }
        while (notAll(checkedSurfaces)) {
            checkedSurfaces[pos1][pos2] = true;
            pos2 = (pos2+1)%4;
            runner = paths[pos1][pos2];
            if (runner >= 1 ) {
                boolean found = false;
                int i = 0;
                int j = 0;
                while (!found & i < paths.length) {
                    j = 0;
                    while (!found & j < 4) {
                        if (paths[i][j] == runner & (!(i == pos1 & j == pos2)))
                            found = true;
                        else j++;
                    }
                    if (!found) i++;
                }
                if (found) {
                    pos1 = i;
                    pos2 = j;
                    twice++;
                }
            }
            if (checkedSurfaces[pos1][pos2]) {
                discs++;
                boolean found = false;
                pos1 = 0;
                while (!found & pos1 < paths.length) {
                    pos2 = 0;
                    while (!found & pos2 < 4) {
                        if (!checkedSurfaces[pos1][pos2]) found = true;
                        else pos2++;
                    }
                    if (!found) pos1++;
                }
            }
        }
        int vme = 0;
        for (int i = 0; i < paths.length; i++) if (paths[i][0]!=0) vme++;
        //System.out.println(vme+" - " +(twice/2)+ " + "+discs);
        return (vme+discs-(twice/2) >= 2);
        //if (max > 2*paths.length) vme = vme - 2 * paths.length;
        //else vme = vme - max + 2;
        //System.out.println(vme+" "+discs);
        //System.out.println(once+" "+twice+" "+discs);
        //return (vme+discs == 2);
    }
    
    boolean notAll(boolean [][] check) {
        boolean notall = false;
        int i = 0;
        int j = 0;
        while (i < check.length & !notall) {
            j = 0;
            while (j < 4 & !notall) {
                if (!check[i][j]) notall = true;
                else j++;
            }
            if (!notall) i++;
        }
        return notall;
    }
    
    void enterTorusLink(JFrame frame) {
        JDialog fram = new JDialog(new JFrame(), "Torus Link", true);
        fram.setBounds(frame.getX()+200, frame.getY()+225, 400,150);
        fram.setResizable(false);//frame.setLocationRelativeTo(null);
        fram.setLayout(new BorderLayout());
        SpinnerNumberModel modelp = new SpinnerNumberModel(5,2,60,1);
        JSpinner spinnerp = new JSpinner(modelp);
        JLabel labep = new JLabel("Enter a p-value :");
        SpinnerNumberModel modelq = new SpinnerNumberModel(3,2,10,1);
        JSpinner spinnerq = new JSpinner(modelq);
        JLabel labeq = new JLabel("Enter a q-value :");
        JPanel panep = new JPanel();
        panep.add(labep);
        panep.add(spinnerp);
        JPanel paneq = new JPanel();
        paneq.add(labeq);
        paneq.add(spinnerq);
        JButton ok = new JButton("OK");
        JButton cancel = new JButton("Cancel");
        JPanel paneb = new JPanel();
        paneb.add(ok);
        paneb.add(cancel);
        fram.add(panep,BorderLayout.NORTH);
        fram.add(paneq,BorderLayout.CENTER);
        fram.add(paneb,BorderLayout.SOUTH);
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                int p = modelp.getNumber().intValue();
                int q = modelq.getNumber().intValue();
                link TLink = torusKnot(p,q);
                LinkData TData = new LinkData("T("+String.valueOf(p)+","+String.valueOf(q)+")", TLink,"", frame);
                theLinks.add(TData);
                listModel.addElement(TData.name.getText());
                cards.add(TData.panel, TData.name.getText());
                if (list.getModel().getSize()>1) 
                    list.setSelectedIndex(list.getModel().getSize()-1);
                fram.setVisible(false);
            }
        });
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        fram.setVisible(true);
    }
    
    LinkData enterGluedDiagram(JFrame frame, String gname, String gcomm, int[] crs, int[][] pts) {
        Knot = null;
        JDialog fram = new JDialog(new JFrame(), "Glued Diagram", true);
        fram.setBounds(frame.getX()+100, frame.getY()+140, 600, 320);
        fram.setResizable(false);
        fram.setLayout(new BorderLayout());
        JLabel name = new JLabel("  Name : ");
        JTextField tname = new JTextField(gname);
        tname.setPreferredSize(new Dimension(240,24));
        JPanel paname = new JPanel();
        paname.add(name);
        paname.add(tname);
        JLabel comments = new JLabel("Comments : ");
        JTextField tcomments = new JTextField(gcomm);
        tcomments.setPreferredSize(new Dimension(400,24));
        JPanel pacomm = new JPanel();
        pacomm.add(comments);
        pacomm.add(tcomments);
        int max = 50;
        if (crs.length > max) max = crs.length;
        JSlider slider = new JSlider(JSlider.HORIZONTAL,1,max,crs.length);
        slider.setMajorTickSpacing(8);
        slider.setMinorTickSpacing(2);
        slider.setPaintTicks(true);
        JLabel numbslides = new JLabel("  Number of crossings : "+slider.getValue());
        JPanel pane = new JPanel(new GridLayout(4,1));
        pane.add(paname);
        pane.add(pacomm);
        pane.add(slider);
        pane.add(numbslides);
        fram.add(pane,BorderLayout.NORTH);
        ArrayList<JCrossing> theCrossings = new ArrayList<JCrossing>();
        JPanel crossings = new JPanel(new GridLayout(1,0));
        for (int i = 0; i < crs.length; i++) {
            JCrossing crossing = new JCrossing(crs[i],pts[i]);
            theCrossings.add(crossing);
            crossings.add(crossing);
        }
        JScrollPane scroll = new JScrollPane(crossings);
        scroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        fram.add(scroll,BorderLayout.CENTER);
        slider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                int value = slider.getValue();
                int currentnumber = theCrossings.size();
                numbslides.setText("  Number of crossings : "+value);
                if (value > currentnumber) {
                    for (int u = currentnumber ; u < value; u++) {
                        int[] bore = {0,0,0,0};
                        JCrossing crossing = new JCrossing(0,bore);
                        theCrossings.add(crossing);
                        crossings.add(crossing);
                    }
                }
                if (value < currentnumber) {
                    for (int u = currentnumber - 1; u >= value; u--) {
                        theCrossings.remove(u);
                        crossings.remove(u);
                    }
                }
            }
        });
        JPanel buttons = new JPanel();
        JButton ok = new JButton("OK");
        JButton cancel = new JButton("Cancel");
        buttons.add(ok);
        buttons.add(cancel);
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
                String name = tname.getText();
                if (name.isEmpty()) name = "Link"+theLinks.size();
                String comments = tcomments.getText();
                int[] crosses = new int[theCrossings.size()];
                int[][] paths = new int[theCrossings.size()][4];
                for (int y = 0; y < theCrossings.size();y++) {
                    try {
                        crosses[y] = Integer.parseInt(theCrossings.get(y).crossing.getText());
                        paths[y][0] = Integer.parseInt(theCrossings.get(y).a.getText());
                        paths[y][1] = Integer.parseInt(theCrossings.get(y).b.getText());
                        paths[y][2] = Integer.parseInt(theCrossings.get(y).c.getText());
                        paths[y][3] = Integer.parseInt(theCrossings.get(y).d.getText());
                    }
                    catch (NumberFormatException e) {
                        
                    }
                }
                link Link = new link(crosses,paths);
                Knot = new LinkData(name,Link,comments,frame);
            }
        });
        
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        
        fram.add(buttons, BorderLayout.SOUTH);
        //fram.pack();
        fram.setVisible(true);
        return Knot;
    }
    
    class STCalc extends Thread {
        boolean runs = true;
        boolean finished = false;
        JDialog frame;
        JLabel LabTask;
        JLabel LabObj;
        JLabel LabLev;
        JLabel LabDis;
        JLabel LabDim;
        link Link;
        int theQ;
        int theN;
        boolean reduced;
        ArrayList<ArrayList<m4_object>> m4Objects;
        
        void cutDown(ArrayList<ArrayList<m4_object>> objects) {
            boolean found = false;
            while (!found) {
                ArrayList<m4_object> objs = objects.get(0);
                if (objs.isEmpty()) objects.remove(objs);
                else found = true;
            }
            while (found) {
                int i = objects.size();
                ArrayList<m4_object> objs = objects.get(i-1);
                if (objs.isEmpty()) objects.remove(i-1);
                else found = false;
            }
        }
        
        ArrayList<ArrayList<m4_object>> Changify(ArrayList<ArrayList<m4_object>> objects) {
            int lv = 0;
            boolean found = false;
            while (!found & lv < objects.size()-2) {
                int u = 0;
                while (!found & u < objects.get(lv).size()) {
                    m4_object obj = objects.get(lv).get(u);
                    if (!obj.moduli1b.isEmpty()) found = true;
                    else u++;
                }
                if (!found) lv++;
            }
            if (found) {
                ArrayList<Integer> goodOnes = new ArrayList<Integer>();
                for (int u = 0; u < objects.get(lv).size(); u++) {
                    if (objects.get(lv).get(u).moduli1b.isEmpty()) goodOnes.add(u);
                }
                while (goodOnes.size() < objects.get(lv).size()) {
                    int max = getMax(objects.get(lv),goodOnes);
                    int combo = maxCombo(objects.get(lv).get(max).moduli1b);
                    m4_object maxObj = objects.get(lv).get(max);
                    while (maxObj.moduli1b.size() > 1) {
                        int u = 0;
                        if (combo == 0) u = 1;
                        slideBelow(maxObj.moduli1b.get(combo),maxObj.moduli1b.get(u));
                        if (combo > u) combo--;
                    }
                    m4_object comObj = maxObj.moduli1b.get(0).topObj;
                    while (comObj.moduli1t.size()> 1) {
                        int u = 0;
                        if (comObj.moduli1t.get(u).botObj == maxObj) u = 1;
                        slideAbove(maxObj.moduli1b.get(0), comObj.moduli1t.get(u));
                    }
                    goodOnes.add(max);
                    for (int u = 0; u < objects.get(lv).size(); u++) {
                        if (objects.get(lv).get(u).moduli1b.isEmpty() & !goodOnes.contains(u)) goodOnes.add(u);
                    }
                }
            }
            return objects;
        }
        
        void slideAbove(m4mod1 modmax, m4mod1 mod) {
            boolean equal = false;
            if (!modmax.botObj.moduli0t.isEmpty()) {
                if (!mod.botObj.moduli0t.isEmpty()) 
                    equal = modmax.botObj.moduli0t.get(0).value == mod.botObj.moduli0t.get(0).value;
            }
            else {
                if (!modmax.botObj.moduli0b.isEmpty()) {
                    if (modmax.botObj.moduli0b.get(0).value == mod.botObj.moduli0b.get(0).value) equal = true;
                }
            }
            moveAlpha(modmax.botObj.moduli1b,mod.botObj);
            moveDelta(mod.botObj.moduli1t,modmax.botObj);
            if (equal) {
                m4_object omObj,onmObj;
                if (!modmax.botObj.moduli0b.isEmpty()) {
                    omObj = modmax.botObj.moduli0b.get(0).topObj;
                    onmObj = mod.botObj.moduli0b.get(0).topObj;
                }
                else {
                    omObj = modmax.botObj.moduli0t.get(0).botObj;
                    onmObj = mod.botObj.moduli0t.get(0).botObj;
                }
                moveDelta(onmObj.moduli1t,omObj);
                moveAlpha(omObj.moduli1b,onmObj);
            }
        }
        
        void slideBelow(m4mod1 modmax, m4mod1 mod) {
            boolean equal = false;
            if (!modmax.topObj.moduli0b.isEmpty()) {
                if (!mod.topObj.moduli0b.isEmpty()) {
                    if (modmax.topObj.moduli0b.get(0).value == mod.topObj.moduli0b.get(0).value) equal = true;
                }
            }
            else {
                if (!modmax.topObj.moduli0t.isEmpty()) {
                    if (modmax.topObj.moduli0t.get(0).value == mod.topObj.moduli0t.get(0).value) equal = true;
                }
            }
            moveDelta(modmax.topObj.moduli1t,mod.topObj);
            moveAlpha(mod.topObj.moduli1b,modmax.topObj);
            if (equal) {
                m4_object omObj,onmObj;
                if (!modmax.topObj.moduli0b.isEmpty()) {
                    omObj = modmax.topObj.moduli0b.get(0).topObj;
                    onmObj = mod.topObj.moduli0b.get(0).topObj;
                }
                else {
                    omObj = modmax.topObj.moduli0t.get(0).botObj;
                    onmObj = mod.topObj.moduli0t.get(0).botObj;
                }
                moveDelta(omObj.moduli1t,onmObj);
                moveAlpha(onmObj.moduli1b,omObj);
            }
        }
        
        void moveDelta(ArrayList<m4mod1> mods, m4_object obj) {
            for (m4mod1 md : mods) {
                boolean found = false;
                int u = 0;
                while (!found & u < obj.moduli1t.size()) {
                    if (obj.moduli1t.get(u).botObj == md.botObj) found = true;
                    else u++;
                }
                if (found) {
                    m4mod1 mrd = obj.moduli1t.get(u);
                    obj.moduli1t.remove(mrd);
                    mrd.botObj.moduli1b.remove(mrd);
                }
                else {
                    m4mod1 mnd = new m4mod1(obj,md.botObj, false);
                    obj.moduli1t.add(mnd);
                    md.botObj.moduli1b.add(mnd);
                }
            }
        }
        
        void moveAlpha(ArrayList<m4mod1> mods, m4_object obj) {
            for (m4mod1 md : mods) {
                boolean found = false;
                int u = 0;
                while (!found & u < obj.moduli1b.size()) {
                    if (obj.moduli1b.get(u).topObj == md.topObj) found = true;
                    else u++;
                }
                if (found) {
                    m4mod1 mrd = obj.moduli1b.get(u);
                    obj.moduli1b.remove(mrd);
                    mrd.topObj.moduli1t.remove(mrd);
                }
                else {
                    m4mod1 mnd = new m4mod1(md.topObj, obj, false);
                    md.topObj.moduli1t.add(mnd);
                    obj.moduli1b.add(mnd);
                }
            }
        }
        
        int maxCombo(ArrayList<m4mod1> mods) {
            int combo = 0;
            int chelp = 0;
            int cval = 0;
            if (!mods.get(0).topObj.moduli0b.isEmpty()) {
                chelp = 1;
                cval = mods.get(0).topObj.moduli0b.get(0).value;
            }
            if (!mods.get(0).topObj.moduli0t.isEmpty()) {
                chelp = -1;
                cval = mods.get(0).topObj.moduli0t.get(0).value;
            }
            int u = 1;
            while (u < mods.size()) {
                m4_object tObj = mods.get(u).topObj;
                int help = 0;
                int val = 0;
                if (!tObj.moduli0b.isEmpty()) {
                    help = 1;
                    val = tObj.moduli0b.get(0).value;
                }
                if (!tObj.moduli0t.isEmpty()) {
                    help = -1;
                    val = tObj.moduli0t.get(0).value;
                }
                boolean bigger = false;
                if (help > chelp) bigger = true;
                if (help == chelp & help * val < chelp * cval) bigger = true;
                if (bigger) {
                    cval = val;
                    chelp = help;
                    combo = u;
                }
                u++;
            }
            return combo;
        }
        
        int getMax(ArrayList<m4_object> objects, ArrayList<Integer> goodOnes) {
            int max = -1;
            int maxhelp = 0;
            int maxval = 0;
            int help = 0;
            int val = 0;
            for (int u = 0; u < objects.size(); u++) {
                if (!goodOnes.contains(u)) {
                    m4_object cObj = objects.get(u);
                    if (max < 0) {
                        max = u;
                        if (!cObj.moduli0t.isEmpty()) {
                            maxhelp = 1;
                            maxval = cObj.moduli0t.get(0).value;
                        }
                        if (!cObj.moduli0b.isEmpty()) {
                            maxval = cObj.moduli0b.get(0).value;
                            maxhelp = -1;
                        }
                    }
                    else {
                        help = 0;
                        if (!cObj.moduli0t.isEmpty()) {
                            help = 1;
                            val = cObj.moduli0t.get(0).value;
                        }
                        if (!cObj.moduli0b.isEmpty()) {
                            val = cObj.moduli0b.get(0).value;
                            help = -1;
                        }
                        boolean bigger = false;
                        if (help > maxhelp) bigger = true;
                        else if (help == maxhelp) {
                            if (help * val < maxhelp * maxval) bigger = true;
                        }
                        if (bigger) {
                            max = u;
                            maxhelp = help;
                            maxval = val;
                            //System.out.println("In max (bigger)"+max+" "+maxval+" "+maxhelp);
                        }
                    }
                }
            }
            return max;
        }
        
        ArrayList<ArrayList<sh_object>> getObjects(link Link, int qdeg, int n,boolean red) {
            int counter = 0;
            if (red) qdeg--;
            LabDis.setText("Diagrams : ");
            ArrayList<sh_object> objects = new ArrayList<sh_object>();
            ArrayList<diagram> diagrams = getDiagrams(Link);
            if (!runs) return null;
            LabTask.setText("Creating Objects");
            LabDis.setText("Objects in Level : ");
            LabDim.setText(" ");
            int qstart = 0;
            int qdig = qstart;
            int hstart = (-1)*nminus(Link);
            int hdeg = hstart;
            int hmin = 0;
            for (int y : Link.crossings) {
                if (y > 0) hmin = hmin + y;
                else hmin = hmin - y;
            }
            int hmax = (-1) * hmin;
            int nstart = (-1)* (n-1);
            for (diagram diag : diagrams) {
                if (!runs) break;
                hdeg = hstart;
                qstart = nplus(Link)-2*nminus(Link);
                for ( int i = 0; i < diag.position.length; i++) {
                    hdeg = hdeg + diag.position[i];
                    qstart = qstart + diag.position[i];
                    if (diag.position[i]>1) qstart = qstart + diag.position[i]-1;
                    if (diag.position[i]<-1) qstart = qstart + diag.position[i]+1;
                    if (Link.crossings[i] < 0) {
                        hdeg = hdeg - Link.crossings[i];
                        qstart = qstart - Link.crossings[i];
                    }
                }
                if (n > 2) qstart = qdegree(n,Link,diag.position);
                boolean full = false;
                ArrayList<Byte> runsym = new ArrayList<Byte>();
                for (int k = 0; k < diag.circles.size(); k++) runsym.add((byte) nstart);
                while (!full) {
                    qdig = qstart;
                    ArrayList<Byte> symbs = new ArrayList<Byte>();
                    for (byte e : runsym) {
                        symbs.add(e);
                        qdig = qdig + e;
                    }
                    boolean redok = true;
                    if (red) redok = oneWithMinus(diag,symbs);
                    if (qdig == qdeg & redok) {
                        sh_object newObject = new sh_object(diag,symbs);
                        objects.add(newObject);
                        counter++;
                        LabObj.setText(String.valueOf(counter));
                        if (hmin > hdeg) hmin = hdeg;
                        if (hmax < hdeg) hmax = hdeg;
                    }
                    boolean increaser = true;
                    int k = 0;
                    while (increaser & k < runsym.size()) {
                        byte t = runsym.get(k);
                        if (t != (byte) n-1) {
                            runsym.set(k, (byte) (t+2));
                            increaser = false;
                        }
                        else {
                            runsym.set(k, (byte) nstart);
                            k++;
                        }
                    }
                    if (increaser & k == runsym.size()) full = true;
                }
            }
            ArrayList<ArrayList<sh_object>> degObjects = new ArrayList<ArrayList<sh_object>>();
            for (int k = 0; k <= hmax - hmin; k++) {
                ArrayList<sh_object> objex = new ArrayList<sh_object>();
                degObjects.add(objex);
            }
            for (sh_object object : objects) {
                hdeg = hstart;
                for ( int i = 0; i < object.dig.position.length; i++) {
                    hdeg = hdeg + object.dig.position[i];
                    if (Link.crossings[i] < 0) hdeg = hdeg - Link.crossings[i];
                }
                degObjects.get(hdeg-hmin).add(object);
            }
            return degObjects;
        }
        
        boolean oneWithMinus(diagram dig, ArrayList<Byte> symbs) {
            boolean found = false;
            int u = 0;
            while (!found) {
                ArrayList<Integer> circle = dig.circles.get(u);
                if (circle.contains(1)) found = true;
                else u++;
            }
            if (symbs.get(u) == -1) return true;
            else return false;
        }
        
        ArrayList<diagram> getDiagrams(link Link) {
            LabTask.setText("Creating Diagrams");
            ArrayList<diagram> diagrams = startDiagrams(Link.crossings, Link.paths[0]);
            int r = Link.crossings.length;
            for (int i = 1; i < r; i++) {
                LabDim.setText(String.valueOf(diagrams.size()));
                diagrams = combineDiagrams(diagrams,Link.crossings[i], Link.paths[i],i);
                if (!runs) break;
            }
            return diagrams;
        }

        ArrayList<diagram> combineDiagrams(ArrayList<diagram> olddiags, int number, int[] ends, int i) {
            ArrayList<diagram> newdiags = new ArrayList<diagram>();
            for (diagram olddiag : olddiags) {
                if (!runs) return null;
                if (number >= 0) {
                    for (int j = 0; j <= number; j++) {
                        int[] pos = new int[olddiag.position.length];
                        for (int y = 0; y < olddiag.position.length; y++) pos[y]=olddiag.position[y];
                        diagram diag = new diagram(olddiag,pos,i,j);
                        int k = disjoint(diag,ends);
                        if (k == 0) {
                            disjointDiag(true,diag,ends,j);
                            newdiags.add(diag);
                        }
                        else {
                            if (j==0) concatenateDiag(true,diag,ends,k);
                            else concatenateDiag(false,diag,ends,k);
                            newdiags.add(diag);
                        }
                    }
                }
                else {
                    for (int j = number; j <= 0; j++) {
                        int[] pos = new int[olddiag.position.length];
                        for (int y = 0; y < olddiag.position.length; y++) pos[y]=olddiag.position[y];
                        diagram diag = new diagram(olddiag,pos,i,j);
                        int k = disjoint(diag,ends);
                        if (k == 0) {
                            disjointDiag(false,diag,ends,j);
                            newdiags.add(diag);
                        }
                        else {
                            if (j==0) concatenateDiag(true,diag,ends,k);
                            else concatenateDiag(false,diag,ends,k);
                            newdiags.add(diag);
                        }
                    }
                }
            }
            return newdiags;
        }
    }
    
    class SteenCalc extends STCalc {
        
        public SteenCalc(JDialog fram, JLabel task, JLabel objn, JLabel levn, JLabel dims,
                JLabel dimn, link Lnk, int q, int n, boolean red) {
            frame = fram;
            LabTask = task;
            LabObj = objn;
            LabLev = levn;
            LabDis = dims;
            LabDim = dimn;
            if (n > 2) Link = Mirror(Lnk);
            else Link = Lnk;
            theQ = q;
            theN = n;
            reduced = red;
        }
        
        public void run() {
            ArrayList<ArrayList<sh_object>> objects = getObjects(Link,theQ,theN,reduced);
            int counter = 0;
            if (runs) for (ArrayList<sh_object> obj : objects) counter = counter + obj.size();
            int nminus = (-1)*nminus(Link);
            if (runs) m4Objects = createm4justObjects(objects,nminus,Link,theN,counter);
            objects = null;
            if (runs) {
                LabTask.setText("Whitneyfication");
                LabObj.setText(String.valueOf(counter));
                int level = 0;
                for (ArrayList<m4_object> objs : m4Objects) {
                    LabLev.setText(String.valueOf(level));
                    level++;
                    m4whitneyfy(objs);
                }
            }
            if (runs) m4Normalise(m4Objects,counter);    
            if (!runs) {
                //System.out.println("Stopped");
                frame.setVisible(false);
                return;
            }
            else {
                //M4CatOutput(m4Objects);
                cutDown(m4Objects);
                Changify(m4Objects);
                //M4CatOutput(m4Objects);
                delay(200);
                frame.setVisible(false);
                finished = true;
            }
        }
        
        void m4Normalise(ArrayList<ArrayList<m4_object>> objects, int counter) {
            LabTask.setText("Mod 4 Normalization");
            LabDis.setText("Dimension : ");
            int ht = objects.size()-1;
            ArrayList<ArrayList<m4_object>> torsionObjects = new ArrayList<ArrayList<m4_object>>();
            for (int k = 0; k < objects.size(); k++) {
                ArrayList<m4_object> tlist = new ArrayList<m4_object>();
                torsionObjects.add(tlist);
            }
            //int count = counter;
            while (ht > 0) {
                LabLev.setText(String.valueOf(ht));
                if (!runs) return;
                m4mod0 mod = m4findOne(objects.get(ht));
                if (mod != null) {
                    m4_object tObj = mod.topObj; 
                    m4_object bObj = mod.botObj;
                    m4shiftTop(mod);
                    m4shiftBot(mod);
                    for (m4mod1 md : tObj.moduli1t) md.botObj.moduli1b.remove(md);
                    for (m4mod1 md : tObj.moduli1b) md.topObj.moduli1t.remove(md);
                    for (m4mod1 md : bObj.moduli1b) md.topObj.moduli1t.remove(md);
                    for (m4mod1 md : bObj.moduli1t) md.botObj.moduli1b.remove(md);
                    objects.get(ht).remove(tObj);
                    objects.get(ht-1).remove(bObj);
                    counter = counter - 2;
                    LabObj.setText(String.valueOf(counter));
                    LabDim.setText(String.valueOf(objects.get(ht).size()+" x "+objects.get(ht-1).size()));
                }
                else {
                    mod = m4findTwo(objects.get(ht));
                    if (mod != null) {
                        m4_object tObj = mod.topObj; 
                        m4_object bObj = mod.botObj;
                        m4shiftTop(mod);
                        m4shiftBot(mod);
                        objects.get(ht).remove(tObj);
                        //objects.get(ht-1).remove(bObj);
                        torsionObjects.get(ht).add(tObj);
                        //torsionObjects.get(ht-1).add(bObj);
                    }
                    else ht--;
                }
            }
            for (int e = 0; e < objects.size(); e++) {
                for (m4_object obj : torsionObjects.get(e)) objects.get(e).add(obj);
            }
        }
        
        void m4whitneyfy(ArrayList<m4_object> objects) {
            for (m4_object obj : objects) {
                int i = 0;
                while (i < obj.moduli0t.size()) {
                    m4mod0 mod = obj.moduli0t.get(i);
                    if (mod.value == 0) {
                        mod.topObj.moduli0t.remove(mod);
                        mod.botObj.moduli0b.remove(mod);
                    }
                    else i++;
                }
                i = 0;
                while (i < obj.moduli1t.size()) {
                    m4mod1 mod = obj.moduli1t.get(i);
                    if (mod.framedCircle & mod.edges.isEmpty()) {
                        mod.topObj.moduli1t.remove(mod);
                        mod.botObj.moduli1b.remove(mod);
                    }
                    else i++;
                }
            }
        }
        
        ArrayList<ArrayList<m4_object>> createm4justObjects(ArrayList<ArrayList<sh_object>> shobjs, int nminus, link Link, 
                int n, int counter) {
            ArrayList<ArrayList<m4_object>> objects = new ArrayList<ArrayList<m4_object>>();
            for (int e = 0; e < shobjs.size(); e++) {
                ArrayList<sh_object> objs = shobjs.get(e);
                ArrayList<m4_object> m4objs = new ArrayList<m4_object>();
                for (int f = 0; f < objs.size(); f++) {
                    int h = nminus+homDegree(objs.get(f).dig.position,Link.crossings);
                    m4_object m4obj = new m4_object(f,h);
                    m4objs.add(m4obj);
                }
                objects.add(m4objs);
            }
            if (runs) createm4Mod0(objects,shobjs,Link,n,counter);
            if (runs) createm4Mod1(objects,shobjs,Link,n,counter);
            return objects;
        }
        
        void createm4Mod1(ArrayList<ArrayList<m4_object>> objects,ArrayList<ArrayList<sh_object>> shobjects, link Link, 
                int n, int count) {
            LabTask.setText("Creating 1-dim moduli spaces");
            int counter = 0;
            String cound = String.valueOf(count);
            int ht = objects.size()-1;
            while (ht > 1) {
                LabLev.setText(String.valueOf(ht));
                String counj = String.valueOf(objects.get(ht).size());
                int couni = 0;
                //if (ht <=1) return;
                for (m4_object object : objects.get(ht)) {
                    counter++;
                    couni++;
                    if (!runs) return;
                    LabObj.setText(String.valueOf(counter)+" of "+cound);
                    LabDim.setText(String.valueOf(couni)+" of "+counj);
                    ArrayList<pairm4Points> pairs = new ArrayList<pairm4Points>();
                    for (m4mod0 mod : object.moduli0t) {
                        for (m4mod0 md : mod.botObj.moduli0t) {
                            pairm4Points pair = new pairm4Points(md,mod);
                            pairs.add(pair);
                        }
                    }
                    while (!pairs.isEmpty()) {
                        pairm4Points first = pairs.get(0);
                        sh_object fftop = shobjects.get(ht-1).get(first.first.topObj.number);
                        sh_object ffbot = shobjects.get(ht-2).get(first.first.botObj.number);
                        sh_object fstop = shobjects.get(ht).get(first.second.topObj.number);
                        sh_object fsbot = shobjects.get(ht-1).get(first.second.botObj.number);
                        int firstChange = diffPosition(fftop.dig,ffbot.dig);
                        int seconChange = diffPosition(fstop.dig,fsbot.dig);
                        ArrayList<pairm4Points> similar = new ArrayList<pairm4Points>();
                        for (pairm4Points next : pairs) {
                            sh_object nft = shobjects.get(ht-1).get(next.first.topObj.number);
                            sh_object nfb = shobjects.get(ht-2).get(next.first.botObj.number);
                            sh_object nst = shobjects.get(ht).get(next.second.topObj.number);
                            sh_object nsb = shobjects.get(ht-1).get(next.second.botObj.number);
                            int nextfChange = diffPosition(nft.dig,nfb.dig);
                            int nextsChange = diffPosition(nst.dig,nsb.dig);
                            if ((nextfChange == firstChange & nextsChange == seconChange) | (nextfChange == seconChange & nextsChange == firstChange)) {
                                if (next.first.botObj == first.first.botObj) similar.add(next);
                            }
                        }
                        for (pairm4Points pr : similar) pairs.remove(pr);
                        if (firstChange == seconChange) {
                            diagram diag = ffbot.dig;
                            if (similar.size()<=2) { // very simple
                                boolean frame = true;
                                int factor = diag.position[firstChange];
                                if (diag.position[firstChange] < 0) factor++;
                                int help = 0;
                                for (int y = 0; y < firstChange; y++) help = help + diag.position[y];
                                help = help * factor;
                                if (help % 2 != 0) frame = !frame;
                                if (first.first.value == 0) {
                                    if (first.second.value%2 == 0) frame = false;
                                    m4mod1 modOne = new m4mod1(first.second.topObj, first.first.botObj, !frame);
                                    first.second.topObj.moduli1t.add(modOne);
                                    first.first.botObj.moduli1b.add(modOne);
                                }
                                if (first.second.value == 0) {
                                    int v = first.first.value;
                                    if (v % 2 != 0 & v < 0) frame = !frame;
                                    m4mod1 modOne = new m4mod1(first.second.topObj, first.first.botObj, !frame);
                                    first.second.topObj.moduli1t.add(modOne);
                                    first.first.botObj.moduli1b.add(modOne);
                                }
                                if ((Math.abs(first.first.value) == 1) & (Math.abs(first.second.value) == 1)) {
                                    m4mod1 modOne = new m4mod1(first.second.topObj, first.first.botObj, frame);
                                    first.second.topObj.moduli1t.add(modOne);
                                    first.first.botObj.moduli1b.add(modOne);
                                    boolean bl = first.first.value == 1;
                                    boolean tl = first.second.value == 1;
                                    boolean br = similar.get(1).first.value == 1;
                                    boolean tr = similar.get(1).second.value == 1;
                                    m4edge eddy = new m4edge(first.first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                    modOne.edges.add(eddy);
                                }
                            }
                            else {
                                if (similar.size()>2) System.out.println("Possible!"); // I don't expect this to happen, if it doesn't, can be removed.
                            }
                        }
                        else {
                            diagram diag = ffbot.dig;
                            int firstval = first.first.value;
                            int seconval = first.second.value;
                            boolean ordnung = true;
                            if (firstChange > seconChange) {
                                int h = firstChange;
                                firstChange = seconChange;
                                seconChange = h;
                                h = firstval;
                                firstval = seconval;
                                seconval = h;
                                ordnung = false;
                            }
                            int a = diag.position[firstChange];
                            int b = diag.position[seconChange];
                            if (similar.size()<=2) { // very simple
                                int firs = similar.get(0).first.value;
                                int secs = similar.get(0).second.value;
                                if ((Math.abs(firs) == 1) & (Math.abs(secs) == 1)) { // one edge between them
                                    boolean frame = m4framing(first.first,first.second,fftop.dig,ffbot.dig,fstop.dig,fsbot.dig);
                                    m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj,frame);
                                    first.second.topObj.moduli1t.add(modone);
                                    first.first.botObj.moduli1b.add(modone);
                                    boolean bl = similar.get(0).first.value == 1;
                                    boolean tl = similar.get(0).second.value == 1;
                                    boolean br = similar.get(1).first.value == 1;
                                    boolean tr = similar.get(1).second.value == 1;
                                    m4edge ed = new m4edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                    modone.edges.add(ed);
                                }
                                if (((Math.abs(firs) == 1) & secs == 0) |
                                        (firs == 0 & (Math.abs(secs) == 1))) {
                                    if (Math.abs(firs) != 1) firs = similar.get(1).first.value;
                                    boolean frame = m4framingZero(first.first,first.second,fftop.dig,ffbot.dig,fstop.dig,fsbot.dig);
                                    m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj,frame);
                                    first.second.topObj.moduli1t.add(modone);
                                    first.first.botObj.moduli1b.add(modone);
                                }
                                if (((Math.abs(firs) == 1) & (Math.abs(secs) == n)) |
                                        ((Math.abs(firs) == n) & (Math.abs(secs) == 1))) {
                                    boolean frame = m4framing(first.first,first.second,fftop.dig,ffbot.dig,fstop.dig,fsbot.dig);
                                    if (n % 2 == 0) frame = true;
                                    boolean eps = true;
                                    if (Math.abs(firs) == 1) {
                                        if ((firs == 1) == ordnung) eps = false;
                                    }
                                    else {
                                        if ((similar.get(1).first.value != 1)!=ordnung) eps = false;
                                    }
                                    if ((n/2)%2 == 0) eps = true;
                                    if (eps == frame) frame = true;
                                    else frame = false;
                                    m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj, frame);
                                    first.second.topObj.moduli1t.add(modone);
                                    first.first.botObj.moduli1b.add(modone);
                                    if (n % 2 != 0) {
                                        boolean bl = similar.get(0).first.value > 0;
                                        boolean tl = similar.get(0).second.value > 0;
                                        boolean br = similar.get(1).first.value > 0;
                                        boolean tr = similar.get(1).second.value > 0;
                                        m4edge ed = new m4edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                        modone.edges.add(ed);
                                    }
                                }
                                //if (firs.abs().equals(BigInteger.ZERO)& secs.abs().equals(BigInteger.ZERO)) {
                                    //no mod1 needed
                                //}
                                if (((firs == 0) & (Math.abs(secs) == n)) |
                                        ((Math.abs(firs) == n) & (secs == 0))) {
                                    if (n%2 != 0) {
                                        m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj, true);
                                        first.second.topObj.moduli1t.add(modone);
                                        first.first.botObj.moduli1b.add(modone);
                                        int y = 0;
                                        if (Math.abs(firs) == n) y = firs;
                                        else y = secs;
                                        boolean frame = m4framingZero(first.first,first.second,fftop.dig,ffbot.dig,fstop.dig,fsbot.dig);
                                        if (y > 0) frame = !frame;
                                        modone.framedCircle = frame;
                                    }// n % 2 == 0 does not require a moduli space
                                }
                                if ((Math.abs(firs) == n) & (Math.abs(secs) == n)) {
                                    if (n%2 == 0) {
                                        if (n%4 != 0) {
                                            m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj,false);
                                            first.second.topObj.moduli1t.add(modone);
                                            first.first.botObj.moduli1b.add(modone);
                                        }
                                    }
                                    else {
                                        m4mod1 modone = new m4mod1(first.second.topObj,first.first.botObj,true);
                                        first.second.topObj.moduli1t.add(modone);
                                        first.first.botObj.moduli1b.add(modone);
                                        boolean bl = similar.get(0).first.value > 0;
                                        boolean tl = similar.get(0).second.value > 0;
                                        boolean br = similar.get(1).first.value > 0;
                                        boolean tr = similar.get(1).second.value > 0;
                                        m4edge ed = new m4edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                        modone.edges.add(ed);
                                        boolean frame = m4framing(first.first,first.second,fftop.dig,ffbot.dig,fstop.dig,fsbot.dig);
                                        if ( (n-1)% 4 != 0) {
                                            frame = !frame;
                                            boolean extra = (similar.get(0).first.value + similar.get(1).first.value == 0);
                                            if (!extra) frame = !frame;
                                        }
                                        modone.framedCircle = frame;
                                    }
                                }
                            }
                            else {
                                m4mod1 modOne = new m4mod1(first.second.topObj,first.first.botObj,true);
                                first.second.topObj.moduli1t.add(modOne);
                                first.first.botObj.moduli1b.add(modOne);
                                if (((a == 0) | (a == -1)) & ((b == 0) | (b == -1))) { // it's a ladybug!!!
                                    if (n%2 !=0) {
                                        int as = 0;
                                        for (int e = 0; e < firstChange; e++) {
                                            as = as + ffbot.dig.position[e];
                                        }
                                        if (as % 2 != 0) {
                                            as = 0;
                                            for (int e = firstChange; e < seconChange; e++) {
                                                as = as + ffbot.dig.position[e];
                                            }
                                            if (as % 2 != 0) modOne.framedCircle = false;
                                        }
                                    }
                                    int rightguy;
                                    if (a == 0) rightguy = Link.paths[firstChange][0];
                                    else rightguy = Link.paths[firstChange][1];
                                    while (!similar.isEmpty()) {
                                        pairm4Points fist = similar.get(0);
                                        m4_object midObj = fist.first.topObj;
                                        sh_object mclObj = shobjects.get(ht-1).get(midObj.number);
                                        int circnum = 0;
                                        boolean found = false;
                                        while (!found) {
                                            ArrayList<Integer> circ = mclObj.dig.circles.get(circnum);
                                            if (circ.contains(rightguy)) found = true;
                                            else circnum++;
                                        }
                                        byte thesign = mclObj.symbols.get(circnum);
                                        found = false;
                                        int v = 1;
                                        while (!found) {
                                            pairm4Points seco = similar.get(v);
                                            m4_object mitObj = seco.first.topObj;
                                            sh_object mctObj = shobjects.get(ht-1).get(mitObj.number);
                                            if (mctObj.dig.position[firstChange] != mclObj.dig.position[firstChange]) {
                                                int cirxnum = 0;
                                                boolean fnd = false;
                                                while (!fnd) {
                                                    ArrayList<Integer> circ = mctObj.dig.circles.get(cirxnum);
                                                    if (circ.contains(rightguy)) fnd = true;
                                                    else cirxnum++;
                                                }
                                                if (mctObj.symbols.get(cirxnum) == thesign) found = true;
                                                else v++;
                                            }
                                            else v++;
                                        }
                                        boolean bl = fist.first.value == 1;
                                        boolean tl = fist.second.value == 1;
                                        boolean br = similar.get(v).first.value == 1;
                                        boolean tr = similar.get(v).second.value == 1;
                                        m4edge ed = new m4edge(fist.first.topObj,similar.get(v).first.topObj,tl,bl,tr,br);
                                        modOne.edges.add(ed);
                                        similar.remove(v);
                                        similar.remove(0);
                                    }
                                }
                                else {
                                    int unpleasant = 0;
                                    int i = 0;
                                    if ((a == 0) | (a == -1) | (b == 0) | (b == -1)) { // check for an unpleasant surgery involved
                                        while (i < similar.size() & unpleasant == 0) {
                                            m4mod0 modf = similar.get(i).first;
                                            m4mod0 mods = similar.get(i).second;
                                            if (Math.abs(modf.value) != 1) unpleasant = 1;
                                            if (Math.abs(mods.value) != 1) unpleasant = 2;
                                            if (unpleasant == 0) i++;
                                        }
                                    }
                                    if (unpleasant > 0) {
                                        boolean extrasign = true;
                                        boolean firstspec = true;
                                        m4mod0 p1 = similar.get(i).first;
                                        m4mod0 p2 = similar.get(i).second;
                                        if (Math.abs(p1.value) == 1) {
                                            extrasign = p1.value > 0;
                                            firstspec = false;
                                        }
                                        int cas;
                                        if (firstspec) cas = p1.value;
                                        else cas = p2.value;
                                        if (cas == 0) {
                                            similar.remove(i);
                                            pairm4Points pts1 = similar.get(0);
                                            pairm4Points pts2 = similar.get(1);
                                            m4edge ed;
                                            boolean bl = pts1.first.value == 1;
                                            boolean tl = pts1.second.value == 1;
                                            boolean br = pts2.first.value == 1;
                                            boolean tr = pts2.second.value == 1;
                                            ed = new m4edge(pts1.first.topObj,pts2.first.topObj,tl,bl,tr,br);
                                            sh_object pts1ft = shobjects.get(ht-1).get(pts1.first.topObj.number);
                                            sh_object pts1fb = shobjects.get(ht-2).get(pts1.first.botObj.number);
                                            sh_object pts1st = shobjects.get(ht).get(pts1.second.topObj.number);
                                            sh_object pts1sb = shobjects.get(ht-1).get(pts1.second.botObj.number);
                                            sh_object pts2ft = shobjects.get(ht-1).get(pts2.first.topObj.number);
                                            sh_object pts2fb = shobjects.get(ht-2).get(pts2.first.botObj.number);
                                            sh_object pts2st = shobjects.get(ht).get(pts2.second.topObj.number);
                                            sh_object pts2sb = shobjects.get(ht-1).get(pts2.second.botObj.number);
                                            boolean fram1 = m4framing(pts1.first,pts1.second,pts1ft.dig,pts1fb.dig,pts1st.dig,pts1sb.dig);
                                            boolean fram2 = m4framing(pts2.first,pts2.second,pts2ft.dig,pts2fb.dig,pts2st.dig,pts2sb.dig);
                                            if (firstspec) {
                                                if (fram1!=fram2) modOne.framedCircle = false;
                                                modOne.edges.add(ed);
                                            }
                                            else {
                                                modOne.framedCircle = extrasign;
                                                modOne.edges.add(ed);
                                                if (fram1!=fram2) modOne.framedCircle = !modOne.framedCircle;
                                            }
                                        }
                                        else {
                                            int c = -1;
                                            sh_object p1clt = shobjects.get(ht-1).get(p1.topObj.number);
                                            sh_object p1clb = shobjects.get(ht-2).get(p1.botObj.number);
                                            sh_object p2clt = shobjects.get(ht).get(p2.topObj.number);
                                            sh_object p2clb = shobjects.get(ht-1).get(p2.botObj.number);
                                            if (firstspec) c = diffPosition(p1clt.dig, p1clb.dig);
                                            else c = diffPosition(p2clt.dig,p2clb.dig);
                                            int v = Link.paths[c][0];
                                            if (n%2!=0) {
                                                int j = 0;
                                                boolean found = false;
                                                while (!found) {
                                                    m4mod0 mod;
                                                    sh_object bcl;
                                                    sh_object tcl;
                                                    if (j == i) j++;
                                                    if (firstspec) {
                                                        mod = similar.get(j).second;
                                                        bcl = shobjects.get(ht-1).get(mod.botObj.number);
                                                        tcl = shobjects.get(ht).get(mod.topObj.number);
                                                    }
                                                    else {
                                                        mod = similar.get(j).first;
                                                        bcl = shobjects.get(ht-2).get(mod.botObj.number);
                                                        tcl = shobjects.get(ht-1).get(mod.topObj.number);
                                                    }
                                                    if (dotIndex(bcl,tcl,v) == n-1) found = true;
                                                    else j++;
                                                }
                                                boolean bl = similar.get(i).first.value > 0;
                                                boolean tl = similar.get(i).second.value > 0;
                                                boolean br = similar.get(j).first.value > 0;
                                                boolean tr = similar.get(j).second.value > 0;
                                                m4edge ed = new m4edge(similar.get(i).first.topObj,similar.get(j).first.topObj,tl,bl,tr,br);
                                                sh_object sjft = shobjects.get(ht-1).get(similar.get(j).first.topObj.number);
                                                sh_object sjfb = shobjects.get(ht-2).get(similar.get(j).first.botObj.number);
                                                sh_object sjst = shobjects.get(ht).get(similar.get(j).second.topObj.number);
                                                sh_object sjsb = shobjects.get(ht-1).get(similar.get(j).second.botObj.number);
                                                modOne.framedCircle = m4framing(similar.get(j).first,similar.get(j).second,sjft.dig,sjfb.dig,sjst.dig,sjsb.dig);
                                                modOne.edges.add(ed);
                                                similar.remove(j);
                                                if (j < i) i--;
                                            }
                                            similar.remove(i);
                                            while (!similar.isEmpty()) {
                                                pairm4Points pts1 = similar.get(0);
                                                int j = 1;
                                                boolean found = false;
                                                m4mod0 mod;
                                                sh_object bcl;
                                                sh_object tcl;
                                                if (firstspec) {
                                                    mod = pts1.second;
                                                    bcl = shobjects.get(ht-1).get(mod.botObj.number);
                                                    tcl = shobjects.get(ht).get(mod.topObj.number);
                                                }
                                                else {
                                                    mod = pts1.first;
                                                    bcl = shobjects.get(ht-2).get(mod.botObj.number);
                                                    tcl = shobjects.get(ht-1).get(mod.topObj.number);
                                                }
                                                int k = dotIndex(bcl,tcl,v);
                                                if ((k%2)!=0) k--;
                                                else k++;
                                                while (!found) {
                                                    if (firstspec) {
                                                        mod = similar.get(j).second;
                                                        bcl = shobjects.get(ht-1).get(mod.botObj.number);
                                                        tcl = shobjects.get(ht).get(mod.topObj.number);
                                                    }
                                                    else {
                                                        mod = similar.get(j).first;
                                                        bcl = shobjects.get(ht-2).get(mod.botObj.number);
                                                        tcl = shobjects.get(ht-1).get(mod.topObj.number);
                                                    }
                                                    if (dotIndex(bcl,tcl,v) == k) found = true;
                                                    else j++;
                                                }
                                                boolean bl = similar.get(0).first.value > 0;
                                                boolean tl = similar.get(0).second.value > 0;
                                                boolean br = similar.get(j).first.value > 0;
                                                boolean tr = similar.get(j).second.value > 0;
                                                m4edge ed = new m4edge(similar.get(0).first.topObj,similar.get(j).first.topObj,tl,bl,tr,br);
                                                if (k%2 == 0) modOne.framedCircle = !modOne.framedCircle;
                                                if (!firstspec & !extrasign) modOne.framedCircle = !modOne.framedCircle;
                                                modOne.edges.add(ed);
                                                similar.remove(j);
                                                similar.remove(0);
                                            }
                                        }
                                    }
                                    else { // it's a combination of dottings,
                                        if (((a > 0) | (a < -1)) & ((b > 0) | (b < -1))) { // just dottings
                                            while (!similar.isEmpty()) {
                                                m4mod0 p1 = similar.get(0).first;
                                                m4mod0 p2 = similar.get(0).second;
                                                sh_object p1b = shobjects.get(ht-2).get(p1.botObj.number);
                                                sh_object p1t = shobjects.get(ht-1).get(p1.topObj.number);
                                                sh_object p2b = shobjects.get(ht-1).get(p2.botObj.number);
                                                sh_object p2t = shobjects.get(ht).get(p2.topObj.number);
                                                int x = diffPosi(p1b.dig,p1t.dig);
                                                int y = diffPosi(p2b.dig,p2t.dig);
                                                int c1 = Link.paths[x][0];
                                                int c2 = Link.paths[y][0];
                                                int k1 = dotIndex(p1b,p1t,c1);
                                                int k2 = dotIndex(p2b,p2t,c2);
                                                boolean found = false;
                                                int e = 1;
                                                while (!found) {
                                                    m4mod0 p3 = similar.get(e).first;
                                                    m4mod0 p4 = similar.get(e).second;
                                                    sh_object p3b = shobjects.get(ht-2).get(p3.botObj.number);
                                                    sh_object p3t = shobjects.get(ht-1).get(p3.topObj.number);
                                                    sh_object p4b = shobjects.get(ht-1).get(p4.botObj.number);
                                                    sh_object p4t = shobjects.get(ht).get(p4.topObj.number);
                                                    if (diffPosi(p3b.dig,p3t.dig) == y) {
                                                        if (dotIndex(p4b,p4t,c1) == k1 & dotIndex(p3b,p3t,c2) == k2) found = true;
                                                    }
                                                    if (!found) e++;
                                                }
                                                m4_object corner = similar.get(e).first.topObj;
                                                boolean bl = p1.value == 1;
                                                boolean tl = p2.value == 1;
                                                boolean br = similar.get(e).first.value == 1;
                                                boolean tr = similar.get(e).second.value == 1;
                                                similar.remove(e);
                                                similar.remove(0);
                                                m4edge ed = new m4edge(p1.topObj,corner,tl,bl,tr,br);
                                                modOne.edges.add(ed);
                                                boolean frame = m4framing(p1,p2,p1t.dig,p1b.dig,p2t.dig,p2b.dig);
                                                if (!frame) modOne.framedCircle = !modOne.framedCircle;
                                            }
                                        }
                                        else {
                                            if (similar.size()!= 3) System.out.println("Possubile "+similar.size());
                                        }
                                    } // */
                                }
                            }
                        }
                    }
                }
                ht--;
            }
        }
        
        void createm4Mod0(ArrayList<ArrayList<m4_object>> objects,ArrayList<ArrayList<sh_object>> shobjects, link Link, int n,
                int count) {
            LabTask.setText("Creating 0-dim moduli spaces");
            int counter = 0;
            String cound = String.valueOf(count);
            int k = objects.size();
            for (int j = k-1; j > 0; j--) {
                LabLev.setText(String.valueOf(j));
                String counj = String.valueOf(objects.get(j).size());
                for (int i = 0; i < objects.get(j).size(); i++) {
                    counter++;
                    LabObj.setText(String.valueOf(counter)+" of "+cound);
                    LabDim.setText(String.valueOf(i)+" of "+counj);
                    m4_object topObj = objects.get(j).get(i);
                    sh_object tclObj = shobjects.get(j).get(i);
                    for (int l = 0; l < objects.get(j-1).size(); l++) {
                        if (!runs) return;
                        m4_object botObj = objects.get(j-1).get(l);
                        sh_object bclObj = shobjects.get(j-1).get(l);
                        int r = diffPosition(tclObj.dig,bclObj.dig);
                        if (r > -1) {
                            if ((bclObj.dig.position[r] == 0) | (bclObj.dig.position[r] == -1)) { // possible surgery
                                boolean surgery = aSurgery(tclObj,bclObj,Link.paths[r]);
                                if (surgery) {
                                    boolean sign = getSign(bclObj.dig.position,r);
                                    int val = 1;
                                    if (!sign) val = -val;
                                    m4mod0 point = new m4mod0(topObj,botObj,val);
                                    topObj.moduli0t.add(point);
                                    botObj.moduli0b.add(point);
                                }
                            }
                            else { // possible dotting of one circle
                                int dotting = aDotting(tclObj,bclObj,Link.paths[r],n);
                                if (dotting == 1) { // produce n points or two points with different sign
                                    boolean sign = getSign(bclObj.dig.position,r);
                                    int checker = bclObj.dig.position[r];
                                    if (checker < 0) checker++;
                                    if (checker%2 ==0) { // points of same sign
                                        int val = 1;
                                        if (!sign) val = -val;
                                        val = val*n;
                                        m4mod0 point = new m4mod0(topObj,botObj,val);
                                        topObj.moduli0t.add(point);
                                        botObj.moduli0b.add(point);
                                    }
                                    else {
                                        int val = 1;
                                        m4mod0 point = new m4mod0(topObj,botObj,val);
                                        point.value = 0;
                                        topObj.moduli0t.add(point);
                                        botObj.moduli0b.add(point);
                                    }
                                }
                                if (dotting >= 2) {
                                    boolean sign = getSign(bclObj.dig.position,r);
                                    int checker = bclObj.dig.position[r];
                                    if (bclObj.dig.position[r]<0) checker++;
                                    if ((checker % 2)!=0) { // need to be careful with signs
                                        if (dotting == 3) sign = !sign; // case when right circle gets dotted and sign should change
                                    }
                                    int val = 1;
                                    if (!sign) val = -val;
                                    m4mod0 point = new m4mod0(topObj,botObj,val);
                                    topObj.moduli0t.add(point);
                                    botObj.moduli0b.add(point);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    class StTyCalc extends STCalc {
                
        public StTyCalc(JDialog fram, JLabel task, JLabel objn, JLabel levn, JLabel dims,
                JLabel dimn, link Lnk, int q, int n, boolean red) {
            frame = fram;
            LabTask = task;
            LabObj = objn;
            LabLev = levn;
            LabDis = dims;
            LabDim = dimn;
            if (n > 2) Link = Mirror(Lnk);
            else Link = Lnk;
            theQ = q;
            theN = n;
            reduced = red;
        }
        
        public void run() {
            ArrayList<ArrayList<sh_object>> objects = getObjects(Link,theQ,theN,reduced);
            int counter = 0;
            if (runs) for (ArrayList<sh_object> obj : objects) counter = counter + obj.size();
            if (runs) createMod0(objects,Link,theN,counter);
            if (runs) createMod1(objects,Link,theN,counter);
            if (runs) {
                LabTask.setText("Whitneyfication");
                LabObj.setText(String.valueOf(counter));
                int level = 0;
                for (ArrayList<sh_object> objs : objects) {
                    LabLev.setText(String.valueOf(level));
                    level++;
                    whitneyfy(objs);
                }
            }
            if (runs) SmithNormalise(objects,counter);    
            if (!runs) {
                //System.out.println("Stopped");
                frame.setVisible(false);
                return;
            }
            else {
                int nminus = (-1)*nminus(Link);
                m4Objects = createm4Objects(objects,nminus,Link);
                cutDown(m4Objects);
                //M4CatOutput(m4Objects);
                Changify(m4Objects);
                //System.out.println(m4Objects.size());
                //M4CatOutput(m4Objects);
                delay(200);
                frame.setVisible(false);
                finished = true;
            }
        }
        
        void SmithNormalise(ArrayList<ArrayList<sh_object>> objects, int counter) {
            LabTask.setText("Smith Normalization");
            LabDis.setText("Dimension : ");
            int ht = objects.size()-1;
            ArrayList<ArrayList<sh_object>> torsionObjects = new ArrayList<ArrayList<sh_object>>();
            for (int k = 0; k < objects.size(); k++) {
                ArrayList<sh_object> tlist = new ArrayList<sh_object>();
                torsionObjects.add(tlist);
            }
            int count = counter;
            while (ht > 0) {
                LabLev.setText(String.valueOf(ht));
                if (!runs) return;
                mod0 mod = shFindOne(objects.get(ht));
                if (mod != null) {
                    sh_object tObj = mod.topObj; 
                    sh_object bObj = mod.botObj;
                    shiftTop(mod);
                    shiftBot(mod);
                    for (mod1 md : tObj.moduli1t) md.botObj.moduli1b.remove(md);
                    for (mod1 md : tObj.moduli1b) md.topObj.moduli1t.remove(md);
                    for (mod1 md : bObj.moduli1b) md.topObj.moduli1t.remove(md);
                    for (mod1 md : bObj.moduli1t) md.botObj.moduli1b.remove(md);
                    objects.get(ht).remove(tObj);
                    objects.get(ht-1).remove(bObj);
                    counter = counter - 2;
                    LabObj.setText(String.valueOf(counter));
                    LabDim.setText(String.valueOf(objects.get(ht).size()+" x "+objects.get(ht-1).size()));
                }
                else {
                    mod = shFindSmallest(objects.get(ht)); 
                    if (mod != null) {
                        String length = String.valueOf(mod.totalValue.abs().toString().length());
                        LabDim.setText(String.valueOf(objects.get(ht).size()+" x "+objects.get(ht-1).size())+", "+length);
                        sh_object tObj = mod.topObj; 
                        sh_object bObj = mod.botObj;
                        slideTop(mod);
                        if (mod.botObj.moduli0b.size()==1) {
                            slideBot(mod);
                            if (mod.topObj.moduli0t.size() == 1) {
                                mod.totalValue = mod.totalValue.abs(); // just make sure it is positive
                                objects.get(ht).remove(tObj);
                                objects.get(ht-1).remove(bObj);
                                torsionObjects.get(ht).add(tObj);
                                torsionObjects.get(ht-1).add(bObj);
                            }
                        }
                    }
                    else ht--;
                }
            }
            for (int e = 0; e < objects.size(); e++) {
                int counts = 0;
                for (sh_object obj : torsionObjects.get(e)) {
                    counts = counts + 10;
                    mod0 mod = obj.moduli0b.get(0);
                    sh_object tbj = mod.topObj;
                    int value = mod.totalValue.intValue();
                    if (value%2 != 0) {
                        while (!obj.moduli1b.isEmpty()) {
                            mod1 md = obj.moduli1b.get(0);
                            obj.moduli1b.remove(md);
                            md.topObj.moduli1t.remove(md);
                        }
                        while (!obj.moduli1t.isEmpty()) {
                            mod1 md = obj.moduli1t.get(0);
                            obj.moduli1t.remove(md);
                            md.botObj.moduli1b.remove(md);
                        }
                        while (!tbj.moduli1b.isEmpty()) {
                            mod1 md = tbj.moduli1b.get(0);
                            tbj.moduli1b.remove(md);
                            md.topObj.moduli1t.remove(md);
                        }
                        while (!tbj.moduli1t.isEmpty()) {
                            mod1 md = tbj.moduli1t.get(0);
                            tbj.moduli1t.remove(md);
                            md.botObj.moduli1b.remove(md);
                        }
                    }
                    ArrayList<Integer> factors = factorize(value);
                    mod.totalValue = BigInteger.valueOf(factors.get(0));
                    objects.get(e).add(obj);
                    objects.get(e+1).add(tbj);
                    torsionObjects.get(e+1).remove(tbj);
                    for (int u = 1; u < factors.size(); u++) {
                        sh_object bObj = new sh_object(obj.dig,null);
                        sh_object tObj = new sh_object(tbj.dig,null);
                        mod0 md = new mod0(tObj,bObj,BigInteger.valueOf(factors.get(u)));
                        bObj.moduli0b.add(md);
                        tObj.moduli0t.add(md);
                        objects.get(e).add(bObj);
                        objects.get(e+1).add(tObj);
                    }
                }
            }
        }
        
        void whitneyfy(ArrayList<sh_object> objects) {
            int count = 0;
            String total = String.valueOf(objects.size());
            for (sh_object obj : objects) {
                count++;
                LabDim.setText(String.valueOf(count)+" of "+total);
                if (!runs) return;
                int i = 0;
                while (i < obj.moduli0t.size()) {
                    mod0 mod = obj.moduli0t.get(i);
                    if (mod.totalValue.equals(BigInteger.ZERO)) {
                        mod.topObj.moduli0t.remove(mod);
                        mod.botObj.moduli0b.remove(mod);
                    }
                    else i++;
                }
                i = 0;
                while (i < obj.moduli1t.size()) {
                    mod1 mod = obj.moduli1t.get(i);
                    if (mod.framedCircle & mod.edges.isEmpty()) {
                        mod.topObj.moduli1t.remove(mod);
                        mod.botObj.moduli1b.remove(mod);
                    }
                    else i++;
                }
            }
        }
        
        void createMod1(ArrayList<ArrayList<sh_object>> objects, link Link, int n, int count) {
            LabTask.setText("Creating 1-dim moduli spaces");
            int counter = 0;
            String cound = String.valueOf(count);
            int ht = objects.size()-1;
            while (ht > 1) {
                LabLev.setText(String.valueOf(ht));
                String counj = String.valueOf(objects.get(ht).size());
                int couni = 0;
                //if (ht <=1) return;
                for (sh_object object : objects.get(ht)) {
                    counter++;
                    couni++;
                    if (!runs) return;
                    LabObj.setText(String.valueOf(counter)+" of "+cound);
                    LabDim.setText(String.valueOf(couni)+" of "+counj);
                    ArrayList<pairPoints> pairs = new ArrayList<pairPoints>(); 
                    for (mod0 mod : object.moduli0t) {
                        for (mod0 md : mod.botObj.moduli0t) {
                            pairPoints pair = new pairPoints(md,mod);
                            pairs.add(pair);
                        }
                    }
                    while (!pairs.isEmpty()) {
                        pairPoints first = pairs.get(0);
                        int firstChange = diffPosition(first.first.topObj.dig,first.first.botObj.dig);
                        int seconChange = diffPosition(first.second.topObj.dig,first.second.botObj.dig);
                        ArrayList<pairPoints> similar = new ArrayList<pairPoints>();
                        for (pairPoints next : pairs) {
                            int nextfChange = diffPosition(next.first.topObj.dig,next.first.botObj.dig);
                            int nextsChange = diffPosition(next.second.topObj.dig,next.second.botObj.dig);
                            if ((nextfChange == firstChange & nextsChange == seconChange) | (nextfChange == seconChange & nextsChange == firstChange)) {
                                if (next.first.botObj == first.first.botObj) similar.add(next);
                            }
                        }
                        for (pairPoints pr : similar) pairs.remove(pr);
                        if (firstChange == seconChange) {
                            diagram diag = first.first.botObj.dig;
                            if (similar.size()<=2) { // very simple
                                boolean frame = true;
                                int factor = diag.position[firstChange];
                                if (diag.position[firstChange] < 0) factor++;
                                int help = 0;
                                for (int y = 0; y < firstChange; y++) help = help + diag.position[y];
                                help = help * factor;
                                if (help % 2 != 0) frame = !frame;
                                if (first.first.totalValue.equals(BigInteger.ZERO)) {
                                    if (first.second.totalValue.intValue()%2 == 0) frame = false;
                                    @SuppressWarnings("unused")
                                    mod1 modOne = new mod1(first.second.topObj, first.first.botObj, !frame);
                                }
                                if (first.second.totalValue.equals(BigInteger.ZERO)) {
                                    int v = first.first.totalValue.intValue();
                                    if (v % 2 != 0 & v < 0) frame = !frame; 
                                    @SuppressWarnings("unused")
                                    mod1 modOne = new mod1(first.second.topObj, first.first.botObj, !frame);
                                }
                                if (first.first.totalValue.abs().equals(BigInteger.ONE) & 
                                        first.second.totalValue.abs().equals(BigInteger.ONE)) {
                                    mod1 modOne = new mod1(first.second.topObj, first.first.botObj, frame);
                                    boolean bl = first.first.totalValue.intValue() == 1;
                                    boolean tl = first.second.totalValue.intValue() == 1;
                                    boolean br = similar.get(1).first.totalValue.intValue() == 1;
                                    boolean tr = similar.get(1).second.totalValue.intValue() == 1;
                                    edge eddy = new edge(first.first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                    modOne.edges.add(eddy);
                                }
                            }
                        }
                        else {
                            diagram diag = first.first.botObj.dig;
                            int firstval = first.first.totalValue.intValue();
                            int seconval = first.second.totalValue.intValue();
                            boolean ordnung = true;
                            if (firstChange > seconChange) {
                                int h = firstChange;
                                firstChange = seconChange;
                                seconChange = h;
                                h = firstval;
                                firstval = seconval;
                                seconval = h;
                                ordnung = false;
                            }
                            int a = diag.position[firstChange];
                            int b = diag.position[seconChange];
                            if (similar.size()<=2) { // very simple
                                BigInteger firs = similar.get(0).first.totalValue;
                                BigInteger secs = similar.get(0).second.totalValue;
                                if (firs.abs().equals(BigInteger.ONE) & secs.abs().equals(BigInteger.ONE)) { // one edge between them
                                    boolean frame = framing(first.first, first.second);
                                    mod1 modone = new mod1(first.second.topObj,first.first.botObj,frame);
                                    boolean bl = similar.get(0).first.totalValue.intValue() == 1;
                                    boolean tl = similar.get(0).second.totalValue.intValue() == 1;
                                    boolean br = similar.get(1).first.totalValue.intValue() == 1;
                                    boolean tr = similar.get(1).second.totalValue.intValue() == 1;
                                    edge ed = new edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                    modone.edges.add(ed);
                                }
                                if ((firs.abs().equals(BigInteger.ONE) & secs.equals(BigInteger.ZERO)) | 
                                        (firs.equals(BigInteger.ZERO) & secs.abs().equals(BigInteger.ONE))) {
                                    if (!firs.abs().equals(BigInteger.ONE)) firs = similar.get(1).first.totalValue;
                                    boolean frame = framingZero(first.first,first.second);
                                    @SuppressWarnings("unused")
                                    mod1 modone = new mod1(first.second.topObj,first.first.botObj,frame);
                                }
                                if ((firs.abs().equals(BigInteger.ONE) & secs.abs().equals(BigInteger.valueOf(n))) | 
                                        (firs.abs().equals(BigInteger.valueOf(n)) & secs.abs().equals(BigInteger.ONE))) {
                                    boolean frame = framing(first.first,first.second);
                                    if (n % 2 == 0) frame = true; 
                                    boolean eps = true;
                                    if (firs.abs().equals(BigInteger.ONE)) {
                                        if (firs.equals(BigInteger.ONE)==ordnung) eps = false;
                                    }
                                    else {
                                        if (!similar.get(1).first.totalValue.equals(BigInteger.ONE)!=ordnung) eps = false;
                                    }
                                    if ((n/2)%2 == 0) eps = true;
                                    if (eps == frame) frame = true;
                                    else frame = false;
                                    mod1 modone = new mod1(first.second.topObj,first.first.botObj, frame);
                                    if (n % 2 != 0) {
                                        boolean bl = similar.get(0).first.totalValue.intValue() > 0;
                                        boolean tl = similar.get(0).second.totalValue.intValue() > 0;
                                        boolean br = similar.get(1).first.totalValue.intValue() > 0;
                                        boolean tr = similar.get(1).second.totalValue.intValue() > 0;
                                        edge ed = new edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                        modone.edges.add(ed);
                                    }
                                }
                                if (firs.abs().equals(BigInteger.ZERO) & secs.abs().equals(BigInteger.ZERO)) {
                                    //no mod1 needed
                                }
                                if ((firs.abs().equals(BigInteger.ZERO) & secs.abs().equals(BigInteger.valueOf(n))) | 
                                        (firs.abs().equals(BigInteger.valueOf(n)) & secs.abs().equals(BigInteger.ZERO))) {
                                    if (n%2 != 0) {
                                        mod1 modone = new mod1(first.second.topObj,first.first.botObj, true);
                                        int y = 0;
                                        if (firs.abs().equals(BigInteger.valueOf(n))) y = firs.intValue();
                                        else y = secs.intValue();
                                        boolean frame = framingZero(first.first,first.second);
                                        if (y > 0) frame = !frame;
                                        modone.framedCircle = frame;
                                    }// n % 2 == 0 does not require a moduli space
                                }
                                if (firs.abs().equals(BigInteger.valueOf(n)) & secs.abs().equals(BigInteger.valueOf(n))) {
                                    if (n%2 == 0) {
                                        if (n%4 != 0) {
                                            mod1 modone = new mod1(first.second.topObj,first.first.botObj,false);
                                        }
                                    }
                                    else {
                                        mod1 modone = new mod1(first.second.topObj,first.first.botObj,true);
                                        boolean bl = similar.get(0).first.totalValue.intValue() > 0;
                                        boolean tl = similar.get(0).second.totalValue.intValue() > 0;
                                        boolean br = similar.get(1).first.totalValue.intValue() > 0;
                                        boolean tr = similar.get(1).second.totalValue.intValue() > 0;
                                        edge ed = new edge(similar.get(0).first.topObj,similar.get(1).first.topObj,tl,bl,tr,br);
                                        modone.edges.add(ed);
                                        boolean frame = framing(first.first,first.second);
                                        if ( (n-1)% 4 != 0) {
                                            frame = !frame;
                                            boolean extra = (similar.get(0).first.totalValue.add(similar.get(1).first.totalValue).equals(BigInteger.ZERO));
                                            if (!extra) frame = !frame;
                                        }
                                        modone.framedCircle = frame;
                                    }
                                }
                            }
                            else {
                                mod1 modOne = new mod1(first.second.topObj,first.first.botObj,true);
                                if (((a == 0) | (a == -1)) & ((b == 0) | (b == -1))) { // it's a ladybug!!!
                                    if (n%2 !=0) {
                                        int as = 0;
                                        for (int e = 0; e < firstChange; e++) {
                                            as = as + first.first.botObj.dig.position[e];
                                        }
                                        if (as % 2 != 0) {
                                            as = 0;
                                            for (int e = firstChange; e < seconChange; e++) {
                                                as = as + first.first.botObj.dig.position[e];
                                            }
                                            if (as % 2 != 0) modOne.framedCircle = false;
                                        }
                                    }
                                    int rightguy;
                                    if (a == 0) rightguy = Link.paths[firstChange][0];
                                    else rightguy = Link.paths[firstChange][1];
                                    while (!similar.isEmpty()) {
                                        pairPoints fist = similar.get(0);
                                        sh_object midObj = fist.first.topObj;
                                        int circnum = 0;
                                        boolean found = false;
                                        while (!found) {
                                            ArrayList<Integer> circ = midObj.dig.circles.get(circnum);
                                            if (circ.contains(rightguy)) found = true;
                                            else circnum++;
                                        }
                                        byte thesign = midObj.symbols.get(circnum);
                                        found = false;
                                        int v = 1;
                                        while (!found) {
                                            pairPoints seco = similar.get(v);
                                            sh_object mitObj = seco.first.topObj;
                                            if (mitObj.dig.position[firstChange] != midObj.dig.position[firstChange])
                                            {
                                                int cirxnum = 0;
                                                boolean fnd = false;
                                                while (!fnd) {
                                                    ArrayList<Integer> circ = mitObj.dig.circles.get(cirxnum);
                                                    if (circ.contains(rightguy)) fnd = true;
                                                    else cirxnum++;
                                                }
                                                if (mitObj.symbols.get(cirxnum) == thesign) found = true;
                                                else v++;
                                            }
                                            else v++;
                                        }
                                        boolean bl = fist.first.totalValue.intValue() == 1;
                                        boolean tl = fist.second.totalValue.intValue() == 1;
                                        boolean br = similar.get(v).first.totalValue.intValue() == 1;
                                        boolean tr = similar.get(v).second.totalValue.intValue() == 1;
                                        edge ed = new edge(fist.first.topObj,similar.get(v).first.topObj,tl,bl,tr,br);
                                        modOne.edges.add(ed);
                                        similar.remove(v);
                                        similar.remove(0);
                                    }
                                }
                                else {
                                    int unpleasant = 0;
                                    int i = 0;
                                    if ((a == 0) | (a == -1) | (b == 0) | (b == -1)) { // check for an unpleasant surgery involved
                                        while (i < similar.size() & unpleasant == 0) {
                                            mod0 modf = similar.get(i).first;
                                            mod0 mods = similar.get(i).second;
                                            if (!modf.totalValue.abs().equals(BigInteger.ONE)) unpleasant = 1;
                                            if (!mods.totalValue.abs().equals(BigInteger.ONE)) unpleasant = 2;
                                            if (unpleasant == 0) i++;
                                        }
                                    }
                                    if (unpleasant > 0) {
                                        boolean extrasign = true;
                                        boolean firstspec = true;
                                        mod0 p1 = similar.get(i).first;
                                        mod0 p2 = similar.get(i).second;
                                        if (p1.totalValue.abs().equals(BigInteger.ONE)) {
                                            extrasign = p1.totalValue.intValue() > 0;
                                            firstspec = false;
                                        }
                                        int cas;
                                        if (firstspec) cas = p1.totalValue.intValue();
                                        else cas = p2.totalValue.intValue();
                                        if (cas == 0) {
                                            similar.remove(i);
                                            pairPoints pts1 = similar.get(0);
                                            pairPoints pts2 = similar.get(1);
                                            edge ed;
                                            boolean bl = pts1.first.totalValue.intValue() == 1;
                                            boolean tl = pts1.second.totalValue.intValue() == 1;
                                            boolean br = pts2.first.totalValue.intValue() == 1;
                                            boolean tr = pts2.second.totalValue.intValue() == 1;
                                            ed = new edge(pts1.first.topObj,pts2.first.topObj,tl,bl,tr,br);
                                            boolean fram1 = framing(pts1.first,pts1.second);
                                            boolean fram2 = framing(pts2.first,pts2.second);
                                            if (firstspec) {
                                                if (fram1!=fram2) modOne.framedCircle = false;
                                                modOne.edges.add(ed);
                                            }
                                            else {
                                                modOne.framedCircle = extrasign;
                                                modOne.edges.add(ed);
                                                if (fram1!=fram2) modOne.framedCircle = !modOne.framedCircle;
                                            }
                                        }
                                        else {
                                            int c = -1;
                                            if (firstspec) c = diffPosition(p1.topObj.dig, p1.botObj.dig);
                                            else c = diffPosition(p2.topObj.dig,p2.botObj.dig);
                                            int v = Link.paths[c][0];
                                            if (n%2!=0) {
                                                int j = 0;
                                                boolean found = false;
                                                while (!found) {
                                                    mod0 mod;
                                                    if (j == i) j++;
                                                    if (firstspec) mod = similar.get(j).second;
                                                    else mod = similar.get(j).first;
                                                    if (dotIndex(mod,v) == n-1) found = true;
                                                    else j++;
                                                }
                                                boolean bl = similar.get(i).first.totalValue.intValue() > 0;
                                                boolean tl = similar.get(i).second.totalValue.intValue() > 0;
                                                boolean br = similar.get(j).first.totalValue.intValue() > 0;
                                                boolean tr = similar.get(j).second.totalValue.intValue() > 0;
                                                edge ed = new edge(similar.get(i).first.topObj,similar.get(j).first.topObj,tl,bl,tr,br);
                                                modOne.framedCircle = framing(similar.get(j).first,similar.get(j).second);
                                                modOne.edges.add(ed);
                                                similar.remove(j);
                                                if (j < i) i--;
                                            }
                                            similar.remove(i);
                                            while (!similar.isEmpty()) {
                                                pairPoints pts1 = similar.get(0);
                                                int j = 1;
                                                boolean found = false;
                                                mod0 mod;
                                                if (firstspec) mod = pts1.second;
                                                else mod = pts1.first;
                                                int k = dotIndex(mod,v);
                                                if ((k%2)!=0) k--;
                                                else k++;
                                                while (!found) {
                                                    if (firstspec) mod = similar.get(j).second;
                                                    else mod = similar.get(j).first;
                                                    if (dotIndex(mod,v) == k) found = true;
                                                    else j++;
                                                }
                                                boolean bl = similar.get(0).first.totalValue.intValue() > 0;
                                                boolean tl = similar.get(0).second.totalValue.intValue() > 0;
                                                boolean br = similar.get(j).first.totalValue.intValue() > 0;
                                                boolean tr = similar.get(j).second.totalValue.intValue() > 0;
                                                edge ed = new edge(similar.get(0).first.topObj,similar.get(j).first.topObj,tl,bl,tr,br);
                                                if (k%2 == 0) modOne.framedCircle = !modOne.framedCircle;
                                                if (!firstspec & !extrasign) modOne.framedCircle = !modOne.framedCircle;
                                                modOne.edges.add(ed);
                                                similar.remove(j);
                                                similar.remove(0);
                                            }
                                        }
                                    }
                                    else { // it's a combination of dottings, 
                                        if (((a > 0) | (a < -1)) & ((b > 0) | (b < -1))) { // just dottings
                                            while (!similar.isEmpty()) {
                                                mod0 p1 = similar.get(0).first;
                                                mod0 p2 = similar.get(0).second;
                                                int x = diffPosi(p1.botObj.dig,p1.topObj.dig);//System.out.println("Ha "+x);
                                                int y = diffPosi(p2.botObj.dig,p2.topObj.dig);
                                                int c1 = Link.paths[x][0];
                                                int c2 = Link.paths[y][0];
                                                int k1 = dotIndex(p1,c1);
                                                int k2 = dotIndex(p2,c2);
                                                boolean found = false;
                                                int e = 1;
                                                while (!found) {
                                                    mod0 p3 = similar.get(e).first;
                                                    mod0 p4 = similar.get(e).second;
                                                    //System.out.println("Ha "+diffPosi(p3.botObj.dig,p3.topObj.dig)+" "+y);
                                                    if (diffPosi(p3.botObj.dig,p3.topObj.dig) == y) {
                                                        if (dotIndex(p4,c1) == k1 & dotIndex(p3,c2) == k2) found = true;
                                                    }
                                                    if (!found) e++;
                                                }
                                                sh_object corner = similar.get(e).first.topObj;
                                                boolean bl = p1.totalValue.intValue() == 1;
                                                boolean tl = p2.totalValue.intValue() == 1;
                                                boolean br = similar.get(e).first.totalValue.intValue() == 1;
                                                boolean tr = similar.get(e).second.totalValue.intValue() == 1;
                                                similar.remove(e);
                                                similar.remove(0);
                                                edge ed = new edge(p1.topObj,corner,tl,bl,tr,br);
                                                modOne.edges.add(ed);
                                                boolean frame = framing(p1,p2);
                                                if (!frame) modOne.framedCircle = !modOne.framedCircle;
                                            }
                                        }
                                    } // */
                                }
                            }
                        }
                    }    
                }
                ht--;
            }
        }
        
        void createMod0(ArrayList<ArrayList<sh_object>> objects, link Link, int n, int count) {
            LabTask.setText("Creating 0-dim moduli spaces");
            int counter = 0;
            String cound = String.valueOf(count);
            int k = objects.size();
            for (int j = k-1; j > 0; j--) {
                LabLev.setText(String.valueOf(j));
                String counj = String.valueOf(objects.get(j).size());
                for (int i = 0; i < objects.get(j).size(); i++) {
                    counter++;
                    LabObj.setText(String.valueOf(counter)+" of "+cound);
                    LabDim.setText(String.valueOf(i)+" of "+counj);
                    sh_object topObj = objects.get(j).get(i);
                    for (int l = 0; l < objects.get(j-1).size(); l++) {
                        if (!runs) return;
                        sh_object botObj = objects.get(j-1).get(l);
                        int r = diffPosition(topObj.dig,botObj.dig);
                        if (r > -1) {
                            if ((botObj.dig.position[r] == 0) | (botObj.dig.position[r] == -1)) { // possible surgery
                                boolean surgery = aSurgery(topObj,botObj,Link.paths[r]);
                                if (surgery) {
                                    boolean sign = getSign(botObj.dig.position,r);
                                    BigInteger val = new BigInteger("1");
                                    if (!sign) val = val.negate();
                                    mod0 point = new mod0(topObj,botObj,val);
                                    topObj.moduli0t.add(point);
                                    botObj.moduli0b.add(point);
                                }
                            }
                            else { // possible dotting of one circle
                                int dotting = aDotting(topObj,botObj,Link.paths[r],n);
                                if (dotting == 1) { // produce n points or two points with different sign
                                    boolean sign = getSign(botObj.dig.position,r);
                                    int checker = botObj.dig.position[r];
                                    if (checker < 0) checker++;
                                    if (checker%2 ==0) { // points of same sign
                                        BigInteger val = new BigInteger("1");
                                        if (!sign) val = val.negate();
                                        val = val.multiply(BigInteger.valueOf(n));
                                        mod0 point = new mod0(topObj,botObj,val);
                                        topObj.moduli0t.add(point);
                                        botObj.moduli0b.add(point);
                                    }
                                    else {
                                        BigInteger val = new BigInteger("1");
                                        mod0 point = new mod0(topObj,botObj,val);
                                        point.totalValue = BigInteger.ZERO;
                                        topObj.moduli0t.add(point);
                                        botObj.moduli0b.add(point);
                                    }
                                }
                                if (dotting >= 2) {
                                    boolean sign = getSign(botObj.dig.position,r);
                                    int checker = botObj.dig.position[r];
                                    if (botObj.dig.position[r]<0) checker++;
                                    if ((checker % 2)!=0) { // need to be careful with signs
                                            if (dotting == 3) sign = !sign; // case when right circle gets dotted and sign should change
                                    }
                                    BigInteger val = new BigInteger("1") ;
                                    if (!sign) val = val.negate();
                                    mod0 point = new mod0(topObj,botObj,val);
                                    topObj.moduli0t.add(point);
                                    botObj.moduli0b.add(point);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    class KhovCalc extends Thread {
        boolean runs = true;
        boolean finished = false;
        boolean reduced = false;
        JDialog frame;
        JLabel LabNam;
        JLabel LabCro;
        JLabel LabObj;
        link Link;
        int theN;
        ArrayList<homology> KhovHom = new ArrayList<homology>();
        
        public KhovCalc(JDialog fram, JLabel labe, JLabel labg, JLabel labh, link Lnk, int n, boolean red) {
            frame = fram;
            LabNam = labe;
            LabCro = labg;
            LabObj = labh;
            Link = Lnk;
            theN = n;
            reduced = red;
        }
        
        @Override
        public void run() {
            ArrayList<Khov_object> objects = getObjects(Link,theN);
            if (reduced) {
                for (Khov_object obj : objects) obj.qdeg++;
            }
            int qmax = objects.get(0).qdeg;
            int qmin = qmax;
            int hmax = objects.get(objects.size()-1).hdeg;
            int hmin = objects.get(0).hdeg;
            for (Khov_object obj: objects) {
                if (obj.qdeg > qmax) qmax = obj.qdeg;
                if (obj.qdeg < qmin) qmin = obj.qdeg;
                if (obj.hdeg > hmax) hmax = obj.hdeg;
                if (obj.hdeg < hmin) hmin = obj.hdeg;
            }
            LabNam.setText("Quantum degree : ");
            for (int q = qmin; q <= qmax; q = q + 2 ) {
                LabCro.setText(" "+q);
                homology Khovq = IntegralHomology(objects,hmax,hmin,q,false);
                if (Khovq != null) {
                    Khovq.quantum = q;
                    KhovHom.add(Khovq);
                }
                if (!runs) break;
            }
            finished = true;
            delay(200);
            frame.setVisible(false);
        }
        
        final homology IntegralHomology(ArrayList<Khov_object> objects, int hmax, int hmin, int qdeg, boolean show) {
            homology hom = new homology();
            ArrayList<ArrayList<Khov_object>> qobjects = new ArrayList<ArrayList<Khov_object>>();
            int h = hmax - hmin;
            for (int k = 0; k <= h; k++) {
                ArrayList<Khov_object> qobjex = new ArrayList<Khov_object>();
                qobjects.add(qobjex);
                homologyGroup newGroup = new homologyGroup();
                newGroup.Betti = 0;
                hom.degrees.add(hmin+k);
                hom.groups.add(newGroup);
            }
            KhovCategory cat = new KhovCategory();
            for (Khov_object object : objects) {
                if (object.qdeg == qdeg) {
                    qobjects.get(object.hdeg-hmin).add(object);
                    cat.objects.add(object);
                }
            }
            SmithNormalise(qobjects);
            if (!runs) return null;
            for (int hq = 0; hq < qobjects.size(); hq++) {
                ArrayList<Khov_object> objex = qobjects.get(hq);
                for (Khov_object obj : objex) {
                    if (obj.botMods.isEmpty() & obj.topMods.isEmpty()) hom.groups.get(hq).Betti++;
                    if (!obj.botMods.isEmpty()) hom.groups.get(hq).Torsion.add(obj.botMods.get(0).totalValue.intValue());
                }
                Collections.sort(hom.groups.get(hq).Torsion);
            }
            return hom;
        }
        
        void SmithNormalise(ArrayList<ArrayList<Khov_object>> objects) {
            int ht = objects.size()-1;
            ArrayList<ArrayList<Khov_object>> torsionObjects = new ArrayList<ArrayList<Khov_object>>();
            for (int k = 0; k < objects.size(); k++) {
                ArrayList<Khov_object> tlist = new ArrayList<Khov_object>();
                torsionObjects.add(tlist);
            }
            while (ht > 0 & runs) {
                int objnumber = 0;
                for (ArrayList<Khov_object> objs : objects) objnumber = objnumber+objs.size();
                LabObj.setText(" "+objnumber);
                removeMod0(objects.get(ht));
                removeMod0(objects.get(ht-1));
                Kmod0 mod = findOne(objects.get(ht));
                if (mod != null) {
                    Khov_object tObj = mod.topObj; 
                    Khov_object bObj = mod.botObj;
                    shiftTop(mod);
                    removeMod0(objects.get(ht));
                    shiftBot(mod);
                    removeMod0(objects.get(ht));
                    removeMod0(objects.get(ht-1));
                    objects.get(ht).remove(tObj);
                    objects.get(ht-1).remove(bObj);
                }
                else {
                    mod = findSmallest(objects.get(ht)); 
                    if (mod != null) {
                        Khov_object tObj = mod.topObj; 
                        Khov_object bObj = mod.botObj;
                        slideTop(mod);
                        removeMod0(objects.get(ht));
                        removeMod0(objects.get(ht-1));
                        if (mod.botObj.topMods.size()==1) {
                            slideBot(mod);
                            removeMod0(objects.get(ht));
                            removeMod0(objects.get(ht-1));
                            if (mod.topObj.botMods.size() == 1) {
                                mod.totalValue = mod.totalValue.abs(); // just make sure it is positive
                                objects.get(ht).remove(tObj);
                                objects.get(ht-1).remove(bObj);
                                torsionObjects.get(ht).add(tObj);
                                torsionObjects.get(ht-1).add(bObj);
                            }
                        }
                    }
                    else ht--;
                }
            }
            for (int e = 0; e < objects.size(); e++) {
                for (Khov_object obj : torsionObjects.get(e)) {
                    if (!obj.topMods.isEmpty()) {
                        Kmod0 mod = obj.topMods.get(0);
                        int p = mod.totalValue.intValue();
                        ArrayList<Integer> factorization = factorize(p);
                        for (int t : factorization) {
                            Khov_object tObj = new Khov_object(mod.topObj.hdeg, mod.topObj.qdeg);
                            Khov_object bObj = new Khov_object(obj.hdeg, obj.qdeg);
                            Kmod0 md = new Kmod0(tObj,bObj,BigInteger.valueOf(t));
                            tObj.botMods.add(md);
                            bObj.topMods.add(md);
                            objects.get(e).add(bObj);
                            objects.get(e+1).add(tObj);
                        }
                    }
                }
            }
	}
        
        ArrayList<Khov_object> getObjects(link Link, int n) {
            ArrayList<Khov_object> objects = new ArrayList<Khov_object>();
            KhovCategory runner = startCat(Link.crossings[0],Link.paths[0],nminus(Link),nplus(Link),n,reduced);
            int u = 1;
            while (u < Link.crossings.length & runs) {
                LabCro.setText(" "+(u+1));
                KhovCategory next = startCat(Link.crossings[u],Link.paths[u],0,0,n,false);
                runner = mergeCat(runner,next,n);
                u++;
                //for (Khov_object obj : runner.objects) {
                //    objectOutput(obj,true);
                //}
            }
            objects = runner.objects;
            return objects;
        }
    
        KhovCategory mergeCat(KhovCategory first, KhovCategory second, int n) {
            KhovCategory merged = new KhovCategory();
            createNewObjects(merged,first,second,n);
            ArrayList<Kmod0> candidates = new ArrayList<Kmod0>(); // list of potential moduli spaces that can be cancelled
            createNewMod0(merged,first,second,candidates,true,n);
            createNewMod0(merged,second,first,candidates,false,n);
            for (Khov_object obj : merged.objects) {
                for (ArrayList<Integer> path : obj.paths) {
                    while (path.size()>2) path.remove(1);
                }
            }// clean up of the paths involved
            LabObj.setText(" "+merged.objects.size());
            cancelObjects(merged,candidates,n);
            mergeMod0(merged.objects);
            whitneyfy(merged.objects);
            return merged;
        }
        
        void createNewObjects(KhovCategory merge, KhovCategory first, KhovCategory second, int n) {
            for (int e : first.endpoints) merge.endpoints.add(e);
            for (int e : second.endpoints) {
                if (merge.endpoints.contains(e)) merge.endpoints.remove((Integer) e);
                else merge.endpoints.add(e);
            }
            for (Khov_object obj0 : first.objects) {
                for (Khov_object obj1 : second.objects) {
                    ArrayList<Khov_object> newObjects = tensorObjects(obj0,obj1, merge.endpoints,n,reduced);
                    for (Khov_object newObj : newObjects) {
                        merge.objects.add(newObj);
                        obj0.descObjects.add(newObj);
                        obj1.descObjects.add(newObj);
                    }
                    if (!runs) break;
                    LabObj.setText(" "+merge.objects.size());
                }
                if (!runs) break;
            }
        }
        
        void createNewMod0(KhovCategory merge, KhovCategory first, KhovCategory second, ArrayList<Kmod0> candidates, boolean order, 
                int n) {
            for (Khov_object obj0 : first.objects) {
                for (Kmod0 mod : obj0.topMods) {
                    for (Khov_object obj : second.objects) {
                        for (Khov_object newObj0 : obj0.descObjects) {
                            if (obj.descObjects.contains(newObj0)) {
                                packige pack = createModsFrom(newObj0,mod,obj,merge.endpoints,n);
                                for (int i = 0; i < pack.objects.size(); i++) {
                                    Khov_object dude = pack.objects.get(i);
                                    ArrayList<ArrayList<Integer>> actions = pack.actions.get(i);
                                    for (Khov_object newObj1 : mod.topObj.descObjects) {
                                        if (obj.descObjects.contains(newObj1)) {
                                            int j = 0;
                                            boolean ok = true;
                                            while (ok & j < newObj1.circles.size()) {
                                                    if (dude.circles.get(j).sign != newObj1.circles.get(j).sign) ok = false;
                                                    else j++;
                                            }
                                            if (ok) {
                                                BigInteger val = mod.totalValue;
                                                if (!order & (obj.hdeg%2) !=0) val = val.negate();
                                                Kmod0 nmod = new Kmod0(newObj1,newObj0,val,actions);
                                                newObj1.botMods.add(nmod);
                                                newObj0.topMods.add(nmod);
                                                mod.descMod0.add(nmod);
                                                if (actions.isEmpty() & (val.abs().equals(BigInteger.ONE))) candidates.add(nmod);
                                            }
                                        }
                                    }
                                }
                            }
                            if (!runs) break;
                        }
                        if (!runs) break;
                    }
                    if (!runs) break;
                }
                if (!runs) break;
            }
        }
        
        void cancelObjects(KhovCategory merge, ArrayList<Kmod0> candidates, int n) {
            for (Kmod0 mod : candidates) {// see what we can cancel
                Khov_object tObj = mod.topObj;
                Khov_object bObj = mod.botObj;
                LabObj.setText(" "+merge.objects.size());
                if (merge.objects.contains(tObj) & merge.objects.contains(bObj)) {
                    boolean cancel = true;
                    if (reduced & bObj.dotpath != tObj.dotpath) cancel = false; 
                    if (mod.totalValue.abs().compareTo(BigInteger.ONE) != 0) cancel = false;
                    if (tObj.botMods.size() < bObj.topMods.size()) {
                        int i = 0;
                        while (cancel & (i< tObj.botMods.size())) {
                            Kmod0 md = tObj.botMods.get(i);
                            if (md != mod) {
                                if (md.botObj == bObj) cancel = false;
                            }
                            i++;
                        }
                    }
                    else {
                        int i = 0;
                        while (cancel & (i < bObj.topMods.size())) {
                            Kmod0 md = bObj.topMods.get(i);
                            if (md != mod) {
                                if (md.topObj == tObj) cancel = false;
                            }
                            i++;
                        }
                    }
                    if (cancel) getRidOf(merge,tObj,bObj, mod,n);
                }
                if (!runs) break;
            }
        }
    }
    
    class HomflyCalc extends Thread {
        boolean runs = true;
        boolean finished = false;
        twoPolynomial Homfly;
        JDialog frame;
        JLabel LabCro;
        JLabel LabObj;
        int writhe;
        int[] crossings;
        int[][] paths;
        int comp;
        
        public HomflyCalc(JDialog fram, JLabel labe, JLabel labf, link Link) {
            frame = fram;
            LabCro = labe;
            LabObj = labf;
            crossings = Link.crossings;
            paths = Link.paths;
            comp = Link.disjointComponents();
            writhe = nplus(Link)-nminus(Link);
        }
        
        @Override
        public void run() {
            HOMCategory runner = startCatHOM(crossings[0],paths[0]);
            int i = 1;
            while(i < crossings.length & runs) {
                HOMCategory next = startCatHOM(crossings[i],paths[i]);
                runner = mergeHOMCat(runner,next);
                LabCro.setText(" "+i);
                i++;
            }
            if (runs) {
                Homfly = runner.objects.get(0).poly;
                finished = true;
                delay(200);
                frame.setVisible(false);
            }
        }
        
        HOMCategory mergeHOMCat(HOMCategory first, HOMCategory second) {
            HOMCategory merge = new HOMCategory();
            for (int e : first.endpoints) merge.endpoints.add(e);
            for (int e : second.endpoints) {
                if (merge.endpoints.contains(e)) merge.endpoints.remove((Integer) e);
                else merge.endpoints.add(e);
            }
            for (HOMObject obj0 : first.objects)
                for (HOMObject obj1 : second.objects) {
                    HOMObject newObj = tensorHOMObject(obj0,obj1,merge.endpoints);
                    merge.objects.add(newObj);
                }
            int u = 0;
            while (u < merge.objects.size()-1) {
                HOMObject obj0 = merge.objects.get(u);
                int top = merge.objects.size()-1;
                for (int v = top; v > u; v--) {
                    HOMObject obj1 = merge.objects.get(v);
                    if (samePaths(obj0.paths,obj1.paths)) {
                        obj0.poly.add(obj1.poly);
                        merge.objects.remove(obj1);
                    }
                }
                u++;
            }
        return merge;
	}
    }
    
    class JonesCalc extends Thread {
        boolean runs = true;
        boolean finished = false;
        polynomial Jones;
        JDialog frame;
        JLabel LabCro;
        JLabel LabObj;
        int writhe;
        int[] crossings;
        int[][] paths;
        int comp;
        
        public JonesCalc(JDialog fram, JLabel labe, JLabel labf, link Link) {
            frame = fram;
            LabCro = labe;
            LabObj = labf;
            crossings = Link.crossings;
            paths = Link.paths;
            comp = Link.disjointComponents();
            writhe = nplus(Link)-nminus(Link);
        }
        
        @Override
        public void run() {
            JonesCategory runner = startCat(crossings[0],paths[0]);
            int i = 1;
            while(i < crossings.length & runs) {
                JonesCategory next = startCat(crossings[i],paths[i]);
                runner = mergeCat(runner,next);
                LabCro.setText(" "+i);
                i++;
            }
            if (runs) {
                polynomial jones = runner.objects.get(0).poly;
                if (comp > 1) {
                    polynomial circlePol = new polynomial();
                    coefficient top = new coefficient(-1,-2);
                    coefficient bot = new coefficient(-1,2);
                    circlePol.entries.add(top);
                    circlePol.entries.add(bot);
                    for (int k = 1; k < comp; k++) jones = multiplyPoly(jones,circlePol);
                }
                int sign = 1;
                if (writhe%2!=0) sign = -1;
                polynomial aux = new polynomial(sign,-3*writhe);
                Jones = multiplyPoly(jones,aux);
                finished = true;
                delay(200);
                frame.setVisible(false);
            }
        }
        
        JonesCategory mergeCat(JonesCategory first, JonesCategory second) {
            JonesCategory merge = new JonesCategory();
            for (int e : first.endpoints) merge.endpoints.add(e);
            for (int e : second.endpoints) {
                if (merge.endpoints.contains(e)) merge.endpoints.remove((Integer) e);
                else merge.endpoints.add(e);
            }
            int i = 0;
            while (runs & i < first.objects.size()) {
                JonesObject obj0 = first.objects.get(i);
                int j = 0;
                while (runs & j < second.objects.size()) {
                    JonesObject obj1 = second.objects.get(j);
                    JonesObject newObj = tensorObject(obj0,obj1,merge.endpoints);
                    merge.objects.add(newObj);
                    LabObj.setText(" "+merge.objects.size());
                    j++;
                }
                i++;
            }
            int u = 0;
            while (u < merge.objects.size()-1 & runs) {
                JonesObject obj0 = merge.objects.get(u);
                int top = merge.objects.size()-1;
                for (int v = top; v > u; v--) {
                    JonesObject obj1 = merge.objects.get(v);
                    if (samePaths(obj0.paths,obj1.paths)) {
                        obj0.poly.add(obj1.poly);
                        merge.objects.remove(obj1);
                        LabObj.setText(" "+merge.objects.size());
                    }
                }
                u++;
            }
            return merge;
        }
    }
    
    @SuppressWarnings("serial")
    class JCrossing extends JPanel {
        JTextField crossing = new JTextField();
        JTextField a = new JTextField();
        JTextField b = new JTextField();
        JTextField c = new JTextField();
        JTextField d = new JTextField();
        JLabel empty = new JLabel(" ");
        JLabel emptz = new JLabel(" ");
        JLabel emptx = new JLabel(" ");
        JLabel emptw = new JLabel(" ");
        
        JPanel north = new JPanel(new GridLayout(1,3));
        JPanel centr = new JPanel(new GridLayout(1,3));
        JPanel south = new JPanel(new GridLayout(1,3));
        
        public JCrossing(int cr, int[] pt) {
            setLayout(new BorderLayout());
            setPreferredSize(new Dimension(100,80));
            crossing.setText(String.valueOf(cr));
            a.setText(String.valueOf(pt[0]));
            b.setText(String.valueOf(pt[1]));
            c.setText(String.valueOf(pt[2]));
            d.setText(String.valueOf(pt[3]));
            centr.add(emptx);
            centr.add(crossing);
            centr.add(emptw);
            add(centr,BorderLayout.CENTER);
            north.add(c);
            north.add(empty);
            north.add(b);
            south.add(d);
            south.add(emptz);
            south.add(a);
            add(north,BorderLayout.NORTH);
            add(south,BorderLayout.SOUTH);
        }
    }
    
    @SuppressWarnings("serial")
    class DiagramComp extends JComponent {
        String[] value;
        String[] a,b,c,d;
        
        public DiagramComp(int[] crossing, int[][] paths) {
            value = new String[crossing.length];
            a = new String[crossing.length];
            b = new String[crossing.length];
            c = new String[crossing.length];
            d = new String[crossing.length];
            for (int i = 0; i < crossing.length; i++) {
                value[i] = String.valueOf(crossing[i]);
                a[i] = String.valueOf(paths[i][0]);
                b[i] = String.valueOf(paths[i][1]);
                c[i] = String.valueOf(paths[i][2]);
                d[i] = String.valueOf(paths[i][3]);
            }
        }
        
        @Override
        public void paintComponent(Graphics g) {
            for (int i = 0; i < value.length; i++) {
                g.drawString(value[i], 40 + 80 * i - value[i].length()*4, 55);
                g.drawString(a[i],60 + 80 * i,88);
                g.drawString(b[i],60 + 80 * i,24);
                g.drawString(c[i],18-c[i].length()*6 + 80 * i,24);
                g.drawString(d[i], 18-d[i].length()*6 + 80 * i, 88);
                g.drawRect(20+80*i,30,40,40);
                g.drawLine(25+80*i, 10, 25+80*i, 30);
                g.drawLine(55+80*i, 10, 55+80*i, 30);
                g.drawLine(25+80*i, 70, 25+80*i, 90);
                g.drawLine(55+80*i, 70, 55+80*i, 90);
            }
        }
    }
    
    class CCFrame {
        boolean ok = false;
        link Lnk;
        JList<String> components;
        int chosen = -1;
        
        CCFrame(JFrame fram, link Link) {
            Lnk = cloneLink(Link);
            JDialog frame = new JDialog(new JFrame(),"Components", true);
            frame.setLayout(new BorderLayout());
            frame.setBounds(fram.getX()+250, fram.getY()+180,300,240);
            frame.setResizable(false);
            JButton orient = new JButton("Change Orientation");
            JButton okay = new JButton("OK");
            JButton cancel = new JButton("Cancel");
            JPanel buttons = new JPanel();
            buttons.add(okay);
            buttons.add(cancel);
            orient.setEnabled(false);
            frame.add(buttons, BorderLayout.SOUTH);
            String[] strings = new String[Link.components.size()];
            for (int i = 0; i < Link.components.size(); i++) {
                strings[i] = Link.components.get(i).toString();
            }
            JComboBox<String> comboBox = new JComboBox<>(strings);
            components = new JList<String>(comboBox.getModel());
            JScrollPane scroller = new JScrollPane(components);
            scroller.setPreferredSize(new Dimension(240,120));
            JPanel center = new JPanel();
            center.add(scroller);
            center.add(orient);
            frame.add(center, BorderLayout.CENTER);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    ok = false;
                    frame.setVisible(false);
                }
            });
            okay.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    ok = true;
                    frame.setVisible(false);
                }
            });
            components.addListSelectionListener(new ListSelectionListener() {
                @Override
                public void valueChanged(ListSelectionEvent le) {
                    @SuppressWarnings("unchecked")
                    JList<String> list = (JList<String>) le.getSource();
                    int u = list.getSelectedIndex();
                    if ( u == -1) orient.setEnabled(false);
                    else {
                        chosen = u;
                        if (Lnk.components.get(u).size()> 1) orient.setEnabled(true);
                        else orient.setEnabled(false);
                    }
                }
            });
            orient.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    int first = Lnk.components.get(chosen).get(0);
                    int secon = Lnk.components.get(chosen).get(1);
                    for (int i = 0; i < Lnk.crossings.length; i++) {
                        for (int j = 0; j < 4; j++) {
                            if (Lnk.paths[i][j] == first) Lnk.paths[i][j] = secon;
                            else if (Lnk.paths[i][j] == secon) Lnk.paths[i][j] = first;
                        }
                    }
                    Lnk = new link(Lnk.crossings,Lnk.paths);
                    for (int i = 0; i < Lnk.components.size(); i++) {
                        strings[i] = Lnk.components.get(i).toString();
                    }
                    JComboBox<String> combBox = new JComboBox<>(strings);
                    components.setModel(combBox.getModel());
                }
            });
            frame.setVisible(true);
        }
    }
    
    class CSFrame {
        int value = -1;
        boolean reduced = false;
        
        CSFrame(String label, JFrame fram, boolean highn) {
            JDialog frame = new JDialog(new JFrame(), label, true);
            frame.setLayout(new BorderLayout());
            frame.setBounds(fram.getX()+250, fram.getY()+225,300,150);
            frame.setResizable(false);
            JLabel choose = new JLabel("Please choose n:");
            SpinnerNumberModel modeln = new SpinnerNumberModel(2,2,100,1);
            JSpinner spinnern = new JSpinner(modeln);
            JPanel cpane = new JPanel();
            JPanel rpane = new JPanel();
            JRadioButton unrButton = new JRadioButton("unreduced",true);
            JRadioButton redButton = new JRadioButton("reduced");
            final ButtonGroup bgroup = new ButtonGroup();
            JButton ok = new JButton("OK");
            JButton cancel = new JButton("Cancel");
            JPanel bpane = new JPanel();
            bpane.add(ok);
            bpane.add(cancel);
            bgroup.add(unrButton);
            bgroup.add(redButton);
            rpane.add(unrButton);
            rpane.add(redButton);
            cpane.add(choose);
            cpane.add(spinnern);
            if (!highn) spinnern.setEnabled(false);
            frame.add(cpane,BorderLayout.NORTH);
            frame.add(rpane,BorderLayout.CENTER);
            frame.add(bpane,BorderLayout.SOUTH);
            spinnern.addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent ae) {
                    if (modeln.getNumber().intValue() > 2) {
                        unrButton.setSelected(true);
                        redButton.setSelected(false);
                        unrButton.setEnabled(false);
                        redButton.setEnabled(false);
                    }
                    else {
                        unrButton.setEnabled(true);
                        redButton.setEnabled(true);
                    }
                }
            });
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    frame.setVisible(false);
                }
            });
            ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    value = modeln.getNumber().intValue();
                    if (redButton.isSelected()) reduced = true;
                    frame.setVisible(false);
                }
            });
            frame.setVisible(true);
        }
    }
    
    class CFrame {
        int value = -1;
        
        CFrame(String label, JFrame fram) {
            JDialog frame = new JDialog(new JFrame(), label, true);
            frame.setLayout(new BorderLayout());
            JRadioButton unrButton = new JRadioButton("unreduced",true);
            JRadioButton redButton = new JRadioButton("reduced");
            final ButtonGroup bgroup = new ButtonGroup();
            bgroup.add(unrButton);
            bgroup.add(redButton);
            JButton ok = new JButton("OK");
            JButton cancel = new JButton("Cancel");
            JPanel pane = new JPanel();
            JPanel panf = new JPanel();
            pane.add(unrButton);
            pane.add(redButton);
            panf.add(ok);
            panf.add(cancel);
            frame.add(pane,BorderLayout.CENTER);
            frame.add(panf,BorderLayout.SOUTH);
            frame.setBounds(fram.getX()+250, fram.getY()+240,300,120);
            frame.setResizable(false);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    frame.setVisible(false);
                }
            });
            ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    if (redButton.isSelected()) value = 0;
                    else value = 1;
                    frame.setVisible(false);
                }
            });
            frame.setVisible(true);
        }
    }
    
    class LinkData {
        JLabel name;
        JLabel comments;
        link Link;
        JPanel panel;
        JPanel compwrith;
        Container diag;
        JScrollPane Scroller;
        JButton khovHom = new JButton("Khovanov Cohomology");
        JButton slnHom = new JButton("sl(n)-Cohomology, n>2");
        JButton stHtpyTp = new JButton("Stable Homotopy Info");
        JButton jonesPol = new JButton("Jones Polynomial");
        JButton homflypt = new JButton("HOMFLYPT Polyomial");
        JButton dtcode = new JButton("DT/Gauss-Code");
        
        LinkData(String str, link Lnk, String cmmnts,JFrame frame) {
            name = new JLabel(str,SwingConstants.CENTER);
            Link = Lnk;
            comments = new JLabel(cmmnts);
            panel = new JPanel(new GridLayout(5,1));
            diag = Box.createHorizontalBox();
            panel.add(name);
            DiagramComp dude = new DiagramComp(Link.crossings,Link.paths);
            JLabel lab = new JLabel(" ");
            lab.setLayout(new BorderLayout());
            Dimension pref = new Dimension(80*Link.crossings.length,30);
            lab.setPreferredSize(pref);
            lab.add(dude,BorderLayout.CENTER);
            Scroller = new JScrollPane(lab);
            panel.add(Scroller);
            compwrith = setComponentPanel(Link);
            panel.add(compwrith);
            panel.add(comments);
            JPanel buttons = new JPanel(new GridLayout(3,3));
            JButton editDiag = new JButton("Edit Diagram");
            JButton mirrLink = new JButton("Mirror Link");
            JButton compChan = new JButton("Components");
            checkButtons(Link);
            khovHom.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    CFrame fram = new CFrame("Options",frame);
                    if (fram.value == 1) calcKhovHom(frame,Link,"Khovanov Cohomology",str,2);
                    if (fram.value == 0) calcKhovHom(frame,Link,"Khovanov Cohomology",str,1);
                }
            });
            slnHom.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    int n = getTheN(frame,"sl(n)-Cohomology",3);
                    if (n > 0) calcKhovHom(frame,Link,"sl("+n+")-Cohomology",str,n);
                }
            });
            stHtpyTp.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    CSFrame fram = new CSFrame("Options",frame,Link.isMatched());
                    if (fram.value > 0) getStableInfo(frame,Link,str,fram.value,fram.reduced);
                }
            });
            homflypt.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    twoPolynomial homfl = obtainHomFly(frame,Link);
                    JFrame fram = new JFrame("HOMFLYPT-Polynomial of "+name.getText());
                    fram.setLayout(new GridLayout(3,1));
                    fram.setBounds(frame.getX()+40, frame.getY()+200, 720, 200);
                    JLabel labe = new JLabel("HOMFLYPT : ", SwingConstants.RIGHT);
                    JLabel labf = new JLabel(homfl.getText(false), SwingConstants.LEFT);
                    JScrollPane scroller = new JScrollPane(labf);
                    scroller.setPreferredSize(new Dimension(600,40));
                    JPanel pane = new JPanel();
                    pane.add(labe);
                    pane.add(scroller);
                    JButton copy = new JButton("Copy");
                    JCheckBox latex = new JCheckBox("LaTeX");
                    JPanel panf = new JPanel();
                    panf.add(copy);
                    panf.add(latex);
                    JButton close = new JButton("Close");
                    JPanel pani = new JPanel();
                    pani.add(close);
                    close.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            fram.setVisible(false);
                        }
                    });
                    copy.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            boolean status = false;
                            if (latex.isSelected()) status = true;
                            StringSelection stringSelection = new StringSelection(homfl.getText(status));
                            Clipboard clpbrd = Toolkit.getDefaultToolkit().getSystemClipboard();
                            clpbrd.setContents(stringSelection, null);
                        }
                    });
                    latex.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            if (latex.isSelected()) labf.setText(homfl.getText(true));
                            else labf.setText(homfl.getText(false));
                        }
                    });
                    fram.add(pane);
                    fram.add(panf);
                    fram.add(pani);
                    fram.setResizable(false);
                    fram.setVisible(true);
                }
            });
            jonesPol.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    polynomial jones = obtainJonesPol(frame,Link);
                    if (jones != null) {
                        polynomial unjones = obtainUnJonesPol(jones);
                        JFrame fram = new JFrame("Jones Polynomial of "+name.getText());
                        fram.setLayout(new GridLayout(5,1));
                        fram.setBounds(frame.getX()+40, frame.getY()+180,720,260);
                        JLabel laba = new JLabel("    reduced : ", SwingConstants.RIGHT);
                        JLabel labb = new JLabel(jones.getJonesPol(true,false), SwingConstants.LEFT);
                        JLabel labc = new JLabel("unreduced : ", SwingConstants.RIGHT);
                        JLabel labd = new JLabel(unjones.getJonesPol(false,false), SwingConstants.LEFT);
                        JScrollPane scroller = new JScrollPane(labb);
                        scroller.setPreferredSize(new Dimension(600,40));
                        JScrollPane scrolleu = new JScrollPane(labd);
                        scrolleu.setPreferredSize(new Dimension(600,40));
                        JPanel pane = new JPanel();
                        pane.add(laba);
                        pane.add(scroller);
                        JButton copyRed = new JButton("Copy");
                        JCheckBox latexRed = new JCheckBox("LaTeX");
                        JPanel panf = new JPanel();
                        panf.add(copyRed);
                        panf.add(latexRed);
                        JPanel pang = new JPanel();
                        pang.add(labc);
                        pang.add(scrolleu);
                        JPanel panh = new JPanel();
                        JButton copyUnr = new JButton("Copy");
                        JCheckBox latexUnr = new JCheckBox("LaTeX");
                        latexUnr.setSelected(false);
                        panh.add(copyUnr);
                        panh.add(latexUnr);
                        JButton close = new JButton("Close");
                        JPanel pani = new JPanel();
                        pani.add(close);
                        close.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent ae) {
                                fram.setVisible(false);
                            }
                        });
                        copyUnr.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent ae) {
                                boolean status = false;
                                if (latexUnr.isSelected()) status = true;
                                StringSelection stringSelection = new StringSelection(unjones.getJonesPol(false,status));
                                Clipboard clpbrd = Toolkit.getDefaultToolkit().getSystemClipboard();
                                clpbrd.setContents(stringSelection, null);
                            }
                        });
                        latexUnr.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent ae) {
                                if (latexUnr.isSelected()) labd.setText(unjones.getJonesPol(false,true));
                                else labd.setText(unjones.getJonesPol(false,false));
                            }
                        });
                        copyRed.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent ae) {
                                boolean status = false;
                                if (latexRed.isSelected()) status = true;
                                StringSelection stringSelection = new StringSelection(jones.getJonesPol(true,status));
                                Clipboard clpbrd = Toolkit.getDefaultToolkit().getSystemClipboard();
                                clpbrd.setContents(stringSelection, null);
                            }
                        });
                        latexRed.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent ae) {
                                if (latexRed.isSelected()) labb.setText(jones.getJonesPol(true,true));
                                else labb.setText(jones.getJonesPol(true,false));
                            }
                        });
                        fram.add(pane);
                        fram.add(panf);
                        fram.add(pang);
                        fram.add(panh);
                        fram.add(pani);
                        close.requestFocusInWindow();
                        fram.setResizable(false);
                        fram.setVisible(true);
                    }
                    dtcode.requestFocusInWindow();
                    jonesPol.requestFocusInWindow();
                }
            });
            dtcode.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    String gauss = GaussCode(Link);
                    String dowker = DowkerThistle(Link);
                    JFrame fram = new JFrame("DT/Gauss-Code of "+name.getText());
                    fram.setLayout(new GridLayout(3,1));
                    fram.setBounds(frame.getX()+60, frame.getY()+200,680,200);
                    JLabel laba = new JLabel("    DT-Code : ", SwingConstants.RIGHT);
                    JLabel labb = new JLabel(dowker, SwingConstants.LEFT);
                    JLabel labc = new JLabel("Gauss-Code : ", SwingConstants.RIGHT);
                    JLabel labd = new JLabel(gauss, SwingConstants.LEFT);
                    JScrollPane scroller = new JScrollPane(labb);
                    scroller.setPreferredSize(new Dimension(560,40));
                    JScrollPane scrolleu = new JScrollPane(labd);
                    scrolleu.setPreferredSize(new Dimension(560,40));
                    JPanel pane = new JPanel();
                    pane.add(laba);
                    pane.add(scroller);
                    JButton copyDT = new JButton("Copy DT-Code");
                    JPanel panf = new JPanel();
                    panf.add(copyDT);
                    JPanel pang = new JPanel();
                    pang.add(labc);
                    pang.add(scrolleu);
                    JButton copyGa = new JButton("Copy Gauss-Code");
                    panf.add(copyGa);
                    JButton close = new JButton("Close");
                    panf.add(close);
                    close.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            fram.setVisible(false);
                        }
                    });
                    copyDT.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            StringSelection stringSelection = new StringSelection(dowker);
                            Clipboard clpbrd = Toolkit.getDefaultToolkit().getSystemClipboard();
                            clpbrd.setContents(stringSelection, null);
                        }
                    });
                    copyGa.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent ae) {
                            StringSelection stringSelection = new StringSelection(gauss);
                            Clipboard clpbrd = Toolkit.getDefaultToolkit().getSystemClipboard();
                            clpbrd.setContents(stringSelection, null);
                        }
                    });
                    fram.add(pane);
                    fram.add(pang);
                    fram.add(panf);
                    fram.setResizable(false);
                    fram.setVisible(true);
                }
            });
            compChan.addActionListener(new ActionListener(){
                public void actionPerformed(ActionEvent ae) {
                    CCFrame compFrame = new CCFrame(frame, Link);
                    if (compFrame.ok) {
                        Link = compFrame.Lnk;
                        LinkData Knot = new LinkData(name.getText(),Link,comments.getText(),frame);
                        comments.setText("Bizarro");
                        panel.remove(Scroller);
                        Scroller = Knot.Scroller;
                        panel.add(Scroller,1);
                        panel.remove(compwrith);
                        compwrith = setComponentPanel(Link);
                        panel.add(compwrith,2);
                        comments.setText(Knot.comments.getText());
                        if (Link.isMatched()) {
                            homflypt.setEnabled(true);
                            slnHom.setEnabled(true);
                        }
                        else {
                            homflypt.setEnabled(false);
                            slnHom.setEnabled(false);
                        }
                    }
                }
            });
            mirrLink.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    Link = Mirror(Link);
                    LinkData Knot = new LinkData(name.getText(),Link,comments.getText(),frame);
                    comments.setText("Bizarro");
                    panel.remove(Scroller);
                    Scroller = Knot.Scroller;
                    panel.add(Scroller,1);
                    panel.remove(compwrith);
                    compwrith = setComponentPanel(Link);
                    panel.add(compwrith,2);
                    comments.setText(Knot.comments.getText());
                }
            });
            
            editDiag.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    LinkData Knott = enterGluedDiagram(frame,name.getText(),comments.getText(),Link.crossings,Link.paths);
                    if (Knott != null) {
                        name.setText(Knott.name.getText());
                        comments.setText(Knott.comments.getText()+" "); // This is bizarre, but without it and no text change the diagram doesn;t get redrawn
                        Link = Knott.Link;
                        int i = list.getSelectedIndex();
                        listModel.remove(i);
                        listModel.add(i, name.getText());
                        list.setSelectedIndex(i);
                        panel.remove(Scroller);
                        Scroller = Knott.Scroller;
                        panel.add(Scroller,1);
                        panel.remove(compwrith);
                        compwrith = setComponentPanel(Link);
                        panel.add(compwrith,2);
                        comments.setText(Knott.comments.getText());
                        checkButtons(Link);
                    }
                }
            });
            buttons.add(khovHom);
            buttons.add(slnHom);
            buttons.add(stHtpyTp);
            buttons.add(jonesPol);
            buttons.add(homflypt);
            buttons.add(dtcode);
            buttons.add(editDiag);
            buttons.add(mirrLink);
            buttons.add(compChan);
            panel.add(buttons);
        }
        
        void checkButtons(link Link) {
            if (!Link.isMatched()) {
                slnHom.setEnabled(false);
                homflypt.setEnabled(false);
            }
            else {
                slnHom.setEnabled(true);
                homflypt.setEnabled(true);
            }
            if (Link.components.isEmpty()|!Link.isPlanar()) {
                khovHom.setEnabled(false);
                slnHom.setEnabled(false);
                homflypt.setEnabled(false);
                stHtpyTp.setEnabled(false);
                jonesPol.setEnabled(false);
            }
            else {
                khovHom.setEnabled(true);
                stHtpyTp.setEnabled(true);
                jonesPol.setEnabled(true);
            }
        }
    }
    
    int value;
    
    int getTheN(JFrame frame, String label, int low) {
        value = -1;
        JDialog fram = new JDialog(new JFrame(), label, true);
        fram.setBounds(frame.getX()+250, frame.getY()+240,300,120);
        JLabel labl = new JLabel("Please choose n : ", SwingConstants.RIGHT);
        JPanel pane = new JPanel();
        JPanel panf = new JPanel();
        JPanel empty = new JPanel();
        JButton ok = new JButton("OK");
        JButton cancel = new JButton("Cancel");
        pane.add(labl);
        SpinnerNumberModel modeln = new SpinnerNumberModel(low,low,100,1);
        JSpinner spinnern = new JSpinner(modeln);
        pane.add(spinnern);
        panf.add(ok);
        panf.add(cancel);
        fram.setLayout(new BorderLayout());
        fram.add(empty, BorderLayout.NORTH);
        fram.add(pane, BorderLayout.CENTER);
        fram.add(panf, BorderLayout.SOUTH);
        fram.setResizable(false);
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                value = modeln.getNumber().intValue();
                fram.setVisible(false);
            }
        });
        fram.setVisible(true);
        return value;
    }
    
    void getStableInfo(JFrame frame, link Link, String name, int n, boolean reduced) {
        String title = "Khovanov Cohomology";
        if (n > 2) title = "sl("+n+")-Cohomology";
        JDialog fram = new JDialog(new JFrame(),title,true);
        fram.setBounds(frame.getX()+250, frame.getY()+225,300, 150);
        JLabel labe = new JLabel("Crossing : ",SwingConstants.RIGHT);
        JLabel labf = new JLabel("Objects : ",SwingConstants.RIGHT);
        JLabel labg = new JLabel(" ", SwingConstants.LEFT);
        JLabel labh = new JLabel(" ", SwingConstants.LEFT);
        JPanel pane = new JPanel(new GridLayout(2,2));
        JPanel panf = new JPanel();
        JButton cancel = new JButton("Cancel");
        pane.add(labe);
        pane.add(labg);
        pane.add(labf);
        pane.add(labh);
        panf.add(cancel);
        fram.setLayout(new BorderLayout());
        fram.add(pane, BorderLayout.CENTER);
        fram.add(panf, BorderLayout.SOUTH);
        fram.setResizable(false);
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        KhovCalc calculate = new KhovCalc(fram,labe,labg,labh,Link,n,reduced);
        calculate.start();
        fram.setVisible(true);
        calculate.runs = false;
        if (calculate.finished) new StableStuff(frame,calculate.KhovHom, Link, "Stable Homotopy Info of "+name,n,reduced);
    }
    
    class gridCombo {
        ArrayList<ArrayList<Integer>> posit;
        ArrayList<m4_object> objects;
        
        gridCombo(ArrayList<ArrayList<Integer>> pos, ArrayList<m4_object> objs) {
            posit = pos;
            objects = objs;
        }
    }
    
    gridCombo theGrid(int level, BHComps objects, int minh) {
        ArrayList<ArrayList<Integer>> posit = new ArrayList<ArrayList<Integer>>();
        for (int i = 0; i < level; i++) {
            ArrayList<Integer> posli = new ArrayList<Integer>();
            posli.add(-1);
            posit.add(posli);
        }
        ArrayList<m4_object> objs = new ArrayList<m4_object>();
        for (BHComponent comp : objects.comps) {
            int lv = comp.objects.get(0).hdeg - minh;
            boolean found = false;
            int i = 0;
            int a = -1;
            int b = -1;
            int c = -1;
            int d = -1;
            if (comp.type == -2 | comp.type == 1) {
                lv--;
                while (!found & i < posit.get(lv).size()) {
                    if ((posit.get(lv).get(i) == -1) & posit.get(lv+1).get(i) == -1) found = true;
                    else i++;
                }
                if (!found) increase(posit,1);
                m4_object bObj = comp.objects.get(1);
                m4_object tObj = comp.objects.get(0);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+1).set(i,tObj.number);
                objs.add(bObj);
                objs.add(tObj);
            }
            if (comp.type == -1 | comp.type == 0) {
                while (!found & i < posit.get(lv).size()) {
                    if (posit.get(lv).get(i) == -1) found = true;
                    else i++;
                }
                if (!found) increase(posit,1);
                m4_object obj = comp.objects.get(0);
                posit.get(lv).set(i,obj.number);
                objs.add(obj);
            }
            if (comp.type == 2) {
                while (!found & i < posit.get(lv).size()-1) {
                    if (posit.get(lv).get(i) == -1 & posit.get(lv+2).get(i+1) == -1) found = true;
                    else i++;
                }
                if (!found) {
                    if (posit.get(lv).get(i) == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(0);
                m4_object tObj = comp.objects.get(1);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                objs.add(bObj);
                objs.add(tObj);
            }
            if (comp.type == 3) {
                while (!found & i < posit.get(lv).size()-1) {
                    a = posit.get(lv).get(i);
                    b = posit.get(lv+2).get(i+1);
                    c = posit.get(lv+1).get(i+1);
                    if (a == -1 & b == -1 & c == -1) found = true;
                    else i++;
                }
                if (!found) {
                    a = posit.get(lv).get(i);
                    if (a == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(0);
                m4_object tObj = comp.objects.get(1);
                m4_object bobj = comp.objects.get(2);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                posit.get(lv+1).set(i+1,bobj.number);
                objs.add(bObj);
                objs.add(tObj);
                objs.add(bobj);
            }
            if (comp.type == 4) {
                lv--;
                while (!found & i < posit.get(lv).size()-1) {
                    a = posit.get(lv).get(i);
                    b = posit.get(lv+2).get(i+1);
                    c = posit.get(lv+1).get(i);
                    if (a == -1 & b == -1 & c == -1) found = true;
                    else i++;
                }
                if (!found) {
                    a = posit.get(lv).get(i);
                    c = posit.get(lv+1).get(i);
                    if (a == -1 & c == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(1);
                m4_object tObj = comp.objects.get(2);
                m4_object bobj = comp.objects.get(0);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                posit.get(lv+1).set(i,bobj.number);
                objs.add(bObj);
                objs.add(tObj);
                objs.add(bobj);
            }
            if (comp.type == 5) {
                lv--;
                while (!found & i < posit.get(lv).size()-1) {
                    a = posit.get(lv).get(i);
                    b = posit.get(lv+2).get(i+1);
                    c = posit.get(lv+1).get(i);
                    d = posit.get(lv+1).get(i+1);
                    if (a == -1 & b == -1 & c == -1 & d == -1) found = true;
                    else i++;
                }
                if (!found) {
                    a = posit.get(lv).get(i);
                    c = posit.get(lv+1).get(i);
                    if (a == -1 & c == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(1);
                m4_object tObj = comp.objects.get(2);
                m4_object bobj = comp.objects.get(3);
                m4_object tobj = comp.objects.get(0);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                posit.get(lv+1).set(i+1,bobj.number);
                posit.get(lv+1).set(i,tobj.number);
                objs.add(bObj);
                objs.add(tObj);
                objs.add(bobj);
                objs.add(tobj);
            }
            if (comp.type == 6) {
                while (!found & i < posit.get(lv).size()-1) {
                    a = posit.get(lv).get(i);
                    b = posit.get(lv+2).get(i+1);
                    c = posit.get(lv+3).get(i+1);
                    if (a == -1 & b == -1 & c == -1) found = true;
                    else i++;
                }
                if (!found) {
                    a = posit.get(lv).get(i);
                    if (a == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(0);
                m4_object tObj = comp.objects.get(1);
                m4_object TObj = comp.objects.get(2);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                posit.get(lv+3).set(i+1,TObj.number);
                objs.add(bObj);
                objs.add(tObj);
                objs.add(TObj);
            }
            if (comp.type == 7) {
                lv--;
                while (!found & i < posit.get(lv).size()-1) {
                    a = posit.get(lv).get(i);
                    b = posit.get(lv+2).get(i+1);
                    c = posit.get(lv+1).get(i);
                    d = posit.get(lv+3).get(i+1);
                    if (a == -1 & b == -1 & c == -1 & d == -1) found = true;
                    else i++;
                }
                if (!found) {
                    a = posit.get(lv).get(i);
                    c = posit.get(lv+1).get(i);
                    if (a == -1 & c == -1) increase(posit,1);
                    else {
                        increase(posit,2);
                        i++;
                    }
                }
                m4_object bObj = comp.objects.get(1);
                m4_object tObj = comp.objects.get(2);
                m4_object TObj = comp.objects.get(3);
                m4_object tobj = comp.objects.get(0);
                posit.get(lv).set(i,bObj.number);
                posit.get(lv+2).set(i+1,tObj.number);
                posit.get(lv+3).set(i+1,TObj.number);
                posit.get(lv+1).set(i,tobj.number);
                objs.add(bObj);
                objs.add(tObj);
                objs.add(TObj);
                objs.add(tobj);
            }
        }
        gridCombo theCombo = new gridCombo(posit,objs);
        return theCombo;
    }
    
    void increase(ArrayList<ArrayList<Integer>> list, int u) {
        for (ArrayList<Integer> lis : list) {
            for (int i = 0; i < u; i++) lis.add(-1);
        }
    }
    
    @SuppressWarnings("serial")
    class picture extends JComponent {
        BHComps objects;
        boolean numbers;
        int level;
        int minh;
        
        public picture(BHComps comps, int lvl, int min, boolean bool) {
            objects = comps;
            numbers = bool;
            level = lvl;
            minh = min;
        }
        
        public void paint(Graphics g) {
            Graphics2D g2 = (Graphics2D) g;
            g2.setFont(new Font("Times New Roman", Font.BOLD,16));
            gridCombo combo = theGrid(level,objects,minh);
            ArrayList<ArrayList<Integer>> posit = combo.posit;
            ArrayList<m4_object> objs = combo.objects;
            int clx = posit.get(0).size() - 1;
            int cly = posit.size() - 1;
            int chgy = 40;
            int chgx = 40;
            int startx = 240 - (chgx/2) * clx;
            if (clx > 12) {
                chgx = 475/clx;
                startx = 3;
            }
            int starty = 130 + (chgy/2) * cly;
            g2.setPaint(Color.blue);
            g2.setStroke(new BasicStroke(4));
            for (m4_object obj : objs) {
                if (!obj.moduli1b.isEmpty()) {
                    int lv = obj.hdeg - minh;
                    boolean found = false;
                    int j = 0;
                    while (!found) {
                        if (posit.get(lv).get(j) == obj.number) found = true;
                        else j++;
                    }
                    int height = starty - chgy * lv;
                    int width = startx + chgx * j;
                    for (m4mod1 mod : obj.moduli1b) {
                        m4_object tObj = mod.topObj;
                        found = false;
                        j = 0;
                        while (!found) {
                            if (posit.get(lv+2).get(j) == tObj.number) found = true;
                            else j++;
                        }
                        int heit = starty - chgy * (lv+2);
                        int widt = startx + chgx * j;
                        g2.drawLine(width+7, height+7, widt+7, heit+7);
                    }
                }
            }
            g2.setPaint(Color.black);
            int height = starty + chgy;
            int width = startx - chgx;
            for (int u = 0; u < level; u++) {
                height = height - chgy;
                width = startx - chgx;
                for (int y = 0; y < posit.get(u).size(); y++) {
                    width = width + chgx;
                    int t = posit.get(u).get(y);
                    if (t != -1) {
                        g2.fillOval(width, height, 14, 14);
                    }
                }
            }
            for (m4_object obj : objs) {
                if (!obj.moduli0b.isEmpty()) {
                    int lv = obj.hdeg - minh;
                    boolean found = false;
                    int j = 0;
                    while (!found) {
                        if (posit.get(lv).get(j) == obj.number) found = true;
                        else j++;
                    }
                    height = starty - chgy * lv;
                    width = startx + chgx * j;
                    g2.drawLine(width+7, height+7, width+7, height+7-chgy);
                    if (numbers) g2.drawString(String.valueOf(obj.moduli0b.get(0).value),width-chgx/10,height-chgy/5);
                }
            }
        }
    }
    
    class StableStuff {
        
        ArrayList<BHComps> theTypes;
        
        boolean alright;
        
        public StableStuff(JFrame frame, ArrayList<homology> knotHom, link Link, String title, int n, boolean red) {
            theTypes = new ArrayList<BHComps>();
            JFrame fram = new JFrame(title);
            fram.setBounds(frame.getX()+100, frame.getY()+100,600,400);
            fram.setResizable(false);
            fram.setLayout(new BorderLayout());
            JPanel paneList = new JPanel(new BorderLayout());
            JLabel labeList = new JLabel("Quantum degrees", SwingConstants.CENTER);
            labeList.setPreferredSize(new Dimension(120,30));
            Card StabCards = new Card();
            ArrayList<JPanel> StabPanels = new ArrayList<JPanel>();
            int u = 0;
            while (u < knotHom.size()) {
                homology hom = knotHom.get(u);
                for (homologyGroup group : hom.groups) {
                    int e = 0;
                    while (e < group.Torsion.size()) {
                        int t = group.Torsion.get(e);
                        if (t%2 != 0) group.Torsion.remove(e);
                        else e++;
                    }
                }
                int max = 0;
                int min = 0;
                boolean setmax = false;
                boolean setmin = false;
                for (int i = 0; i < hom.degrees.size();i++) {
                    int deg = hom.degrees.get(i);
                    boolean nontriv = false;
                    if (hom.groups.get(i).Betti>0 | !hom.groups.get(i).Torsion.isEmpty()) nontriv = true;
                    if ((deg > max | !setmax) & nontriv) {
                        max = deg;
                        if (!setmax) min = deg;
                        setmax = true;
                    }
                    if ((deg <= min | !setmin) & nontriv) {
                        min = deg;
                        setmin = true;
                        boolean extra = false;
                        int y = 0;
                        while (y < hom.groups.get(i).Torsion.size() & !extra) {
                            int t = hom.groups.get(i).Torsion.get(y);
                            if (t % 2 == 0) extra = true;
                            else y++;
                        }
                        if (extra) min--;
                    }
                }
                if (max - min < 2) knotHom.remove(u);
                else u++;
            }
            String[] items = new String[knotHom.size()];
            int m = knotHom.size()-1;
            for (int i = m; i >= 0; i--) {
                items[m-i] = "q = "+knotHom.get(i).quantum;
                JPanel stPanel1 = addstableInfo(knotHom.get(i).quantum,n,red,Link,fram,true);
                JPanel stPanel2 = addstableInfo(knotHom.get(i).quantum,n,red,Link,fram,false);
                StabPanels.add(stPanel1);
                StabPanels.add(stPanel2);
                StabCards.add(stPanel1);
                StabCards.add(stPanel2);
            }
            JList<String> quantumList = new JList<>(items);
            JPanel buttonPane = new JPanel();
            JButton latexBut = new JButton("Save as LaTeX file");
            JButton closeBut = new JButton("Close");
            closeBut.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            latexBut.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    saveStabType(fram,title,n,red,Link);
                }
            });
            buttonPane.add(latexBut);
            buttonPane.add(closeBut);
            JScrollPane scrollPane = new JScrollPane(quantumList);
            scrollPane.setPreferredSize(new Dimension(100,300));
            paneList.add(scrollPane);
            JPanel middlePane = new JPanel(new BorderLayout());
            middlePane.add(StabCards, BorderLayout.CENTER);
            JLabel labe = new JLabel(" ");
            middlePane.add(labe, BorderLayout.NORTH);
            final ButtonGroup options = new ButtonGroup();
            JRadioButton stabType = new JRadioButton("Stable Type",true);
            JPanel optionPanel = new JPanel();
            optionPanel.add(stabType);
            stabType.setActionCommand("stabT");
            options.add(stabType);
            JRadioButton SteenAlg = new JRadioButton("Steenrod Action");
            optionPanel.add(SteenAlg);
            SteenAlg.setActionCommand("Steen");
            options.add(SteenAlg);
            middlePane.add(optionPanel, BorderLayout.SOUTH);
            quantumList.setSelectedIndex(0);
            quantumList.addListSelectionListener(new ListSelectionListener() {
                @Override
                public void valueChanged(ListSelectionEvent le) {
                    @SuppressWarnings("unchecked")
                    JList<String> list = (JList<String>) le.getSource();
                    int u = list.getSelectedIndex();
                    if ( fram.isActive() & u >= 0 ) {
                        for (JPanel pan : StabPanels) pan.setVisible(false);
                        String entree = options.getSelection().getActionCommand();
                        quantumList.setSelectedIndex(u);
                        if ("stabT".equals(entree)) u = 2 * u;
                        else  u = 2 * u + 1;
                        StabPanels.get(u).setVisible(true);
                    }
                }
            });
            stabType.addItemListener(new ItemListener() {
                @Override
                public void itemStateChanged(ItemEvent ae) {
                    boolean found = false;
                    int u = 0;
                    while (!found) {
                        JPanel pan = StabPanels.get(u);
                        if (pan.isVisible()) found = true;
                        else u++;
                    }
                    if (u % 2 != 0) {
                        StabPanels.get(u).setVisible(false);
                        u--;
                        StabPanels.get(u).setVisible(true);
                    }
                }
            });
            SteenAlg.addItemListener(new ItemListener() {
                @Override
                public void itemStateChanged(ItemEvent ae) {
                    boolean found = false;
                    int u = 0;
                    while (!found) {
                        JPanel pan = StabPanels.get(u);
                        if (pan.isVisible()) found = true;
                        else u++;
                    }
                    if (u % 2 == 0) {
                        StabPanels.get(u).setVisible(false);
                        u++;
                        StabPanels.get(u).setVisible(true);
                    }
                }
            });
            if (knotHom.isEmpty()) {
                latexBut.setEnabled(false);
                stabType.setEnabled(false);
                SteenAlg.setEnabled(false);
            }
            fram.add(middlePane, BorderLayout.CENTER);
            fram.add(paneList, BorderLayout.WEST);
            fram.add(buttonPane, BorderLayout.SOUTH);
            fram.setVisible(true);
        }

        void saveStabType(JFrame frame, String title, int n, boolean red, link Link) {
            JDialog fram = new JDialog(new JFrame(),"Options", true);
            fram.setBounds(frame.getX()+70, frame.getY()+50, 460, 300);
            fram.setResizable(false);
            fram.setLayout(new BorderLayout());
            alright = false;
            JPanel paneList1 = new JPanel(new BorderLayout());
            JPanel paneList2 = new JPanel(new BorderLayout());
            Collections.sort(theTypes);
            String[] items = new String[theTypes.size()];
            for (int i = 0; i < theTypes.size(); i++) {
                BHComps comps = theTypes.get(i);
                items[i] = "q = "+comps.qdeg+", ";
                if (comps.stab) items[i] = items[i]+"Stable Type";
                else items[i] = items[i] + "Steenrod Algebra";
            }
            JList<String> quantumList = new JList<>(items);
            quantumList.setSelectedIndex(0);
            DefaultListModel<String> model = new DefaultListModel<>();
            JList<String> chosenList = new JList<String>(model);
            ArrayList<Integer> chosen = new ArrayList<Integer>();
            JScrollPane scrollPane1 = new JScrollPane(quantumList);
            JScrollPane scrollPane2 = new JScrollPane(chosenList);
            scrollPane1.setPreferredSize(new Dimension(100,100));
            scrollPane2.setPreferredSize(new Dimension(100,100));
            paneList1.add(scrollPane1);
            paneList2.add(scrollPane2);
            JPanel paneButton = new JPanel(new GridLayout(5,1));
            JPanel butAdd = new JPanel();
            JButton buttonAdd = new JButton(" --> ");
            butAdd.add(buttonAdd);
            JPanel butAll = new JPanel();
            JButton buttonAll = new JButton(" - All > ");
            butAll.add(buttonAll);
            JLabel empty1 = new JLabel("  ");
            JLabel empty2 = new JLabel("  ");
            JLabel empty3 = new JLabel("  ");
            JPanel butSub = new JPanel();
            JButton buttonSub = new JButton(" <-- ");
            butSub.add(buttonSub);
            paneButton.add(empty1);
            paneButton.add(butAdd);
            paneButton.add(butAll);
            paneButton.add(butSub);
            paneButton.add(empty3);
            JPanel middleStuff = new JPanel(new GridLayout(1,3));
            middleStuff.add(paneList1);
            middleStuff.add(paneButton);
            middleStuff.add(paneList2);
            JLabel topLine = new JLabel("Please select from the list", SwingConstants.CENTER);
            topLine.setPreferredSize(new Dimension(100,24));
            JButton cancel = new JButton("Cancel");
            JButton save = new JButton("Save");
            save.setEnabled(false);
            JPanel butPanel = new JPanel();
            JLabel emptyl = new JLabel("  ");
            JLabel emptyr = new JLabel("  ");
            butPanel.add(cancel);
            butPanel.add(save);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            save.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    alright = true;
                    fram.setVisible(false);
                }
            });
            buttonAll.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    for (int r = 0; r < theTypes.size(); r++) {
                        if (!chosen.contains(r)) {
                            chosen.add(r);
                            int i = model.size();
                            model.add(i,items[r]);
                        }
                    }
                    if (!model.isEmpty()) save.setEnabled(true);
                }
            });
            buttonAdd.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    for (int r : quantumList.getSelectedIndices()) {
                        if (!chosen.contains(r)) {
                            chosen.add(r);
                            int i = model.size();
                            model.add(i,items[r]);
                        }
                    }
                    if (!model.isEmpty()) save.setEnabled(true);
                }
            });
            buttonSub.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    int[] lists = chosenList.getSelectedIndices();
                    for (int r = lists.length-1; r >= 0; r--)  {
                        model.remove(lists[r]);
                        chosen.remove(lists[r]);
                    }
                    if (model.isEmpty()) save.setEnabled(false);
                }
            });
            fram.add(topLine, BorderLayout.NORTH);
            fram.add(middleStuff, BorderLayout.CENTER);
            fram.add(butPanel, BorderLayout.SOUTH);
            fram.add(emptyl, BorderLayout.WEST);
            fram.add(emptyr, BorderLayout.EAST);
            fram.setVisible(true);
            if (alright) {
                ArrayList<String> commands = new ArrayList<String>();
                commands.add("\\documentclass[border=1bp]{standalone}");
                commands.add("\\usepackage{tikz}");
                commands.add("\\begin{document}");
                commands.add("\\begin{tabular}{l}");
                for (int u : chosen) {
                    BHComps comps = theTypes.get(u);
                    String command;
                    if (comps.stab) command = "Stable type for $q="+comps.qdeg+"$\\\\";
                    else command = "Steenrod Action for $q="+comps.qdeg+"$\\\\";
                    commands.add(command);
                    commands.add("\\begin{tikzpicture}");
                    int hmin = minHom(comps);
                    gridCombo combo = theGrid(comps.level,comps,hmin);
                    ArrayList<ArrayList<Integer>> posit = combo.posit;
                    ArrayList<m4_object> objs = combo.objects;
                    for (int w = 0; w < comps.level; w++) {
                        String com = "\\node at (0,"+(1+w)+") {$"+(hmin+w)+"$};";
                        commands.add(com);
                    }
                    for (m4_object obj : objs) {
                        if (!obj.moduli1b.isEmpty()) {
                            boolean found = false;
                            int lv = obj.hdeg - hmin;
                            int j = 0;
                            while (!found) {
                                if (posit.get(lv).get(j) == obj.number) found = true;
                                else j++;
                            }
                            int xpos = j + 1;
                            for (m4mod1 mod : obj.moduli1b) {
                                m4_object tObj = mod.topObj;
                                found = false;
                                j = 0;
                                while (!found) {
                                    if (posit.get(lv+2).get(j) == tObj.number) found = true;
                                    else j++;
                                }
                                String com = "\\draw[-,blue,shorten >=2pt,shorten <=2pt] (";
                                com = com + xpos +","+(lv+1)+") -- ("+(j+1)+","+(lv+3)+");";
                                commands.add(com);
                            }
                        }
                    }
                    for (int v = 0; v < comps.level; v++) {
                        int xpos = 1;
                        int ypos = 1+v;
                        for (int w = 0; w < posit.get(v).size(); w++) {
                            xpos = 1+w;
                            if (posit.get(v).get(w) != -1) {
                                commands.add("\\node[circle, draw, inner sep=0pt, minimum size=4pt] at ("+xpos+","+ypos+") {$ $};");
                            }
                        }
                    }
                    for (m4_object obj : objs) {
                        if (!obj.moduli0b.isEmpty()) {
                            boolean found = false;
                            int lv = obj.hdeg - hmin;
                            int j = 0;
                            while (!found) {
                                if (posit.get(lv).get(j) == obj.number) found = true;
                                else j++;
                            }
                            int xpos = j + 1;
                            String com = "\\draw[-,shorten >=2pt,shorten <=2pt] (";
                            com = com + xpos +","+(lv+1)+") -- ("+ xpos +","+(lv+2)+");";
                            commands.add(com);
                            if (comps.stab) {
                                int val = obj.moduli0b.get(0).value;
                                String cmd = "\\node at ("+(xpos-1)+".9,"+(lv+1)+".5) {$"+val+"$};";
                                commands.add(cmd);
                            }
                        }
                    }
                    commands.add("\\end{tikzpicture}");
                    commands.add("\\\\");
                    commands.add("\\\\");
                }
                commands.add("\\end{tabular}");
                commands.add("\\end{document}");
                saveCommands("Stable_Info_"+title.substring(24), commands, frame);
            }
        }
        
        int minHom(BHComps comps) {
            int minh = comps.comps.get(0).objects.get(0).hdeg;
            for (BHComponent comp : comps.comps) {
                for (m4_object obj : comp.objects) {
                    if (obj.hdeg < minh) minh = obj.hdeg;
                }
            }
            return minh;
        }
        
        JPanel addstableInfo(int q,int n,boolean red,link Link,JFrame frame,boolean stab) {
            JPanel panelo = new JPanel();
            panelo.setLayout(new BorderLayout());
            JButton action = new JButton("Action q = "+q);
            JPanel panact = new JPanel();
            panact.add(action);
            JLabel empty = new JLabel(" ");
            empty.setPreferredSize(new Dimension(50,120));
            panelo.add(empty, BorderLayout.NORTH);
            action.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent ae) {
                    ArrayList<ArrayList<m4_object>> objects;
                    if (!stab) {
                        boolean found = false;
                        int i = 0;
                        while (i < theTypes.size() & !found) {
                            BHComps comp = theTypes.get(i);
                            if (comp.qdeg == q) found = true;
                            else i++;
                        }
                        if (found) objects = SteenFromType(theTypes.get(i));
                        else objects = calcSteenAlg(frame,Link,q,n,red);
                    }
                    else objects = calcStabType(frame,Link,q,n,red);
                    if (objects != null) {
                        int levels = objects.size();
                        int hmin = objects.get(0).get(0).hdeg;
                        BHComps components = BHCompOf(objects);
                        components.qdeg = q;
                        components.stab = stab;
                        components.level = levels;
                        theTypes.add(components);
                        picture pic = new picture(components,levels,hmin,stab);
                        action.setVisible(false);
                        panelo.remove(panact);
                        panelo.remove(empty);
                        panelo.add(pic,BorderLayout.CENTER);
                    }
                }
            });
            panelo.add(panact, BorderLayout.CENTER);
            return panelo;
        }

        ArrayList<ArrayList<m4_object>> SteenFromType(BHComps comps) {
            ArrayList<ArrayList<m4_object>> objects = new ArrayList<ArrayList<m4_object>>();
            int maxh = comps.comps.get(0).objects.get(0).hdeg;
            int minh = maxh;
            for (BHComponent comp : comps.comps) {
                for (m4_object obj : comp.objects) {
                    int h = obj.hdeg;
                    if (h > maxh) maxh = h;
                    if (h < minh) minh = h;
                }
            }
            for (int y = minh; y <= maxh; y++) {
                ArrayList<m4_object> objs = new ArrayList<m4_object>();
                objects.add(objs);
            }
            for (BHComponent comp : comps.comps) {
                boolean ok = true;
                if (comp.type == 1 | comp.type == -2) {
                    if (comp.objects.get(0).moduli0t.get(0).value % 2 != 0) ok = false;
                }
                if (ok) {
                    for (m4_object obj : comp.objects) {
                        int h = obj.hdeg - minh;
                        objects.get(h).add(obj);
                    }
                }
            }
            objects = clonem4Objects(objects);
            for (ArrayList<m4_object> objs : objects) {
                for (m4_object obj : objs) {
                    if (!obj.moduli0b.isEmpty()) {
                        m4mod0 mod = obj.moduli0b.get(0);
                        if (mod.value % 4 == 0) {
                            mod.botObj.moduli0b.remove(mod);
                            mod.topObj.moduli0t.remove(mod);
                        }
                    }
                }
            }
            return objects;
        }
        
        BHComps BHCompOf(ArrayList<ArrayList<m4_object>> objects) {
            BHComps comps = new BHComps();
            int u = ChangLevel(objects);
            boolean aboveChang = false;
            for (int j = 0; j < objects.size(); j++) {
                if (j > u) aboveChang = true;
                ArrayList<m4_object> objs = objects.get(j);
                while (!objs.isEmpty()) {
                    m4_object obj = objs.get(0);
                    objs.remove(obj);
                    BHComponent comp = new BHComponent();
                    comps.comps.add(comp);
                    if (!aboveChang) {
                        if (obj.moduli0b.isEmpty()) {
                            comp.objects.add(obj);
                            if (obj.moduli1b.isEmpty()) {
                                comp.type = 0;
                            }
                            else {
                                m4_object tobj = obj.moduli1b.get(0).topObj;
                                objects.get(j+2).remove(tobj);
                                comp.objects.add(tobj);
                                if (tobj.moduli0t.isEmpty()) {
                                    if (tobj.moduli0b.isEmpty()) comp.type = 2;
                                    else {
                                        m4_object tObj = tobj.moduli0b.get(0).topObj;
                                        objects.get(j+3).remove(tObj);
                                        comp.objects.add(tObj);
                                        comp.type = 6;
                                    }
                                }
                                else {
                                    m4_object bobj = tobj.moduli0t.get(0).botObj;
                                    objects.get(j+1).remove(bobj);
                                    comp.objects.add(bobj);
                                    comp.type = 3;
                                }
                            }
                        }
                        else {
                            m4_object tobj = obj.moduli0b.get(0).topObj;
                            objects.get(j+1).remove(tobj);
                            comp.objects.add(tobj);
                            comp.objects.add(obj);
                            if (obj.moduli1b.isEmpty()) comp.type = 1;
                            else {
                                m4_object tObj = obj.moduli1b.get(0).topObj;
                                objects.get(j+2).remove(tObj);
                                comp.objects.add(tObj);
                                if (tObj.moduli0t.isEmpty()) {
                                    if (tObj.moduli0b.isEmpty()) comp.type = 4;
                                    else {
                                        m4_object TObj = tObj.moduli0b.get(0).topObj;
                                        objects.get(j+3).remove(TObj);
                                        comp.objects.add(TObj);
                                        comp.type = 7;
                                    }
                                }
                                else {
                                    m4_object bObj = tObj.moduli0t.get(0).botObj;
                                    objects.get(j+1).remove(bObj);
                                    comp.objects.add(bObj);
                                    comp.type = 5;
                                }
                            }
                        }
                    }
                    else {
                        if (obj.moduli0b.isEmpty()) comp.type = -1;
                        else {
                            m4_object tobj = obj.moduli0b.get(0).topObj;
                            objects.get(j+1).remove(tobj);
                            comp.objects.add(tobj);
                            comp.type = -2;
                        }
                        comp.objects.add(obj);
                    }
                }
            }
            return comps;
        }
        
        ArrayList<ArrayList<m4_object>> calcStabType(JFrame frame,link Link, int q, int n, boolean red) {
            JDialog fram = new JDialog(new JFrame(), "Stable Homotopy Calculation", true);
            fram.setBounds(frame.getX()+50,frame.getY()+100,500,200);
            JLabel task = new JLabel("Creating Objects", SwingConstants.CENTER);
            JLabel objs = new JLabel("Objects : ", SwingConstants.RIGHT);
            JLabel objn = new JLabel(" ", SwingConstants.LEFT);
            JLabel levl = new JLabel("Level : ", SwingConstants.RIGHT);
            JLabel levn = new JLabel(" ", SwingConstants.LEFT);
            JLabel dims = new JLabel("Dimension : ", SwingConstants.RIGHT);
            JLabel dimn = new JLabel(" ", SwingConstants.LEFT);
            JPanel pand = new JPanel();
            JPanel pane = new JPanel(new GridLayout(3,2));
            JPanel panf = new JPanel();
            JButton cancel = new JButton("Cancel");
            pand.add(task);
            pane.add(objs);
            pane.add(objn);
            pane.add(levl);
            pane.add(levn);
            pane.add(dims);
            pane.add(dimn);
            panf.add(cancel);
            fram.setLayout(new BorderLayout());
            fram.add(pand, BorderLayout.NORTH);
            fram.add(pane, BorderLayout.CENTER);
            fram.add(panf, BorderLayout.SOUTH);
            fram.setResizable(false);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            StTyCalc calculate = new StTyCalc(fram,task,objn,levn,dims,dimn,Link,q,n,red);
            calculate.start();
            fram.setVisible(true);
            calculate.runs = false;
            if (calculate.finished) return calculate.m4Objects;
            else return null;
        }

        ArrayList<ArrayList<m4_object>> calcSteenAlg(JFrame frame,link Link,int q,int n,boolean red) {
            JDialog fram = new JDialog(new JFrame(), "Steenrod Algebra Calculation", true);
            fram.setBounds(frame.getX()+50,frame.getY()+100,500,200);
            JLabel task = new JLabel("Creating Objects", SwingConstants.CENTER);
            JLabel objs = new JLabel("Objects : ", SwingConstants.RIGHT);
            JLabel objn = new JLabel(" ", SwingConstants.LEFT);
            JLabel levl = new JLabel("Level : ", SwingConstants.RIGHT);
            JLabel levn = new JLabel(" ", SwingConstants.LEFT);
            JLabel dims = new JLabel("Dimension : ", SwingConstants.RIGHT);
            JLabel dimn = new JLabel(" ", SwingConstants.LEFT);
            JPanel pand = new JPanel();
            JPanel pane = new JPanel(new GridLayout(3,2));
            JPanel panf = new JPanel();
            JButton cancel = new JButton("Cancel");
            pand.add(task);
            pane.add(objs);
            pane.add(objn);
            pane.add(levl);
            pane.add(levn);
            pane.add(dims);
            pane.add(dimn);
            panf.add(cancel);
            fram.setLayout(new BorderLayout());
            fram.add(pand, BorderLayout.NORTH);
            fram.add(pane, BorderLayout.CENTER);
            fram.add(panf, BorderLayout.SOUTH);
            fram.setResizable(false);
            cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    fram.setVisible(false);
                }
            });
            SteenCalc calculate = new SteenCalc(fram,task,objn,levn,dims,dimn,Link,q,n,red);
            calculate.start();
            fram.setVisible(true);
            calculate.runs = false;
            if (calculate.finished) return calculate.m4Objects;
            else return null;
        }
    }
    
    void calcKhovHom(JFrame frame, link Link, String title, String name, int n) {
        boolean reduced = false;
        if (n == 1) {
            n = 2;
            reduced = true;
        }
        JDialog fram = new JDialog(new JFrame(), title, true);
        fram.setBounds(frame.getX()+250, frame.getY()+225,300, 150);
        JLabel labe = new JLabel("Crossing : ",SwingConstants.RIGHT);
        JLabel labf = new JLabel("Objects : ",SwingConstants.RIGHT);
        JLabel labg = new JLabel(" ", SwingConstants.LEFT);
        JLabel labh = new JLabel(" ", SwingConstants.LEFT);
        JPanel pane = new JPanel(new GridLayout(2,2));
        JPanel panf = new JPanel();
        JButton cancel = new JButton("Cancel");
        pane.add(labe);
        pane.add(labg);
        pane.add(labf);
        pane.add(labh);
        panf.add(cancel);
        fram.setLayout(new BorderLayout());
        fram.add(pane, BorderLayout.CENTER);
        fram.add(panf, BorderLayout.SOUTH);
        fram.setResizable(false);
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        KhovCalc calculate = new KhovCalc(fram,labe,labg,labh,Link,n,reduced);
        calculate.start();
        fram.setVisible(true);
        calculate.runs = false;
        if (calculate.finished) representHomology(frame,calculate.KhovHom, title+" of "+name,n,reduced,Link);
    }
    
    void representHomology(JFrame frame, ArrayList<homology> KhovHom, String title, int n,boolean red, link Link) {
        JFrame fram = new JFrame(title);
        fram.setBounds(frame.getX()+100, frame.getY()+100,600,400);
        fram.setResizable(false);
        fram.setLayout(new BorderLayout());
        JPanel paneList = new JPanel(new BorderLayout());
        JLabel labeList = new JLabel("Quantum degrees", SwingConstants.CENTER);
        labeList.setPreferredSize(new Dimension(120,30));
        int u = 0;
        while (u < KhovHom.size()) {
            homology hom = KhovHom.get(u);
            boolean nontrivial = false;
            int v = 0;
            while (!nontrivial & v < hom.degrees.size()) {
                if (hom.groups.get(v).Betti > 0 | !hom.groups.get(v).Torsion.isEmpty()) nontrivial = true;
                else v++;
            }
            if (!nontrivial) KhovHom.remove(u);
            else u++;
        }
        Card HomCards = new Card();
        ArrayList<JPanel> HomPanels = new ArrayList<JPanel>();
        String[] items = new String[KhovHom.size()];
        int m = KhovHom.size()-1;
        for (int i = m; i >= 0; i--) {
            items[m-i] = "q = "+KhovHom.get(i).quantum;
            boolean nontrivial = false;
            while (!nontrivial) {
                if (KhovHom.get(i).groups.get(0).Betti > 0 | !KhovHom.get(i).groups.get(0).Torsion.isEmpty()) 
                    nontrivial = true;
                else {
                    KhovHom.get(i).degrees.remove(0);
                    KhovHom.get(i).groups.remove(0);
                }
            }
            int k = KhovHom.get(i).groups.size()-1;
            nontrivial = false;
            while (!nontrivial) {
                if (KhovHom.get(i).groups.get(k).Betti > 0 | !KhovHom.get(i).groups.get(k).Torsion.isEmpty())
                    nontrivial = true;
                else {
                    KhovHom.get(i).degrees.remove(k);
                    KhovHom.get(i).groups.remove(k);
                    k--;
                }
            }
            int diff = HomDiff(KhovHom.get(i),0);
            int max = KhovHom.get(i).degrees.get(0)+diff;
            JPanel homPanel = addLabels2Panel(max,diff,KhovHom.get(i),0); // creates Integral cards
            HomCards.add(homPanel);
            HomPanels.add(homPanel);
            diff = HomDiff(KhovHom.get(i),1);
            max = maxHomDeg(KhovHom.get(i),1);
            homPanel = addLabels2Panel(max,diff,KhovHom.get(i),1); // creates rational cards
            HomCards.add(homPanel);
            HomPanels.add(homPanel);
            diff = HomDiff(KhovHom.get(i),n);
            max = maxHomDeg(KhovHom.get(i),n);
            homPanel = addLabels2Panel(max,diff,KhovHom.get(i),n); // creates mod cards
            HomCards.add(homPanel);
            HomPanels.add(homPanel);
        }
        JList<String> quantumList = new JList<>(items);
        JPanel buttonPane = new JPanel();
        JButton latexBut = new JButton("Save as LaTeX file");
        JButton closeBut = new JButton("Close");
        closeBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        buttonPane.add(latexBut);
        buttonPane.add(closeBut);
        paneList.add(labeList, BorderLayout.NORTH);
        final ButtonGroup coefficients = new ButtonGroup();
        JRadioButton radioButtonZ = new JRadioButton("Z",true);
        JPanel coeffPanel = new JPanel(new GridLayout(2,2));
        coeffPanel.add(radioButtonZ);
        radioButtonZ.setActionCommand("Z");
        coefficients.add(radioButtonZ);
        JRadioButton radioButtonQ = new JRadioButton("Q");
        coeffPanel.add(radioButtonQ);
        radioButtonQ.setActionCommand("Q");
        coefficients.add(radioButtonQ);
        JRadioButton radioButtonZm = new JRadioButton("Z/");
        coeffPanel.add(radioButtonZm);
        radioButtonZm.setActionCommand("Z/");
        coefficients.add(radioButtonZm);
        SpinnerNumberModel modelc = new SpinnerNumberModel(n,2,100,1);
        JSpinner spinnerc = new JSpinner(modelc);
        coeffPanel.add(spinnerc);
        JScrollPane scrollPane = new JScrollPane(quantumList);
        scrollPane.setPreferredSize(new Dimension(100,300));
        paneList.add(scrollPane, BorderLayout.CENTER);
        paneList.add(coeffPanel, BorderLayout.SOUTH);
        quantumList.setSelectedIndex(0);
        quantumList.addListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent le) {
                @SuppressWarnings("unchecked")
                JList<String> list = (JList<String>) le.getSource();
                int u = list.getSelectedIndex();
                if ( fram.isActive() & u >= 0 ) {
                    for (JPanel pan : HomPanels) pan.setVisible(false);
                    String entree = coefficients.getSelection().getActionCommand();
                    quantumList.setSelectedIndex(u);
                    if ("Z".equals(entree)) u = 3 * u;
                    else if ("Q".equals(entree)) u = 3 * u + 1;
                    else u = 3 * u + 2;
                    HomPanels.get(u).setVisible(true);
                }
            }
        });
        spinnerc.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent ce) {
                int k = (int) spinnerc.getValue();
                int l = HomPanels.size()/3;
                for (int u = 0; u < l; u++) {
                    JPanel tut = HomPanels.get(3*u+2);
                    boolean tutvis = tut.isVisible();
                    if (tutvis) tut.setVisible(false);
                    tut.removeAll();
                    int diff = HomDiff(KhovHom.get(l-1-u),k);
                    int max = maxHomDeg(KhovHom.get(l-1-u),k);
                    JPanel homPanel = addLabels2Panel(max,diff,KhovHom.get(l-1-u),k);
                    for (Component ob : homPanel.getComponents()) tut.add(ob);
                    if (tutvis) tut.setVisible(true);
                }
            }
        });
        radioButtonZ.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                boolean found = false;
                int u = 0;
                while (!found) {
                    JPanel pan = HomPanels.get(u);
                    if (pan.isVisible()) found = true;
                    else u++;
                }
                if (u % 3 != 0) {
                    HomPanels.get(u).setVisible(false);
                    u--;
                    if (u % 3 != 0) u--;
                    HomPanels.get(u).setVisible(true);
                }
            }
        });
        radioButtonQ.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                boolean found = false;
                int u = 0;
                while (!found) {
                    JPanel pan = HomPanels.get(u);
                    if (pan.isVisible()) found = true;
                    else u++;
                }
                if (u % 3 != 1) {
                    HomPanels.get(u).setVisible(false);
                    if (u % 3 == 0) u++;
                    else u--;
                    HomPanels.get(u).setVisible(true);
                }
            }
        });
        radioButtonZm.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                boolean found = false;
                int u = 0;
                while (!found) {
                    JPanel pan = HomPanels.get(u);
                    if (pan.isVisible()) found = true;
                    else u++;
                }
                if (u % 3 != 2) {
                    HomPanels.get(u).setVisible(false);
                    u++;
                    if (u % 3 != 2) u++;
                    HomPanels.get(u).setVisible(true);
                }
            }
        });
        latexBut.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                int k = 0;
                if (radioButtonQ.isSelected()) k = 1;
                if (radioButtonZm.isSelected()) k = (int) spinnerc.getValue();
                saveKhovHom(fram,KhovHom,title,n,k,red,Link);
            }
        });
        fram.add(paneList, BorderLayout.WEST);
        fram.add(buttonPane, BorderLayout.SOUTH);
        fram.add(HomCards, BorderLayout.CENTER);
        fram.setVisible(true);
    }
    
    boolean alright;
    
    void saveKhovHom(JFrame frame, ArrayList<homology> KhovHom, String title, int n, int k, boolean red, link Link) {
        ArrayList<Integer> diffs = new ArrayList<Integer>();
        ArrayList<Integer> maxs = new ArrayList<Integer>();
        int qmin = KhovHom.get(0).quantum;
        int qmax = KhovHom.get(KhovHom.size()-1).quantum;
        for (homology hom : KhovHom) {
            int diff = HomDiff(hom,k);
            int max;
            if (k==0) max = hom.degrees.get(0)+diff;
            else max = maxHomDeg(hom,k);
            diffs.add(diff);
            maxs.add(max);
        }
        boolean lame = true;
        int v = 0;
        int starter = 0;
        v = KhovHom.size()-1;
        int ender = v;
        JDialog fram = new JDialog(new JFrame(),"Options", true);
        fram.setBounds(frame.getX()+150, frame.getY()+90, 300, 220);
        fram.setResizable(false);
        fram.setLayout(new GridLayout(5,1));
        JLabel orient = new JLabel("Orientation :");
        final ButtonGroup orientations = new ButtonGroup();
        JRadioButton qslashh = new JRadioButton("q \\ h",true);
        JRadioButton hslashq = new JRadioButton("h \\ q");
        boolean qh = true;
        orientations.add(qslashh);
        orientations.add(hslashq);
        JPanel orButtons = new JPanel();
        orButtons.add(orient);
        orButtons.add(qslashh);
        orButtons.add(hslashq);
        JLabel qrange = new JLabel("Quantum range :");
        int qlow = qmin;
        int qhih = qmax;
        SpinnerNumberModel modelmin = new SpinnerNumberModel(qlow,qmin,qhih,2);
        JSpinner spinnermin = new JSpinner(modelmin);
        SpinnerNumberModel modelmax = new SpinnerNumberModel(qhih,qlow,qmax,2);
        JSpinner spinnermax = new JSpinner(modelmax);
        JPanel rangePan = new JPanel();
        rangePan.add(qrange);
        rangePan.add(spinnermin);
        rangePan.add(spinnermax);
        JCheckBox nameCheck = new JCheckBox("Add Link Name");
        JCheckBox gausCheck = new JCheckBox("Add DT-Code");
        if (Link.components.size()>1) gausCheck.setEnabled(false);
        JPanel namePan = new JPanel();
        namePan.add(nameCheck);
        JPanel gausPan = new JPanel();
        gausPan.add(gausCheck);
        JButton cancel = new JButton("Cancel");
        JButton save = new JButton("Save");
        JPanel butPanel = new JPanel();
        butPanel.add(cancel);
        butPanel.add(save);
        alright = false;
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        save.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                alright = true;
                fram.setVisible(false);
            }
        });
        spinnermin.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent ce) {
                int k = (int) spinnermin.getValue();
                modelmax.setMinimum(k);
            }
        });
        spinnermax.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent ce) {
                int k = (int) spinnermax.getValue();
                modelmin.setMaximum(k);
            }
        });
        fram.add(orButtons);
        fram.add(rangePan);
        fram.add(namePan);
        fram.add(gausPan);
        fram.add(butPanel);
        fram.setVisible(true);
        if (alright) {
            boolean addName = nameCheck.isSelected();
            boolean addGaus = gausCheck.isSelected();
            qlow = (int) spinnermin.getValue();
            qhih = (int) spinnermax.getValue();
            if (hslashq.isSelected()) qh = false;
            int r = 0;
            while (KhovHom.get(r).quantum < qlow) {
                r++;
                starter++;
            }
            r = KhovHom.size()-1;
            while (KhovHom.get(r).quantum > qhih) {
                r--;
                ender--;
            }
            int min = minHom(diffs,maxs,starter,ender);
            int max = maxHom(diffs,maxs,starter,ender);
            ArrayList<String> commands = new ArrayList<String>();
            commands.add("\\documentclass[border=1bp]{standalone}");
            commands.add("\\usepackage{amssymb,amsmath,array,diagbox}");
            commands.add("\\newcommand{\\Z}{\\mathbb{Z}}");
            commands.add("\\newcommand{\\Q}{\\mathbb{Q}}");
            commands.add("\\begin{document}");
            commands.add("\\setlength\\extrarowheight{2pt}");
            if (addName | addGaus) commands.add("\\begin{tabular}{l}");
            if (addName) {
                String intro = "";
                if (red) intro = "Reduced "+intro;
                commands.add(intro+title+"\\\\");
            }
            if (addGaus) {
                String DTCode = DowkerThistle(Link);
                commands.add("DT-Code : "+DTCode+"\\\\");
            }
            String next = "\\begin{tabular}{|c||";
            String after;
            if (qh) {
                after = "\\backslashbox{\\!$q$\\!}{\\!$h$\\!} ";
                for (int i = min; i <= max; i++) {
                    next = next + "c|";
                    after = after + "& $"+i+"$ ";
                }
                next = next +"}";
                commands.add(next);
                after = after + "\\\\";
                commands.add("\\hline");
                commands.add(after);
                commands.add("\\hline");
                commands.add("\\hline");
                v = ender;
                for (int q = qhih; q >= qlow; q = q - 2) {
                    String qline = "$"+q+"$ "; 
                    if (KhovHom.get(v).quantum != q) {
                        for (int i = min; i <= max; i++) qline = qline +"& ";
                        v++;
                    }
                    else {
                        for (int i = min; i <= max; i++) {
                            qline = qline + "& ";
                            if (i >= maxs.get(v)-diffs.get(v) & i <= maxs.get(v)) 
                                qline = qline + "$"+LatexHom(KhovHom.get(v),i,k)+"$ ";
                        }
                    }
                    qline = qline + "\\\\";
                    commands.add(qline);
                    commands.add("\\hline");
                    v--;
                }
            }
            else {
                after = "\\backslashbox{\\!$h$\\!}{\\!$q$\\!} ";
                for (int q = qlow; q <= qhih; q = q + 2) {
                    next = next + "c|";
                    after = after + "& $"+q+"$ ";
                }
                next = next +"}";
                commands.add(next);
                after = after + "\\\\";
                commands.add("\\hline");
                commands.add(after);
                commands.add("\\hline");
                commands.add("\\hline");
                for (int i = max; i >= min; i--) {
                    v = starter;
                    String qline = "$"+i+"$ ";
                    for (int q = qlow; q <= qhih; q = q + 2) {
                        qline = qline + "& ";
                        if (KhovHom.get(v).quantum == q) {
                            if (i >= maxs.get(v)-diffs.get(v) & i <= maxs.get(v)) 
                                qline = qline + "$"+LatexHom(KhovHom.get(v),i,k)+"$ ";
                            v++;
                        }
                    }
                    qline = qline + "\\\\";
                    commands.add(qline);
                    commands.add("\\hline");
                }
            }
            commands.add("\\end{tabular}");
            if (addName | addGaus) commands.add("\\end{tabular}");
            commands.add("\\end{document}");
            if (title.startsWith("K")) title = "Kh_"+title.substring(23);
            else title = title.replaceFirst("-Cohomology of ", "_");
            if (k == 1) title = title + "_Q";
            if (k >= 2) title = title + "_Z_"+k;
            if (qhih - qlow < qmax - qmin) title = title + "_"+qlow+"_"+qhih;
            if (!qh) title = title +"^";
            if (red) title = title +"_red";
            saveCommands(title,commands,frame);
        }
    }
    
    void saveCommands(String title, ArrayList<String> commands, JFrame frame) {
        JFileChooser chooser = new JFileChooser();
        if (saveKhovanov != null) chooser.setCurrentDirectory(saveKhovanov);
        FileNameExtensionFilter filter = new FileNameExtensionFilter("LaTeX files (*.tex)", "tex");
        chooser.setFileFilter(filter);
        chooser.setSelectedFile(new File(title+".tex"));
        int val = chooser.showSaveDialog(frame);
        if (val == JFileChooser.APPROVE_OPTION) {
            try {
                File file = chooser.getSelectedFile();
                saveKhovanov = chooser.getCurrentDirectory();
                String fname = file.getAbsolutePath();
                if(!fname.endsWith(".tex") ) {
                    file = new File(fname + ".tex");
                }
                FileWriter fw = new FileWriter(file);
                PrintWriter pw = new PrintWriter(fw);
                for (String command : commands) pw.println(command);
                pw.close();
            }
            catch (IOException e) {

            }
        }
    }
    
    String LatexHom(homology Hom, int deg, int n) {
        boolean found = false;
        int u = 0;
        while (!found & u < Hom.degrees.size()) {
            if (Hom.degrees.get(u) == deg) found = true;
            else u++;
        }
        if (!found) u = -1;
        return getHomString(Hom,u,n,true);
    }
    
    int minHom(ArrayList<Integer> diffs, ArrayList<Integer> maxs, int start, int end) {
        boolean minfound = false;
        int min = 0;
        for (int i = start; i <= end; i++) {
            if (diffs.get(i)!=-1) {
                int here = maxs.get(i)-diffs.get(i);
                if (!minfound) {
                    minfound = true;
                    min = here;
                }
                else if (min > here) min = here;
            }
        }
        return min;
    }
    
    int maxHom(ArrayList<Integer> diffs, ArrayList<Integer> maxs, int start, int end) {
        boolean maxfound = false;
        int max = 0;
        for (int i = start; i <= end; i++) {
            if (diffs.get(i) != -1) {
                int here = maxs.get(i);
                if (!maxfound) {
                    maxfound = true;
                    max = here;
                }
                else if (max < here) max = here;
            }
        }
        return max;
    }
    
    JPanel addLabels2Panel(int max,int diff, homology KhovHom, int n) {
        int panelSize = 7;
        if (diff >= panelSize) panelSize = diff+1;
        JPanel homPanel = new JPanel(new GridLayout(panelSize,2));
        int s = KhovHom.degrees.size()-1;
        boolean ok = false;
        while (!ok & s >=0 ) {
            if (max == KhovHom.degrees.get(s)) ok = true;
            else s--;
        }
        for (int j = 0; j < (panelSize - diff - 1)/2; j++) {
            JLabel blah = new JLabel(" ");
            JLabel blahh = new JLabel(" ");
            homPanel.add(blah);
            homPanel.add(blahh);
        }
        for (int j = 0; j <= diff; j++) {
            JLabel homLab = new JLabel("H^"+(max-j)+" = ", SwingConstants.RIGHT);
            String grpStr = "0";
            if (s >=0) {
                if (max-j == KhovHom.degrees.get(s)) {
                    grpStr = getHomString(KhovHom,s,n,false);
                    s--;
                }
            }
            else if (max - j + 1 == KhovHom.degrees.get(0)) {
                grpStr = getHomString(KhovHom,-1,n,false);
            }
            JLabel grpLab = new JLabel(grpStr, SwingConstants.LEFT);
            homPanel.add(homLab);
            homPanel.add(grpLab);
        }
        return homPanel;
    }
    
    String modCoeffString(int betti, ArrayList<Integer> torsion, int n, boolean latex) {
        String modStr = "0";
        String lbrack = "";
        String rbrack = "";
        String lpa = "(";
        String rpa = ")";
        String Zm = "Z/";
        String plus = " + ";
        if (latex) {
            lbrack = "{";
            rbrack = "}";
            lpa = "";
            rpa = "";
            Zm = "\\Z_";
            plus = " \\oplus ";
        }
        ArrayList<Integer> tor = new ArrayList<Integer>();
        ArrayList<Integer> pow = new ArrayList<Integer>();
        if (!torsion.isEmpty()) {
            int u = -1;
            for (int t : torsion) {
                if (tor.contains(t)) {
                    int p = pow.get(u);
                    pow.set(u, p+1);
                }
                else {
                    tor.add(t);
                    pow.add(1);
                    u++;
                }
            }
            ArrayList<Integer> appear = new ArrayList<Integer>();
            for (int v = 0; v < tor.size(); v++) {
                int val = pow.get(v);
                for (int w = v+1; w < tor.size(); w++) if (tor.get(w)%tor.get(v) == 0) val = val + pow.get(w);
                appear.add(val);
            }
            int max = 0;
            for (int r : appear) if (r > max) max = r;
            ArrayList<Integer> tors = new ArrayList<Integer>();
            for (int r = 0; r < max; r++) tors.add(1);
            for (int j = 0; j < tor.size(); j++) {
                int a = tor.get(j);
                int b = pow.get(j);
                int start = max - appear.get(j);
                for (int r = start; r < start+b; r++) {
                    int p = tors.get(r) * a;
                    tors.set(r,p);
                }
            }
            tor = new ArrayList<Integer>();
            pow = new ArrayList<Integer>();
            u = -1;
            for (int t : tors) {
                if (tor.contains(t)) {
                    int p = pow.get(u);
                    pow.set(u, p+1);
                }
                else {
                    tor.add(t);
                    pow.add(1);
                    u++;
                }
            }
            u = 0;
            while (u < tor.size()) {
                int a = gcd(tor.get(u),n);
                if (a == 1) {
                    tor.remove(u);
                    pow.remove(u);
                    u--;
                }
                if (a == n) {
                    betti = betti + pow.get(u);
                    tor.remove(u);
                    pow.remove(u);
                    u--;
                }
                if (a > 1 & a < n) {
                    tor.set(u, a);
                }
                u++;
            }
            u = 0;
            while (u < tor.size()-1) {
                int v = u+1;
                if (Objects.equals(tor.get(u), tor.get(v))) {
                    int r = pow.get(u)+pow.get(v);
                    pow.set(u,r);
                    tor.remove(v);
                    pow.remove(v);
                }
                else u++;
            }
        }
        if (betti == 1) modStr = Zm+lbrack+n+rbrack;
        if (betti > 1) modStr = lpa+Zm+lbrack+n+rbrack+rpa+"^"+lbrack+betti+rbrack;
        if (!tor.isEmpty()) {
            if (betti != 0) modStr = modStr + plus;
            else modStr = "";
            for (int i = 0; i < tor.size(); i++) {
                if (pow.get(i) == 1) modStr = modStr + Zm+lbrack+tor.get(i)+rbrack;
                else modStr = modStr + lpa+Zm+lbrack+tor.get(i)+rbrack+rpa+"^"+lbrack+pow.get(i)+rbrack;
                if (i < tor.size()-1) modStr = modStr + plus;
            }
        }
        return modStr;
    }
    
    String torStr(ArrayList<Integer> torsion, boolean latex) {
        String torStr = "";
        String lbrack = "";
        String rbrack = "";
        String lpa = "(";
        String rpa = ")";
        String Zm = "Z/";
        String plus = " + ";
        if (latex) {
            lbrack = "{";
            rbrack = "}";
            lpa = "";
            rpa = "";
            Zm = "\\Z_";
            plus = " \\oplus ";
        }
        if (!torsion.isEmpty()) {
            ArrayList<Integer> tor = new ArrayList<Integer>();
            ArrayList<Integer> pow = new ArrayList<Integer>();
            int u = -1;
            for (int t : torsion) {
                if (tor.contains(t)) {
                    int p = pow.get(u);
                    pow.set(u, p+1);
                }
                else {
                    tor.add(t);
                    pow.add(1);
                    u++;
                }
            }
            for (int i = 0; i < tor.size(); i++) {
                if (pow.get(i) == 1) torStr = torStr + Zm+lbrack+tor.get(i)+rbrack;
                else torStr = torStr + lpa+Zm+lbrack+tor.get(i)+rbrack+rpa+"^"+lbrack+pow.get(i)+rbrack;
                if (i < tor.size()-1) torStr = torStr + plus;
            }
        }
        return torStr;
    }
    
    String getHomString(homology KhovHom, int s, int n, boolean latex) {
        String grpStr = "0";
        String lbrack = "";
        String rbrack = "";
        String Z = "Z";
        String Q = "Q";
        String plus = " + ";
        if (latex) {
            lbrack = "{";
            rbrack = "}";
            Z = "\\Z";
            Q = "\\Q";
            plus = " \\oplus ";
        }
        if (n == 0) {
            if (KhovHom.groups.get(s).Betti == 1) grpStr = Z;
            if (KhovHom.groups.get(s).Betti > 1) grpStr = Z+"^"+lbrack+KhovHom.groups.get(s).Betti+rbrack;
            if (KhovHom.groups.get(s).Betti > 0 & !KhovHom.groups.get(s).Torsion.isEmpty()) 
                grpStr = grpStr + plus;
            if (KhovHom.groups.get(s).Betti == 0 & !KhovHom.groups.get(s).Torsion.isEmpty())
                grpStr = "";
            grpStr = grpStr + torStr(KhovHom.groups.get(s).Torsion,latex);
        }
        if (n == 1) {
            if (KhovHom.groups.get(s).Betti == 1) grpStr = Q;
            if (KhovHom.groups.get(s).Betti > 1) grpStr = Q+"^"+lbrack+KhovHom.groups.get(s).Betti+rbrack;
        }
        if (n > 1) {
            ArrayList<Integer> tor = new ArrayList<Integer>();
            int betti = 0;
            if (s >=0) betti = KhovHom.groups.get(s).Betti;
            if (s != KhovHom.groups.size()-1) {
                for (int t : KhovHom.groups.get(s+1).Torsion) tor.add(t);
            }
            if (s != -1)
                for (int t : KhovHom.groups.get(s).Torsion) tor.add(t);
            Collections.sort(tor);
            if (betti > 0 | !tor.isEmpty()) grpStr = modCoeffString(betti,tor,n,latex);
        }
        return grpStr;
    }
    
    int maxHomDeg(homology KhovHom, int n) {
        int max = -1;
        if (n == 1) {
            int u = KhovHom.degrees.size()-1;
            boolean found = false;
            while (!found & u >=0 ) {
                if (KhovHom.groups.get(u).Betti > 0) found = true;
                else u--;
            }
            if (found) max = KhovHom.degrees.get(u);
        }
        else {
            int u = KhovHom.degrees.size()-1;
            boolean found = false;
            while (!found & u >=0 ) {
                if (KhovHom.groups.get(u).Betti > 0 | conTorsion(KhovHom.groups.get(u).Torsion,n)) found = true;
                else u--;
            }
            if (found) max = KhovHom.degrees.get(u);
        }
        return max;
    }
    
    int HomDiff(homology KhovHom, int theCase) {
        int diff = -1;
        if (theCase == 0) { // integer coefficients
            diff = KhovHom.degrees.get(KhovHom.degrees.size()-1) - KhovHom.degrees.get(0);
        }
        if (theCase == 1) { // rational coefficients
            int u = KhovHom.degrees.size()-1;
            boolean found = false;
            while (!found & u >=0 ) {
                if (KhovHom.groups.get(u).Betti > 0) found = true;
                else u--;
            }
            if (u > 0) {
                int v = 0;
                found = false;
                while (!found & v < u) {
                    if (KhovHom.groups.get(v).Betti > 0) found = true;
                    else v++;
                }
                diff = u - v;
                found = false;
            }
            if (found) diff = 0;
        }
        if (theCase >= 2) { //coefficients in Z/theCase
            int u = KhovHom.degrees.size()-1;
            boolean found = false;
            while (!found & u >=0 ) {
                if (KhovHom.groups.get(u).Betti > 0 | conTorsion(KhovHom.groups.get(u).Torsion,theCase)) found = true;
                else u--;
            }
            if (u > 0) {
                int v = 0;
                found = false;
                while (!found & v < u) {
                    if (KhovHom.groups.get(v).Betti > 0 | conTorsion(KhovHom.groups.get(v).Torsion,theCase)) found = true;
                    else v++;
                }
                diff = u - v;
                if (conTorsion(KhovHom.groups.get(v).Torsion, theCase)) diff++;
                found = false;
            }
            else if (conTorsion(KhovHom.groups.get(0).Torsion, theCase)) {
                diff = 1;
                found = false; 
            }
            if (found) diff = 0;
        }
        return diff;
    }
    
    boolean conTorsion(ArrayList<Integer> torsion, int n) {
        boolean sure = false;
        int i = 0;
        while (!sure & i < torsion.size()) {
            if (gcd(torsion.get(i),n) > 1) sure = true;
            else i++;
        }
        return sure;
    }
    
    int gcd(int a, int b) {
        if (a % b == 0) return b;
        if (a > b) return gcd(b,a%b);
        if (b % a == 0) return a;
        return gcd(a,b%a);
    }
    
    twoPolynomial obtainHomFly(JFrame frame, link Link) {
        JDialog fram = new JDialog(new JFrame(),"HOMFLYPT Polynomial", true);
        fram.setBounds(frame.getX()+250, frame.getY()+225,300, 150);
        JLabel labe = new JLabel("Crossing : ",SwingConstants.RIGHT);
        JLabel labf = new JLabel("Objects : ",SwingConstants.RIGHT);
        JLabel labg = new JLabel(" ", SwingConstants.LEFT);
        JLabel labh = new JLabel(" ", SwingConstants.LEFT);
        JPanel pane = new JPanel(new GridLayout(2,2));
        JPanel panf = new JPanel();
        JButton cancel = new JButton("Cancel");
        pane.add(labe);
        pane.add(labg);
        pane.add(labf);
        pane.add(labh);
        panf.add(cancel);
        fram.setLayout(new BorderLayout());
        fram.add(pane, BorderLayout.CENTER);
        fram.add(panf, BorderLayout.SOUTH);
        fram.setResizable(false);
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        HomflyCalc calculate = new HomflyCalc(fram,labg,labh,Link);
        calculate.start();
        fram.setVisible(true);
        calculate.runs = false;
        return calculate.Homfly;
    }
    
    polynomial obtainUnJonesPol(polynomial poly) {
        int[] ent = {1,1};
        int[] coe = {-2,2};
        polynomial pol = new polynomial(ent,coe);
        polynomial res = multiplyPoly(poly,pol);
        for (coefficient cof : res.entries) {
            cof.power = cof.power * 2;
        }
        return res;
    }
    
    polynomial obtainJonesPol(JFrame frame, link Link) {
        JDialog fram = new JDialog(new JFrame(),"Jones Polynomial",true);
        fram.setBounds(frame.getX()+250, frame.getY()+225,300, 150);
        JLabel labe = new JLabel("Crossing : ",SwingConstants.RIGHT);
        JLabel labf = new JLabel("Objects : ",SwingConstants.RIGHT);
        JLabel labg = new JLabel(" ", SwingConstants.LEFT);
        JLabel labh = new JLabel(" ", SwingConstants.LEFT);
        JPanel pane = new JPanel(new GridLayout(2,2));
        JPanel panf = new JPanel();
        JButton cancel = new JButton("Cancel");
        pane.add(labe);
        pane.add(labg);
        pane.add(labf);
        pane.add(labh);
        panf.add(cancel);
        fram.setLayout(new BorderLayout());
        fram.add(pane, BorderLayout.CENTER);
        fram.add(panf, BorderLayout.SOUTH);
        fram.setResizable(false);
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                fram.setVisible(false);
            }
        });
        JonesCalc calculate = new JonesCalc(fram,labg,labh,Link);
        calculate.start();
        fram.setVisible(true);
        calculate.runs = false;
        return calculate.Jones;
    }
    
    JPanel setComponentPanel(link Link) {
        JPanel compPanel = new JPanel(new GridLayout(1,3));
        JLabel comp = new JLabel("Components : "+Link.components.size());
        int writhe = nplus(Link)-nminus(Link);
        JLabel writ = new JLabel("Writhe : "+writhe);
        int crosss = 0;
        for (int r : Link.crossings) crosss = crosss + Math.abs(r);
        JLabel cross = new JLabel("Crossings : "+crosss+" / "+Link.crossings.length);
        JLabel plan = new JLabel(" ");
        if (!Link.isPlanar()) plan = new JLabel("Not planar");
        compPanel.add(comp);
        compPanel.add(writ);
        compPanel.add(cross);
        compPanel.add(plan);
        return compPanel;
    }
    
    private class link {
        int [] crossings;
        int [] [] paths;
        ArrayList<ArrayList<Integer>> components;
        
        link(int [] crossins, int [] [] path) {
            crossings = crossins;
            paths = path;
            components = components(crossins,path);
        }
        
        int compOf(int k,ArrayList<ArrayList<Integer>> lists) {
            int u = 0;
            boolean found = false;
            while (!found) {
                if (lists.get(u).contains(k)) found = true;
                else u++;
            }
            return u;
        }
        
        int disjointComponents() {
            int i = components.size();
            if (i <=1) return i;
            ArrayList<ArrayList<Integer>> comps = new ArrayList<ArrayList<Integer>>();
            for (ArrayList<Integer> comp : components) {
                ArrayList<Integer> cp = new ArrayList<Integer>();
                for (int j : comp) cp.add(j);
                comps.add(cp);
            }
            int k = 0;
            while (comps.size()>1 & k < crossings.length) {
                int a = compOf(paths[k][0],comps);
                int b;
                if (crossings[k]%2 == 0) b = compOf(paths[k][2],comps);
                else b = compOf(paths[k][1],comps);
                if (a!=b) {
                    for (int e : comps.get(b)) comps.get(a).add(e);
                    comps.remove(b);
                }
                k++;
            }
            i = comps.size();
            return i;
        }
        
        boolean isPlanar() {
            if (components.isEmpty()) return false;
            boolean [][] checkedSurfaces = new boolean[crossings.length][4];
            int surfaces = 0;
            int k = 0;
            int runner = 0;
            while (k < crossings.length) {
                for (int j = 0; j < 4; j++) {
                    if (!checkedSurfaces[k][j]) {
                        checkedSurfaces[k][j] = true;
                        int r = (j+1)%4;
                        runner = paths[k][r];
                        int co1 = k;
                        int co2 = r;
                        boolean bremain = true;
                        while (bremain) {
                            boolean foundNext = false;
                            int u = 0;
                            int v = 0;
                            while (!foundNext) {
                                v = 0;
                                while (!foundNext & v < 4) {
                                    if (!(co1 == u & co2 == v)) {
                                        if (paths[u][v] == runner) {
                                            foundNext = true;
                                            if (checkedSurfaces[u][v]) bremain = false;
                                            checkedSurfaces[u][v] = true;
                                        }
                                        else v++;
                                    }
                                    else v++;
                                }
                                if (!foundNext) u++;
                            }
                            v = (v+1) % 4;
                            runner = paths[u][v];
                            co1 = u;
                            co2 = v;
                        }
                        surfaces++;
                    }
                }
                k++;
            }
            int Euler = surfaces - crossings.length;
            return (Euler == 2 * disjointComponents());
        }
        
        boolean isMatched() {
            boolean matched = true;
            if (components.isEmpty()) return false;
            for (int i : crossings) if ((i%2)!=0) matched = false;
            if (components.size() == 1) return matched;
            int i = 0;
            ArrayList<Integer> used = new ArrayList<Integer>();
            while (matched & i < crossings.length) {
                int a = paths[i][0];
                int b = paths[i][1];
                int c = paths[i][2];
                int d = paths[i][3];
                boolean rup = true;
                boolean ldn = true;
                boolean found = false;
                int u = 0;
                while (!found) {
                    if (components.get(u).contains(a)) found = true;
                    else u++;
                }
                found = false;
                int v = 0;
                while (!found) {
                    if (components.get(v).contains(c)) found = true;
                    else v++;
                }
                int p1 = components.get(u).indexOf(a);
                int p2 = components.get(u).indexOf(b);
                int p3 = components.get(v).indexOf(c);
                int p4 = components.get(v).indexOf(d);
                if (p1 == 0 | p2 == 0) {
                    if (p1 == 0) {
                        if (used.contains(a)) rup = false;
                        else {
                            rup = true;
                            used.add(a);
                        }
                    }
                    else {
                        if (used.contains(b)) rup = true;
                        else {
                            rup = false;
                            used.add(b);
                        }
                    }
                }
                else rup = p1 < p2;
                if (p3 == 0 | p4 == 0) {
                    if (p3 == 0) {
                        if (used.contains(c)) ldn = false;
                        else {
                            ldn = true;
                            used.add(c);
                        }
                    }
                    else {
                        if (used.contains(d)) ldn = true;
                        else {
                            ldn = false;
                            used.add(d);
                        }
                    }
                }
                else ldn = p3 < p4;
                if (rup != ldn) matched = false;
                else i++;
            }
            return matched;
        }
        
        final ArrayList<ArrayList<Integer>> components(int [] crossings,
                int [] [] paths) {
            ArrayList<ArrayList<Integer>> compts = new ArrayList<ArrayList<Integer>>();
            boolean[] notcheckedpaths = new boolean[2*crossings.length];
            int max = 5 * crossings.length;
            int security = 0;
            for (int u = 0; u < notcheckedpaths.length; u++) 
                notcheckedpaths[u] = true;
            int path = 1;
            int endpath, runpath;
            int pos = 0;
            int [] [] pafs = new int [paths.length][paths[0].length];
            for (int u = 0;u < pafs.length; u++) {
                System.arraycopy(paths[u], 0, pafs[u], 0, pafs[0].length);
            }
            try {
                while (path <= 2 * crossings.length) {
                    ArrayList<Integer> comp = new ArrayList<Integer>();
                    comp.add(path);
                    notcheckedpaths[path - 1] = false;
                    boolean comprun = true;
                    runpath = path;
                    security = 0;
                    while (comprun) {
                        security++;
                        if (security > 4 * crossings.length) throw new Exception();
                        int i = 0;
                        while (i < pafs.length) {
                            boolean found = false;
                            int k = 0;
                            while (!found & (k < 4)) {
                                if (pafs[i][k] == runpath) found = true;
                                else k++;
                            }
                            if (found) {
                                if ((crossings[i] % 2) == 0) {
                                    if (k == 0) pos = 1;
                                    if (k == 1) pos = 0;
                                    if (k == 2) pos = 3;
                                    if (k == 3) pos = 2;
                                }
                                else {
                                    if (k == 0) pos = 2;
                                    if (k == 1) pos = 3;
                                    if (k == 2) pos = 0;
                                    if (k == 3) pos = 1;
                                }
                                endpath = pafs[i][pos];
                                pafs[i][k] = 0;
                                pafs[i][pos] = 0;
                                if (endpath != path) {
                                    notcheckedpaths[endpath - 1] = false;
                                    comp.add(endpath);
                                    runpath = endpath;
                                }
                                else comprun = false;
                                break;
                            }
                            i++;
                        }
                    }
                    compts.add(comp);
                    int j = 1;
                    while (j < 2 * crossings.length) {
                        if (notcheckedpaths[j]) break;
                        j++;
                    }
                    path = j+1;
                }
            }
            catch (Exception e) {
                compts = new ArrayList<ArrayList<Integer>>();
            }
            return compts;
        }
    }
    
    link torusKnot(int p, int q) {
        int prod = p * (q-1);
        int [] crossings = new int[prod];
        for (int k = 0 ; k < prod; k++) {
            crossings[k] = 1;
        }
        int [][] paths = new int[prod][4];
        for (int i = 0; i < p; i++) {
            int u = i * (q-1);
            paths[u][0] = (2*i*(q-1)) + 2;
            paths[u][1] = ((2*i+1)*(q-1)) + 2;
            paths[u][2] = (2*(i+1)*(q-1)) + 1;
            paths[u][3] = (2*i*(q-1)) + 1;
        }
        paths[(p-1)*(q-1)][2] = 1;
        for (int j = 1; j < q-2; j++) {
            for (int i = 0; i < p ; i++) {
                int u = i * (q-1) + j;
                paths[u][0] = (2*i*(q-1)) + j + 2;
                paths[u][1] = ((2*i)+1)*(q-1) + j + 2;
                paths[u][2] = ((2*i)+2)*(q-1) + j + 1;
                paths[u][3] = ((2*i)+1)*(q-1) + j + 1;
            }
            paths[(p-1)*(q-1)+j][2] = j + 1;
        }
        for (int i = 0; i < p; i++) {
            int u = i * (q-1) + (q - 2);
            paths[u][0] = (2*i*(q-1)) + q;
            paths[u][1] = (2*(i+1)*(q-1)) + q;
            paths[u][2] = ((2*i+2)*(q-1)) + q - 1;
            if (q > 2) paths[u][3] = ((2*i+1)*(q-1)) + q - 1;
            else paths[u][3] = ((2*i+1)*(q-1));
        }
        paths[p*(q-1)-1][1] = q;
        paths[p*(q-1)-1][2] = q-1;
        link tLink = new link(crossings,paths);
        return tLink;
    }
    
    int nplus(link Link) {
        if (Link.components.isEmpty()) return 0;
        int n = 0;
        int a,b,c,d;
        int co1,co2;
        int in1,in2,in3,in4;
        boolean rup = true;
        boolean lup = true;
        ArrayList<Integer> used = new ArrayList<Integer>();
        for (int r = 0; r < Link.crossings.length; r++) {
            a = Link.paths[r][0];
            b = Link.paths[r][1];
            c = Link.paths[r][2];
            d = Link.paths[r][3];
            if (Link.crossings[r]%2 == 0) {
                co1=0;
                co2=0;
                while (true) {
                    if (Link.components.get(co1).contains(a)) break;
                    co1++;
                }
                in1 = Link.components.get(co1).indexOf(a);
                in2 = Link.components.get(co1).indexOf(b);
                if ((in1 == 0) | (in2 == 0)) {
                    if (in1 == 0) {
                        if (!used.contains(a)) {
                            used.add(a);
                            rup = true;
                        }
                        else rup = false;
                    }
                    else {
                        if (!used.contains(b)) {
                            used.add(b);
                            rup = false;
                        }
                        else rup = true;
                    }
                }
                else {
                    if (in1 < in2) rup = true;
                    else rup = false;
                }
                while (true) {
                    if (Link.components.get(co2).contains(c)) break;
                    co2++;
                }
                in3 = Link.components.get(co2).indexOf(c);
                in4 = Link.components.get(co2).indexOf(d);
                if ((in3 == 0) | (in4 == 0)) {
                    if (in3 == 0) {
                        if (!used.contains(c)) {
                            used.add(c);
                            lup = false;
                        }
                        else lup = true;
                    }
                    else {
                        if (!used.contains(d)) {
                            used.add(d);
                            lup = true;
                        }
                        else lup = false;
                    }
                }
                else {
                    if (in3 < in4) lup = false;
                    else lup = true;
                }
                int factor = 1;
                if (!rup) factor = (-1);
                if (!lup) factor = factor * (-1);
                if (factor * Link.crossings[r] > 0) n = n + factor * 
                    Link.crossings[r];
            }
            else {
                co1=0;
                co2=0;
                while (true) {
                    if (Link.components.get(co1).contains(a)) break;
                    co1++;
                }
                in1 = Link.components.get(co1).indexOf(a);
                in2 = Link.components.get(co1).indexOf(c);
                if ((in1 == 0) | (in2 == 0)) {
                    if (in1 == 0) {
                        if (!used.contains(a)) {
                            used.add(a);
                            rup = true;
                        }
                        else rup = false;
                    }
                    else {
                        if (!used.contains(c)) {
                            used.add(c);
                            rup = false;
                        }
                        else rup = true;
                    }
                }
                else {
                    if (in1 < in2) rup = true;
                    else rup = false;
                }
                while (true) {
                    if (Link.components.get(co2).contains(b)) break;
                    co2++;
                }
                in3 = Link.components.get(co2).indexOf(b);
                in4 = Link.components.get(co2).indexOf(d);
                if ((in3 == 0) | (in4 == 0)) {
                    if (in3 == 0) {
                        if (!used.contains(b)) {
                            used.add(b);
                            lup = false;
                        }
                        else lup = true;
                    }
                    else {
                        if (!used.contains(d)) {
                            used.add(d);
                            lup = true;
                        }
                        else lup = false;
                    }
                }
                else {
                    if (in3 < in4) lup = false;
                    else lup = true;
                }
                int factor = 1;
                if (!rup) factor = (-1);
                if (!lup) factor = factor * (-1);
                if (factor * Link.crossings[r] > 0) n = n + factor * 
                    Link.crossings[r];
            }
        }
        return n;
    }
    
    link Mirror(link Link) {
        int [] mcross = new int[Link.crossings.length];
        for (int u=0; u < Link.crossings.length; u++) 
            mcross[u] = -Link.crossings[u];
        link mirror = new link(mcross, Link.paths);
        return mirror;
    }
	
    int nminus(link Link) {
        return nplus(Mirror(Link));
    }
    
    
	
    boolean samePaths(ArrayList<ArrayList<Integer>> paths0, ArrayList<ArrayList<Integer>> paths1) {
        if (paths0.size()!=paths1.size()) return false;
        boolean ok = true;
        int i = 0;
        while (ok & i < paths0.size()) {
            ArrayList<Integer> path = paths0.get(i);
            boolean found = false;
            int j = 0;
            while (!found) {
                if (paths1.get(j).contains(path.get(0))) found = true;
                else j++;
            }
            if (!paths1.get(j).contains(path.get(1))) ok = false; 
            i++;
        }
        return ok;
    }
	
    JonesObject tensorObject(JonesObject obj0, JonesObject obj1, ArrayList<Integer> endpoints) {
        JonesObject newObj;
        ArrayList<ArrayList<Integer>> paths = pathsFrom(obj0.paths,obj1.paths,endpoints);
        ArrayList<Integer> newcircs = getNewCircles(paths);
        for (int y : newcircs) {
            boolean found = false;
            int z = 0;
            while (!found & z < paths.size()) {
                ArrayList<Integer> path = paths.get(z);
                if (path.contains(y)) found = true;
                else z++;
            }
            if (found) paths.remove(paths.get(z));
        }
        for (ArrayList<Integer> path : paths) {
            while (path.size()>2) path.remove(1);
        }
        polynomial a = obj0.poly;
        polynomial b = obj1.poly;
        polynomial ab = multiplyPoly(a,b);
        if (!newcircs.isEmpty()) {
            polynomial circlePol = new polynomial();
            coefficient top = new coefficient(-1,-2);
            coefficient bot = new coefficient(-1,2);
            circlePol.entries.add(top);
            circlePol.entries.add(bot);
            if (endpoints.isEmpty()) {
                if (newcircs.size() == 2) ab = multiplyPoly(ab,circlePol);
            }
            else for (int r = 0; r < newcircs.size(); r++) ab = multiplyPoly(ab,circlePol);
        }
        newObj = new JonesObject(paths,ab);
        return newObj;
    }
    
    int smallest(ArrayList<Integer> path) {
        int min = path.get(0);
        for (int i = 1; i < path.size(); i++) 
            if (path.get(i)<min) min = path.get(i);
        return min;
    }
    
    ArrayList<Integer> getNewCircles(ArrayList<ArrayList<Integer>> paths) {
        ArrayList<Integer> newcircs = new ArrayList<Integer>();
        int i = 0;
        while (i < paths.size()) {
            if (startPoint(paths.get(i)) == endPoint(paths.get(i))) {
                newcircs.add(smallest(paths.get(i)));
                paths.remove(i);
            }
            else i++;
        }
        Collections.sort(newcircs);
        return newcircs;
    }

    ArrayList<ArrayList<Integer>> pathsFrom(ArrayList<ArrayList<Integer>> paths0,ArrayList<ArrayList<Integer>> paths1,
            ArrayList<Integer> endpoints) {
        ArrayList<ArrayList<Integer>> newpaths = new ArrayList<ArrayList<Integer>>();
        for (ArrayList<Integer> path : paths0) {
            ArrayList<Integer> npath = new ArrayList<Integer>();
            for (int e : path) npath.add(e);
            newpaths.add(npath);
        }
        for (ArrayList<Integer> path : paths1) {
            if (endpoints.contains(path.get(0)) & endpoints.contains(path.get(1))) newpaths.add(path);
            else {
                boolean found1 = false;
                boolean found2 = false;
                if (endpoints.contains(path.get(0))) found1 = true;
                if (endpoints.contains(path.get(1))) found2 = true;
                int i = 0;
                ArrayList<Integer> pth = new ArrayList<Integer>();
                pth.add(path.get(0));
                pth.add(path.get(1));
                while (!found1 | !found2) {
                    ArrayList<Integer> pat = newpaths.get(i);
                    if (!found2 & pat.contains(path.get(1))) {
                        found2 = true;
                        pth = mergePath(pth,pat);
                        if (pat.contains(pth.get(0))) found1 = true;
                        newpaths.remove(pat);
                    }
                    else if (!found1 & pat.contains(path.get(0))) {
                        found1 = true;
                        pth = mergePath(pth,pat);
                        newpaths.remove(pat);
                        i = 0;
                    }
                    else i++;
                    if (i >= newpaths.size()) break;
                }
                newpaths.add(pth);
            }
        }
        return newpaths;
    }

    ArrayList<Integer> mergePath(ArrayList<Integer> first, ArrayList<Integer> second) {
        ArrayList<Integer> path = new ArrayList<Integer>();
        if (startPoint(first) == startPoint(second)) {
            for (int u = first.size()-1; u >=0; u--) path.add(first.get(u));
            for (int u = 1; u < second.size(); u++) path.add(second.get(u));
        }
        else if (startPoint(first) == endPoint(second)) {
            for (int u = first.size()-1; u >0; u--) path.add(first.get(u));
            for (int u = second.size()-1; u >= 0; u--) path.add(second.get(u));
        }
        else if (endPoint(first) == startPoint(second)) {
            for (int u = 0; u < first.size()-1; u++) path.add(first.get(u));
            for (int u = 0; u < second.size(); u++) path.add(second.get(u));
        }
        else if (endPoint(first) == endPoint(second)) {
            for (int u = 0; u < first.size()-1; u++) path.add(first.get(u));
            for (int u = second.size()-1; u >= 0; u--) path.add(second.get(u));
        }

        return path;
    }
    
    int startPoint(ArrayList<Integer> path) {
        return path.get(0);
    }

    int endPoint(ArrayList<Integer> path) {
        int w = path.size()-1;
        return path.get(w);
    }

    JonesCategory startCat(int u, int[] path) {
        JonesCategory cat = new JonesCategory();
        ArrayList<Integer> path1 = new ArrayList<Integer>();
        path1.add(path[0]);
        path1.add(path[1]);
        ArrayList<Integer> path2 = new ArrayList<Integer>();
        path2.add(path[2]);
        path2.add(path[3]);
        ArrayList<ArrayList<Integer>> paths1 = new ArrayList<ArrayList<Integer>>();
        paths1.add(path1);
        paths1.add(path2);
        JonesObject first = new JonesObject(paths1,1,u);
        cat.objects.add(first);
        if (u != 0) {
            ArrayList<Integer> path3 = new ArrayList<Integer>();
            path3.add(path[0]);
            path3.add(path[3]);
            ArrayList<Integer> path4 = new ArrayList<Integer>();
            path4.add(path[2]);
            path4.add(path[1]);
            ArrayList<ArrayList<Integer>> paths2 = new ArrayList<ArrayList<Integer>>();
            paths2.add(path3);
            paths2.add(path4);
            ArrayList<Integer> entries = new ArrayList<Integer>();
            ArrayList<Integer> powers = new ArrayList<Integer>();
            int entry = 1, pwr = Math.abs(u)-2,factor = 1;
            if (u<0) factor = -1;
            pwr = pwr * factor;
            for (int v = 0; v < Math.abs(u); v++) {
                entries.add(entry);
                entry = entry * (-1);
                powers.add(pwr-4*factor*v);
            }
            JonesObject second = new JonesObject(paths2,entries,powers);
            cat.objects.add(second);
        }
        for (int y = 0; y < 4; y++) {
            if (!cat.endpoints.contains(path[y])) cat.endpoints.add(path[y]);
            else cat.endpoints.remove(cat.endpoints.indexOf(path[y]));
        }
        polynomial circlePol = new polynomial();
            coefficient top = new coefficient(-1,-2);
            coefficient bot = new coefficient(-1,2);
            circlePol.entries.add(top);
            circlePol.entries.add(bot);
        for (JonesObject obj : cat.objects) {
            ArrayList<Integer> circles = getNewCircles(obj.paths);
            for (int r = obj.paths.size()-1; r >=0 ; r--) {
                int st = startPoint(obj.paths.get(r));
                int en = endPoint(obj.paths.get(r));
                if (!cat.endpoints.contains(st) & !cat.endpoints.contains(en))
                    obj.paths.remove(r);
            }
            if (!circles.isEmpty()) obj.poly = multiplyPoly(obj.poly,circlePol);
        }
        int w = 0;
            while (w < cat.objects.size()-1) {
                JonesObject obj0 = cat.objects.get(w);
                int topp = cat.objects.size()-1;
                for (int v = topp; v > w; v--) {
                    JonesObject obj1 = cat.objects.get(v);
                    if (samePaths(obj0.paths,obj1.paths)) {
                        obj0.poly.add(obj1.poly);
                        cat.objects.remove(obj1);
                    }
                }
                w++;
            }
        return cat;
    }

    polynomial multiplyPoly(polynomial a, polynomial b) {
        polynomial c = new polynomial();
        for (int i = 0; i < a.entries.size(); i++) {
            polynomial d = new polynomial();
            for (int j = 0; j < b.entries.size(); j++) {
                coefficient x = a.entries.get(i);
                coefficient y = b.entries.get(j);
                coefficient rag = new coefficient(x.entry * y.entry, x.power + y.power);
                d.entries.add(rag);
            }
            c.add(d);
        }
        c.sort();
        return c;
    }
    
    int[][] copyOf(int[][] orig) {
        int[][] copy = new int[orig.length][4];
        for (int i = 0; i < orig.length; i++) {
            for (int j = 0; j < 4; j++) copy[i][j] = orig[i][j];
        }
        return copy;
    }
    
    String GaussCode(link Link) {
        String gauss = "";
        int[] crossings = Link.crossings;
        int[][] paths = new int[crossings.length][4];
        for (int y = 0; y < crossings.length; y++) {
            for (int t = 0; t < 4 ; t++) paths[y][t] = Link.paths[y][t];
        }
        ArrayList<ArrayList<Integer>> comps = Link.components;
        for (ArrayList<Integer> comp : comps) {
            for (int c : comp) {
                boolean found = false;
                int i = 0;
                int j = 0;
                while (!found) {
                    j = 0;
                    while (!found & j < 4) {
                        if (paths[i][j] == c) found = true;
                        else j++;
                    }
                    if (!found) i++;
                }
                paths[i][j] = 0;
                if (crossings[i]%2 == 0) {
                    if (j == 0) paths[i][1] = 0;
                    if (j == 1) paths[i][0] = 0;
                    if (j == 2) paths[i][3] = 0;
                    if (j == 3) paths[i][2] = 0;
                }
                else{
                    if (j == 0) paths[i][2] = 0;
                    if (j == 1) paths[i][3] = 0;
                    if (j == 2) paths[i][0] = 0;
                    if (j == 3) paths[i][1] = 0;
                }
                if (crossings[i]!=0) {
                    int pos = gaussPos(crossings,i,j);
                    boolean up = (j == 0 | j == 3);
                    boolean plus = false;
                    if (crossings[i] < 0) {
                        if (j == 3 | j == 1) gauss = gauss+"-";
                        else plus = true;
                    }
                    else {
                        if (j == 0 | j == 2) gauss = gauss+"-";
                        else plus = true;
                    }
                    gauss = gauss+String.valueOf(pos);
                    int factor = -1;
                    if (up) factor = 1;
                    for (int t = 1; t < Math.abs(crossings[i]); t++) {
                        gauss = gauss +",";
                        if (plus) {
                            gauss = gauss + "-";
                        }
                        plus = !plus;
                        gauss = gauss + String.valueOf(pos+factor*t);
                    }
                    if (comp.indexOf(c) != comp.size()-1) gauss = gauss+",";
                }
            }
            if (comps.indexOf(comp) != comps.size()-1) gauss = gauss+":";
        }
        return gauss;
    }
    
    int gaussPos(int[] cross, int i, int j) {
        int pos = 0;
        for (int k = 0; k < i; k++) pos = pos + Math.abs(cross[k]);
        if (j == 1 | j == 2) pos = pos + Math.abs(cross[i]);
        else pos++;
        return pos;
    }
    
    String DowkerThistle(link Link) { 
        String dowker = "";
        if (Link.components.size()> 1) return "Not implemented for links";
        int[] crossings = Link.crossings;
        int[][] paths = Link.paths;
        int[] checked = new int[paths.length];
        boolean[] beenThere = new boolean[paths.length];
        int[][] pats = copyOf(paths);
        int start = paths[0][0];
        int runner;
        int v;
        if (crossings[0]%2 == 0) v = 1;
        else v = 2;
        runner = paths[0][v];
        int cross = Math.abs(crossings[0]);
        int cros = cross;
        ArrayList<coefficient> crosses = new ArrayList<coefficient>();
        for (int y = 1; y <= cross; y++) {
            int z = y;
            if (z%2 == 0 & crossings[0]>0) z = -z;
            coefficient first = new coefficient(z,0);
            crosses.add(first);
        }
        pats[0][0] = -1;
        if (crossings[0]%2 == 0) pats[0][1] = 0;
        else pats[0][2] = 0;
        checked[0] = cros;
        beenThere[0] = true;
        while (runner != start) {
            boolean found = false;
            int u = 0;
            v = 0;
            while (!found) {
                int[] pts = pats[u];
                v = 0;
                while (!found & v < 4) {
                    if (pts[v] == runner) found = true;
                    else v++;
                }
                if (!found) u++;
            }
            boolean beenHere = false;
            int here = 0;
            if (beenThere[u]) {
                beenHere = true;
                here = checked[u];
            }
            else {
                cros = cros + Math.abs(crossings[u]);
                checked[u] = cros;
                beenThere[u] = true;
            }
            for (int y = 1; y <= Math.abs(crossings[u]); y++) {
                int z = cross + y;
                if (crossings[u] < 0) {
                    if (v == 1 | v == 3) { //for odd y get undercrossing
                        if (z % 2 == 0 & y % 2 == 0) z = -z;
                    }
                    else { // for even y get undercrossing
                        if (z % 2 == 0 & y % 2 != 0) z = -z;
                    }
                }
                else {
                    if (v == 1 | v == 3) {
                        if (z % 2 == 0 & y % 2 != 0) z = -z;
                    }
                    else { // for even y get undercrossing
                        if (z % 2 == 0 & y % 2 == 0) z = -z;
                    }
                }
                if (!beenHere) { // just get a new entry
                    coefficient next = new coefficient(z,0);
                        crosses.add(next);
                }
                else { // more involved
                    boolean opposite = true;
                    switch(v) {
                        case 0 : if (pats[u][3] == -1) opposite = false; break;
                        case 1 : if (pats[u][2] == -1) opposite = false; break;
                        case 2 : if (pats[u][1] == -1) opposite = false; break;
                        case 3 : if (pats[u][0] == -1) opposite = false; break;
                    }
                    if (opposite) {
                        coefficient next = crosses.get(here-y);
                        next.power = z;
                    }
                    else {
                        coefficient next = crosses.get(here - Math.abs(crossings[u]) - 1 + y);
                        next.power = z;
                    }
                }
            }
            cross = cross + Math.abs(crossings[u]);
            pats[u][v] = -1;
            switch(v) {
                case 0 : if (crossings[u]%2 == 0) v = 1; else v = 2; break;
                case 1 : if (crossings[u]%2 == 0) v = 0; else v = 3; break;
                case 2 : if (crossings[u]%2 == 0) v = 3; else v = 0; break;
                case 3 : if (crossings[u]%2 == 0) v = 2; else v = 1; break;
            }
            pats[u][v] = 0;
            runner = paths[u][v];
        }
        int run = 3;
        dowker = String.valueOf(crosses.get(0).power);
        while (run < 2 * crosses.size()) {
            boolean found = false;
            int y = 0;
            while (!found) {
                coefficient ct = crosses.get(y);
                if (Math.abs(ct.entry) == run) found = true;
                if (Math.abs(ct.power) == run) found = true;
                if (!found) y++;
            }
            int str = 0;
            if (crosses.get(y).entry == run ) str = crosses.get(y).power;
            if (crosses.get(y).entry == -run ) str = -crosses.get(y).power;
            if (crosses.get(y).power == run ) str = crosses.get(y).entry;
            if (crosses.get(y).power == -run ) str = -crosses.get(y).entry;
            dowker = dowker + " " + String.valueOf(str);
            run = run + 2;
        }
        return dowker;
    }
    
    private class coefficient implements Comparable<coefficient>{
        int entry;
        int power;
		
        coefficient(int ent, int pow) {
            entry = ent;
            power = pow;
        }

        boolean sameEntry(coefficient comp) {
            boolean same = true;
            if (entry != comp.entry) same = false;
            if (power != comp.power) same = false;
            return same;
        }

        @Override
        public int compareTo(coefficient coeff) {
            return power - coeff.power;
        }
    }
	
    private class polynomial {
        ArrayList<coefficient> entries;

        polynomial () {
            entries = new ArrayList<coefficient>();
        }

        polynomial(int ent, int pow) {
            entries = new ArrayList<coefficient>();
            coefficient entry = new coefficient(ent,pow);
            entries.add(entry);
        }

        polynomial(int[] ent, int[] pow) {
            entries = new ArrayList<coefficient>();
            for (int u = 0; u < ent.length; u++) {
                coefficient entry = new coefficient(ent[u],pow[u]);
                entries.add(entry);
            }
        }

        polynomial(ArrayList<Integer> ent, ArrayList<Integer> pow) {
            entries = new ArrayList<coefficient>();
            for (int u = 0; u < ent.size();u++) {
                coefficient entry = new coefficient(ent.get(u),pow.get(u));
                entries.add(entry);
            }
        }

        void add(polynomial c) {
            for (coefficient entry : c.entries) {
                boolean found = false;
                int u = 0;
                while (!found & u < this.entries.size()) {
                    if (this.entries.get(u).power == entry.power) found = true;
                    else u++;
                }
                if (!found) this.entries.add(entry);
                else {
                    this.entries.get(u).entry = this.entries.get(u).entry + entry.entry;
                    if (this.entries.get(u).entry == 0) this.entries.remove(u);
                }
            }
            this.sort();
        }

        String getJonesPol(boolean reduced, boolean latex) {
            String word;
            char JonesPolLabel = 't';
            if (!reduced) JonesPolLabel = 'q';
            String rub = "";
            String power = "";
            int divider = 4;
            String divid = "";
            String latop = "";
            String labot = "";
            if (latex) {
                latop = "{";
                labot = "}";
            }
            if (entries.get(0).power % 4 != 0) {
                divider = 2;
                divid = "/2";
            }
            if (entries.get(0).entry == 1) rub = "";
            if (entries.get(0).entry == -1) rub = "-";
            if (Math.abs(entries.get(0).entry)>1) rub = String.valueOf(entries.get(0).entry);
            if (entries.get(0).power != 0) 
                power = String.valueOf(JonesPolLabel)+"^"+latop+String.valueOf(((-1)*entries.get(0).power)/divider)+divid+labot;
            else {
                if (Math.abs(entries.get(0).entry) == 1) power = "1";
                    else power = "";
            }
            word = rub+power;
            for (int u = 1; u < entries.size(); u++) {
                int coef = entries.get(u).entry;
                int pwr = entries.get(u).power;
                if (coef == 1) word = word +" +";
                if (coef > 1 ) word = word +" +"+String.valueOf(coef);
                if (coef == -1) word = word +" -";
                if (coef < -1 ) word = word +" "+String.valueOf(coef);
                if (entries.get(u).power != 0)
                    power = String.valueOf(JonesPolLabel)+"^"+latop+String.valueOf(((-1)*pwr)/divider)+divid+labot;
                else {
                    if (Math.abs(entries.get(u).entry) == 1) power = "1";
                    else power = "";
                }
                word = word + power;
            }
            return word;
        }

        boolean samePoly(polynomial comp) {
            boolean same = true;
            if (entries.size()!=comp.entries.size()) return false;
            int u = 0;
            while (same & u < entries.size()) {
                if (!entries.get(u).sameEntry(comp.entries.get(u))) same = false;
                else u++;
            }
            return same;
        }

        void sort() {
            Collections.sort(entries);
        }
    }
	
    private class JonesCategory{
        ArrayList<JonesObject> objects = new ArrayList<JonesObject>();
        ArrayList<Integer> endpoints = new ArrayList<Integer>();
    }
	
    private class JonesObject{
        ArrayList<ArrayList<Integer>> paths;
        polynomial poly;

        JonesObject(ArrayList<ArrayList<Integer>> pts, int ent, int pow) {
            paths = new ArrayList<ArrayList<Integer>>(pts);
            poly = new polynomial(ent,pow);
        }

        JonesObject(ArrayList<ArrayList<Integer>> pts, polynomial pol) {
            paths = pts;
            poly = pol;
        }

        JonesObject(ArrayList<ArrayList<Integer>> pts, ArrayList<Integer> ent, ArrayList<Integer> pow) {
            paths = new ArrayList<ArrayList<Integer>>(pts);
            poly = new polynomial(ent,pow);
        }
    }
    
    KhovCategory startCat(int crossing, int[] paths, int nminus, int nplus, int n, boolean red) {
        KhovCategory start = new KhovCategory();
        for (int m : paths) {
            if (!start.endpoints.contains(m)) start.endpoints.add(m);
            else start.endpoints.remove((Integer) m);
        }
        switch (crossing) {
        case 1 : start = posCross(start.endpoints,paths,nminus,nplus,true);
        break;
        case 0 : start = zeroCross(start.endpoints,paths,nminus,nplus,false);
            break;
        case -1 : start = posCross(start.endpoints,paths,nminus,nplus,false);
        break;
        }
        if (crossing > 1) {
            boolean a = true;
            if (n > 2) a = false;
            start = posCross(start.endpoints,paths,nminus,nplus,a);
            if (n>2) {
                start.objects.get(0).hdeg = crossing - 1;
                start.objects.get(0).qdeg = ((-1) * crossing +1) * n -1;
                start.objects.get(1).hdeg = crossing;
                start.objects.get(1).qdeg = (-1) * crossing * n;
            }
            extendCat(start,crossing,nminus,nplus,n);
        }
        if (crossing < -1) {
            boolean b = false;
            if (n > 2) b = true;
            start = posCross(start.endpoints,paths,nminus,nplus,b);
            if (n>2) {
                start.objects.get(0).hdeg = crossing;
                start.objects.get(0).qdeg = -crossing * n;
                start.objects.get(1).hdeg = crossing+1;
                start.objects.get(1).qdeg = -crossing * n - n + 1;
            }
            extendCat(start,crossing,nminus,nplus,n);
        }
        if (red) {
            for (Khov_object obj : start.objects) obj.dotpath = 1;
        }
        if (start.endpoints.size()!= 4) {
            KhovCategory newCat = new KhovCategory();
            newCat.endpoints = start.endpoints;
            for (Khov_object obj : start.objects) {
                ArrayList<Integer> newcircs = getNewCircles(obj.paths);
                if (newcircs.isEmpty() & start.endpoints.isEmpty()) newcircs.add(smallest(obj.paths.get(0)));
                ArrayList<circle> circles = circlesFrom(newcircs,n);
                ArrayList<ArrayList<Integer>> pats = new ArrayList<ArrayList<Integer>>();
                ArrayList<Khov_object> newObjects = new ArrayList<Khov_object>();
                if (newcircs.isEmpty()) {
                    pats.add(start.endpoints);
                    Khov_object newobj = new Khov_object(pats,obj.hdeg,obj.qdeg);
                    newObjects.add(newobj);
                }
                else {
                    for (ArrayList<Integer> path : obj.paths) {
                        boolean found = false;
                        int u = 0;
                        while (!found & u < newcircs.size()) {
                            if (path.contains(newcircs.get(u))) found = true;
                            else u++;
                        }
                        if (!found) pats.add(path);
                    }
                    newObjects = newKhovObjects(pats,newcircs,circles,n,obj.hdeg,obj.qdeg,0); // something that might need change in reduced case
                }
                obj.descObjects = newObjects;
                for (Khov_object ob : newObjects) newCat.objects.add(ob);
            }
            for (Khov_object obj0 : start.objects) {
                for (Kmod0 mod : obj0.topMods) {
                    Khov_object obj1 = mod.topObj;
                    for (Khov_object newObj0 : obj0.descObjects) {
                        for (Khov_object newObj1 : obj1.descObjects) {
                            ArrayList<Integer> modif = mod.actions.get(0);
                            boolean noway = false;
                            if (mod.actions.size() > 1) noway = true;
                            if (modif.size() == 4) { // surgery
                                if (newObj0.circles.size() > newObj1.circles.size()) {
                                    if (newObj0.circles.size() == 1) {
                                        int sign = 1;
                                        if (n == 2) sign = -1;
                                        int u = -(sign * newObj0.circles.get(0).sign + n-1)/2;
                                        ArrayList<Integer> action = new ArrayList<Integer>();
                                        action.add(start.endpoints.get(0));
                                        action.add(start.endpoints.get(1));
                                        action.add(u);
                                        ArrayList<ArrayList<Integer>> akschns = new ArrayList<ArrayList<Integer>>();
                                        akschns.add(action);
                                        Kmod0 mod0;
                                        if (u == 0) mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                        else mod0 = new Kmod0(newObj1,newObj0,mod.totalValue,akschns);
                                        newObj0.topMods.add(mod0);
                                        newObj1.botMods.add(mod0);
                                    }
                                    else {
                                        if (newObj0.qdeg == newObj1.qdeg) {
                                            Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                            newObj0.topMods.add(mod0);
                                            newObj1.botMods.add(mod0);
                                        }
                                    }
                                }
                                else {
                                    if (newObj1.circles.size() == 1) {
                                        int sign = 1;
                                        if (n == 2) sign = -1;
                                        int u = (sign * newObj1.circles.get(0).sign + 1 - n)/2;
                                        ArrayList<Integer> action = new ArrayList<Integer>();
                                        action.add(start.endpoints.get(0));
                                        action.add(start.endpoints.get(1));
                                        action.add(u);
                                        ArrayList<ArrayList<Integer>> akschns = new ArrayList<ArrayList<Integer>>();
                                        akschns.add(action);
                                        Kmod0 mod0;
                                        if (u == 0) mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                        else mod0 = new Kmod0(newObj1,newObj0,mod.totalValue,akschns);
                                        newObj0.topMods.add(mod0);
                                        newObj1.botMods.add(mod0);
                                    }
                                    else {
                                        if (newObj0.qdeg == newObj1.qdeg) {
                                            Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                            newObj0.topMods.add(mod0);
                                            newObj1.botMods.add(mod0);
                                        }
                                    }
                                }
                            }
                            else { // dotting
                                if (newObj0.circles.isEmpty()) {
                                    ArrayList<Integer> action = new ArrayList<Integer>();
                                    action.add(start.endpoints.get(0));
                                    action.add(start.endpoints.get(1));
                                    action.add(mod.actions.get(0).get(2));
                                    ArrayList<ArrayList<Integer>> akschns = new ArrayList<ArrayList<Integer>>();
                                    akschns.add(action);
                                    if (noway) action.set(2, -(n-1));
                                    Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue,akschns);
                                    newObj0.topMods.add(mod0);
                                    newObj1.botMods.add(mod0);
                                }
                                if (newObj0.circles.size() == 1) {
                                    int circ = newObj0.circles.get(0).comp;
                                    int sign = newObj0.circles.get(0).sign;
                                    int newsign;
                                    if (modif.contains(circ)) {
                                        if (n == 2) newsign = sign - 2;
                                        else newsign = sign - 2 * modif.get(2);
                                        if (noway & start.endpoints.isEmpty()) newsign = newsign - 2 * mod.actions.get(1).get(2);
                                        if (newObj1.circles.get(0).sign == newsign) {
                                            Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                            newObj0.topMods.add(mod0);
                                            newObj1.botMods.add(mod0);
                                            if (noway & !start.endpoints.isEmpty()) {
                                                ArrayList<Integer> accion = new ArrayList<Integer>();
                                                accion.add(start.endpoints.get(0));
                                                accion.add(start.endpoints.get(1));
                                                accion.add(mod.actions.get(1).get(2));
                                                mod0.actions.add(accion);
                                            }
                                        }
                                    }
                                    else {
                                        int mover = newObj1.circles.get(0).sign;
                                        if (noway) mover = mover + 2 * mod.actions.get(1).get(2);
                                        if (newObj0.circles.get(0).sign == mover) {
                                            ArrayList<Integer> accion = new ArrayList<Integer>();
                                            accion.add(start.endpoints.get(0));
                                            accion.add(start.endpoints.get(1));
                                            accion.add(mod.actions.get(0).get(2));
                                            ArrayList<ArrayList<Integer>> akschns = new ArrayList<ArrayList<Integer>>();
                                            akschns.add(accion);
                                            Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue,akschns);
                                            newObj0.topMods.add(mod0);
                                            newObj1.botMods.add(mod0);
                                        }
                                    }
                                }
                                if (newObj0.circles.size() == 2) {
                                    int circ0 = newObj0.circles.get(0).comp;
                                    int sign0 = newObj0.circles.get(0).sign;
                                    int sign1 = newObj0.circles.get(1).sign;
                                    int newsign0,newsign1;
                                    if (modif.contains(circ0)) {
                                        if (n == 2) newsign0 = sign0 - 2;
                                        else newsign0 = sign0 - 2 * modif.get(2);
                                        if (!noway) newsign1 = sign1;
                                        else newsign1 = sign1 - 2 * mod.actions.get(1).get(2);
                                    }
                                    else {
                                        if (n == 2) newsign1 = sign1 - 2;
                                        else newsign1 = sign1 - 2 * modif.get(2);
                                        if (!noway) newsign0 = sign0;
                                        else newsign0 = sign0 - 2 * mod.actions.get(1).get(2);
                                    }
                                    if (newObj1.circles.get(0).sign == newsign0 & newObj1.circles.get(1).sign == newsign1) {
                                        Kmod0 mod0 = new Kmod0(newObj1,newObj0,mod.totalValue);
                                        newObj0.topMods.add(mod0);
                                        newObj1.botMods.add(mod0);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            start = newCat;
            startWhitneyfy(start.objects);
            whitneyfy(start.objects);
            if (red) {
                for (Khov_object obj : start.objects) obj.dotpath = 1;
                ArrayList<Integer> badboys = new ArrayList<Integer>();
                for (int i = 0; i < start.objects.size(); i++) {
                    Khov_object obj = start.objects.get(i);
                    for (circle cir : obj.circles) if (cir.comp == 1 & cir.sign == 1) badboys.add(i);
                }
                for (int r : badboys) {
                    Khov_object obj = start.objects.get(r);
                    for (Kmod0 mod : obj.botMods) {
                        Khov_object ob = mod.botObj;
                        ob.topMods.remove(mod);
                    }
                    for (Kmod0 mod : obj.topMods) {
                        Khov_object ob = mod.topObj;
                        ob.botMods.remove(mod);
                    }
                }
                for (int r = badboys.size()-1; r >=0 ; r--) {
                    Khov_object obj = start.objects.get(badboys.get(r));
                    start.objects.remove(obj);
                }
            }
        }
        return start;
    }

    void allCatOutput(KhovCategory cat) {
        if (cat.objects.isEmpty()) return;
        ArrayList<ArrayList<Khov_object>> objects = new ArrayList<ArrayList<Khov_object>>();
        int hmin = cat.objects.get(0).hdeg;
        int hmax = hmin;
        for (Khov_object obj : cat.objects) {
            if (hmin > obj.hdeg) hmin = obj.hdeg;
            if (hmax < obj.hdeg) hmax = obj.hdeg;
        }
        for (int i = 0; i <= hmax - hmin; i++) {
            ArrayList<Khov_object> objs = new ArrayList<Khov_object>();
            objects.add(objs);
        }
        for (Khov_object obj : cat.objects) {
            int ind = obj.hdeg-hmin;
            objects.get(ind).add(obj);
        }
        for (int ht = hmax - hmin; ht > 0; ht--) {
            System.out.println(ht);
            catOutput(objects.get(ht),objects.get(ht-1),true);
            catOutput(objects.get(ht),objects.get(ht-1),false);
        }
        System.out.println("Done");
    }
    
    void catOutput(ArrayList<Khov_object> objtop, ArrayList<Khov_object> objbot, boolean down) {
        if (down) {
            for (int i = 0; i < objtop.size(); i++) {
                System.out.println("T("+i+")");objectOutput(objtop.get(i));
                for (Kmod0 mod : objtop.get(i).botMods) {
                    int p = objbot.indexOf(mod.botObj);
                    System.out.println(mod.totalValue+" B("+p+")");

                }
                System.out.println();
            }
        }
        else {
            for (int i = 0; i < objbot.size(); i++) {
                System.out.println("B("+i+")");
                for (Kmod0 mod : objbot.get(i).topMods) {
                    int p = objtop.indexOf(mod.topObj);
                    System.out.println(mod.totalValue+" T("+p+")");
                    if (p==-1) {
                        objectOutput(mod.topObj);
                        objectOutput(mod.botObj);
                    }
                }
                System.out.println();
            }
        }
        System.out.println("- - - - - "+down);
    }
    
    void objectOutput(Khov_object obj, boolean mods) {
		objectOutput(obj);
		if (mods) {
                    for (Kmod0 point : obj.botMods) {
                        Khov_object botObj = point.botObj;
                        System.out.print(point.totalValue);
                        for (ArrayList<Integer> act : point.actions) System.out.print(" "+act);
                        System.out.println();
                        objectOutput(botObj);
                    }
		}
		System.out.println();
	}
	
	void objectOutput(Khov_object obj) {
            System.out.print("  [ "+obj.hdeg+" "+obj.qdeg+" ]");
            System.out.print(obj.paths+" ");
        for (circle circ : obj.circles) System.out.print("("+circ.comp+") "+circ.sign+" ");
        System.out.println("h = "+obj.hdeg+" q = "+obj.qdeg);
	}
    
    void extendCatn(KhovCategory cat, int crossing, int n) { // the case n > 2 will look rather different from n = 2
        int hd = 0;
        int qd = 0;
        if (crossing > 1) {
            Khov_object tomod = cat.objects.get(0);
            qd = tomod.qdeg;
            for (int i = 2; i <= crossing; i++) {
                hd = crossing - i;
                if (i % 2 == 0) qd = qd+2;
                else qd = qd + 2 * (n-1);
                ArrayList<ArrayList<Integer>> paths = cloneActions(tomod.paths);
                Khov_object newObj = new Khov_object(paths,hd,qd);
                cat.objects.add(0,newObj);
                if (i % 2 == 0) {
                    ArrayList<ArrayList<Integer>> posAction = new ArrayList<ArrayList<Integer>>();
                    ArrayList<ArrayList<Integer>> negAction = new ArrayList<ArrayList<Integer>>();
                    ArrayList<Integer> action0 = new ArrayList<Integer>();
                    for (int e : tomod.paths.get(0)) action0.add(e);
                    action0.add(-1);
                    ArrayList<Integer> action1 = new ArrayList<Integer>();
                    for (int e : tomod.paths.get(1)) action1.add(e);
                    action1.add(-1);
                    posAction.add(action0);
                    negAction.add(action1);
                    Kmod0 modp = new Kmod0(cat.objects.get(1), cat.objects.get(0), BigInteger.ONE, posAction);
                    Kmod0 modn = new Kmod0(cat.objects.get(1), cat.objects.get(0), BigInteger.ONE.negate(), negAction);
                    cat.objects.get(1).botMods.add(modp);
                    cat.objects.get(1).botMods.add(modn);
                    cat.objects.get(0).topMods.add(modp);
                    cat.objects.get(0).topMods.add(modn);
                }
                else {
                    for (int e = 0; e < n; e++) {
                        ArrayList<Integer> action0 = new ArrayList<Integer>();
                        for (int f : tomod.paths.get(0)) action0.add(f);
                        action0.add(-e);
                        ArrayList<Integer> action1 = new ArrayList<Integer>();
                        for (int f : tomod.paths.get(1)) action1.add(f);
                        action1.add(-(n-1-e));
                        ArrayList<ArrayList<Integer>> posAction = new ArrayList<ArrayList<Integer>>();
                        if (e != 0) posAction.add(action0);
                        if (e != n-1) posAction.add(action1);
                        Kmod0 mod = new Kmod0(cat.objects.get(1), cat.objects.get(0), BigInteger.ONE, posAction);
                        cat.objects.get(1).botMods.add(mod);
                        cat.objects.get(0).topMods.add(mod);
                    }
                }
            }
        }
        else {
            Khov_object tomod = cat.objects.get(1);
            qd = tomod.qdeg;
            for (int i = -2; i >= crossing; i--) {
                hd = -i+crossing;
                if (i % 2 == 0) qd = qd-2;
                else qd = qd - 2 * (n-1);
                ArrayList<ArrayList<Integer>> paths = cloneActions(tomod.paths);
                Khov_object newObj = new Khov_object(paths,hd,qd);
                cat.objects.add(newObj);
                if (i%2 == 0) {
                    ArrayList<ArrayList<Integer>> posAction = new ArrayList<ArrayList<Integer>>();
                    ArrayList<ArrayList<Integer>> negAction = new ArrayList<ArrayList<Integer>>();
                    ArrayList<Integer> action0 = new ArrayList<Integer>();
                    for (int e : tomod.paths.get(0)) action0.add(e);
                    action0.add(-1);
                    ArrayList<Integer> action1 = new ArrayList<Integer>();
                    for (int e : tomod.paths.get(1)) action1.add(e);
                    action1.add(-1);
                    posAction.add(action0);
                    negAction.add(action1);
                    Kmod0 modp = new Kmod0(cat.objects.get(-i), cat.objects.get(-i-1), BigInteger.ONE, posAction);
                    Kmod0 modn = new Kmod0(cat.objects.get(-i), cat.objects.get(-i-1), BigInteger.ONE.negate(), negAction);
                    cat.objects.get(-i).botMods.add(modp);
                    cat.objects.get(-i).botMods.add(modn);
                    cat.objects.get(-i-1).topMods.add(modp);
                    cat.objects.get(-i-1).topMods.add(modn);
                }
                else {
                    for (int e = 0; e < n; e++) {
                        ArrayList<Integer> action0 = new ArrayList<Integer>();
                        for (int f : tomod.paths.get(0)) action0.add(f);
                        action0.add(-e);
                        ArrayList<Integer> action1 = new ArrayList<Integer>();
                        for (int f : tomod.paths.get(1)) action1.add(f);
                        action1.add(-(n-1-e));
                        ArrayList<ArrayList<Integer>> posAction = new ArrayList<ArrayList<Integer>>();
                        if (e != 0) posAction.add(action0);
                        if (e != n-1) posAction.add(action1);
                        Kmod0 mod = new Kmod0(cat.objects.get(-i), cat.objects.get(-i-1), BigInteger.ONE, posAction);
                        cat.objects.get(-i).botMods.add(mod);
                        cat.objects.get(-i-1).topMods.add(mod);
                    }
                }
            }
        }
    }

    void extendCat(KhovCategory cat, int crossing, int nminus, int nplus, int n) { // need to check the case when there are less than 4 endpoints.
        if (n > 2) {
            extendCatn(cat,crossing,n);
            return;
        }
        if (crossing > 1) {
            Khov_object tomod = cat.objects.get(1);
            for (int i = 2 ; i <= crossing; i++) {
                ArrayList<ArrayList<Integer>> paths = cloneActions(tomod.paths);
                Khov_object newObj = new Khov_object(paths,i-nminus,2*i-1+nplus-2*nminus);
                cat.objects.add(newObj);
            }
            addMorphisms(tomod,cat,1,crossing,true);
        }
        else {
            Khov_object lastObj = cat.objects.get(1);
            Khov_object nearLObj = cat.objects.get(0);
            cat.objects.remove(lastObj);
            cat.objects.remove(nearLObj);
            lastObj.hdeg = (-1)*crossing-nminus;
            lastObj.qdeg = lastObj.hdeg + nplus - nminus;
            nearLObj.hdeg = lastObj.hdeg - 1;
            nearLObj.qdeg = lastObj.qdeg - 1;
            for (int i = 0; i < (-1)*crossing -1; i++) {
                ArrayList<ArrayList<Integer>> paths = cloneActions(nearLObj.paths);
                Khov_object newObj = new Khov_object(paths,i-nminus,nearLObj.qdeg + 2*(crossing + i + 1));
                cat.objects.add(newObj);
            }
            cat.objects.add(nearLObj);
            cat.objects.add(lastObj);
            addMorphisms(nearLObj,cat,0, (-1)* crossing -1, false);
        }
    }

    void addMorphisms(Khov_object tomod, KhovCategory cat, int start, int end, boolean truth) {
        for (int i = start; i < end; i++) {
            BigInteger pos = BigInteger.ONE;
            ArrayList<Integer> action0 = new ArrayList<Integer>();
            for (int e : tomod.paths.get(0)) action0.add(e);
            action0.add(-1);
            ArrayList<Integer> action1 = new ArrayList<Integer>();
            for (int e : tomod.paths.get(1)) action1.add(e);
            action1.add(-1);
            ArrayList<ArrayList<Integer>> actions0 = new ArrayList<ArrayList<Integer>>();
            ArrayList<ArrayList<Integer>> actions1 = new ArrayList<ArrayList<Integer>>();
            actions0.add(action0);
            actions1.add(action1);
            Kmod0 modup = new Kmod0(cat.objects.get(i+1),cat.objects.get(i),pos,actions0);
            if (truth & (i%2 != 0)) pos = pos.negate();
            if (!truth & ((i+end)%2)!= 0) pos = pos.negate();
            Kmod0 moddn = new Kmod0(cat.objects.get(i+1),cat.objects.get(i),pos,actions1);
            cat.objects.get(i+1).botMods.add(modup);
            cat.objects.get(i+1).botMods.add(moddn);
            cat.objects.get(i).topMods.add(modup);
            cat.objects.get(i).topMods.add(moddn);
        }
    }

    KhovCategory zeroCross(ArrayList<Integer> endpoints, int[] paths, int nminus, int nplus,boolean pos) {
        KhovCategory cat = new KhovCategory();
        cat.endpoints = endpoints;
        ArrayList<Integer> path1 = new ArrayList<Integer>();
        ArrayList<Integer> path2 = new ArrayList<Integer>();
        path1.add(paths[0]);
        path1.add(paths[1]);
        path2.add(paths[2]);
        path2.add(paths[3]);
        ArrayList<ArrayList<Integer>> pats = new ArrayList<ArrayList<Integer>>();
        pats.add(path1);
        pats.add(path2);
        Khov_object obj0;
        obj0 = new Khov_object(pats,-nminus,nplus-2*nminus);
        cat.objects.add(obj0);
        return cat;
    }
    
    KhovCategory posCross(ArrayList<Integer> endpoints, int[] paths, int nminus, int nplus,boolean pos) {
        KhovCategory cat = new KhovCategory();
        cat.endpoints = endpoints;
        ArrayList<Integer> path1 = new ArrayList<Integer>();
        ArrayList<Integer> path2 = new ArrayList<Integer>();
        path1.add(paths[0]);
        path1.add(paths[1]);
        path2.add(paths[2]);
        path2.add(paths[3]);
        ArrayList<ArrayList<Integer>> pats = new ArrayList<ArrayList<Integer>>();
        pats.add(path1);
        pats.add(path2);
        ArrayList<Integer> path3 = new ArrayList<Integer>();
        ArrayList<Integer> path4 = new ArrayList<Integer>();
        path3.add(paths[0]);
        path3.add(paths[3]);
        path4.add(paths[2]);
        path4.add(paths[1]);
        ArrayList<ArrayList<Integer>> patz = new ArrayList<ArrayList<Integer>>();
        patz.add(path3);
        patz.add(path4);
        Khov_object obj0;
        Khov_object obj1;
        if (pos) {
            obj0 = new Khov_object(pats,-nminus,nplus-2*nminus);
            obj1 = new Khov_object(patz,1-nminus,1+nplus-2*nminus);
        }
        else {
            obj0 = new Khov_object(patz,-nminus,nplus-2*nminus);
            obj1 = new Khov_object(pats,1-nminus,1+nplus-2*nminus);
        }
        Kmod0 mod = new Kmod0(obj1,obj0,BigInteger.ONE);
        ArrayList<Integer> surgery = new ArrayList<Integer>();
        int x = 1;
        if (pos) x = 0;
        int a = obj1.paths.get(x).get(x);
        int b = obj1.paths.get(x).get(1-x);
        int c = obj1.paths.get(1-x).get(x);
        int d = obj1.paths.get(1-x).get(1-x);
        surgery.add(a);
        surgery.add(b);
        if (obj0.paths.get(0).contains(a)) {
            if (obj0.paths.get(0).contains(c)) {
                surgery.add(c);
                surgery.add(d);
            }
            else {
                surgery.add(d);
                surgery.add(c);
            }
        }
        else {
            if (obj0.paths.get(1).contains(c)) {
                surgery.add(c);
                surgery.add(d);
            }
            else {
                surgery.add(d);
                surgery.add(c);
            }
        }
        mod.actions.add(surgery);
        obj0.topMods.add(mod);
        obj1.botMods.add(mod);
        cat.objects.add(obj0);
        cat.objects.add(obj1);
        return cat;
    }
    
    void whitneyfy(ArrayList<Khov_object> objects) {
        for (Khov_object obj0 : objects) {
            int i = obj0.topMods.size()-1;
            while (i >= 0) {
                Kmod0 mod = obj0.topMods.get(i);
                i--;
                if (mod.totalValue.equals(BigInteger.ZERO)) {
                    Khov_object obj1 = mod.topObj;
                    obj0.topMods.remove(mod);
                    obj1.botMods.remove(mod);
                }
            }
        }
    }
    
    void startWhitneyfy(ArrayList<Khov_object> objects) {
        for (Khov_object obj0 : objects) {
            int i = 0;
            while (i < obj0.topMods.size()) {
                Kmod0 mod = obj0.topMods.get(i);
                int j = i + 1;
                while (j < obj0.topMods.size()) {
                    Kmod0 mod1 = obj0.topMods.get(j);
                    if (mod.topObj == mod1.topObj) {
                        mod.totalValue = mod.totalValue.add(mod1.totalValue);
                        obj0.topMods.remove(mod1);
                        Khov_object obj1 = mod.topObj;
                        obj1.botMods.remove(mod1);
                    }
                    else j++;
                }
                i++;
            }
        }
    }
    
    ArrayList<Khov_object> tensorObjects(Khov_object obj0, Khov_object obj1, ArrayList<Integer> endpoints, int n, boolean red) {
        ArrayList<ArrayList<Integer>> paths = pathsFrom(obj0.paths,obj1.paths,endpoints);
        int dot = 0;
        if (red) {
            boolean found = false;
            int i = 0;
            while (!found & i < paths.size()) {
                ArrayList<Integer> path = paths.get(i);
                if (path.contains(obj0.dotpath)) found = true;
                else i++;
            }
            if (found) {
                ArrayList<Integer> path = paths.get(i);
                int a = path.get(0);
                int b = path.get(path.size()-1);
                if (a < b) dot = a;
                if (a > b) dot = b;
                if (a == b) dot = smallestOf(path);
            }
            else dot = obj0.dotpath;
        }
        ArrayList<Integer> newcircs = getNewCircles(paths);
        ArrayList<circle> circles = circlesFrom(newcircs,n);
        ArrayList<Khov_object> newObjects = newKhovObjects(paths,newcircs,circles,n,obj0.hdeg+obj1.hdeg,obj0.qdeg+obj1.qdeg,dot);
        return newObjects;
    }
    
    int smallestOf(ArrayList<Integer> path) {
        int u = -1;
        for (int y : path) {
            if (u == -1) u = y;
            if (y < u) u = y;
        }
        return u;
    }
    
    ArrayList<Khov_object> newKhovObjects(ArrayList<ArrayList<Integer>> paths, ArrayList<Integer> newcircs, 
            ArrayList<circle> circles, int n, int hdeg, int qdeg, int dot) {
        ArrayList<Khov_object> newObjects = new ArrayList<Khov_object>();
        int numb = 1;
        int k = newcircs.size();
        for (int j = 0; j < k; j++) numb = n*numb;
        for (int j = 0; j < numb; j++) {
            Khov_object newObj = new Khov_object(paths,hdeg,qdeg);
            if (numb > 1) {
                int u = j;
                for (int o = 0; o < k; o++) {
                    newObj.circles.add(circles.get(n*o+(u%n)));
                    newObj.qdeg = newObj.qdeg + circles.get(n*o+(u%n)).sign;
                    u = u/n;
                }
            }
            if (dot == 0) newObjects.add(newObj);
            else { // this is reduced business
                boolean ok = true;
                newObj.dotpath = dot;
                for (circle cir : newObj.circles) {
                    if (cir.comp == dot & cir.sign == 1) ok = false;
                }
                if (ok) newObjects.add(newObj);
            }
        }
        return newObjects;
    }
    
    ArrayList<circle> circlesFrom(ArrayList<Integer> circs, int n) {
        ArrayList<circle> circles = new ArrayList<circle>();
        for (int i = 0; i < circs.size() ; i++ ) {
            for (int j = -(n-1); j <= (n-1); j = j+2) {
                circle circ = new circle(circs.get(i),j);
                circles.add(circ);
            }
        }
        return circles;
    }
    
    packige createModsFrom(Khov_object newObj0, Kmod0 mod, Khov_object obj, ArrayList<Integer> endpoints, int n) {
    	ArrayList<ArrayList<Integer>> currentpaths = mod.botObj.paths;
        Khov_object startObj = mod.botObj;
        Khov_object potObj = new Khov_object(newObj0.paths,newObj0.hdeg+1, newObj0.qdeg+1);
        potObj.circles = newObj0.circles;
        ArrayList<Khov_object> dudes = new ArrayList<Khov_object>();
        ArrayList<ArrayList<ArrayList<Integer>>> newActionList = new ArrayList<ArrayList<ArrayList<Integer>>>();
        ArrayList<ArrayList<Integer>> newActions = new ArrayList<ArrayList<Integer>>();
        int h = startObj.hdeg;
        int q = startObj.qdeg;
        if (mod.actions.isEmpty()) {
            ArrayList<Khov_object> special = tensorObjects(mod.topObj,obj,endpoints,n,false);
            for (Khov_object ob : special) {
                if (ob.qdeg == newObj0.qdeg) {
                    boolean circlesOk = true;
                    int i = 0;
                    while (circlesOk & i < ob.circles.size()) {
                        if (ob.circles.get(i).sign != newObj0.circles.get(i).sign) circlesOk = false;
                        else i++;
                    }
                    if (circlesOk) {
                        dudes.add(ob);
                        newActionList.add(newActions);
                    }
                }
            }
        }
        else {
            dudes.add(potObj);
            newActionList.add(newActions);
        }
        for (ArrayList<Integer> action : mod.actions) {
            ArrayList<Khov_object> newDudes = new ArrayList<Khov_object>();
            ArrayList<ArrayList<ArrayList<Integer>>> futActionList = new ArrayList<ArrayList<ArrayList<Integer>>>();
            ArrayList<ArrayList<Integer>> inBetweenPaths = new ArrayList<ArrayList<Integer>>();
            if (action.size() == 4) { // surgery
                for (ArrayList<Integer> endpath : currentpaths) {
                    if (!action.contains(endpath.get(0))) inBetweenPaths.add(endpath);
                    else {
                        ArrayList<Integer> changepath = new ArrayList<Integer>();
                        if (endpath.contains(action.get(0))) {
                            changepath.add(action.get(0));
                            changepath.add(action.get(1));
                        }
                        else {
                            changepath.add(action.get(2));
                            changepath.add(action.get(3));
                        }
                        inBetweenPaths.add(changepath);
                    }
                }
            }
            else { // dotting
                inBetweenPaths = currentpaths;
            }
            if (n == 2) q++;
            else q--;
            Khov_object inBetweenObj = new Khov_object(inBetweenPaths,h+1,q);
            ArrayList<Khov_object> inBetObjects = tensorObjects(inBetweenObj,obj,endpoints,n,false);
            for (int i = 0; i < dudes.size(); i++) {
                Khov_object dude = dudes.get(i);
                for (Khov_object newdude : inBetObjects) {
                    ArrayList<ArrayList<Integer>> currActions = cloneActions(newActionList.get(i));
                    ArrayList<Integer> actionType = typeOfRel(dude,newdude,action,n);
                    if (actionType.size()!=1) {
                        newDudes.add(newdude);
                        if (!actionType.isEmpty()) currActions.add(actionType);
                        futActionList.add(currActions);
                    }
                }
            }
            currentpaths = inBetweenPaths;
            dudes = newDudes;
            newActionList = futActionList;
        }
        packige pac = new packige(newActionList, dudes);
        return pac;
    }
    
    ArrayList<ArrayList<Integer>> cloneActions(ArrayList<ArrayList<Integer>> actions) {
    	ArrayList<ArrayList<Integer>> clone = new ArrayList<ArrayList<Integer>>();
    	int i = 0;
    	while (i < actions.size()) {
            ArrayList<Integer> action = new ArrayList<Integer>();
            for (int j : actions.get(i)) {
                action.add(j);
            }
            clone.add(action);
            i++;
    	}
    	return clone;
    }
    
    ArrayList<Integer> typeOfRel(Khov_object first, Khov_object second,  
    		ArrayList<Integer> action, int n) {
    	ArrayList<Integer> newAction = new ArrayList<Integer>();
    	int factor = 1;
    	if (n>2) factor = -1;
    	if (action.size() == 4) { //surgery
            ArrayList<Integer> path0 = getOrientedNewPath(action.get(0),first.paths,action);
                    ArrayList<Integer> path1 = getOrientedNewPath(action.get(3),first.paths,action);
                    ArrayList<Integer> path2 = getOrientedNewPath(action.get(0),second.paths,action);
                    ArrayList<Integer> path3 = getOrientedNewPath(action.get(3),second.paths,action);
            if (first.circles.size() == second.circles.size()) { //surgery does not affect circles
                    newAction = fromSurge(path0,path1,path2,path3,first,second);
            }
            if (first.circles.size() > second.circles.size()) { // surgery reduces number of circles
                if (path0.isEmpty() & path1.isEmpty()) {// two circles melt into one
                    int ok = first.circles.get(0).sign + first.circles.get(1).sign - second.circles.get(0).sign;
                    if (ok != factor * (n-1)) newAction.add(-1);
                }
                else {
                    ArrayList<Integer> path;
                    if (path0.isEmpty()) path = path1;
                    else path = path0;
                    newAction = fromSurge(path,first,second,true,n);
                }
            }
            if (first.circles.size() < second.circles.size()) { // surgery increases number of circles
                if (path0.isEmpty() & path1.isEmpty()) {// one circle turns into two
                    int ok = second.circles.get(1).sign + second.circles.get(0).sign - first.circles.get(0).sign;
                    if (ok != factor * (1-n)) newAction.add(-1);
                }
                else {
                    ArrayList<Integer> path;
                    if (path0.isEmpty()) path = path1;
                    else path = path0;
                    newAction = fromSurge(path,second,first,false,n);
                }
            }
    	}
    	else { //dotting
            ArrayList<Integer> path = getNewPath(action.get(0),first.paths);
            boolean circlesOk = true;
            if (path.isEmpty()) {
                int i = 0;
                if (first.circles.size() > 1) {
                    if (action.contains(first.circles.get(0).comp)) {
                        if (first.circles.get(1).sign != second.circles.get(1).sign) circlesOk = false;
                    }
                    else {
                        i = 1;
                        if (first.circles.get(0).sign != second.circles.get(0).sign) circlesOk = false;
                    }
                }
                if (circlesOk) {
                    if (second.circles.get(i).sign-first.circles.get(i).sign!=2 * factor * action.get(2)) circlesOk = false;
                }
                if (!circlesOk) newAction.add(-1);
            }
            else {
                for (int i = 0; i < first.circles.size(); i++) {
                    if (first.circles.get(i).sign != second.circles.get(i).sign) circlesOk = false;
                }
                if (circlesOk) {
                    newAction.add(path.get(0));
                    newAction.add(path.get(path.size()-1));
                    newAction.add(action.get(2));
                }
                else newAction.add(-1);
            }
    	}
    	return newAction;
    }
    
    ArrayList<Integer> fromSurge(ArrayList<Integer> path0, ArrayList<Integer> path1, ArrayList<Integer> path2,
            ArrayList<Integer> path3, Khov_object first, Khov_object second) {
    	ArrayList<Integer> newAction = new ArrayList<Integer>();
    	boolean circlesOk = true;
            if (first.circles.size() == 1) {
                if (first.circles.get(0).sign != second.circles.get(0).sign) circlesOk = false;
            }
            if (first.circles.size() == 2) {
                circle firstcirc0 = first.circles.get(0);
                circle firstcirc1 = first.circles.get(1);
                circle seconcirc0 = second.circles.get(0);
                circle seconcirc1 = second.circles.get(1);
                if (seconcirc0.comp == firstcirc0.comp) {
                    if ((seconcirc0.sign != firstcirc0.sign) | (seconcirc1.sign != firstcirc1.sign)) circlesOk = false;
                }
                else {
                    if ((seconcirc1.sign != firstcirc0.sign) | (seconcirc0.sign != firstcirc1.sign)) circlesOk = false;
                }
            }
            if (circlesOk) {
                int a = path2.get(0);
                int b = path3.get(0);
                int c = path3.get(path3.size()-1);
                int d = path2.get(path2.size()-1);
                newAction.add(a);
                newAction.add(d);
                ArrayList<Integer> path;
                if (path0.contains(a)) path = path0;
                else path = path1;
                if (path.contains(b)) {
                    newAction.add(b);
                    newAction.add(c);
                }
                else {
                    newAction.add(c);
                    newAction.add(b);
                }
            }
            else newAction.add(-1); // aborting this possibility
    	return newAction;
    }
    
    ArrayList<Integer> fromSurge(ArrayList<Integer> path, Khov_object first, Khov_object second,
            boolean direction, int n) { // direction checks whether surgery reduces circles (true) or increases (false)
    	ArrayList<Integer> newAction = new ArrayList<Integer>();
    	boolean circlesOk = true;
    	int factor = 1;
    	if (n>2) factor = -1;
            int i = 0;
            if (second.circles.size() == 1) {
                int comp = second.circles.get(0).comp;
                int sgn = second.circles.get(0).sign;
                if (first.circles.get(0).comp == comp) {
                    i = 1;
                    if (first.circles.get(0).sign != sgn) circlesOk = false;
                }
                else {
                    if (first.circles.get(1).sign != sgn) circlesOk = false;
                }
            }
            if (circlesOk) {
                if (direction) {
                    if (first.circles.get(i).sign * factor != (n-1)) {
                        newAction.add(path.get(0));
                        newAction.add(path.get(path.size()-1));
                        newAction.add((factor * first.circles.get(i).sign - (n-1))/2);
                    }
                }
                else {
                    if (first.circles.get(i).sign * factor != (1-n)) {
                        newAction.add(path.get(0));
                        newAction.add(path.get(path.size()-1));
                        newAction.add((1-n-factor * first.circles.get(i).sign)/2);
                    }
                }
            }
            else newAction.add(-1);
    	return newAction;
    }
    
    ArrayList<Integer> getOrientedNewPath(int comp, ArrayList<ArrayList<Integer>> paths, ArrayList<Integer> action) {
        ArrayList<Integer> newpath = getNewPath(comp,paths);
        if (!newpath.isEmpty()) {
            int y = action.indexOf(comp);
            int z = 1;
            boolean forward = true;
            switch (y) {
            case 0 : z = 1; break;
            case 1 : z = 0; forward = false; break;
            case 2 : z = 3; forward = false; break;
            case 3 : z = 2; break;
            }
            int n = newpath.indexOf(comp);
            int m = newpath.indexOf(action.get(z));
            boolean switcheroo = false;
            if (forward & (n > m)) switcheroo = true;
            if (!forward & (n < m)) switcheroo = true;
            if (switcheroo) newpath = reversePath(newpath);
        }
        return newpath;
    }
    
    ArrayList<Integer> reversePath(ArrayList<Integer> path) {
        ArrayList<Integer> newpath = new ArrayList<Integer>();
        for (int i = path.size()-1; i >= 0; i--) newpath.add(path.get(i));
        return newpath;
    }
	
    ArrayList<Integer> getNewPath(int comp, ArrayList<ArrayList<Integer>> paths) {
        ArrayList<Integer> newpath = new ArrayList<Integer>();
        boolean found = false;
        int i = 0;
        while (!found & (i < paths.size())) {
            if (paths.get(i).contains(comp)) found = true;
            else i++;
        }
        if (found) newpath = paths.get(i);
        return newpath;
    }
	
    ArrayList<Integer> getNewPath(ArrayList<Integer> path, 
            ArrayList<ArrayList<Integer>> paths) {
        ArrayList<Integer> newpath = new ArrayList<Integer>();
        int a = path.get(0);
        boolean found = false;
        int i = 0;
        while (!found & (i < paths.size())) {
            if (paths.get(i).contains(a)) found = true;
            else i++;
        }
        if (found) newpath = paths.get(i);
        return newpath;
    }
    
    void getRidOf(KhovCategory merge, Khov_object tObj,Khov_object bObj, Kmod0 mod, int n) {
    	if (tObj.botMods.size()>1 & bObj.topMods.size()>1) { // check to redirect 0-dim mods
            ArrayList<Khov_object> changers = new ArrayList<Khov_object>();
            for (Kmod0 md : tObj.botMods) {
                if (md != mod) {
                    Khov_object lObj = md.botObj;
                    if (!changers.contains(lObj)) changers.add(lObj);
                    for (Kmod0 mbd : bObj.topMods) {
                        if (mbd != mod) {
                            Khov_object rObj = mbd.topObj;
                            BigInteger sign = mod.totalValue.negate();
                            BigInteger val = sign.multiply(md.totalValue).multiply(mbd.totalValue);
                            ArrayList<ArrayList<ArrayList<Integer>>> newactions = combineActions(md.actions,mbd.actions,n);
                            for (ArrayList<ArrayList<Integer>> newaction : newactions) {
                            	int mtr = find(newaction,rObj,lObj);
                            	if (mtr == -1) {
                                    Kmod0 mot = new Kmod0(rObj,lObj,val,newaction);
                                    rObj.botMods.add(mot);
                                    lObj.topMods.add(mot);
                            	}
                            	else {
                                    Kmod0 mot = rObj.botMods.get(mtr);
                                    mot.totalValue = mot.totalValue.add(val);
                            	}
                            }
                        }
                    }
                }
            }
            mergeMod0(changers);
        }
        for (Kmod0 md : tObj.botMods) md.botObj.topMods.remove(md);
        for (Kmod0 md : tObj.topMods) md.topObj.botMods.remove(md);
        for (Kmod0 md : bObj.topMods) md.topObj.botMods.remove(md);
        for (Kmod0 md : bObj.botMods) md.botObj.topMods.remove(md);
        merge.objects.remove(tObj);
        merge.objects.remove(bObj);
    }
    
    void mergeMod0(ArrayList<Khov_object> objects) {
        for (Khov_object obj : objects) {
            for (int i = 0 ; i < obj.topMods.size(); i++) {
                Kmod0 mod = obj.topMods.get(i);
                for (int j = obj.topMods.size()-1; j > i; j--) {
                    Kmod0 md = obj.topMods.get(j);
                    if (md.topObj == mod.topObj) {
                        if (same(md.actions,mod.actions)) {
                            mod.totalValue = mod.totalValue.add(md.totalValue);
                            obj.topMods.remove(md);
                            md.topObj.botMods.remove(md);
                        }
                    }
                }
            }
        }
    }
    
    boolean same(ArrayList<ArrayList<Integer>> list0, ArrayList<ArrayList<Integer>> list1) {
        boolean same = true;
        int i = 0;
        if (list0.size() != list1.size()) return false;
        while (same & (i < list0.size())) {
            if (!samePath(list0.get(i),list1.get(i))) same = false;
            else i++;
        }
        return same;
    }
    
    boolean samePath(ArrayList<Integer> path0, ArrayList<Integer> path1) {
        boolean same = true;
        if (path0.size()!=path1.size()) return false;
        int i = 0;
        while (same & i < path0.size()) {
                if (!path1.contains(path0.get(i))) same = false;
                else i++;
        }
        return same;
    }
    
    int find(ArrayList<ArrayList<Integer>> action, Khov_object rObj, Khov_object lObj) {
            boolean found = false;
            int i = 0;
            int mod = -1;
            while (!found & i < rObj.botMods.size()) {
                    Kmod0 md = rObj.botMods.get(i);
                    if (md.botObj == lObj) {
                            if (same(action,md.actions)) found = true;
                    }
                    if (!found) i++;
            }
            if (!found) return -1;
            mod = i;
            return mod;
	}
	
    ArrayList<ArrayList<ArrayList<Integer>>> combineActions(ArrayList<ArrayList<Integer>> factions, 
            ArrayList<ArrayList<Integer>> sactions, int n) {
    	ArrayList<ArrayList<ArrayList<Integer>>> newactions = new ArrayList<ArrayList<ArrayList<Integer>>>();
    	ArrayList<ArrayList<Integer>> newaction = new ArrayList<ArrayList<Integer>>();
    	for (ArrayList<Integer> action : factions) newaction.add(action);
        for (ArrayList<Integer> action : sactions) newaction.add(action);
        newaction = sortActions(newaction);
        if (noDoubleDot(newaction,n)) newactions.add(newaction);
        boolean nochange = false;
        while (!nochange & !newactions.isEmpty()) {
            nochange = check4DoubleSurg(newactions,n);
        }
    	return newactions;
    }
	
    boolean check4DoubleSurg(ArrayList<ArrayList<ArrayList<Integer>>> actionlist, int n) {
    	boolean nochange = true;
    	int i = 0;
    	while (i < actionlist.size()) {
            ArrayList<ArrayList<Integer>> actions = actionlist.get(i);
            int j = 0;
            int k = 1;
            boolean giveup = false;
            while (!giveup & j < actions.size()-1) {
                k = j+1;
                ArrayList<Integer> action = actions.get(j);
                ArrayList<Integer> accion = actions.get(k);
                if (doubleSurg(action,accion)) giveup = true;
                else {
                    k++;
                    if (k < actions.size()) accion = actions.get(k);
                    if (doubleSurg(action,accion)) giveup = true;
                    if (!giveup) j++;
                }
            }
            if (giveup) {
                actionlist.remove(actions);
                for (int e = 0; e < n ; e++) {
                    ArrayList<ArrayList<Integer>> akschens = new ArrayList<ArrayList<Integer>>();
                    for (int f = 0; f < j; f++) akschens.add(actions.get(f));
                    if (k!=j+1) akschens.add(actions.get(j+1));
                    ArrayList<Integer> help1 = new ArrayList<Integer>();
                    ArrayList<Integer> help2 = new ArrayList<Integer>();
                    help1.add(actions.get(k).get(0));
                    help1.add(actions.get(k).get(1));
                    help1.add(-e);
                    help2.add(actions.get(k).get(2));
                    help2.add(actions.get(k).get(3));
                    help2.add(-(n-1-e));
                    if (e!=0) akschens.add(help1);
                    if (e!=n-1) akschens.add(help2);
                    for (int f = k+1; f < actions.size(); f++) akschens.add(actions.get(f));
                    akschens = sortActions(akschens);
                    actionlist.add(akschens);
                }
                nochange = false;
            }
            else i++;
    	}
    	return nochange;
    }
	
    boolean doubleSurg(ArrayList<Integer> action, ArrayList<Integer> accion) {
    	boolean doublesurg = false;
    	if ((action.size()==4) & (accion.size() == 4)) {
            doublesurg = true;
            int i = 0;
            while (doublesurg & i < 4) {
                if (!accion.contains(action.get(i))) doublesurg = false;
                else i++;
            }
    	}
    	return doublesurg;
    }
	
    boolean noDoubleDot(ArrayList<ArrayList<Integer>> actions, int n) {
        if (n > 2) return true;
        boolean nodouble = true;
    	int taintlevel = 0;
    	ArrayList<Integer> tainted = new ArrayList<Integer>();
    	int i = 0;
    	while (i < actions.size() & taintlevel < n) {
            ArrayList<Integer> action = actions.get(i);
            if (action.size() == 3) {
                if (tainted.contains(action.get(0)) | tainted.contains(action.get(1))) 
                    taintlevel = taintlevel - action.get(2);
                else {
                    tainted.add(action.get(0));
                    tainted.add(action.get(1));
                }
            }
            i++;
    	}
    	if (taintlevel >=n) nodouble = false;
    	return nodouble;
    }
    
    ArrayList<ArrayList<Integer>> sortActions(ArrayList<ArrayList<Integer>> actions) {
    	for (int i = 0 ; i < actions.size()-1; i++) {
            ArrayList<Integer> action = actions.get(i);
            if (action.size() == 3) {
                int j = i+1;
                boolean found = false;
                while (!found & j < actions.size()) {
                    ArrayList<Integer> accion = actions.get(j);
                    if (accion.size() == 4) {
                        found = true;
                        while (j > i) {
                            flipDotSurg(actions,j);
                            j--;
                        }
                    }
                    else j++;
                }
            }
    	}
    	return actions;
    }
    
    void flipDotSurg(ArrayList<ArrayList<Integer>> actions, int j) {
    	ArrayList<Integer> action = actions.get(j-1);
    	ArrayList<Integer> accion = actions.get(j);
    	ArrayList<Integer> help = new ArrayList<Integer>();
    	help.add(action.get(0));
    	help.add(action.get(1));
    	help.add(action.get(2));
    	actions.set(j-1, accion);
    	if (accion.contains(help.get(0))) {
            int e = accion.indexOf(help.get(0));
            switch (e) {
            case 0 : help.set(1, accion.get(1)); break;
            case 1 : help.set(1, accion.get(0)); break;
            case 2 : help.set(1, accion.get(3)); break;
            case 3 : help.set(1, accion.get(2));
            }
            if (e == -1) System.out.println("What???");
    	}
    	actions.set(j, help);
    }
    
    ArrayList<Integer> factorize(int p) {
        ArrayList<Integer> factorise = new ArrayList<Integer>();
        int prime = 2;
        int factor = 1;
        while (p > 0) {
            if (p % prime == 0) {
                p = p / prime;
                factor = factor * prime; 
            }
            else {
                prime++;
                if (factor > 1) factorise.add(factor);
                factor = 1;
                if (p == 1) p = 0;
            }
        }
        return factorise;
    }
	
    void removeMod0(ArrayList<Khov_object> objects) {
        for (Khov_object obj : objects) {
            int i = obj.botMods.size();
            while (i > 0) {
                Kmod0 mod = obj.botMods.get(i-1);
                if (mod.totalValue.equals(BigInteger.ZERO)) {
                        obj.botMods.remove(mod);
                        mod.botObj.topMods.remove(mod);
                }
                i--;
            }
            i = obj.topMods.size();
            while (i > 0) {
                Kmod0 mod = obj.topMods.get(i-1);
                if (mod.totalValue.equals(BigInteger.ZERO)) {
                    obj.topMods.remove(mod);
                    mod.topObj.botMods.remove(mod);
                }
                i--;
            }
        }
    }

    void shiftBot(Kmod0 mod) {
        Khov_object tObj = mod.topObj;
        Khov_object bObj = mod.botObj;
        for (Kmod0 rmod : tObj.botMods) {
            if (rmod != mod) {
                BigInteger factor = rmod.totalValue.multiply(mod.totalValue).negate();
                handleSlide(bObj,rmod.botObj,factor); 
            }
        }
    }

    void shiftTop(Kmod0 mod) {
        Khov_object tObj = mod.topObj;
        Khov_object bObj = mod.botObj;
        for (Kmod0 rmod : bObj.topMods) {
            if (rmod != mod) {
                BigInteger factor = rmod.totalValue.multiply(mod.totalValue);
                handleSlide(rmod.topObj,tObj,factor);
            }
        }
    }

    void slideBot(Kmod0 mod) {
            Khov_object tObj = mod.topObj;
            Khov_object bObj = mod.botObj;
            for (Kmod0 rmod : tObj.botMods) {
                if (rmod != mod) {
                    BigInteger div = rmod.totalValue.divide(mod.totalValue);
                    div = div.negate();
                    handleSlide(bObj,rmod.botObj,div);
                }
            }
    }

    void slideTop(Kmod0 mod) {
            Khov_object tObj = mod.topObj;
            Khov_object bObj = mod.botObj;
            for (Kmod0 rmod : bObj.topMods) {
                if (rmod != mod) {
                    BigInteger div = rmod.totalValue.divide(mod.totalValue);
                    handleSlide(rmod.topObj,tObj,div);
                }
            }
    }

    void handleSlide(Khov_object xObj, Khov_object yObj, BigInteger factor) {
        for (Kmod0 mod : xObj.topMods) { // now let those mod0 going into xObj also go into yObj (with factor)
            Khov_object tObj = mod.topObj;
            boolean found = false;
            int i = 0;
            while (!found & i < tObj.botMods.size()) {
                Kmod0 md = tObj.botMods.get(i);
                if (md.botObj == yObj) found = true;
                else i++;
            }
            BigInteger fact = mod.totalValue.multiply(factor);
            Kmod0 myd;
            if (found) {
                myd = tObj.botMods.get(i);
                myd.totalValue = myd.totalValue.add(fact);
            }
            else {
                myd = new Kmod0(tObj,yObj,fact);
                tObj.botMods.add(myd);
                yObj.topMods.add(myd);
            }
        }
        for (Kmod0 mod : yObj.botMods) { // now let those mod0 going out of yObj also go out of xObj (with factor)
            Khov_object bObj = mod.botObj;
            boolean found = false;
            int i = 0;
            while (!found & i < bObj.topMods.size()) {
                Kmod0 md = bObj.topMods.get(i);
                if (md.topObj == xObj) found = true;
                else i++;
            }
            BigInteger fact = mod.totalValue.multiply(factor).negate();
            Kmod0 mxd;
            if (found) {
                mxd = bObj.topMods.get(i);
                mxd.totalValue = mxd.totalValue.add(fact);
            }
            else {
                mxd = new Kmod0(xObj,bObj,fact);
                xObj.botMods.add(mxd);
                bObj.topMods.add(mxd);
            }
        } 
    }
    
    Kmod0 findSmallest(ArrayList<Khov_object> objects) {
        Kmod0 returner = null;
        boolean foundsmth = false;
        BigInteger minvalue = BigInteger.ZERO;
        for (Khov_object obj : objects) {
            if (obj.botMods.size()>0) {
                int u = 0;
                if (!foundsmth) {
                    returner = obj.botMods.get(0);
                    minvalue = returner.totalValue.abs();
                    foundsmth = true;
                }
                while (u < obj.botMods.size()) {
                    BigInteger compare = obj.botMods.get(u).totalValue.abs(); 
                    int comp = minvalue.compareTo(compare);
                    if (comp == 1) {
                        returner = obj.botMods.get(u);
                        minvalue = compare;
                    }
                    u++;
                }
            }
        }
        return returner;
    }

    Kmod0 findOne(ArrayList<Khov_object> objects) {
        int u = 0;
        Kmod0 returner = null;
        boolean foundone = false;
        while (!foundone & u < objects.size()) {
            Khov_object tObj = objects.get(u);
            int v = 0;
            while (!foundone & v < tObj.botMods.size()) {
                Kmod0 mod = tObj.botMods.get(v);
                BigInteger tester = mod.totalValue; 
                if ((tester.abs().equals(BigInteger.ONE))) {
                    foundone = true;
                    returner = mod;
                }
                else v++;
            }
            u++;
        }
        return returner;
    }
    
    int disjoint(diagram diag, int[] ends) {
        int dj = 0;
        for (ArrayList<Integer> path : diag.paths) {
            boolean found = false;
            int i = 0;
            while (!found & i < 4) {
                if (path.contains(ends[i])) found = true;
                i++;
            }
            if (found) dj++;
        }
        return dj;
    }

    int qdegree(int n, link Link, int[] pos) {
        int q = 0;
        int qhelp;
        for (int k = 0; k < pos.length; k++) {
            q = q+Link.crossings[k]*n;
            if (pos[k] < 0) {
                q = q +(n-1);
                for (int j = -2; j >= pos[k]; j--) {
                    if (j%2==0) q = q + 2;
                    else q = q + 2*(n-1);
                }
            }
            if (pos[k] > 0) {
                q = q -(n-1);
                for (int j = 2; j <= pos[k]; j++) {
                    if (j%2==0) q = q - 2;
                    else q = q - 2*(n-1);
                }
            }
        }
        return q;
    }
    
    void disjointDiag(boolean right, diagram diag, int[] ends, int i) {
        ArrayList<Integer> firstpath = new ArrayList<Integer>();
        ArrayList<Integer> secondpath = new ArrayList<Integer> ();
        firstpath.add(ends[0]);
        if (i==0) addRight(diag,firstpath,secondpath,ends);
        else addLeft(diag,firstpath,secondpath,ends);
    }
    
    ArrayList<diagram> startDiagrams(int[] numbers, int[] ends) {
        ArrayList<diagram> diagrams = new ArrayList<diagram>();
        if (numbers[0] >= 0) {
            for (int i = 0; i <= numbers[0]; i++) {
                diagram diag = new diagram(i,numbers.length);
                disjointDiag(true,diag,ends,i);
                diagrams.add(diag);
            }
        }
        else {
            for (int i = numbers[0]; i <= 0; i++) {
                diagram diag = new diagram(i,numbers.length);
                disjointDiag(false,diag,ends,i);
                diagrams.add(diag);
            }
        }
        return diagrams;
    }
    
    void addRight(diagram diag, ArrayList<Integer> firstpath, ArrayList<Integer> secondpath, int[] ends) {
        firstpath.add(ends[1]);
        if (ends[1] == ends[0]) {
            diag.circles.add(firstpath);
            secondpath.add(ends[2]);
            secondpath.add(ends[3]);
            if (ends[2] == ends[3]) diag.circles.add(secondpath);
            else diag.paths.add(secondpath);
        }
        else {
            if (ends[0] != ends[3]) {
                if (ends[1] == ends[2]) {
                    firstpath.add(ends[2]);
                    firstpath.add(ends[3]);
                    diag.paths.add(firstpath);
                }
                else {
                    diag.paths.add(firstpath);
                    secondpath.add(ends[2]);
                    secondpath.add(ends[3]);
                    if (ends[2] == ends[3]) diag.circles.add(secondpath);
                    else diag.paths.add(secondpath);
                }
            }
            else {
                if (ends[1] == ends[2]) { //one big circle
                    firstpath.add(ends[2]);
                    firstpath.add(ends[3]);
                    diag.circles.add(firstpath);
                }
                else { 
                    secondpath.add(ends[2]);
                    secondpath.add(ends[3]);
                    secondpath.add(ends[0]);
                    secondpath.add(ends[1]);
                    diag.paths.add(secondpath);
                }
            }
        }
    }

    void addLeft(diagram diag, ArrayList<Integer> firstpath, ArrayList<Integer> secondpath, int[] ends) {
        firstpath.add(ends[3]);
        if (ends[3] == ends[0]) {
            diag.circles.add(firstpath);
            secondpath.add(ends[2]);
            secondpath.add(ends[1]);
            if (ends[2] == ends[1]) diag.circles.add(secondpath);
            else diag.paths.add(secondpath);
        }
        else {
            if (ends[0] != ends[1]) {
                if (ends[3] == ends[2]) {
                    firstpath.add(ends[2]);
                    firstpath.add(ends[1]);
                    diag.paths.add(firstpath);
                }
                else {
                    diag.paths.add(firstpath);
                    secondpath.add(ends[2]);
                    secondpath.add(ends[1]);
                    if (ends[2] == ends[1]) diag.circles.add(secondpath);
                    else diag.paths.add(secondpath);
                }
            }
            else {
                if (ends[3] == ends[2]) { //one big circle
                    firstpath.add(ends[2]);
                    firstpath.add(ends[1]);
                    diag.circles.add(firstpath);
                }
                else { 
                    secondpath.add(ends[2]);
                    secondpath.add(ends[1]);
                    secondpath.add(ends[0]);
                    secondpath.add(ends[3]);
                    diag.paths.add(secondpath);
                }
            }
        }
    }
    
    void concatenateDiag(boolean right, diagram diag, int[] ends, int k) {
        ArrayList<Integer> firstpath = new ArrayList<Integer>();
        ArrayList<Integer> secondpath = new ArrayList<Integer>();
        boolean justone = false;
        if (right) {
            firstpath.add(ends[0]);
            firstpath.add(ends[1]);
            secondpath.add(ends[2]);
            secondpath.add(ends[3]);
            if (ends[1]==ends[2]) {
                justone = true;
                firstpath = combine(firstpath,secondpath);
            }
            if (ends[0]==ends[3]) {
                justone = true;
                firstpath = combine(secondpath,firstpath);
            }
        }
        else {
            firstpath.add(ends[0]);
            firstpath.add(ends[3]);
            secondpath.add(ends[2]);
            secondpath.add(ends[1]);
            if (ends[0]==ends[1]) {
                justone = true;
                firstpath = combine(secondpath,firstpath);
            }
            if (ends[2]==ends[3]) {
                justone = true;
                firstpath = combine(firstpath,secondpath);
            }
        }
        int start = firstpath.get(0);
        int end = firstpath.get(firstpath.size()-1);
        boolean startfound = false;
        boolean endfound = false;
        boolean circle = false;
        int j = 0;
        while (!startfound  & (j < diag.paths.size())) {
            ArrayList<Integer> path = diag.paths.get(j);
            if (path.contains(start)) startfound = true;
            else j++;
        }
        if (startfound) {
            ArrayList<Integer> path = diag.paths.get(j);
            path = combinePaths(diag.paths.get(j),firstpath,true);
            diag.paths.remove(j);
            if (path.get(0) == end) {
                diag.circles.add(path);
                circle = true;
            }
            else firstpath = path;
        }
        j = 0;
        while (!endfound  & (j < diag.paths.size())) {
            ArrayList<Integer> path = diag.paths.get(j); 
            if (path.contains(end)) endfound = true;
            else j++;
        }
        if (endfound) {
            firstpath = combinePaths(diag.paths.get(j),firstpath,false);
            diag.paths.remove(j);
        }
        if (!circle) diag.paths.add(firstpath);
        if (!justone) {
            start = secondpath.get(0);
            end = secondpath.get(secondpath.size()-1);
            startfound = false;
            endfound = false;
            circle = false;
            j = 0;
            while (!startfound  & (j < diag.paths.size())) {
                ArrayList<Integer> path = diag.paths.get(j);
                if (path.contains(start)) startfound = true;
                else j++;
            }
            if (startfound) {
                ArrayList<Integer> path = diag.paths.get(j);
                path = combinePaths(diag.paths.get(j),secondpath,true);
                diag.paths.remove(j);
                if (path.get(0) == end) {
                    diag.circles.add(path);
                    circle = true;
                }
                else secondpath = path;
            }
            j = 0;
            while (!endfound  & (j < diag.paths.size())) {
                ArrayList<Integer> path = diag.paths.get(j);
                if (path.contains(end)) endfound = true;
                else j++;
            }
            if (endfound) {
                secondpath = combinePaths(diag.paths.get(j),secondpath,false);
                diag.paths.remove(j);
            }
            if (!circle) diag.paths.add(secondpath);
        }
    }
    
    ArrayList<Integer> combinePaths(ArrayList<Integer> first, ArrayList<Integer> second, boolean front) {
        ArrayList<Integer> path = new ArrayList<Integer>();
        if (front) {
            if (first.get(0) == second.get(0)) {
                for (int u = first.size()-1; u >= 0; u--) path.add(first.get(u));
                for (int u = 1; u < second.size(); u++) path.add(second.get(u));
            }
            else {
                for (int u = 0; u < first.size(); u++) path.add(first.get(u));
                for (int u = 1; u < second.size(); u++) path.add(second.get(u));
            }
        }
        else {
            if (first.get(0) == second.get(second.size()-1)) {
                for (int u = first.size()-1; u > 0; u--) path.add(first.get(u));
                for (int u = second.size()-1; u >= 0; u--) path.add(second.get(u));
            }
            else {
                for (int u = 0; u < first.size()-1; u++) path.add(first.get(u));
                for (int u = second.size()-1; u >= 0; u--) path.add(second.get(u));
            }
        }
        return path;
    }

    ArrayList<Integer> combine(ArrayList<Integer> first, ArrayList<Integer> second) {
        for (int k : second) {
            first.add(k);
        }
        return first;
    }
    
    int diffPosition(diagram dig1, diagram dig2) {
        int diff = -1;
        int u = 0;
        int v = 0;
        boolean ok = true;
        int duf = 0;
        int dif = 0;
        while (ok & u < dig1.position.length) {
            dif = dig1.position[u] - dig2.position[u];
            if (dif == 1) {
                duf++;
                v = u;
            }
            if ((dif > 1) | (dif < 0)) duf = 2;
            if (duf > 1) ok = false;
            u++;
        }
        if (duf == 1) diff = v;
        return diff;
    }
    
    boolean getSign(int[] position, int r) {
        boolean sign = true;
        int sum = 0;
        for (int i = 0; i < r; i++) sum = sum+position[i];
        if ((sum%2)!=0) sign = false;
        return sign;
    }

    int aDotting(sh_object top, sh_object bot, int[] path, int n) {
        int dotts = 0;
        int circlea = 0;
        boolean cafound = false;
        while (!cafound) {
                if (bot.dig.circles.get(circlea).contains(path[0])) cafound = true;
                else circlea++;
        }
        int circlec = 0;
        boolean ccfound = false;
        while (!ccfound) {
                if (bot.dig.circles.get(circlec).contains(path[2])) ccfound = true;
                else circlec++;
        }
        if (circlea == circlec) dotts = 1;
        else dotts = 2;
        int u = 0;
        while ((dotts >= 1) & (u < bot.dig.circles.size())) {
                if ((u == circlea) | (u == circlec)) u++;
                else {
                        if (top.symbols.get(u)!=bot.symbols.get(u)) dotts = 0;
                        else u++;
                }
        }
        if (dotts == 2) { // work out exactly which circle gets dotted
            if (n > 2) { // will have to work harder
                int symbdiff = top.symbols.get(circlea)+top.symbols.get(circlec)-bot.symbols.get(circlea)-bot.symbols.get(circlec);
                if (symbdiff == 2) {
                    if (top.symbols.get(circlec) == bot.symbols.get(circlec)) dotts = 3;
                    if ((top.symbols.get(circlea)!= bot.symbols.get(circlea)) & (top.symbols.get(circlec)!= bot.symbols.get(circlec))) dotts = 0;
                }
            }
            else {
                if ((bot.symbols.get(circlea) == 1) & (bot.symbols.get(circlec) == 1)) { 
                    if (top.symbols.get(circlea) == 1) dotts = 3; // this means the right circle is getting dotted.
                }
                else {
                    if (bot.symbols.get(circlea) != 1) dotts = 3; // again, right circle gets dotted.
                }
            }
        }
        return dotts;
    }

    boolean aSurgery(sh_object top, sh_object bot, int[] path) {
        boolean surge = true;
        int circlea = 0;
        boolean cafound = false;
        while (!cafound) {
            if (bot.dig.circles.get(circlea).contains(path[0])) cafound = true;
            else circlea++;
        } // circlea is the circle containing bottom right endpoint of r-th crossing
        int circlec = 0;
        boolean ccfound = false;
        while (!ccfound) {
            if (bot.dig.circles.get(circlec).contains(path[2])) ccfound = true;
            else circlec++;
        } // circlec is the circle containing top left endpoint of r-th crossing
        if (circlea == circlec) { // one circle to two circles
            int u = 0;
            while (surge & (u < bot.dig.circles.size())) {
                if (u == circlea) u++;
                else {
                    int k = bot.dig.circles.get(u).get(0);
                    boolean found = false;
                    int v = 0;
                    while (!found) {
                        if (top.dig.circles.get(v).contains(k)) found = true;
                        else v++;
                    }
                    if (bot.symbols.get(u)!=top.symbols.get(v)) surge = false;
                    else u++;
                }
            }
        }
        else { // two circles to one circle
            int u = 0;
            while (surge & (u < bot.dig.circles.size())) {
                if (u == circlea | u == circlec) u++;
                else {
                    int k = bot.dig.circles.get(u).get(0);
                    boolean found = false;
                    int v = 0;
                    while (!found) {
                        if (top.dig.circles.get(v).contains(k)) found = true;
                        else v++;
                    }
                    if (bot.symbols.get(u)!=top.symbols.get(v)) surge = false;
                    else u++;
                }
            }
        }
        return surge;
    }
    
    boolean framing(mod0 first, mod0 second) {
        int factor1 = 0;
        int factor2 = 0;
        int firstChange = diffPosition(first.topObj.dig,first.botObj.dig);
        int seconChange = diffPosition(second.topObj.dig,second.botObj.dig);
        boolean m1 = false;
        boolean m2 = false;
        int i1 = first.totalValue.intValue();
        if (Math.abs(i1)==1) {
            boolean sign1 = getSign(first.botObj.dig.position,firstChange);
            if (sign1 != (i1 > 0)) m1 = true;
        }
        i1 = second.totalValue.intValue();
        if (Math.abs(i1)==1) {
            boolean sign2 = getSign(second.botObj.dig.position,seconChange);
            if (sign2 != (i1>0)) m2 = true;
        }
        boolean order = true;
        if (firstChange > seconChange) {
            int h = firstChange;
            firstChange = seconChange;
            seconChange = h;
            order = false;
        }
        for (int y = 0; y < firstChange; y++) factor1 = factor1 + first.botObj.dig.position[y];
        for (int y = firstChange; y < seconChange; y++) factor2 = factor2 + first.botObj.dig.position[y];
        int b = first.botObj.dig.position[firstChange];
        if ((order & m1) | (!order & m2)) {
            factor1++;
            factor2 = factor2 + first.botObj.dig.position[firstChange];
            if (b < 0) factor2++;
        }
        if ((order & m2) | (!order & m1)) factor2++;
        factor1 = factor1 * factor2;
        boolean frame = true;
        if (factor1 % 2 != 0) frame = false; // framing of interval has been determined.
        return frame;
    }
    
    boolean framingZero(mod0 first, mod0 second) {
        int firstChange = diffPosition(first.topObj.dig,first.botObj.dig);
        int seconChange = diffPosition(second.topObj.dig,second.botObj.dig);
        boolean order = true;
        if (firstChange > seconChange) {
            int h = firstChange;
            firstChange = seconChange;
            seconChange = h;
            order = false;
        }
        boolean firstZero = true;
        boolean frame = true;
        boolean m = false;
        if (!first.totalValue.equals(BigInteger.ZERO)) firstZero = false;
        if ( order == firstZero ) { // this means 0 then something
            int i1;
            if (order) i1 = second.totalValue.intValue();
            else i1 = first.totalValue.intValue();
            if (Math.abs(i1) == 1) { // check if it's an M point
                boolean sign;
                if (order) sign = getSign(second.botObj.dig.position,seconChange);
                else sign = getSign(first.botObj.dig.position,seconChange);
                if (sign != (i1 > 0)) m = true;
            }
            int a = 0;
            for (int y = 0; y < firstChange; y++) a = a + first.botObj.dig.position[y];
            int b = 0;
            for (int y = firstChange; y < seconChange; y++) b = b + first.botObj.dig.position[y];
            int d = 0;
            if (first.botObj.dig.position[firstChange] < 0) d = 1;
            int c = d + first.botObj.dig.position[firstChange];
            a = 1 + (a+1) * c+b;
            if ((i1 == 1) == order) a++;
            if (m) a++;
            if (a%2 != 0) frame = false;
        }
        else { // this means something then 0
            int i1;
            if (order) i1 = first.totalValue.intValue();
            else i1 = second.totalValue.intValue();
            if (Math.abs(i1) == 1) { // check if it's an M point
                boolean sign;
                if (order) sign = getSign(first.botObj.dig.position,firstChange);
                else sign = getSign(second.botObj.dig.position,firstChange);
                if (sign != (i1 > 0)) m = true;
            }
            int a = 0;
            for (int y = 0; y < firstChange; y++) a = a + first.botObj.dig.position[y];
            a = 1+a;
            if (i1 < 0) a++;
            if (m) a++;
            if (a%2 !=0) frame = false;
        }
        return frame;
    }
    
    int dotIndex(mod0 mod, int label) {
        int ind = -1;
        sh_object bObj = mod.botObj;
        sh_object tObj = mod.topObj;
        boolean found = false;
        int i = 0;
        while (!found) {
            if (tObj.dig.circles.get(i).contains(label)) found = true;
            else i++;
        }
        ind = (tObj.symbols.get(i)-bObj.symbols.get(i))/2;
        return Math.abs(ind);
    }
    
    int diffPosi(diagram dig1, diagram dig2) {
        int u = 0;
        boolean found = false;
        while (!found) {
            if (dig1.position[u] != dig2.position[u]) found = true;
            else u++;
        }
        return u;
    }
    
    mod0 shFindOne(ArrayList<sh_object> objects) {
        int u = 0;
        mod0 returner = null;
        boolean foundone = false;
        while (!foundone & u < objects.size()) {
            sh_object tObj = objects.get(u);
            int v = 0;
            while (!foundone & v < tObj.moduli0t.size()) {
                mod0 mod = tObj.moduli0t.get(v);
                BigInteger tester = mod.totalValue;
                if ((tester.abs().equals(BigInteger.ONE))) {
                    foundone = true;
                    returner = mod;
                }
                else v++;
            }
            u++;
        }
        return returner;
    }
    
    void shiftTop(mod0 mod) {
        sh_object tObj = mod.topObj;
        sh_object bObj = mod.botObj;
        int i = 0;
        while (i < bObj.moduli0b.size()) {
            mod0 rmod = bObj.moduli0b.get(i);
            if (rmod != mod) {
                BigInteger factor = rmod.totalValue.multiply(mod.totalValue).negate();
                handleSlide(rmod.topObj,tObj,factor);
                int j = 0;
                while (j <rmod.topObj.moduli1t.size()) {
                    mod1 md = rmod.topObj.moduli1t.get(j);
                    if (md.framedCircle & md.edges.isEmpty()) {
                        rmod.topObj.moduli1t.remove(md);
                        md.botObj.moduli1b.remove(md);
                    }
                    else j++;
                }
            }
            else i++;
        }
    }

    void slideTop(mod0 mod) {
        sh_object tObj = mod.topObj;
        sh_object bObj = mod.botObj;
        int i = 0;
        while (i < bObj.moduli0b.size()) {
            mod0 rmod = bObj.moduli0b.get(i);
            if (rmod != mod) {
                BigInteger div = rmod.totalValue.divide(mod.totalValue);
                div = div.negate();
                handleSlide(rmod.topObj,tObj,div);
                if (!rmod.totalValue.equals(BigInteger.ZERO)) i++;
                int j = 0;
                while (j <rmod.topObj.moduli1t.size()) {
                    mod1 md = rmod.topObj.moduli1t.get(j);
                    if (md.framedCircle & md.edges.isEmpty()) {
                        rmod.topObj.moduli1t.remove(md);
                        md.botObj.moduli1b.remove(md);
                    }
                    else j++;
                }
            }
            else i++;
        }
    }
    
    void shiftBot(mod0 mod) {
        sh_object tObj = mod.topObj;
        sh_object bObj = mod.botObj;
        int i = 0;
        while (i < tObj.moduli0t.size()) {
            mod0 rmod = tObj.moduli0t.get(i);
            if (rmod != mod) {
                BigInteger factor = rmod.totalValue.divide(mod.totalValue);
                handleSlide(bObj,rmod.botObj,factor);
            }
            else i++;
        }
        i = 0;
        while (i < tObj.moduli1t.size()) {
            mod1 md = tObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                tObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
        i = 0;
        while (i < bObj.moduli1t.size()) {
            mod1 md = bObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                bObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
    }

    void slideBot(mod0 mod) {
        sh_object tObj = mod.topObj;
        sh_object bObj = mod.botObj;
        int i = 0;
        while (i < tObj.moduli0t.size()) {
            mod0 rmod = tObj.moduli0t.get(i);
            if (rmod != mod) {
                BigInteger div = rmod.totalValue.divide(mod.totalValue);
                handleSlide(bObj,rmod.botObj,div);
                if (!rmod.totalValue.equals(BigInteger.ZERO)) i++;
            }
            else i++;
        }
        i = 0;
        while (i < tObj.moduli1t.size()) {
            mod1 md = tObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                tObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
        i = 0;
        while (i < bObj.moduli1t.size()) {
            mod1 md = bObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                bObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
    }
    
    void handleSlide(sh_object xObj, sh_object yObj, BigInteger factor) {
        BigInteger two = new BigInteger("2");
        BigInteger four = two.multiply(two);
        boolean evenfactor = factor.remainder(two).equals(BigInteger.ZERO);
        boolean minusslide = factor.compareTo(BigInteger.ZERO) < 0;
        if (!evenfactor) { // this is just about circles going into xObj
            for (mod1 mod : xObj.moduli1b) { // the existence of such a mod implies that there is a circle into xObj
                sh_object tObj = mod.topObj;
                boolean found = false;
                int i = 0;
                while (!found & i < tObj.moduli1t.size()) {
                    mod1 md = tObj.moduli1t.get(i);
                    if (md.botObj == yObj) found = true;
                    else i++;
                }
                if (found) { // there is a circle from tObj to yObj, so remove it
                    mod1 md = tObj.moduli1t.get(i);
                    yObj.moduli1b.remove(md);
                    tObj.moduli1t.remove(md);
                }
                else {
                    mod1 md = new mod1(tObj,yObj,false);
                }
            }
        }
        for (mod0 mod : xObj.moduli0b) { // in fact, there should be only one here
            BigInteger u = new BigInteger("0");
            boolean caution = false;
            if (mod.totalValue.abs().remainder(four).intValue()>1) caution = true;
            boolean found = false;
            int i = 0;
            while (!found & i < mod.topObj.moduli0t.size()) {
                mod0 md = mod.topObj.moduli0t.get(i);
                if (md.botObj == yObj) found = true;
                else i++;
            }
            if (found) {
                mod0 mby = mod.topObj.moduli0t.get(i);
                u = u.add(mby.totalValue);
                mby.totalValue = mby.totalValue.subtract(mod.totalValue.multiply(factor));
                if (mby.totalValue.equals(BigInteger.ZERO)) {
                        mby.topObj.moduli0t.remove(mby);
                        mby.botObj.moduli0b.remove(mby);
                }
            }
            else {
                BigInteger r = new BigInteger("-1");
                mod0 mby = new mod0(mod.topObj,yObj,r.multiply(factor.multiply(mod.totalValue)));
                mod.topObj.moduli0t.add(mby);
                yObj.moduli0b.add(mby);
            }
            for (mod0 md : yObj.moduli0t) {
                found = false;
                BigInteger v = new BigInteger("0");
                boolean morecaution = false;
                boolean extra = false;
                if (caution & !md.totalValue.abs().remainder(two).equals(BigInteger.ZERO)) morecaution = true; 
                i = 0;
                while (!found & i < xObj.moduli0t.size()) {
                    mod0 mxd = xObj.moduli0t.get(i);
                    if (mxd.botObj == md.botObj) found = true;
                    else i++;
                }
                if (found) {
                    mod0 mxd = xObj.moduli0t.get(i);
                    v = v.add(mxd.totalValue);
                    if (!v.remainder(two).equals(BigInteger.ZERO)) extra = true;
                }
                boolean nonewedge = true;
                found = false;
                i = 0;
                while (!found & i < mod.topObj.moduli1t.size()) {
                    if (mod.topObj.moduli1t.get(i).botObj == md.botObj) found = true;
                    else i++;
                }
                mod1 mbd;
                if (found) mbd = mod.topObj.moduli1t.get(i);
                else mbd = new mod1(mod.topObj,md.botObj,true);
                if (!evenfactor) {// produce new edges
                    if (morecaution & extra) mbd.framedCircle = !mbd.framedCircle;
                    if (!mod.totalValue.remainder(two).equals(BigInteger.ZERO) & 
                            !md.totalValue.remainder(two).equals(BigInteger.ZERO)) {
                        boolean tl = mod.totalValue.compareTo(BigInteger.ZERO) > 0;
                        boolean bl = md.totalValue.multiply(factor).compareTo(BigInteger.ZERO) > 0;
                        boolean tr = mod.totalValue.multiply(factor).compareTo(BigInteger.ZERO) < 0;
                        boolean br = md.totalValue.compareTo(BigInteger.ZERO) > 0;
                        edge ed = new edge(xObj,yObj,tl,bl,tr,br);
                        if (br != minusslide) mbd.framedCircle = !mbd.framedCircle;
                        mbd.edges.add(ed);
                        nonewedge = false;
                    }
                }
                // cleanup edges
                if (nonewedge) {
                    if (!found) {
                        if (mbd.framedCircle) {
                            mod.topObj.moduli1t.remove(mbd);
                            md.botObj.moduli1b.remove(mbd);
                        }
                    }
                    else {
                        boolean foundx = false;
                        boolean foundy = false;
                        if (u.abs().compareTo(factor.multiply(mod.totalValue).abs())>0) foundy = true;
                        else if (u.multiply(factor.multiply(mod.totalValue)).compareTo(BigInteger.ZERO)<=0) foundy = true;
                        if (v.abs().compareTo(factor.multiply(md.totalValue).abs())>0) foundx = true;
                        else if (v.multiply(factor.multiply(md.totalValue)).compareTo(BigInteger.ZERO) >=0) foundx = true;
                        i = 0;
                        while (i < mbd.edges.size() & (!foundx | !foundy)) { // will change the sign of the relevant points
                            edge ed = mbd.edges.get(i);
                            if (!foundx) {
                                if (ed.leftPoint == xObj) {
                                    foundx = true;
                                    ed.botleft = !ed.botleft;
                                }
                                if (ed.rightPoint == xObj) {
                                    foundx = true;
                                    ed.botright = !ed.botright;
                                    mbd.framedCircle = !mbd.framedCircle;
                                }
                            }
                            if (!foundy) {
                                if (ed.leftPoint == yObj) {
                                    foundy = true;
                                    ed.topleft = !ed.topleft;
                                }
                                if (ed.rightPoint == yObj) {
                                    foundy = true;
                                    ed.topright = !ed.topright;
                                    mbd.framedCircle = !mbd.framedCircle;
                                }
                            }
                            i++;
                        }
                    }
                }
                else { // there are new edges, in fact just one between x and y
                    if (found) {
                        i = 0;
                        int foundx = -1;
                        int foundy = -1;
                        while ((foundx == -1 | foundy == -1) & i < mbd.edges.size()-1) {
                            edge ed = mbd.edges.get(i);
                            if (ed.leftPoint == xObj | ed.rightPoint == xObj) foundx = i;
                            if (ed.leftPoint == yObj | ed.rightPoint == yObj) foundy = i;
                            i++;
                        }
                        // go through the various cases of combining edges
                        if (foundx == foundy) {
                            if (foundx == -1) { // only need to check that the signs of the edge are correct
                                boolean uppery = (u.subtract(factor.multiply(mod.totalValue)).compareTo(BigInteger.ZERO) > 0);
                                boolean lowerx = (v.add(factor.multiply(md.totalValue)).compareTo(BigInteger.ZERO) > 0);
                                edge ed = mbd.edges.get(mbd.edges.size()-1);
                                ed.botleft = lowerx;
                                ed.topright = uppery;
                                if (uppery != factor.multiply(mod.totalValue).compareTo(BigInteger.ZERO) < 0) 
                                    mbd.framedCircle = !mbd.framedCircle;
                            }
                            else { // two edges producing a circle
                                edge enew = mbd.edges.get(mbd.edges.size()-1);
                                edge eold = mbd.edges.get(foundx);
                                boolean directed = ((eold.botleft ^ eold.topleft) == (eold.botright ^ eold.topright));
                                boolean value = enew.botright;
                                if (enew.botleft != v.compareTo(BigInteger.ZERO) > 0) value = !value;
                                if (directed & eold.leftPoint != xObj) value = !value;
                                if (!value) mbd.framedCircle = !mbd.framedCircle;
                                mbd.edges.remove(enew);
                                mbd.edges.remove(eold);
                            }
                        }
                        else {
                            edge ex = null;
                            edge ey = null;
                            edge enew = mbd.edges.get(mbd.edges.size()-1);
                            if (foundx >=0) ex = mbd.edges.get(foundx);
                            if (foundy >=0) ey = mbd.edges.get(foundy);
                            sh_object ox = null;
                            sh_object oy = null;
                            boolean ty = true;
                            boolean tl = enew.topleft;
                            boolean bl = v.add(factor.multiply(md.totalValue)).compareTo(BigInteger.ZERO) > 0;
                            boolean tr = u.subtract(factor.multiply(mod.totalValue)).compareTo(BigInteger.ZERO) > 0;
                            boolean br = enew.botright;
                            boolean cy = (tr != enew.topright);
                            if (foundx >=0) {
                                ox = ex.leftPoint;
                                tl = ex.topleft;
                                bl = ex.botleft;
                                if (ox == xObj) {
                                        ox = ex.rightPoint;
                                        tl = ex.topright;
                                        bl = ex.botright;
                                }
                            }
                            else ox = xObj;
                            if (foundy >=0) {
                                oy = ey.leftPoint;
                                tr = ey.topleft;
                                br = ey.botleft;
                                ty = ey.topright;
                                if (oy == yObj) {
                                    oy = ey.rightPoint;
                                    tr = ey.topright;
                                    br = ey.botright;
                                    ty = ey.topleft;
                                }
                            }
                            else oy = yObj;
                            edge newed = new edge(ox,oy,tl,bl,tr,br);
                            mbd.edges.remove(ex);
                            mbd.edges.remove(ey);
                            mbd.edges.remove(enew);
                            mbd.edges.add(newed);
                            boolean value = mbd.framedCircle;
                            if (foundy >= 0) {
                                if (ty == enew.topright) value = !value;
                                if (ey.rightPoint == yObj) value = !value;
                            }
                            else if (cy) value = !value;
                            if (foundx >= 0) {
                                if (ex.rightPoint == xObj) value = !value;
                            }
                            mbd.framedCircle = value;
                        }
                    }// if not found, all good with the edges
                }
            }
        }
        // fix the mod0 below x
        for (mod0 md : yObj.moduli0t) {
            boolean found = false;
            int i = 0;
            while (!found & i < xObj.moduli0t.size()) {
                    mod0 mxd = xObj.moduli0t.get(i);
                    if (mxd.botObj == md.botObj) found = true;
                    else i++;
            }
            if (found) {
                    mod0 mxd = xObj.moduli0t.get(i);
                    mxd.totalValue = mxd.totalValue.add(factor.multiply(md.totalValue));
                    if (mxd.totalValue.equals(BigInteger.ZERO)) {
                            mxd.topObj.moduli0t.remove(mxd);
                            mxd.botObj.moduli0b.remove(mxd);
                    }
            }
            else {
                    mod0 mxd = new mod0(xObj,md.botObj,factor.multiply(md.totalValue));
                    md.botObj.moduli0b.add(mxd);
                    xObj.moduli0t.add(mxd);
            }
        }
        // now fix the mod1 below x
        int fourfactor = (factor.abs().remainder(four).intValue());
        for (mod1 mod : yObj.moduli1t) { // find mod1 from x with same end and modify it
            boolean found = false;
            int i = 0;
            while (!found & i < xObj.moduli1t.size()) {
                if (xObj.moduli1t.get(i).botObj == mod.botObj) found = true;
                else i++;
            }
            mod1 md;
            if (found) md = xObj.moduli1t.get(i);
            else md = new mod1(xObj,mod.botObj,true);
            if (fourfactor > 1) {
                int u = mod.edges.size();
                if (u % 2 != 0) md.framedCircle = !md.framedCircle;
            }
            int oldedges = md.edges.size();
            if (!evenfactor) {
                md.framedCircle = !(md.framedCircle ^ mod.framedCircle);
                for (edge ed : mod.edges) {
                    boolean tl = !(!minusslide ^ ed.topleft);
                    boolean tr = !(!minusslide ^ ed.topright);
                    edge eddx = new edge(ed.leftPoint,ed.rightPoint,tl,ed.botleft,tr,ed.botright);
                    md.edges.add(eddx);
                }
            }
            // clean up the edges, turning them to circles if nec
            i = 0;
            while (i < oldedges) {
                edge oed = md.edges.get(i);
                found = false;
                boolean modify = false;
                int j = oldedges;
                edge ned = md.edges.get(0);
                while (!found & j < md.edges.size()) {
                    ned = md.edges.get(j);
                    if (ned.leftPoint == oed.leftPoint | ned.rightPoint == oed.leftPoint) found = true;
                    else j++;
                }
                boolean finished = false;
                if (found) {
                    modify = true;
                    sh_object otherPoint = ned.leftPoint;
                    boolean to = ned.topleft;
                    boolean bo = ned.botleft;
                    boolean tc = ned.topright;
                    boolean ow = false;
                    if (ned.leftPoint == oed.leftPoint) {
                        otherPoint = ned.rightPoint;
                        to = ned.topright;
                        bo = ned.botright;
                        tc = ned.topleft;
                        ow = true;
                    }
                    boolean downarrow = (tc == oed.topleft);
                    oed.leftPoint = otherPoint;
                    oed.topleft = to;
                    if (!oed.botleft) md.framedCircle = !md.framedCircle;
                    if (directEdge(ned) & ow) md.framedCircle = !md.framedCircle;
                    oed.botleft = bo;
                    md.edges.remove(ned);
                    if (downarrow) md.framedCircle = !md.framedCircle;
                    if (otherPoint == oed.rightPoint) { // it's a circle
                        md.framedCircle = !md.framedCircle;
                        if (!oed.botright) md.framedCircle = !md.framedCircle;
                        md.edges.remove(oed);
                        oldedges--;
                        i--;
                        finished = true;
                    }
                }
                if (!finished) {
                    found = false;
                    j = oldedges;
                    while (!found & j < md.edges.size()) {
                        ned = md.edges.get(j);
                        if (ned.leftPoint == oed.rightPoint | ned.rightPoint == oed.rightPoint) found = true;
                        else j++;
                    }
                    if (found) {
                        modify = true;
                        sh_object otherPoind = ned.leftPoint;
                        boolean to = ned.topleft;
                        boolean bo = ned.botleft;
                        boolean ow = true;
                        if (ned.leftPoint == oed.rightPoint) {
                            otherPoind = ned.rightPoint;
                            to = ned.topright;
                            bo = ned.botright;
                            ow = false;
                        }
                        oed.rightPoint = otherPoind;
                        oed.topright = to;
                        if (!oed.botright) md.framedCircle = !md.framedCircle; 
                        oed.botright = bo;
                        md.edges.remove(ned);
                        if (ow & directEdge(ned)) md.framedCircle = !md.framedCircle;
                    }
                    if (modify) {
                        md.edges.remove(oed);
                        md.edges.add(oed);
                        oldedges--;
                        i--;
                    }
                }
                i++;
            }
        }
        // make sure edges have the right signs
        for (mod1 md : xObj.moduli1t) {
            for (edge ed : md.edges) {
                boolean foundl = false;
                boolean foundr = false;
                int fleft = -1;
                int fright = -1;
                int i = 0;
                while (i < xObj.moduli0t.size() & (!foundl | !foundr)) {
                    mod0 mrd = xObj.moduli0t.get(i);
                    if (mrd.botObj == ed.leftPoint) {
                        foundl = true;
                        fleft = i;
                    }
                    if (mrd.botObj == ed.rightPoint) {
                        foundr = true;
                        fright = i;
                    }
                    i++;
                }
                mod0 modl = xObj.moduli0t.get(fleft);
                mod0 modr = xObj.moduli0t.get(fright);
                if (ed.topleft == (modl.totalValue.compareTo(BigInteger.ZERO) < 0)) {
                    ed.topleft = !ed.topleft;
                }
                if (ed.topright == (modr.totalValue.compareTo(BigInteger.ZERO) < 0)) {
                    ed.topright = !ed.topright;
                    md.framedCircle = !md.framedCircle;
                }
            }
        }
    }
    
    boolean directEdge(edge ed) {
        boolean decision = false;
        if(ed.botleft ^ ed.topleft == ed.botright ^ ed.topright) decision = true;
        return decision;
    }
    
    mod0 shFindSmallest(ArrayList<sh_object> objects) {
        mod0 returner = null;
        boolean foundsmth = false;
        BigInteger minvalue = BigInteger.ZERO;
        for (sh_object obj : objects) {
            if (obj.moduli0t.size()>0) {
                int u = 0;
                if (!foundsmth) {
                    returner = obj.moduli0t.get(0);
                    minvalue = returner.totalValue.abs();
                    foundsmth = true;
                }
                while (u < obj.moduli0t.size()) {
                    BigInteger compare = obj.moduli0t.get(u).totalValue.abs(); 
                    int comp = minvalue.compareTo(compare);
                    if (comp == 1) {
                        returner = obj.moduli0t.get(u);
                        minvalue = compare;
                    }
                    u++;
                }
            }
        }
        return returner;
    }
    
    ArrayList<ArrayList<m4_object>> createm4Objects(ArrayList<ArrayList<sh_object>> objects,int nminus, link Link) {
        ArrayList<ArrayList<m4_object>> m4Objects = new ArrayList<ArrayList<m4_object>>();
        for (int e = 0; e < objects.size(); e++) {
            ArrayList<sh_object> objs = objects.get(e);
            ArrayList<m4_object> m4objs = new ArrayList<m4_object>();
            for (int f = 0; f < objs.size(); f++) {
                int h = nminus+homDegree(objs.get(f).dig.position,Link.crossings);
                m4_object m4obj = new m4_object(f,h);
                m4objs.add(m4obj);
            }
            m4Objects.add(m4objs);
        }
        for (int e = 0; e < objects.size(); e++) {
            ArrayList<sh_object> objs = objects.get(e);
            ArrayList<m4_object> m4objs = m4Objects.get(e);
            for (int f = 0; f < objs.size(); f++) {
                sh_object bObj = objs.get(f);
                m4_object bOb4 = m4objs.get(f);
                for (mod0 mod : bObj.moduli0b) {
                    int tPos = objects.get(e+1).indexOf(mod.topObj);
                    m4_object tOb4 = m4Objects.get(e+1).get(tPos);
                    m4mod0 m4d = new m4mod0(tOb4,bOb4,mod.totalValue.intValue());
                    tOb4.moduli0t.add(m4d);
                    bOb4.moduli0b.add(m4d);
                }
                for (mod1 mod : bObj.moduli1b) {
                    int tPos = objects.get(e+2).indexOf(mod.topObj);
                    m4_object tOb4 = m4Objects.get(e+2).get(tPos);
                    m4mod1 m4d = new m4mod1(tOb4,bOb4,mod.framedCircle);
                    tOb4.moduli1t.add(m4d);
                    bOb4.moduli1b.add(m4d);
                    // deal with edges in general
                    for (edge ed : mod.edges) {
                        int lPos = objects.get(e+1).indexOf(ed.leftPoint);
                        int rPos = objects.get(e+1).indexOf(ed.rightPoint);
                        m4_object lOb4 = m4Objects.get(e+1).get(lPos);
                        m4_object rOb4 = m4Objects.get(e+1).get(rPos);
                        m4edge med = new m4edge(lOb4,rOb4,ed.topleft,ed.botleft,ed.topright,ed.botright);
                        m4d.edges.add(med);
                    }
                }
            }
        }
        return m4Objects;
    }
    
    ArrayList<ArrayList<m4_object>> clonem4Objects(ArrayList<ArrayList<m4_object>> objects) {
        ArrayList<ArrayList<m4_object>> m4Objects = new ArrayList<ArrayList<m4_object>>();
        for (int e = 0; e < objects.size(); e++) {
            ArrayList<m4_object> objs = objects.get(e);
            ArrayList<m4_object> m4objs = new ArrayList<m4_object>();
            for (int f = 0; f < objs.size(); f++) {
                int h = objs.get(f).hdeg;
                m4_object m4obj = new m4_object(f,h);
                m4objs.add(m4obj);
            }
            m4Objects.add(m4objs);
        }
        for (int e = 0; e < objects.size(); e++) {
            ArrayList<m4_object> objs = objects.get(e);
            ArrayList<m4_object> m4objs = m4Objects.get(e);
            for (int f = 0; f < objs.size(); f++) {
                m4_object bObj = objs.get(f);
                m4_object bOb4 = m4objs.get(f);
                for (m4mod0 mod : bObj.moduli0b) {
                    int tPos = objects.get(e+1).indexOf(mod.topObj);
                    m4_object tOb4 = m4Objects.get(e+1).get(tPos);
                    m4mod0 m4d = new m4mod0(tOb4,bOb4,mod.value);
                    tOb4.moduli0t.add(m4d);
                    bOb4.moduli0b.add(m4d);
                }
                for (m4mod1 mod : bObj.moduli1b) {
                    int tPos = objects.get(e+2).indexOf(mod.topObj);
                    m4_object tOb4 = m4Objects.get(e+2).get(tPos);
                    m4mod1 m4d = new m4mod1(tOb4,bOb4,mod.framedCircle);
                    tOb4.moduli1t.add(m4d);
                    bOb4.moduli1b.add(m4d);
                    // deal with edges in general
                    for (m4edge ed : mod.edges) {
                        int lPos = objects.get(e+1).indexOf(ed.leftPoint);
                        int rPos = objects.get(e+1).indexOf(ed.rightPoint);
                        m4_object lOb4 = m4Objects.get(e+1).get(lPos);
                        m4_object rOb4 = m4Objects.get(e+1).get(rPos);
                        m4edge med = new m4edge(lOb4,rOb4,ed.topleft,ed.botleft,ed.topright,ed.botright);
                        m4d.edges.add(med);
                    }
                }
            }
        }
        return m4Objects;
    }
    
    int homDegree(int[] pos, int[] abs) {
        int hdeg = 0;
        for (int j = 0; j < pos.length; j++) {
            if (abs[j]>=0) hdeg = hdeg +pos[j];
            else hdeg = hdeg+pos[j]-abs[j];
        }
        return hdeg;
    }

    void M4CatOutput(ArrayList<ArrayList<m4_object>> objects) {
        int j = objects.size()-1;
        while (j >= 0) {
            if (!objects.get(j).isEmpty()) {
                System.out.println("-----");
                for (m4_object obj : objects.get(j)) {
                    System.out.println("Level "+obj.hdeg+" ("+obj.number+")");
                    if (!obj.moduli0t.isEmpty()) {
                        System.out.println("Mod0 :");
                        for (m4mod0 mod : obj.moduli0t) {
                            System.out.println(mod.value+" to ("+mod.botObj.number+")");
                        }
                    }
                    if (!obj.moduli1t.isEmpty()) {
                        System.out.println("Mod1 :");
                        for (m4mod1 mod : obj.moduli1t) {
                            System.out.println(mod.framedCircle+" to ("+mod.botObj.number+")");
                            if (!mod.edges.isEmpty()) {
                                System.out.println("Edges");
                                for (m4edge ed : mod.edges) {
                                    System.out.println(ed.botleft+" "+ed.topleft+" via ("+ed.leftPoint.number+")");
                                    System.out.println(ed.botright+" "+ed.topright+" via ("+ed.rightPoint.number+")");
                                }
                            }
                        }
                    }
                    System.out.println();
                }
            }
            j--;
        }
    }
    
    m4mod0 m4findOne(ArrayList<m4_object> objects) {
        int u = 0;
        m4mod0 returner = null;
        boolean foundone = false;
        while (!foundone & u < objects.size()) {
            m4_object tObj = objects.get(u);
            int v = 0;
            while (!foundone & v < tObj.moduli0t.size()) {
                m4mod0 mod = tObj.moduli0t.get(v);
                if ((mod.value)%2 != 0) {
                    foundone = true;
                    returner = mod;
                }
                else v++;
            }
            u++;
        }
        return returner;
    }
    
    m4mod0 m4findTwo(ArrayList<m4_object> objects) {// note that this method is only used when there are no one's
        m4mod0 returner = null;
        boolean foundtwo = false;
        int u = 0;
        while (!foundtwo & u < objects.size()) {
            m4_object tObj = objects.get(u);
            int v = 0;
            while (!foundtwo & v < tObj.moduli0t.size()) {
                m4mod0 mod = tObj.moduli0t.get(v);
                if ((mod.value)%4 != 0) {
                    foundtwo = true;
                    returner = mod;
                }
                else v++;
            }
            u++;
        }
        return returner;
    }
    
    int dotIndex(sh_object bObj, sh_object tObj, int label) {
        int ind = -1;
        boolean found = false;
        int i = 0;
        while (!found) {
            if (tObj.dig.circles.get(i).contains(label)) found = true;
            else i++;
        }
        ind = (tObj.symbols.get(i)-bObj.symbols.get(i))/2;
        return Math.abs(ind);
    }
    
    boolean m4framing(m4mod0 first, m4mod0 second,diagram ftdig, diagram fbdig, diagram stdig, diagram sbdig) {
        int factor1 = 0;
        int factor2 = 0;
        int firstChange = diffPosition(ftdig,fbdig);
        int seconChange = diffPosition(stdig,sbdig);
        boolean m1 = false;
        boolean m2 = false;
        int i1 = first.value;
        if (Math.abs(i1)==1) {
            boolean sign1 = getSign(fbdig.position,firstChange);
            if (sign1 != (i1 > 0)) m1 = true;
        }
        i1 = second.value;
        if (Math.abs(i1)==1) {
            boolean sign2 = getSign(sbdig.position,seconChange);
            if (sign2 != (i1>0)) m2 = true;
        }
        boolean order = true;
        if (firstChange > seconChange) {
            int h = firstChange;
            firstChange = seconChange;
            seconChange = h;
            order = false;
        }
        for (int y = 0; y < firstChange; y++) factor1 = factor1 + fbdig.position[y];
        for (int y = firstChange; y < seconChange; y++) factor2 = factor2 + fbdig.position[y];
        int b = fbdig.position[firstChange];
        if ((order & m1) | (!order & m2)) {
            factor1++;
            factor2 = factor2 + fbdig.position[firstChange];
            if (b < 0) factor2++;
        }
        if ((order & m2) | (!order & m1)) factor2++;
        factor1 = factor1 * factor2;
        boolean frame = true;
        if (factor1 % 2 != 0) frame = false; // framing of interval has been determined.
        return frame;
    }
    
    boolean m4framingZero(m4mod0 first, m4mod0 second,diagram ftdig, diagram fbdig, diagram stdig, diagram sbdig) {
        int firstChange = diffPosition(ftdig,fbdig);
        int seconChange = diffPosition(stdig,sbdig);
        boolean order = true;
        if (firstChange > seconChange) {
            int h = firstChange;
            firstChange = seconChange;
            seconChange = h;
            order = false;
        }
        boolean firstZero = true;
        boolean frame = true;
        boolean m = false;
        if (first.value != 0) firstZero = false;
        if ( order == firstZero ) { // this means 0 then something
            int i1;
            if (order) i1 = second.value;
            else i1 = first.value;
            if (Math.abs(i1) == 1) { // check if it's an M point
                boolean sign;
                if (order) sign = getSign(sbdig.position,seconChange);
                else sign = getSign(fbdig.position,seconChange);
                if (sign != (i1 > 0)) m = true;
            }
            int a = 0;
            for (int y = 0; y < firstChange; y++) a = a + fbdig.position[y];
            int b = 0;
            for (int y = firstChange; y < seconChange; y++) b = b + fbdig.position[y];
            int d = 0;
            if (fbdig.position[firstChange] < 0) d = 1;
            int c = d + fbdig.position[firstChange];
            a = 1 + (a+1) * c+b;
            if ((i1 == 1) == order) a++;
            if (m) a++;
            if (a%2 != 0) frame = false;
        }
        else { // this means something then 0
            int i1;
            if (order) i1 = first.value;
            else i1 = second.value;
            if (Math.abs(i1) == 1) { // check if it's an M point
                boolean sign;
                if (order) sign = getSign(fbdig.position,firstChange);
                else sign = getSign(sbdig.position,firstChange);
                if (sign != (i1 > 0)) m = true;
            }
            int a = 0;
            for (int y = 0; y < firstChange; y++) a = a + fbdig.position[y];
            a = 1+a;
            if (i1 < 0) a++;
            if (m) a++;
            if (a%2 !=0) frame = false;
        }
        return frame;
    }
    
    void m4shiftBot(m4mod0 mod) {
        m4_object tObj = mod.topObj;
        m4_object bObj = mod.botObj;
        int i = 0;
        while (i < tObj.moduli0t.size()) {
            m4mod0 rmod = tObj.moduli0t.get(i);
            if (rmod != mod) {
                int factor = 1;
                if ((mod.value+4)%4 == 1) factor = (rmod.value+4)%4;
                if ((mod.value+4)%4 == 3) factor = 4-(rmod.value+4)%4;
                if (factor ==3) factor = -1;
                m4handleSlide(bObj,rmod.botObj,factor);
            }
            else i++;
        }
        i = 0;
        while (i < tObj.moduli1t.size()) {
            m4mod1 md = tObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                tObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
        i = 0;
        while (i < bObj.moduli1t.size()) {
            m4mod1 md = bObj.moduli1t.get(i);
            if (md.framedCircle & md.edges.isEmpty()) {
                bObj.moduli1t.remove(md);
                md.botObj.moduli1b.remove(md);
            }
            else i++;
        }
    }
    
    void m4shiftTop(m4mod0 mod) {
        m4_object tObj = mod.topObj;
        m4_object bObj = mod.botObj;
        int i = 0;
        while (i < bObj.moduli0b.size()) {
            m4mod0 rmod = bObj.moduli0b.get(i);
            if (rmod != mod) {
                int factor = 1;
                if ((mod.value+40)%4 == 1) factor = 4-(rmod.value+4)%4;
                if ((mod.value+40)%4 == 3) factor = (rmod.value+4)%4;
                if (factor ==3) factor = -1;
                m4handleSlide(rmod.topObj,tObj,factor);
                int j = 0;
                while (j <rmod.topObj.moduli1t.size()) {
                    m4mod1 md = rmod.topObj.moduli1t.get(j);
                    if (md.framedCircle & md.edges.isEmpty()) {
                        rmod.topObj.moduli1t.remove(md);
                        md.botObj.moduli1b.remove(md);
                    }
                    else j++;
                }
            }
            else i++;
        }
    }
    
    boolean m4directEdge(m4edge ed) {
        boolean decision = false;
        if(ed.botleft ^ ed.topleft == ed.botright ^ ed.topright) decision = true;
        return decision;
    }
    
    void m4handleSlide(m4_object xObj, m4_object yObj, int factor) {
        boolean evenfactor = factor==2;
        boolean minusslide = factor==-1;
        if (!evenfactor) { // this is just about circles going into xObj
            for (m4mod1 mod : xObj.moduli1b) { // the existence of such a mod implies that there is a circle into xObj
                m4_object tObj = mod.topObj;
                boolean found = false;
                int i = 0;
                while (!found & i < tObj.moduli1t.size()) {
                    m4mod1 md = tObj.moduli1t.get(i);
                    if (md.botObj == yObj) found = true;
                    else i++;
                }
                if (found) { // there is a circle from tObj to yObj, so remove it
                    m4mod1 md = tObj.moduli1t.get(i);
                    yObj.moduli1b.remove(md);
                    tObj.moduli1t.remove(md);
                }
                else {
                    m4mod1 md = new m4mod1(tObj,yObj,false);
                    tObj.moduli1t.add(md);
                    yObj.moduli1b.add(md);
                }
            }
        }
        for (m4mod0 mod : xObj.moduli0b) { // in fact, there should be only one here
            boolean caution = false;
            if ((mod.value+40)%4==2) caution = true;
            boolean found = false;
            int i = 0;
            int u = 0;
            while (!found & i < mod.topObj.moduli0t.size()) {
                m4mod0 md = mod.topObj.moduli0t.get(i);
                if (md.botObj == yObj) found = true;
                else i++;
            }
            if (found) {
                m4mod0 mby = mod.topObj.moduli0t.get(i);
                u = (mby.value+40)%4;
                mby.topObj.moduli0t.remove(mby);
                mby.botObj.moduli0b.remove(mby);
            }
            else {
                if (factor != 2) {
                    System.out.println("I don't believe this happens "+mod.value+" "+mod.topObj.moduli0t.size()+" "+mod.topObj.number);
                    System.out.println(mod.topObj.number);
                }
            }
            for (m4mod0 md : yObj.moduli0t) {
                found = false;
                boolean morecaution = false;
                boolean extra = false;
                if (caution & !((md.value+40)%4 ==2)) morecaution = true; 
                i = 0;
                int v = 0;
                while (!found & i < xObj.moduli0t.size()) {
                    m4mod0 mxd = xObj.moduli0t.get(i);
                    if (mxd.botObj == md.botObj) found = true;
                    else i++;
                }
                if (found) {
                    m4mod0 mxd = xObj.moduli0t.get(i);
                    v = (mxd.value+40)%4;
                    if (!(v%2==0)) extra = true;
                }
                boolean nonewedge = true;
                found = false;
                i = 0;
                while (!found & i < mod.topObj.moduli1t.size()) {
                    if (mod.topObj.moduli1t.get(i).botObj == md.botObj) found = true;
                    else i++;
                }
                m4mod1 mbd;
                if (found) mbd = mod.topObj.moduli1t.get(i);
                else {
                    mbd = new m4mod1(mod.topObj,md.botObj,true);
                    mod.topObj.moduli1t.add(mbd);
                    md.botObj.moduli1b.add(mbd);
                }
                if (!evenfactor) {// produce new edges
                    if (morecaution & extra) mbd.framedCircle = !mbd.framedCircle;
                    if (!(mod.value%2==0) & 
                        !(md.value%2==0)) {
                        boolean tl = (mod.value+40)%4 == 1;
                        boolean bl = (md.value*factor+40)%4 == 1;
                        boolean tr = (mod.value*factor+40)%4 == 3;
                        boolean br = (md.value+40)%4 == 1;
                        m4edge ed = new m4edge(xObj,yObj,tl,bl,tr,br);
                        if (br != minusslide) mbd.framedCircle = !mbd.framedCircle;
                        mbd.edges.add(ed);
                        nonewedge = false;
                    }
                }
                // cleanup edges
                if (nonewedge) {
                    if (!found) {
                        if (mbd.framedCircle) {
                            mod.topObj.moduli1t.remove(mbd);
                            md.botObj.moduli1b.remove(mbd);
                        }
                    }
                    else {
                        boolean foundx = false;
                        if ((v%2==0) | (mod.value%2)==0) foundx = true;
                        else if ((factor*md.value+40)%4!=2) foundx = true;
                        i = 0;
                        while (i < mbd.edges.size() & !foundx ) { // will change the sign of the relevant points
                            m4edge ed = mbd.edges.get(i);
                            if (ed.leftPoint == xObj) {
                                foundx = true;
                                ed.botleft = !ed.botleft;
                            }
                            if (ed.rightPoint == xObj) {
                                foundx = true;
                                ed.botright = !ed.botright;
                                mbd.framedCircle = !mbd.framedCircle;
                            }
                            i++;
                        }
                    }
                }
                else { // there are new edges, in fact just one between x and y
                    if (found) {
                        i = 0;
                        int foundx = -1;
                        int foundy = -1;
                        while ((foundx == -1 | foundy == -1) & i < mbd.edges.size()-1) {
                            m4edge ed = mbd.edges.get(i);
                            if (ed.leftPoint == xObj | ed.rightPoint == xObj) foundx = i;
                            if (ed.leftPoint == yObj | ed.rightPoint == yObj) foundy = i;
                            i++;
                        }
                        // go through the various cases of combining edges
                        if (foundx == foundy) {
                            m4edge enew = mbd.edges.get(mbd.edges.size()-1);
                            m4edge eold = mbd.edges.get(foundx);
                            boolean directed = ((eold.botleft ^ eold.topleft) == (eold.botright ^ eold.topright));
                            boolean value = enew.botright;
                            if (enew.botleft != v > 0) value = !value;
                            if (directed & eold.leftPoint != xObj) value = !value;
                            if (!value) mbd.framedCircle = !mbd.framedCircle;
                            mbd.edges.remove(enew);
                            mbd.edges.remove(eold);
                        }
                        else {
                            m4edge ex = null;
                            m4edge ey = null;
                            m4edge enew = mbd.edges.get(mbd.edges.size()-1);
                            if (foundx >=0) ex = mbd.edges.get(foundx);
                            if (foundy >=0) ey = mbd.edges.get(foundy);
                            m4_object ox = null;
                            m4_object oy = null;
                            boolean ty = true;
                            boolean tl = enew.topleft;
                            boolean bl = (v+ factor*md.value+40)%4 == 1;
                            boolean tr = (u-factor*mod.value+40)%4 == 1;
                            boolean br = enew.botright;
                            boolean cy = (tr != enew.topright);
                            if (foundx >=0) {
                                ox = ex.leftPoint;
                                tl = ex.topleft;
                                bl = ex.botleft;
                                if (ox == xObj) {
                                    ox = ex.rightPoint;
                                    tl = ex.topright;
                                    bl = ex.botright;
                                }
                            }
                            else ox = xObj;
                            if (foundy >=0) {
                                oy = ey.leftPoint;
                                tr = ey.topleft;
                                br = ey.botleft;
                                ty = ey.topright;
                                if (oy == yObj) {
                                    oy = ey.rightPoint;
                                    tr = ey.topright;
                                    br = ey.botright;
                                    ty = ey.topleft;
                                }
                            }
                            else oy = yObj;
                            m4edge newed = new m4edge(ox,oy,tl,bl,tr,br);
                            mbd.edges.remove(ex);
                            mbd.edges.remove(ey);
                            mbd.edges.remove(enew);
                            mbd.edges.add(newed);
                            boolean value = mbd.framedCircle;
                            if (foundy >= 0) {
                                if (ty == enew.topright) value = !value;
                                if (ey.rightPoint == yObj) value = !value;
                            }
                            else if (cy) value = !value;
                            if (foundx >= 0) {
                                if (ex.rightPoint == xObj) value = !value;
                            }
                            mbd.framedCircle = value;
                        }
                    }// if not found, all good with the edges
                }
            }
        }
        // fix the mod0 below x
        for (m4mod0 md : yObj.moduli0t) {
            boolean found = false;
            int i = 0;
            while (!found & i < xObj.moduli0t.size()) {
                m4mod0 mxd = xObj.moduli0t.get(i);
                if (mxd.botObj == md.botObj) found = true;
                else i++;
            }
            if (found) {
                m4mod0 mxd = xObj.moduli0t.get(i);
                mxd.value = (mxd.value+factor*md.value+40)%4;//mxd.totalValue.add(factor.multiply(md.totalValue));
                if (mxd.value==0) {
                    mxd.topObj.moduli0t.remove(mxd);
                    mxd.botObj.moduli0b.remove(mxd);
                }
            }
            else {
                    if ((40+factor*md.value)%4!=0) {
                            m4mod0 mxd = new m4mod0(xObj,md.botObj,(40+factor*md.value)%4);
                            md.botObj.moduli0b.add(mxd);
                            xObj.moduli0t.add(mxd);
                    }
            }
        }
        // now fix the mod1 below x
        for (m4mod1 mod : yObj.moduli1t) { // find mod1 from x with same end and modify it
                boolean found = false;
                int i = 0;
                while (!found & i < xObj.moduli1t.size()) {
                    if (xObj.moduli1t.get(i).botObj == mod.botObj) found = true;
                    else i++;
                }
                m4mod1 md;
                if (found) md = xObj.moduli1t.get(i);
                else {
                    md = new m4mod1(xObj,mod.botObj,true);
                    xObj.moduli1t.add(md);
                    mod.botObj.moduli1b.add(md);
                }
                if (factor == 2) {
                    int u = mod.edges.size();
                    if (u % 2 != 0) md.framedCircle = !md.framedCircle;
                }
                int oldedges = md.edges.size();
                if (!evenfactor) {
                    md.framedCircle = !(md.framedCircle ^ mod.framedCircle);
                    for (m4edge ed : mod.edges) {
                        boolean tl = !(!minusslide ^ ed.topleft);
                        boolean tr = !(!minusslide ^ ed.topright);
                        m4edge eddx = new m4edge(ed.leftPoint,ed.rightPoint,tl,ed.botleft,tr,ed.botright);
                        md.edges.add(eddx);
                    }
                }
                // clean up the edges, turning them to circles if nec
                i = 0;
                while (i < oldedges) {
                    m4edge oed = md.edges.get(i);
                    found = false;
                    boolean modify = false;
                    int j = oldedges;
                    m4edge ned = md.edges.get(0);
                    while (!found & j < md.edges.size()) {
                        ned = md.edges.get(j);
                        if (ned.leftPoint == oed.leftPoint | ned.rightPoint == oed.leftPoint) found = true;
                        else j++;
                    }
                    boolean finished = false;
                    if (found) {
                        modify = true;
                        m4_object otherPoint = ned.leftPoint;
                        boolean to = ned.topleft;
                        boolean bo = ned.botleft;
                        boolean tc = ned.topright;
                        boolean ow = false;
                        if (ned.leftPoint == oed.leftPoint) {
                            otherPoint = ned.rightPoint;
                            to = ned.topright;
                            bo = ned.botright;
                            tc = ned.topleft;
                            ow = true;
                        }
                        boolean downarrow = (tc == oed.topleft);
                        oed.leftPoint = otherPoint;
                        oed.topleft = to;
                        if (!oed.botleft) md.framedCircle = !md.framedCircle;
                        if (m4directEdge(ned) & ow) md.framedCircle = !md.framedCircle;
                        oed.botleft = bo;
                        md.edges.remove(ned);
                        if (downarrow) md.framedCircle = !md.framedCircle;
                        if (otherPoint == oed.rightPoint) { // it's a circle
                            md.framedCircle = !md.framedCircle;
                            if (!oed.botright) md.framedCircle = !md.framedCircle;
                            md.edges.remove(oed);
                            oldedges--;
                            i--;
                            finished = true;
                        }
                    }
                    if (!finished) {
                            found = false;
                            j = oldedges;
                            while (!found & j < md.edges.size()) {
                                    ned = md.edges.get(j);
                                    if (ned.leftPoint == oed.rightPoint | ned.rightPoint == oed.rightPoint) found = true;
                                    else j++;
                            }
                            if (found) {
                                    modify = true;
                                    m4_object otherPoind = ned.leftPoint;
                                    boolean to = ned.topleft;
                                    boolean bo = ned.botleft;
                                    boolean ow = true;
                                    if (ned.leftPoint == oed.rightPoint) {
                                            otherPoind = ned.rightPoint;
                                            to = ned.topright;
                                            bo = ned.botright;
                                            ow = false;
                                    }
                                    oed.rightPoint = otherPoind;
                                    oed.topright = to;
                                    if (!oed.botright) md.framedCircle = !md.framedCircle; 
                                    oed.botright = bo;
                                    md.edges.remove(ned);
                                    if (ow & m4directEdge(ned)) md.framedCircle = !md.framedCircle;
                            }
                            if (modify) {
                                    md.edges.remove(oed);
                                    md.edges.add(oed);
                                    oldedges--;
                                    i--;
                            }
                    }
                    i++;
                }
        }
        // make sure edges have the right signs
        for (m4mod1 md : xObj.moduli1t) {
            for (m4edge ed : md.edges) {
                boolean foundl = false;
                boolean foundr = false;
                int fleft = -1;
                int fright = -1;
                int i = 0;
                while (i < xObj.moduli0t.size() & (!foundl | !foundr)) {
                    m4mod0 mrd = xObj.moduli0t.get(i);
                    if (mrd.botObj == ed.leftPoint) {
                        foundl = true;
                        fleft = i;
                    }
                    if (mrd.botObj == ed.rightPoint) {
                        foundr = true;
                        fright = i;
                    }
                    i++;
                }
                m4mod0 modl = xObj.moduli0t.get(fleft);
                m4mod0 modr = xObj.moduli0t.get(fright);
                if (ed.topleft == ((40+modl.value)%4 ==3)) {
                    ed.topleft = !ed.topleft;
                }
                if (ed.topright == ((40+modr.value)%4 ==3)) {
                    ed.topright = !ed.topright;
                    md.framedCircle = !md.framedCircle;
                }
            }
        }
    }
    
    int ChangLevel(ArrayList<ArrayList<m4_object>> objects) {
        int i = 0;
        boolean found = false;
        while (!found & i < objects.size()) {
            ArrayList<m4_object> objs = objects.get(i);
            int j = 0;
            while (!found & j < objs.size()) {
                m4_object obj = objs.get(j);
                if (!obj.moduli1b.isEmpty()) found = true;
                else j++;
            }
            if (!found) i++;
        }
        return i;
    }
    
    void delay(int k) {
        try {
                Thread.sleep(k);
            }
            catch(InterruptedException ex) {
                Thread.currentThread().interrupt();
            }
    }
    
    private class pairPoints {
    	mod0 first,second;
    	
    	pairPoints(mod0 f, mod0 s) {
    		first = f;
    		second = s;
    	}
    }
    
    private class pairm4Points {
        m4mod0 first,second;
        
        pairm4Points(m4mod0 f, m4mod0 s) {
            first = f;
            second = s;
        }
    }
    
    private class packige {
   	ArrayList<ArrayList<ArrayList<Integer>>> actions;
   	ArrayList<Khov_object> objects;
   	
   	packige(ArrayList<ArrayList<ArrayList<Integer>>> akschn, ArrayList<Khov_object> objecs) {
   		actions = akschn;
   		objects = objecs;
   	}
    }
    
    private class homology {
        int quantum = 0;
        ArrayList<Integer> degrees = new ArrayList<Integer>();
        ArrayList<homologyGroup> groups = new ArrayList<homologyGroup>();
    }
    
    private class homologyGroup {
        int Betti;
        ArrayList<Integer> Torsion = new ArrayList<Integer>();
    }
    
    public class KhovCategory {
   	ArrayList<Khov_object> objects = new ArrayList<Khov_object>();
   	ArrayList<Integer> endpoints = new ArrayList<Integer>();
    }
    
    public class Khov_object {
        ArrayList<ArrayList<Integer>> paths;
        int hdeg;
        int qdeg;
        int dotpath = 0; // that's unreduced, reduced usually 1
        ArrayList<Khov_object> descObjects;
        ArrayList<circle> circles;
        ArrayList<Kmod0> botMods;
        ArrayList<Kmod0> topMods;
	
        Khov_object(int h, int q) {
            paths = new ArrayList<ArrayList<Integer>>();
            circles = new ArrayList<circle>();
            botMods = new ArrayList<Kmod0>();
            topMods = new ArrayList<Kmod0>();
            hdeg = h;
            qdeg = q;
        }
	   
        Khov_object(ArrayList<ArrayList<Integer>> pats, int h, int q) {
            paths = pats;
            hdeg = h;
            qdeg = q;
            descObjects = new ArrayList<Khov_object>();
            circles = new ArrayList<circle>();
            botMods = new ArrayList<Kmod0>();
            topMods = new ArrayList<Kmod0>();
        }
    }
    
    private class Kmod0 {
        BigInteger totalValue;
        Khov_object topObj, botObj;
        ArrayList<ArrayList<Integer>> actions;
        ArrayList<Kmod0> descMod0;

        Kmod0(Khov_object tobj, Khov_object bobj, BigInteger value) {
            topObj = tobj;
            botObj = bobj;
            totalValue = value;
            actions = new ArrayList<ArrayList<Integer>>();
            descMod0 = new ArrayList<Kmod0>();
        }

        Kmod0(Khov_object tobj, Khov_object bobj, BigInteger value, ArrayList<ArrayList<Integer>> akschn) {
            topObj = tobj;
            botObj = bobj;
            totalValue = value;
            actions = akschn;
            descMod0 = new ArrayList<Kmod0>();
        }
    }
    
    class ReidemeisterMove implements Comparable<ReidemeisterMove>{
        int x,y,z;
        int u,v,t,s;
        
        ReidemeisterMove() {
            x = 0;
            y = 0;
            z = 0;
            u = 0;
            v = 0;
            t = 0;
            s = 0;
        }
        
        ReidemeisterMove(int a, int b, int c, int d, int e, int f, int g) {
            x = a;
            y = b;
            z = c;
            u = d;
            v = e;
            t = f;
            s = g;
        }
        
        public int compareTo(ReidemeisterMove mov) {
            int w = x - mov.x;
            if (w == 0) w = y - mov.y;
            if (w == 0) w = z - mov.z;
            return w;
        }
        
        void printOut() {
            System.out.println("Crossings : "+x+" "+y+" "+z);
            System.out.println("Path positions : "+u+" "+v+" "+t+" "+s);
            System.out.println();
        }
    }
    
    private class circle {
        int comp;
        int sign;

        circle(int c, int s) {
            comp = c;
            sign = s;
        }
        
        circle(ArrayList<Integer> path, int s) {
            comp = path.get(0);
            for (int i = 1; i < path.size(); i++) if (comp > path.get(i)) comp = path.get(i);
            sign = s;
        }
    }
    
    private class dcoeff implements Comparable<dcoeff> {
        int entry;
        int pwra;
        int pwrz;

        dcoeff(int ent, int pow1, int pow2) {
            entry = ent;
            pwra = pow1;
            pwrz = pow2;
        }

        boolean sameEntry(dcoeff comp) {
            boolean same = true;
            if (entry != comp.entry) same = false;
            if (pwra != comp.pwra) same = false;
            if (pwrz != comp.pwrz) same = false; 
            return same;
        }
		
        public int compareTo(dcoeff coff) {
            int i = pwra - coff.pwra;
            if (i == 0) i = pwrz - coff.pwrz;
            return i;
        }
    }
    
    twoPolynomial multiplyTwoPoly(twoPolynomial a, twoPolynomial b) {
        twoPolynomial c = new twoPolynomial();
        for (int i = 0; i < a.entries.size(); i++) {
            twoPolynomial d = new twoPolynomial();
            for (int j = 0; j < b.entries.size(); j++) {
                dcoeff x = a.entries.get(i);
                dcoeff y = b.entries.get(j);
                dcoeff rag = new dcoeff(x.entry * y.entry, x.pwra + y.pwra, x.pwrz + y.pwrz);
                d.entries.add(rag);
            }
            c.add(d);
        }
        c.sort();
        return c;
    }
    
    HOMObject tensorHOMObject(HOMObject obj0, HOMObject obj1, ArrayList<Integer> endpoints) {
        HOMObject newObj;
        ArrayList<ArrayList<Integer>> paths = pathsFrom(obj0.paths,obj1.paths,endpoints);
        ArrayList<Integer> newcircs = getNewCircles(paths);
        for (int y : newcircs) {
            boolean found = false;
            int z = 0;
            while (!found & z < paths.size()) {
                ArrayList<Integer> path = paths.get(z);
                if (path.contains(y)) found = true;
                else z++;
            }
            if (found) paths.remove(paths.get(z));
        }
        for (ArrayList<Integer> path : paths) {
            while (path.size()>2) path.remove(1);
        }
        twoPolynomial a = obj0.poly;
        twoPolynomial b = obj1.poly;
        twoPolynomial ab = multiplyTwoPoly(a,b);
        if (!newcircs.isEmpty()) {
            twoPolynomial circlePol = new twoPolynomial();
            dcoeff top = new dcoeff(1,1,-1);
            dcoeff bot = new dcoeff(-1,-1,-1);
            circlePol.entries.add(top);
            circlePol.entries.add(bot);
            if (endpoints.isEmpty()) {
                if (newcircs.size() == 2) ab = multiplyTwoPoly(ab,circlePol);
            }
            else for (int r = 0; r < newcircs.size(); r++) ab = multiplyTwoPoly(ab,circlePol);
        }
        newObj = new HOMObject(paths,ab);
        return newObj;
    }
    
    HOMCategory startCatHOM(int u, int[] path) {
        HOMCategory cat = new HOMCategory();
        ArrayList<Integer> path1 = new ArrayList<Integer>();
        path1.add(path[0]);
        path1.add(path[1]);
        ArrayList<Integer> path2 = new ArrayList<Integer>();
        path2.add(path[2]);
        path2.add(path[3]);
        ArrayList<ArrayList<Integer>> paths1 = new ArrayList<ArrayList<Integer>>();
        paths1.add(path1);
        paths1.add(path2);
        HOMObject first = new HOMObject(paths1,1,u,0);
        cat.objects.add(first);
        if (u!=0) {
            ArrayList<Integer> path3 = new ArrayList<Integer>();
            path3.add(path[0]);
            path3.add(path[3]);
            ArrayList<Integer> path4 = new ArrayList<Integer>();
            path4.add(path[2]);
            path4.add(path[1]);
            ArrayList<ArrayList<Integer>> paths2 = new ArrayList<ArrayList<Integer>>();
            paths2.add(path3);
            paths2.add(path4);
            ArrayList<Integer> entries = new ArrayList<Integer>();
            ArrayList<Integer> powera = new ArrayList<Integer>();
            ArrayList<Integer> powerz = new ArrayList<Integer>();
            int entry = 1;
            if (u > 0) entry = -1;
            for (int v = 1; v < Math.abs(u); v = v + 2) {
                entries.add(entry);
                powera.add(v * (-1) * entry);
                powerz.add(1);
            }
            HOMObject second = new HOMObject(paths2,entries,powera,powerz);
            cat.objects.add(second);
        }
        for (int y = 0; y < 4; y++) cat.endpoints.add(path[y]);
        return cat;
    }
    
    private class twoPolynomial {
        ArrayList<dcoeff> entries;

        twoPolynomial () {
            entries = new ArrayList<dcoeff>();
        }

        twoPolynomial(int ent, int pow1, int pow2) {
            entries = new ArrayList<dcoeff>();
            dcoeff entry = new dcoeff(ent,pow1,pow2);
            entries.add(entry);
        }

        twoPolynomial(ArrayList<Integer> ent, ArrayList<Integer> powa, ArrayList<Integer> powz) {
            entries = new ArrayList<dcoeff>();
            for (int u = 0; u < ent.size();u++) {
                dcoeff dcoff = new dcoeff(ent.get(u),powa.get(u),powz.get(u));
                entries.add(dcoff);
            }
        }

        void add(twoPolynomial c) {
            for (dcoeff entry : c.entries) {
                boolean found = false;
                int u = 0;
                while (!found & u < this.entries.size()) {
                    if (this.entries.get(u).pwra == entry.pwra & this.entries.get(u).pwrz == entry.pwrz) found = true;
                    else u++;
                }
                if (!found) this.entries.add(entry);
                else {
                    this.entries.get(u).entry = this.entries.get(u).entry + entry.entry;
                    if (this.entries.get(u).entry == 0) this.entries.remove(u);
                }
            }
            this.sort();
        }

        boolean samePoly(twoPolynomial comp) {
            boolean same = true;
            if (entries.size()!=comp.entries.size()) return false;
            int u = 0;
            while (same & u < entries.size()) {
                if (!entries.get(u).sameEntry(comp.entries.get(u))) same = false;
                else u++;
            }
            return same;
        }
		
        String getText(boolean latex) {
            String word = "";
            String latop = "";
            String labot = "";
            if (latex) {
                latop = "{";
                labot = "}";
            }
            String number = "";
            boolean start = true;
            for (dcoeff coff : entries) {
                String apwr = "";
                String zpwr = "";
                if (start) {
                    start = false;
                    number = numberOf(coff.entry,coff.pwra,coff.pwrz);
                }
                else {
                    if (coff.entry > 0) number = "+";
                    else number = "";
                    number = number + numberOf(coff.entry,coff.pwra,coff.pwrz); 
                }
                word = word + number+" ";
                if (coff.pwra == 1) apwr = "a ";
                else if (coff.pwra != 0) apwr = "a^"+latop+String.valueOf(coff.pwra)+labot+" ";
                if (coff.pwrz == 1) zpwr = "z ";
                else if (coff.pwrz != 0) zpwr = "z^"+latop+String.valueOf(coff.pwrz)+labot+" ";
                word = word + apwr + zpwr;
            }
            return word.substring(0,word.length()-1);
        }

        String numberOf(int entry, int pwra, int pwrz) {
            String number = "";
            if (entry < 0) number = "-";
            if (pwra == 0 & pwrz == 0) number = number + String.valueOf(Math.abs(entry));
            else if (Math.abs(entry)>1) number = number + String.valueOf(Math.abs(entry));
            return number;
        }
        
        void sort() {
            Collections.sort(entries);
        }
    }
    
    private class HOMCategory {
        ArrayList<HOMObject> objects = new ArrayList<HOMObject>();
        ArrayList<Integer> endpoints = new ArrayList<Integer>();
    }
	
    private class HOMObject {
        ArrayList<ArrayList<Integer>> paths;
        twoPolynomial poly;

        HOMObject(ArrayList<ArrayList<Integer>> pts, int ent, int pow1, int pow2) {
            paths = new ArrayList<ArrayList<Integer>>(pts);
            poly = new twoPolynomial(ent,pow1,pow2);
        }

        HOMObject(ArrayList<ArrayList<Integer>> pts, twoPolynomial pol) {
            paths = new ArrayList<ArrayList<Integer>>(pts);
            poly = pol;
        }

        HOMObject(ArrayList<ArrayList<Integer>> pts, ArrayList<Integer> ent, ArrayList<Integer> powa,
                ArrayList<Integer> powz) {
            paths = new ArrayList<ArrayList<Integer>>(pts);
            poly = new twoPolynomial(ent,powa,powz);
        }
    }
    
    private class diagram {
        int [] position;
        ArrayList<ArrayList<Integer>> circles;
        ArrayList<ArrayList<Integer>> paths;
        
        diagram(int n, int k) {
        	int [] pos = new int[k];
        	pos[0] = n;
        	for (int j = 1; j < k; j++) pos[j] = -1;
        	position = pos;
        	circles = new ArrayList<ArrayList<Integer>>();
        	paths = new ArrayList<ArrayList<Integer>>();
        }
        
        diagram(diagram olddiag, int[] pos, int i, int j) {
        	position = pos;
        	position[i] = j;
        	circles = new ArrayList<ArrayList<Integer>>();
        	paths = new ArrayList<ArrayList<Integer>>();
        	copyComp(circles,olddiag.circles);
        	copyComp(paths,olddiag.paths);
        }
        
    void copyComp(ArrayList<ArrayList<Integer>> newguys, ArrayList<ArrayList<Integer>> oldguys) {
            for (ArrayList<Integer> oldguy : oldguys) {
                ArrayList<Integer> newguy = new ArrayList<Integer>();
                for (int i : oldguy) newguy.add(i);
                newguys.add(newguy);
            }
        }
    }
    
    private class sh_object {
        diagram dig;
        ArrayList<Byte> symbols;
        ArrayList<mod0> moduli0t,moduli0b;
        ArrayList<mod1> moduli1t,moduli1b;
        
        sh_object(diagram dg, ArrayList<Byte> symbs) {
            dig = dg;
            symbols = symbs;
            moduli0t = new ArrayList<mod0>(); // List of 0-dim moduli spaces of which this object is the top-object
            moduli0b = new ArrayList<mod0>(); // List of 0-dim moduli spaces of which this object is the bot-object
            moduli1t = new ArrayList<mod1>();
            moduli1b = new ArrayList<mod1>();
        }
    }
    
    private class mod1 {
    	sh_object topObj, botObj;
    	boolean framedCircle;
    	ArrayList<edge> edges;
    	
    	mod1(sh_object top, sh_object bot, boolean frame) {
    		topObj = top;
    		botObj = bot;
    		topObj.moduli1t.add(this);
    		botObj.moduli1b.add(this);
    		edges = new ArrayList<edge>();
    		framedCircle = frame; // true means there is no framed circle (or rather, an even number of circles)
    	}
    }
    
    private class edge {
    	sh_object leftPoint, rightPoint;
    	boolean topleft,botleft,topright,botright;
    	
    	edge(sh_object left, sh_object right, boolean tl,boolean bl, boolean tr,boolean br) {
    		leftPoint = left;
    		rightPoint = right;
    		topleft = tl;
    		botleft = bl;
    		topright = tr;
    		botright = br;
    	}
    }
    
    private class mod0 {
    	BigInteger totalValue;
    	sh_object topObj, botObj;
    	mod0(sh_object top, sh_object bot, BigInteger val) {
    		topObj = top;
    		botObj = bot;
    		totalValue = val;
    		int i = val.intValue();
    		BigInteger start = new BigInteger("1");
    		if (i < 0) {
    			start = start.negate();
    			i = (-1)* i;
    		}
    	}
    }
    
    private class BHComps implements Comparable<BHComps> {
        ArrayList<BHComponent> comps;
        int qdeg;
        int level;
        boolean stab;
        
        public BHComps() {
            comps = new ArrayList<BHComponent>();
        }
        
        public int compareTo(BHComps comp) {
            int i = qdeg - comp.qdeg;
            if (i == 0) {
                if (stab) i = 1;
                else i = -1;
            }
            return -i;
        
        }
    }
    
    private class BHComponent {
        ArrayList<m4_object> objects;
        int type;
        
        public BHComponent() {
            objects = new ArrayList<m4_object>();
        }
    }
    
    private class m4_object {
    	int number;
    	int hdeg;
    	ArrayList<m4mod0> moduli0t,moduli0b;
    	ArrayList<m4mod1> moduli1t,moduli1b;
    	
    	m4_object(int n, int h) {
    		number = n;
    		hdeg = h;
    		moduli0t = new ArrayList<m4mod0>();
    		moduli0b = new ArrayList<m4mod0>();
    		moduli1t = new ArrayList<m4mod1>();
    		moduli1b = new ArrayList<m4mod1>();
    	}
    }
    
    private class m4mod0 {
    	int value; // really an integer mod4
    	m4_object topObj,botObj;
    	
    	m4mod0(m4_object tObj, m4_object bObj, int v) {
    		value = v;
    		topObj = tObj;
    		botObj = bObj;
    	}
    }
    
    private class m4mod1 {
    	boolean framedCircle;
    	m4_object topObj,botObj;
    	ArrayList<m4edge> edges;
    	
    	m4mod1(m4_object tObj, m4_object bObj, boolean v) {
    		framedCircle = v;
    		topObj = tObj;
    		botObj = bObj;
    		edges = new ArrayList<m4edge>();
    	}
    }
    
    private class m4edge {
    	m4_object leftPoint,rightPoint;
    	boolean topleft,botleft,topright,botright;
    	
    	m4edge(m4_object lp, m4_object rp, boolean tl, boolean bl, boolean tr, boolean br) {
    		leftPoint = lp;
    		rightPoint = rp;
    		topleft = tl;
    		botleft = bl;
    		topright = tr;
    		botright = br;
    	}
    }
}
